@@ -1,5 +1,5 @@
 /**
- *  Copyright (C) 2002-2015   The FreeCol Team
+ *  Copyright (C) 2002-2022   The FreeCol Team
  *
  *  This file is part of FreeCol.
  *
@@ -20,12 +20,12 @@
 package net.sf.freecol.client.gui.panel;
 
 import java.awt.Component;
+import java.awt.Container;
+import java.awt.Dimension;
 import java.awt.GridLayout;
-import java.awt.event.ActionEvent;
 import java.awt.event.ActionListener;
 import java.awt.event.KeyEvent;
 import java.awt.event.MouseListener;
-import java.util.Iterator;
 import java.util.List;
 import java.util.logging.Level;
 import java.util.logging.Logger;
@@ -45,12 +45,15 @@ import javax.swing.text.StyleConstants;
 import javax.swing.text.StyledDocument;
 
 import net.miginfocom.swing.MigLayout;
-
 import net.sf.freecol.client.ClientOptions;
 import net.sf.freecol.client.FreeColClient;
-import net.sf.freecol.client.control.InGameController.BoycottAction;
-import net.sf.freecol.client.gui.ImageLibrary;
+import net.sf.freecol.client.gui.FontLibrary;
+import net.sf.freecol.client.gui.label.GoodsLabel;
+import net.sf.freecol.client.gui.label.MarketLabel;
+import net.sf.freecol.client.gui.label.UnitLabel;
+import net.sf.freecol.client.gui.panel.FreeColButton.ButtonStyle;
 import net.sf.freecol.common.i18n.Messages;
+import net.sf.freecol.common.model.Constants.BoycottAction;
 import net.sf.freecol.common.model.Europe;
 import net.sf.freecol.common.model.Goods;
 import net.sf.freecol.common.model.GoodsType;
@@ -59,6 +62,7 @@ import net.sf.freecol.common.model.Location;
 import net.sf.freecol.common.model.Map;
 import net.sf.freecol.common.model.Market;
 import net.sf.freecol.common.model.MarketData;
+import net.sf.freecol.common.model.Player;
 import net.sf.freecol.common.model.StringTemplate;
 import net.sf.freecol.common.model.TransactionListener;
 import net.sf.freecol.common.model.Unit;
@@ -83,6 +87,9 @@ public final class EuropePanel extends PortPanel {
 
         /**
          * Initialize this DestinationPanel.
+         *
+         * @param destination The destination {@code Location}
+         *     for this panel.
          */
         public void initialize(Location destination) {
             this.destination = destination;
@@ -129,7 +136,7 @@ public final class EuropePanel extends PortPanel {
             }
 
             // "ship" is a tag, not a key
-            Utility.localizeBorder(this, Unit.getDestinationLabel("ship",
+            Utility.localizeBorder(this, Unit.getUnitDestinationLabel("ship",
                     destination, getMyPlayer()));
             revalidate();
         }
@@ -140,13 +147,7 @@ public final class EuropePanel extends PortPanel {
         /**
          * {@inheritDoc}
          */
-        public boolean accepts(Unit unit) {
-            return unit.isNaval() && !unit.isDamaged();
-        }
-
-        /**
-         * {@inheritDoc}
-         */
+        @Override
         public boolean accepts(Goods goods) {
             return false;
         }
@@ -154,6 +155,23 @@ public final class EuropePanel extends PortPanel {
         /**
          * {@inheritDoc}
          */
+        @Override
+        public boolean accepts(GoodsType goodsType) {
+            return false;
+        }
+
+        /**
+         * {@inheritDoc}
+         */
+        @Override
+        public boolean accepts(Unit unit) {
+            return unit.isNaval() && !unit.isDamaged();
+        }
+
+        /**
+         * {@inheritDoc}
+         */
+        @Override
         public Component add(Component comp, boolean editState) {
             if (editState) {
                 if (!(comp instanceof UnitLabel)) {
@@ -177,7 +195,7 @@ public final class EuropePanel extends PortPanel {
                 if (!co.getBoolean(ClientOptions.AUTOLOAD_EMIGRANTS)
                     && unit.isInEurope()
                     && !(destination instanceof Europe)
-                    && docksPanel.getComponentCount() > 0
+                    && europeanDocksPanel.getComponentCount() > 0
                     && unit.hasSpaceLeft()) {
                     StringTemplate locName = destination
                         .getLocationLabelFor(unit.getOwner());
@@ -187,10 +205,17 @@ public final class EuropePanel extends PortPanel {
                             unit, "ok", "cancel")) return null;
                 }
 
-                comp.getParent().remove(comp);
                 igc().moveTo(unit, dest);
+                if (dest instanceof Europe) {
+                    ; // Assume this works
+                } else {
+                    // Fail if still in Europe if told to leave
+                    if (unit.isInEurope()) return null;
+                }
+                Container parent = comp.getParent();
+                if (parent != null) parent.remove(comp);
                 inPortPanel.update();
-                docksPanel.update();
+                europeanDocksPanel.update();
                 cargoPanel.update();
                 if (unit == cargoPanel.getCarrier()) {
                     cargoPanel.setCarrier(null);
@@ -206,6 +231,7 @@ public final class EuropePanel extends PortPanel {
         /**
          * {@inheritDoc}
          */
+        @Override
         public int suggested(GoodsType type) { return -1; } // N/A
     }
 
@@ -213,28 +239,55 @@ public final class EuropePanel extends PortPanel {
      * A panel that holds UnitLabels that represent Units that are
      * waiting on the docks in Europe.
      */
-    public final class DocksPanel extends UnitPanel implements DropTarget {
+    public final class EuropeanDocksPanel extends UnitPanel
+        implements DropTarget {
 
-        public DocksPanel() {
-            super(EuropePanel.this, "Europe - docks", true);
-
-            setLayout(new MigLayout("wrap 6"));
+        public EuropeanDocksPanel() {
+            super(null, new MigLayout("wrap 6"), EuropePanel.this,
+                  "Europe - docks", true);
         }
 
 
+        /**
+         * {@inheritDoc}
+         */
         @Override
         public void addPropertyChangeListeners() {
             europe.addPropertyChangeListener(this);
         }
 
+        /**
+         * {@inheritDoc}
+         */
         @Override
         public void removePropertyChangeListeners() {
             europe.removePropertyChangeListener(this);
         }
 
+        /**
+         * {@inheritDoc}
+         */
+        public void selectLabel() {
+            // do nothing
+        }
 
         // Interface DropTarget
 
+        /**
+         * {@inheritDoc}
+         */
+        @Override
+        public boolean accepts(Goods goods) {
+            return false;
+        }
+
+        /**
+         * {@inheritDoc}
+         */
+        public boolean accepts(GoodsType goodsType) {
+            return false;
+        }
+
         /**
          * {@inheritDoc}
          */
@@ -245,13 +298,7 @@ public final class EuropePanel extends PortPanel {
         /**
          * {@inheritDoc}
          */
-        public boolean accepts(Goods goods) {
-            return false;
-        }
-
-        /**
-         * {@inheritDoc}
-         */
+        @Override
         public Component add(Component comp, boolean editState) {
             Component c = add(comp);
             update();
@@ -261,6 +308,7 @@ public final class EuropePanel extends PortPanel {
         /**
          * {@inheritDoc}
          */
+        @Override
         public int suggested(GoodsType type) { return -1; } // N/A
 
 
@@ -275,7 +323,7 @@ public final class EuropePanel extends PortPanel {
         }
     }
 
-    private static final class EuropeButton extends JButton {
+    private static final class EuropeButton extends FreeColButton {
 
         public EuropeButton(String text, int keyEvent, String command,
                             ActionListener listener) {
@@ -301,7 +349,7 @@ public final class EuropePanel extends PortPanel {
     private final class EuropeInPortPanel extends InPortPanel {
 
         public EuropeInPortPanel() {
-            super(EuropePanel.this, "Europe - port", true);
+            super(new MigLayout(), EuropePanel.this, "Europe - port", true);
         }
 
 
@@ -357,10 +405,10 @@ public final class EuropePanel extends PortPanel {
         public void initialize() {
             removeAll();
 
+            final FreeColClient fcc = getFreeColClient();
             final Market market = getMyPlayer().getMarket();
-            ImageLibrary lib = getImageLibrary();
             for (GoodsType goodsType : getSpecification().getStorableGoodsTypeList()) {
-                MarketLabel label = new MarketLabel(lib, goodsType, market);
+                MarketLabel label = new MarketLabel(fcc, goodsType, market);
                 label.setTransferHandler(defaultTransferHandler);
                 label.addMouseListener(pressListener);
                 MarketData md = market.getMarketData(goodsType);
@@ -380,13 +428,7 @@ public final class EuropePanel extends PortPanel {
         /**
          * {@inheritDoc}
          */
-        public boolean accepts(Unit unit) {
-            return false;
-        }
-
-        /**
-         * {@inheritDoc}
-         */
+        @Override
         public boolean accepts(Goods goods) {
             return true;
         }
@@ -394,6 +436,23 @@ public final class EuropePanel extends PortPanel {
         /**
          * {@inheritDoc}
          */
+        @Override
+        public boolean accepts(GoodsType goodsType) {
+            return false;
+        }
+
+        /**
+         * {@inheritDoc}
+         */
+        @Override
+        public boolean accepts(Unit unit) {
+            return false;
+        }
+
+        /**
+         * {@inheritDoc}
+         */
+        @Override
         public Component add(Component comp, boolean editState) {
             if (editState) {
                 if (!(comp instanceof GoodsLabel)) {
@@ -409,10 +468,10 @@ public final class EuropePanel extends PortPanel {
                         .getBoycottChoice(goods, europe);
                     if (act != null) {
                         switch (act) {
-                        case PAY_ARREARS:
+                        case BOYCOTT_PAY_ARREARS:
                             igc().payArrears(goods.getType());
                             break;
-                        case DUMP_CARGO:
+                        case BOYCOTT_DUMP_CARGO:
                             igc().unloadCargo(goods, true);
                             break;
                         default:
@@ -430,13 +489,6 @@ public final class EuropePanel extends PortPanel {
             return comp;
         }
 
-        /**
-         * {@inheritDoc}
-         */
-        public int suggested(GoodsType type) {
-            return -1; // No good choice
-        }
-
 
         // Override Container
 
@@ -454,7 +506,7 @@ public final class EuropePanel extends PortPanel {
      */
     private final class TransactionLog extends JTextPane
         implements TransactionListener {
-
+        
         /**
          * Creates a transaction log.
          */
@@ -487,6 +539,7 @@ public final class EuropePanel extends PortPanel {
             StyledDocument doc = getStyledDocument();
             try {
                 if (doc.getLength() > 0) text = "\n\n" + text;
+                
                 doc.insertString(doc.getLength(), text, null);
             } catch (Exception e) {
                 logger.log(Level.WARNING, "Transaction log update failure", e);
@@ -550,7 +603,7 @@ public final class EuropePanel extends PortPanel {
 
     private DestinationPanel toEuropePanel;
 
-    private DocksPanel docksPanel;
+    private EuropeanDocksPanel europeanDocksPanel;
 
     private MarketPanel marketPanel;
 
@@ -565,7 +618,7 @@ public final class EuropePanel extends PortPanel {
     /**
      * The constructor for a EuropePanel.
      *
-     * @param freeColClient The <code>FreeColClient</code> for the game.
+     * @param freeColClient The {@code FreeColClient} for the game.
      * @param header True when a header should be added.
      */
     public EuropePanel(FreeColClient freeColClient, boolean header) {
@@ -573,23 +626,29 @@ public final class EuropePanel extends PortPanel {
                                            "[30%:][30%:][15%:]"));
 
         exitButton = new EuropeButton(Messages.message("close"),
-            KeyEvent.VK_ESCAPE, EuropeAction.EXIT.toString(), this);
+            KeyEvent.VK_ESCAPE, EuropeAction.EXIT.toString(),
+            ae -> exitAction()).withButtonStyle(ButtonStyle.IMPORTANT);
         trainButton = new EuropeButton(Messages.message("train"),
-            KeyEvent.VK_T, EuropeAction.TRAIN.toString(), this);
+            KeyEvent.VK_T, EuropeAction.TRAIN.toString(),
+            ae -> getGUI().showTrainPanel());
         purchaseButton = new EuropeButton(Messages.message("purchase"),
-            KeyEvent.VK_P, EuropeAction.PURCHASE.toString(), this);
+            KeyEvent.VK_P, EuropeAction.PURCHASE.toString(),
+            ae -> getGUI().showPurchasePanel());
         recruitButton = new EuropeButton(Messages.message("recruit"),
-            KeyEvent.VK_R, EuropeAction.RECRUIT.toString(), this);
+            KeyEvent.VK_R, EuropeAction.RECRUIT.toString(),
+            ae -> getGUI().showRecruitPanel());
         unloadButton = new EuropeButton(Messages.message("unload"),
-            KeyEvent.VK_U, EuropeAction.UNLOAD.toString(), this);
+            KeyEvent.VK_U, EuropeAction.UNLOAD.toString(),
+            ae -> unloadAction());
         sailButton = new EuropeButton(Messages.message("setSail"),
-            KeyEvent.VK_S, EuropeAction.SAIL.toString(), this);
+            KeyEvent.VK_S, EuropeAction.SAIL.toString(),
+            ae -> sailAction());
 
         toAmericaPanel = new DestinationPanel();
         toEuropePanel = new DestinationPanel();
         inPortPanel = new EuropeInPortPanel();
         cargoPanel = new CargoPanel(freeColClient, true);
-        docksPanel = new DocksPanel();
+        europeanDocksPanel = new EuropeanDocksPanel();
         marketPanel = new MarketPanel(this);
         log = new TransactionLog();
         europe = freeColClient.getMyPlayer().getEurope();
@@ -598,6 +657,7 @@ public final class EuropePanel extends PortPanel {
         StyleConstants.setAlignment(attributes, StyleConstants.ALIGN_RIGHT);
         //StyleConstants.setForeground(attributes, Color.WHITE);
         StyleConstants.setBold(attributes, true);
+        StyleConstants.setFontSize(attributes, (int) FontLibrary.getMainFontSize());
         log.setParagraphAttributes(attributes, true);
 
         defaultTransferHandler
@@ -606,7 +666,7 @@ public final class EuropePanel extends PortPanel {
         toEuropePanel.setTransferHandler(defaultTransferHandler);
         inPortPanel.setTransferHandler(defaultTransferHandler);
         cargoPanel.setTransferHandler(defaultTransferHandler);
-        docksPanel.setTransferHandler(defaultTransferHandler);
+        europeanDocksPanel.setTransferHandler(defaultTransferHandler);
         marketPanel.setTransferHandler(defaultTransferHandler);
 
         pressListener = new DragListener(freeColClient, this);
@@ -615,14 +675,14 @@ public final class EuropePanel extends PortPanel {
         toEuropePanel.addMouseListener(releaseListener);
         inPortPanel.addMouseListener(releaseListener);
         cargoPanel.addMouseListener(releaseListener);
-        docksPanel.addMouseListener(releaseListener);
+        europeanDocksPanel.addMouseListener(releaseListener);
         marketPanel.addMouseListener(releaseListener);
 
         toAmericaPanel.setLayout(new GridLayout(1, 0));
         toEuropePanel.setLayout(new GridLayout(1, 0));
         inPortPanel.setLayout(new GridLayout(1, 0));
         cargoPanel.setLayout(new GridLayout(1, 0));
-        docksPanel.setLayout(new GridLayout(0, 5));
+        europeanDocksPanel.setLayout(new GridLayout(0, 5));
 
         JScrollPane toAmericaScroll = new JScrollPane(toAmericaPanel,
             ScrollPaneConstants.VERTICAL_SCROLLBAR_NEVER,
@@ -640,7 +700,7 @@ public final class EuropePanel extends PortPanel {
             ScrollPaneConstants.VERTICAL_SCROLLBAR_NEVER,
             ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
         cargoScroll.getVerticalScrollBar().setUnitIncrement(16);
-        JScrollPane docksScroll = new JScrollPane(docksPanel,
+        JScrollPane docksScroll = new JScrollPane(europeanDocksPanel,
             ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED,
             ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
         docksScroll.getVerticalScrollBar().setUnitIncrement(16);
@@ -654,7 +714,7 @@ public final class EuropePanel extends PortPanel {
 
         toAmericaPanel.setBorder(Utility.localizedBorder("sailingToAmerica"));
         toEuropePanel.setBorder(Utility.localizedBorder("sailingToEurope"));
-        docksPanel.setBorder(Utility.localizedBorder("docks"));
+        europeanDocksPanel.setBorder(Utility.localizedBorder("docks"));
         inPortPanel.setBorder(Utility.localizedBorder("inPort"));
         marketPanel.setBorder(Utility.blankBorder(10, 10, 10, 10));
         log.setBorder(Utility.localizedBorder("sales"));
@@ -668,7 +728,7 @@ public final class EuropePanel extends PortPanel {
         cargoScroll.getViewport().setOpaque(false);
         cargoPanel.setOpaque(false);
         docksScroll.getViewport().setOpaque(false);
-        docksPanel.setOpaque(false);
+        europeanDocksPanel.setOpaque(false);
         marketScroll.getViewport().setOpaque(false);
         marketPanel.setOpaque(false);
         logScroll.getViewport().setOpaque(false);
@@ -677,7 +737,8 @@ public final class EuropePanel extends PortPanel {
         initialize(europe);
 
         if(header) {
-            add(Utility.localizedHeader(europe.getNameKey(), false),
+            add(Utility.localizedHeader(europe.getNameKey(),
+                                        Utility.FONTSPEC_TITLE),
                 "span, top, center");
         }
         add(toAmericaScroll, "sg, height 15%:, grow");
@@ -695,16 +756,20 @@ public final class EuropePanel extends PortPanel {
         add(sailButton);
         add(exitButton, "tag ok");
 
-        setSelectedUnitLabel(null);
+        // Select a unit, preferring most recently added
+        Unit u = europe.getLastUnit();
+        if (u == null) setSelectedUnitLabel(null); else setSelectedUnit(u);
 
-        float scale = getImageLibrary().getScaleFactor();
-        getGUI().restoreSavedSize(this, 200 + (int)(scale*850), 200 + (int)(scale*525));
+        final int width = getImageLibrary().scaleInt(1050);
+        final int height = getImageLibrary().scaleInt(725);
+        
+        getGUI().restoreSavedSize(this, new Dimension(width, height));
     }
 
     /**
      * Initialize this EuropePanel.
      *
-     * @param europe The <code>Europe</code> this panel should display.
+     * @param europe The {@code Europe} this panel should display.
      */
     private void initialize(Europe europe) {
         // Initialize the subpanels.
@@ -714,16 +779,16 @@ public final class EuropePanel extends PortPanel {
         cargoPanel.initialize();
         inPortPanel.initialize();
         marketPanel.initialize();
-        docksPanel.initialize();
+        europeanDocksPanel.initialize();
         log.initialize();
     }
 
     /**
      * Cleans up this EuropePanel.
      */
-    public void cleanup() {
+    private void cleanup() {
         log.cleanup();
-        docksPanel.cleanup();
+        europeanDocksPanel.cleanup();
         marketPanel.cleanup();
         inPortPanel.cleanup();
         cargoPanel.cleanup();
@@ -749,7 +814,7 @@ public final class EuropePanel extends PortPanel {
     /**
      * Selects a unit that is located somewhere on this panel.
      *
-     * @param unitLabel The <code>UnitLabel</code> for the unit that
+     * @param unitLabel The {@code UnitLabel} for the unit that
      *     is being selected.
      */
     @Override
@@ -775,7 +840,7 @@ public final class EuropePanel extends PortPanel {
      */
     private void exitAction() {
         cleanup();
-        getGUI().removeFromCanvas(this);
+        getGUI().removeComponent(this);
         igc().nextModelMessage();
     }
 
@@ -783,24 +848,21 @@ public final class EuropePanel extends PortPanel {
      * Unload the contents of the currently selected carrier.
      */
     private void unloadAction() {
+        final Player player = getMyPlayer();
         Unit unit = getSelectedUnit();
         if (unit != null && unit.isCarrier()) {
-            Iterator<Goods> goodsIterator = unit.getGoodsIterator();
-            while (goodsIterator.hasNext()) {
-                Goods goods = goodsIterator.next();
-                if (getMyPlayer().canTrade(goods.getType())) {
-                    igc().sellGoods(goods);
+            for (Goods g : unit.getCompactGoodsList()) {
+                if (player.canTrade(g.getType())) {
+                    igc().sellGoods(g);
                 } else {
-                    igc().payArrears(goods.getType());
+                    igc().payArrears(g.getType());
                 }
             }
-            Iterator<Unit> unitIterator = unit.getUnitIterator();
-            while (unitIterator.hasNext()) {
-                Unit newUnit = unitIterator.next();
-                igc().leaveShip(newUnit);
+            for (Unit u : unit.getUnitList()) {
+                igc().leaveShip(u);
             }
             cargoPanel.update();
-            docksPanel.update();
+            europeanDocksPanel.update();
         }
         requestFocus();
     }
@@ -810,9 +872,8 @@ public final class EuropePanel extends PortPanel {
      */
     private void sailAction() {
         Unit unit = getSelectedUnit();
-        if (unit != null && unit.isNaval()) {
-            UnitLabel unitLabel = getSelectedUnitLabel();
-            toAmericaPanel.add(unitLabel, true);
+        if (unit != null && toAmericaPanel.accepts(unit)) {
+            toAmericaPanel.add(getSelectedUnitLabel(), true);
         }
         requestFocus();
     }
@@ -831,41 +892,6 @@ public final class EuropePanel extends PortPanel {
     }
 
 
-    // Interface ActionListener
-
-    /**
-     * {@inheritDoc}
-     */
-    @Override
-    public void actionPerformed(ActionEvent ae) {
-        final String command = ae.getActionCommand();
-        EuropeAction act = EuropeAction.valueOf(command);
-        switch (act) {
-        case EXIT:
-            exitAction();
-            break;
-        case PURCHASE:
-            getGUI().showPurchasePanel();
-            break;
-        case RECRUIT:
-            getGUI().showRecruitPanel();
-            break;
-        case SAIL:
-            sailAction();
-            break;
-        case TRAIN:
-            getGUI().showTrainPanel();
-            break;
-        case UNLOAD:
-            unloadAction();
-            break;
-        default:
-            super.actionPerformed(ae);
-            break;
-        }
-    }
-
-
     // Override Component
 
     /**
@@ -878,7 +904,7 @@ public final class EuropePanel extends PortPanel {
         removeAll();
         toAmericaPanel = null;
         toEuropePanel = null;
-        docksPanel = null;
+        europeanDocksPanel = null;
         marketPanel = null;
         log = null;
         exitButton = trainButton = purchaseButton = recruitButton
