@@ -1,5 +1,5 @@
 /**
- *  Copyright (C) 2002-2015   The FreeCol Team
+ *  Copyright (C) 2002-2022   The FreeCol Team
  *
  *  This file is part of FreeCol.
  *
@@ -22,9 +22,9 @@ package net.sf.freecol.client.gui.panel;
 import java.awt.Color;
 import java.awt.Component;
 import java.awt.event.ActionEvent;
-import java.awt.event.ActionListener;
 import java.awt.event.ItemEvent;
 import java.awt.event.ItemListener;
+import java.net.InetAddress;
 import java.util.List;
 import java.util.logging.Logger;
 
@@ -38,17 +38,18 @@ import javax.swing.JSeparator;
 import javax.swing.JTextField;
 
 import net.miginfocom.swing.MigLayout;
-
 import net.sf.freecol.FreeCol;
 import net.sf.freecol.client.ClientOptions;
 import net.sf.freecol.client.FreeColClient;
 import net.sf.freecol.client.control.ConnectController;
-import net.sf.freecol.client.gui.SwingGUI;
+import net.sf.freecol.client.gui.GUI;
 import net.sf.freecol.client.gui.plaf.FreeColComboBoxRenderer;
 import net.sf.freecol.common.i18n.Messages;
-import net.sf.freecol.common.ServerInfo;
+import net.sf.freecol.common.io.FreeColModFile;
+import net.sf.freecol.common.io.FreeColRules;
 import net.sf.freecol.common.io.FreeColTcFile;
-import net.sf.freecol.common.io.Mods;
+import net.sf.freecol.common.metaserver.MetaServerUtils;
+import net.sf.freecol.common.metaserver.ServerInfo;
 import net.sf.freecol.common.model.NationOptions.Advantages;
 import net.sf.freecol.common.model.Specification;
 import net.sf.freecol.common.option.OptionGroup;
@@ -59,8 +60,7 @@ import net.sf.freecol.common.option.OptionGroup;
  * game, to join a running game, and to fetch a list of games from the
  * meta-server.
  */
-public final class NewPanel extends FreeColPanel
-    implements ActionListener, ItemListener {
+public final class NewPanel extends FreeColPanel implements ItemListener {
 
     private static final Logger logger = Logger.getLogger(NewPanel.class.getName());
 
@@ -113,12 +113,14 @@ public final class NewPanel extends FreeColPanel
 
     /** Start server port number label and field to input through. */
     private final JTextField serverPortField;
+    
+    private final JComboBox<InetAddress> serverAddressBox;
 
     /** The label for the rules selection. */
     private final JLabel rulesLabel;
 
     /** A box to choose the rules from. */
-    private final JComboBox<FreeColTcFile> rulesBox;
+    private final JComboBox<FreeColModFile> rulesBox;
 
     /** The check box to select a public server with. */
     private final JCheckBox publicServer;
@@ -157,7 +159,7 @@ public final class NewPanel extends FreeColPanel
     /**
      * Creates a new game panel.
      *
-     * @param freeColClient The <code>FreeColClient</code> for the game.
+     * @param freeColClient The {@code FreeColClient} for the game.
      */
     public NewPanel(FreeColClient freeColClient) {
         this(freeColClient, null);
@@ -166,12 +168,12 @@ public final class NewPanel extends FreeColPanel
     /**
      * Creates a new game panel.
      *
-     * @param freeColClient The <code>FreeColClient</code> for the game.
-     * @param specification An optional <code>Specification</code> value for
+     * @param freeColClient The {@code FreeColClient} for the game.
+     * @param specification An optional {@code Specification} value for
      *     the new game.
      */
     public NewPanel(FreeColClient freeColClient, Specification specification) {
-        super(freeColClient, new MigLayout("wrap 6", "[15]", ""));
+        super(freeColClient, null, new MigLayout("wrap 6", "[15]", ""));
 
         this.fixedSpecification = specification;
 
@@ -185,19 +187,19 @@ public final class NewPanel extends FreeColPanel
             start = new JRadioButton(Messages.message("newPanel.startMultiplayerGame"),
                                      false),
             meta = new JRadioButton(Messages.message("newPanel.getServerList")
-                + " (" + FreeCol.META_SERVER_ADDRESS + ")", false);
+                + " (" + FreeCol.getMetaServerAddress() + ")", false);
         this.buttonGroup.add(single);
         single.setActionCommand(String.valueOf(NewPanelAction.SINGLE));
-        single.addActionListener(this);
+        single.addActionListener(ae -> enableComponents());
         this.buttonGroup.add(join);
         join.setActionCommand(String.valueOf(NewPanelAction.JOIN));
-        join.addActionListener(this);
+        join.addActionListener(ae -> enableComponents());
         this.buttonGroup.add(start);
         start.setActionCommand(String.valueOf(NewPanelAction.START));
-        start.addActionListener(this);
+        start.addActionListener(ae -> enableComponents());
         this.buttonGroup.add(meta);
         meta.setActionCommand(String.valueOf(NewPanelAction.META_SERVER));
-        meta.addActionListener(this);
+        meta.addActionListener(ae -> enableComponents());
         single.setSelected(true);
 
         String name = getClientOptions().getText(ClientOptions.NAME);
@@ -220,31 +222,47 @@ public final class NewPanel extends FreeColPanel
         this.serverPortField.addActionListener((ActionEvent ae) -> {
                 getSelectedPort(NewPanel.this.serverPortField);
             });
+        
+        this.serverAddressBox = Utility.createServerInetAddressBox();
 
         this.rulesLabel = Utility.localizedLabel("rules");
         this.rulesBox = new JComboBox<>();
+        String selectRules;
         if (this.fixedSpecification == null) { // Allow TC selection
-            String selectTC = FreeCol.getTC();
-            for (FreeColTcFile tc : Mods.getRuleSets()) {
+            selectRules = FreeCol.getRules();
+            for (FreeColModFile tc : FreeColRules.getRulesList()) {
                 this.rulesBox.addItem(tc);
-                if (selectTC.equals(tc.getId())) {
+                if (selectRules.equals(tc.getId())) {
                     this.rulesBox.setSelectedItem(tc);
                 }
             }
         } else { // Force the use of the TC that contains the given spec
-            String selectTC = this.fixedSpecification.getId();
-            for (FreeColTcFile tc : Mods.getRuleSets()) {
-                if (selectTC.equals(tc.getId())) {
+            selectRules = this.fixedSpecification.getId();
+            for (FreeColModFile tc : FreeColRules.getRulesList()) {
+                if (selectRules.equals(tc.getId())) {
                     this.rulesBox.addItem(tc);
                     this.rulesBox.setSelectedItem(tc);
                 }
             }
         }
+        if (this.rulesBox.getSelectedItem() == null) {
+            this.rulesBox.setSelectedItem(this.rulesBox.getItemCount()-1);
+            logger.warning("No TC found for: " + selectRules
+                + ", failling back to " + this.rulesBox.getSelectedItem());
+        }
         this.rulesBox
-            .setRenderer(new FreeColComboBoxRenderer<FreeColTcFile>("mod."));
+            .setRenderer(new FreeColComboBoxRenderer<FreeColModFile>("mod."));
         this.rulesBox.addItemListener(this);
 
-        this.publicServer = new JCheckBox(Messages.message("newPanel.publicServer"));
+        this.publicServer
+            = new JCheckBox(Messages.message("newPanel.publicServer"));
+        
+        serverAddressBox.addActionListener(event -> {
+            final InetAddress address = (InetAddress) serverAddressBox.getSelectedItem();
+            publicServer.setEnabled(!address.isLoopbackAddress());
+        });
+        final InetAddress address = (InetAddress) serverAddressBox.getSelectedItem();
+        publicServer.setEnabled(!address.isLoopbackAddress());
 
         this.difficultyLabel = Utility.localizedLabel("difficulty");
         this.difficultyBox = new JComboBox<>();
@@ -252,9 +270,13 @@ public final class NewPanel extends FreeColPanel
             .setRenderer(new FreeColComboBoxRenderer<OptionGroup>());
         this.difficultyBox.addItemListener(this);
         this.difficultyButton = Utility.localizedButton("newPanel.showDifficulty");
-        this.difficultyButton
-            .setActionCommand(String.valueOf(NewPanelAction.SHOW_DIFFICULTY));
-        this.difficultyButton.addActionListener(this);
+        this.difficultyButton.addActionListener(ae -> {
+            getGUI().showDifficultyDialog(this.specification, this.difficulty, this.difficulty.isEditable(), (group) -> {
+                if (group != null) {
+                    update(true);
+                }
+            });
+        });
 
         this.joinNameLabel = Utility.localizedLabel("host");
         this.joinNameField = new JTextField(FreeCol.getServerHost());
@@ -269,18 +291,29 @@ public final class NewPanel extends FreeColPanel
 
         JButton cancel = Utility.localizedButton("cancel");
         cancel.setActionCommand(String.valueOf(NewPanelAction.CANCEL));
-        cancel.addActionListener(this);
+        cancel.addActionListener(ae -> {
+                final GUI gui = getGUI();
+                gui.removeComponent(this);
+                if (getFreeColClient().isMapEditor()) {
+                    gui.startMapEditorGUI();
+                } else {
+                    gui.showMainPanel(null);
+                }
+            });
         setCancelComponent(cancel);
 
         // Add all the components
-        add(Utility.localizedHeader("newPanel.newGamePanel", false), "span 6, center");
+        add(Utility.localizedHeader("newPanel.newGamePanel",
+                                    Utility.FONTSPEC_TITLE),
+            "span 6, center");
         add(single, "newline, span 3");
         add(new JSeparator(JSeparator.VERTICAL), "spany 7, grow");
-        add(Utility.localizedLabel("name"), "span, split 2");
+        add(Utility.localizedLabel("name"), SPAN_SPLIT_2);
         add(this.nameBox, "growx");
         add(start, "newline, span 3");
         add(this.advantagesLabel);
         add(this.advantagesBox, "growx");
+        add(this.serverAddressBox, "newline, skip");
         add(this.serverPortLabel, "newline, skip");
         add(this.serverPortField, "width 60:");
         add(this.rulesLabel);
@@ -302,7 +335,7 @@ public final class NewPanel extends FreeColPanel
             this.joinPortLabel, this.joinPortField
         };
         serverComponents = new Component[] {
-            this.serverPortLabel, this.serverPortField, this.publicServer
+            this.serverAddressBox, this.serverPortLabel, this.serverPortField, this.publicServer
         };
         gameComponents = new Component[] {
             this.advantagesLabel, this.advantagesBox,
@@ -311,21 +344,36 @@ public final class NewPanel extends FreeColPanel
         };
 
         this.specification = getSpecification();
-        this.difficulty = null;
-        updateDifficulty();
+        if (this.specification == null) {
+            throw new RuntimeException("No specification found for: " + this);
+        }
+        this.difficulty = this.specification
+            .getDifficultyOptionGroup(FreeCol.getDifficulty());
+        updateDifficultyBox();
+        if (this.difficulty == null) {
+            int index = this.difficultyBox.getItemCount() / 2;
+            this.difficultyBox.setSelectedItem(index);
+            this.difficulty = getSelectedDifficulty();
+        }
+        if (this.difficulty == null) {
+            throw new RuntimeException("No difficulty found for: " + this);
+        }
+        logger.info("NewPanel initialized with " + this.specification.getId()
+            + "/" + this.difficulty.getId());
         enableComponents();
         setSize(getPreferredSize());
     }
 
-    
+        
     /**
-     * If the TC box changed, update the specification.
+     * Update specification and difficulty as needed.
+     *
+     * @param changed If true force an update.
      */
-    private void updateSpecification() {
-        if (this.specification.getId().equals(getSelectedTC().getId())) return;
-        this.specification = getSpecification();
-        // Spec changed.  If using a custom difficulty, preserve it.
-        // Otherwise reset the difficulty wrt the new spec.
+    private void update(boolean changed) {
+        if (!changed) return;
+        // If using a custom difficulty, update it within the
+        // spec, otherwise read the difficulty from the new spec.
         if (this.difficulty.isEditable()) {
             this.specification.prepare(null, this.difficulty);
         } else {
@@ -333,47 +381,55 @@ public final class NewPanel extends FreeColPanel
             this.difficulty = this.specification
                 .getDifficultyOptionGroup(oldId);
         }
-        updateDifficulty();
+        updateDifficultyBox();
+    }
+
+    /**
+     * Check if the TC changed.
+     *
+     * @return True if the TC changed.
+     */
+    private boolean checkTC() {
+        if (this.specification.getId()
+            .equals(getSelectedRules().getId())) return false;
+        this.specification = getSpecification();
+        return true;
     }
         
     /**
-     * Update the contents of the difficulty level box depending on
-     * the specification currently selected.
+     * Check if the difficulty box changed.
+     *
+     * @return True if the difficulty changed.
      */
-    private void updateDifficulty() {
+    private boolean checkDifficulty() {
+        OptionGroup difficulty = getSelectedDifficulty();
+        if (this.difficulty.getId().equals(difficulty.getId())) return false;
+        this.difficulty = difficulty;
+        return true;
+    }
+    
+    /**
+     * Update just the difficulty box state.
+     */
+    private void updateDifficultyBox() {
+        // Update the contents of the difficulty level box depending on
+        // the specification currently selected.
         this.difficultyBox.removeItemListener(this);
         this.difficultyBox.removeAllItems();
         for (OptionGroup og : this.specification.getDifficultyLevels()) {
             this.difficultyBox.addItem(og);
         }
-        if (this.difficulty == null) {
-            this.difficulty = this.specification
-                .getDifficultyOptionGroup(FreeCol.getDifficulty());
-            if (this.difficulty == null) {
-                int index = this.difficultyBox.getItemCount() / 2;
-                this.difficulty = this.difficultyBox.getItemAt(index);
-            }
+        if (this.difficulty != null) {
+            this.difficultyBox.setSelectedItem(this.difficulty);
         }
-        this.difficultyBox.setSelectedItem(this.difficulty);
-        updateShowButton();
+        this.difficultyButton.setEnabled(this.difficulty != null);
+        String text = (this.difficulty != null && this.difficulty.isEditable())
+            ? "newPanel.editDifficulty"
+            : "newPanel.showDifficulty";
+        this.difficultyButton.setText(Messages.message(text));
         this.difficultyBox.addItemListener(this);
     }
 
-    /**
-     * Update the show button.
-     */
-    private void updateShowButton() {
-        if (this.difficulty == null) {
-            this.difficultyButton.setEnabled(false);
-        } else {
-            this.difficultyButton.setEnabled(true);
-            String text = (this.difficulty.isEditable())
-                ? "newPanel.editDifficulty"
-                : "newPanel.showDifficulty";
-            this.difficultyButton.setText(Messages.message(text));
-        }
-    }
-
     /**
      * Get the selected player name from the nameBox.
      *
@@ -398,14 +454,14 @@ public final class NewPanel extends FreeColPanel
      *
      * @return The selected TC.
      */
-    private FreeColTcFile getSelectedTC() {
-        return (FreeColTcFile)this.rulesBox.getSelectedItem();
+    private FreeColModFile getSelectedRules() {
+        return (FreeColModFile)this.rulesBox.getSelectedItem();
     }
 
     /**
      * Gets the currently selected difficulty from the difficultyBox.
      *
-     * @return The difficulty <code>OptionGroup</code>.
+     * @return The difficulty {@code OptionGroup}.
      */
     private OptionGroup getSelectedDifficulty() {
         return (OptionGroup)this.difficultyBox.getSelectedItem();
@@ -468,7 +524,7 @@ public final class NewPanel extends FreeColPanel
     /**
      * Dis/Enable a group of components.
      *
-     * @param components The <code>Component</code>s to set.
+     * @param components The {@code Component}s to set.
      * @param enable Enable if true.
      */
     private void enableComponents(Component[] components, boolean enable) {
@@ -484,12 +540,12 @@ public final class NewPanel extends FreeColPanel
      * Get the specification.  Either the one set for this panel, or the
      * one implied by the currently selected TC.
      *
-     * @return The current <code>Specification</code>.
+     * @return The current {@code Specification}.
      */
     @Override
     public Specification getSpecification() {
         if (this.fixedSpecification != null) return this.fixedSpecification;
-        return FreeCol.loadSpecification(getSelectedTC(), null, null);
+        return FreeCol.loadSpecification(getSelectedRules(), null, null);
     }
 
 
@@ -501,14 +557,14 @@ public final class NewPanel extends FreeColPanel
     @Override
     public void actionPerformed(ActionEvent ae) {
         final ConnectController cc = getFreeColClient().getConnectController();
-        final SwingGUI gui = getGUI();
+        final GUI gui = getGUI();
         final String command = ae.getActionCommand();
 
         switch (Enum.valueOf(NewPanelAction.class, command)) {
         case OK:
             FreeCol.setName(getSelectedName());
             FreeCol.setAdvantages(getSelectedAdvantages());
-            FreeCol.setTC(getSelectedTC().getId());
+            FreeCol.setRules(getSelectedRules().getId());
 
             NewPanelAction action = Enum.valueOf(NewPanelAction.class,
                 buttonGroup.getSelection().getActionCommand());
@@ -516,7 +572,7 @@ public final class NewPanel extends FreeColPanel
             case SINGLE:
                 this.specification.prepare(getSelectedAdvantages(),
                                            this.difficulty);
-                if (cc.startSinglePlayerGame(this.specification, false)) return;
+                if (cc.startSinglePlayerGame(this.specification)) return;
                 break;
             case JOIN:
                 int joinPort = getSelectedPort(this.joinPortField);
@@ -529,29 +585,20 @@ public final class NewPanel extends FreeColPanel
                 if (serverPort < 0) break;
                 this.specification.prepare(getSelectedAdvantages(),
                                            this.difficulty);
+                final InetAddress serverAddress = (InetAddress) this.serverAddressBox.getSelectedItem();
+                final boolean publicServerValue = this.publicServer.isSelected()
+                        && !serverAddress.isLoopbackAddress();
                 if (cc.startMultiplayerGame(this.specification,
-                        this.publicServer.isSelected(), serverPort)) return;
+                        publicServerValue, serverAddress, serverPort)) return;
                 break;
             case META_SERVER:
-                List<ServerInfo> servers = cc.getServerList();
+                List<ServerInfo> servers = MetaServerUtils.getServerList();
                 if (servers != null) gui.showServerListPanel(servers);
                 break;
             default:
                 break;
             }
             break;
-        case CANCEL:
-            gui.removeFromCanvas(this);
-            gui.showMainPanel(null);
-            break;
-        case SHOW_DIFFICULTY:
-            this.difficulty = gui.showDifficultyDialog(this.specification,
-                                                       this.difficulty);
-            updateDifficulty();
-            break;
-        case SINGLE: case JOIN: case START: case META_SERVER:
-            enableComponents();
-            break;
         default:
             super.actionPerformed(ae);
             break;
@@ -566,11 +613,6 @@ public final class NewPanel extends FreeColPanel
      */
     @Override
     public void itemStateChanged(ItemEvent e) {
-        if (e.getSource() == this.rulesBox) {
-            updateSpecification();
-        } else if (e.getSource() == this.difficultyBox) {
-            this.difficulty = getSelectedDifficulty();
-            updateShowButton();
-        }
+        update(checkTC() || checkDifficulty());
     }
 }
