@@ -1,5 +1,5 @@
 /**
- *  Copyright (C) 2002-2015   The FreeCol Team
+ *  Copyright (C) 2002-2022   The FreeCol Team
  *
  *  This file is part of FreeCol.
  *
@@ -19,9 +19,14 @@
 
 package net.sf.freecol.client.gui.panel;
 
+import static net.sf.freecol.common.util.CollectionUtils.any;
+import static net.sf.freecol.common.util.CollectionUtils.matchKey;
+import static net.sf.freecol.common.util.CollectionUtils.matchKeyEquals;
+import static net.sf.freecol.common.util.CollectionUtils.removeInPlace;
+import static net.sf.freecol.common.util.CollectionUtils.transform;
+
 import java.awt.Component;
 import java.awt.GridLayout;
-import java.awt.Image;
 import java.awt.datatransfer.DataFlavor;
 import java.awt.datatransfer.Transferable;
 import java.awt.datatransfer.UnsupportedFlavorException;
@@ -31,7 +36,6 @@ import java.awt.event.KeyListener;
 import java.awt.event.MouseListener;
 import java.io.IOException;
 import java.util.ArrayList;
-import java.util.Arrays;
 import java.util.List;
 import java.util.logging.Level;
 import java.util.logging.Logger;
@@ -48,15 +52,16 @@ import javax.swing.JPanel;
 import javax.swing.JScrollPane;
 import javax.swing.JTextField;
 import javax.swing.ListCellRenderer;
+import javax.swing.SwingConstants;
 import javax.swing.TransferHandler;
 import javax.swing.event.ListSelectionEvent;
 import javax.swing.event.ListSelectionListener;
 import javax.swing.plaf.PanelUI;
 
 import net.miginfocom.swing.MigLayout;
-
 import net.sf.freecol.client.FreeColClient;
 import net.sf.freecol.client.gui.ImageLibrary;
+import net.sf.freecol.client.gui.label.GoodsTypeLabel;
 import net.sf.freecol.client.gui.plaf.FreeColSelectedPanelUI;
 import net.sf.freecol.common.i18n.Messages;
 import net.sf.freecol.common.model.Colony;
@@ -67,9 +72,9 @@ import net.sf.freecol.common.model.GoodsType;
 import net.sf.freecol.common.model.Location;
 import net.sf.freecol.common.model.Player;
 import net.sf.freecol.common.model.StringTemplate;
+import net.sf.freecol.common.model.TradeLocation;
 import net.sf.freecol.common.model.TradeRoute;
 import net.sf.freecol.common.model.TradeRouteStop;
-import static net.sf.freecol.common.util.CollectionUtils.*;
 
 
 /**
@@ -83,152 +88,6 @@ public final class TradeRouteInputPanel extends FreeColPanel
     public static final DataFlavor STOP_FLAVOR
         = new DataFlavor(TradeRouteStop.class, "Stop");
 
-    /**
-     * Special label for cargo-to-carry type.
-     */
-    private class CargoLabel extends JLabel {
-
-        private final GoodsType goodsType;
-
-
-        public CargoLabel(GoodsType type) {
-            super(new ImageIcon(getImageLibrary().getIconImage(type)));
-
-            this.goodsType = type;
-            setDisabledIcon(getDisabledIcon());
-            setTransferHandler(TradeRouteInputPanel.this.cargoHandler);
-            addMouseListener(TradeRouteInputPanel.this.dragListener);
-        }
-
-        public GoodsType getType() {
-            return this.goodsType;
-        }
-    }
-
-    /**
-     * Panel for the cargo the carrier is supposed to take on board at
-     * a certain stop.
-     *
-     * FIXME: create a single cargo panel for this purpose and the use
-     * in the ColonyPanel, the EuropePanel and the CaptureGoodsDialog?
-     */
-    private class CargoPanel extends JPanel {
-
-        public CargoPanel() {
-            super();
-
-            setOpaque(false);
-            setBorder(Utility.localizedBorder("cargoOnCarrier"));
-            addMouseListener(TradeRouteInputPanel.this.dropListener);
-        }
-
-        public void initialize(TradeRouteStop newStop) {
-            removeAll();
-            if (newStop != null) {
-                // stop = newStop;
-                for (GoodsType goodsType : newStop.getCargo()) {
-                    add(new CargoLabel(goodsType));
-                }
-            }
-            revalidate();
-            repaint();
-        }
-    }
-
-    /**
-     * TransferHandler for CargoLabels.
-     *
-     * FIXME: check whether this could/should be folded into the
-     * DefaultTransferHandler.
-     */
-    private class CargoHandler extends TransferHandler {
-
-        /**
-         * {@inheritDoc}
-         */
-        @Override
-        protected Transferable createTransferable(JComponent c) {
-            return new ImageSelection((CargoLabel) c);
-        }
-
-        /**
-         * {@inheritDoc}
-         */
-        @Override
-        public int getSourceActions(JComponent c) {
-            return COPY_OR_MOVE;
-        }
-
-        /**
-         * {@inheritDoc}
-         */
-        @Override
-        public boolean importData(JComponent target, Transferable data) {
-            if (!canImport(target, data.getTransferDataFlavors()))
-                return false;
-            try {
-                CargoLabel label = (CargoLabel)data.getTransferData(DefaultTransferHandler.flavor);
-                if (target instanceof CargoPanel) {
-                    CargoLabel newLabel = new CargoLabel(label.getType());
-                    TradeRouteInputPanel.this.cargoPanel.add(newLabel);
-                    TradeRouteInputPanel.this.cargoPanel.revalidate();
-                    int[] indices = TradeRouteInputPanel.this.stopList
-                        .getSelectedIndices();
-                    for (int index : indices) {
-                        TradeRouteStop stop = TradeRouteInputPanel.this
-                            .stopListModel.get(index);
-                        stop.addCargo(label.getType());
-                    }
-                    TradeRouteInputPanel.this.stopList.revalidate();
-                    TradeRouteInputPanel.this.stopList.repaint();
-                }
-                return true;
-            } catch (IOException|UnsupportedFlavorException ex) {
-                logger.log(Level.WARNING, "CargoHandler import", ex);
-            }
-            return false;
-        }
-
-        /**
-         * {@inheritDoc}
-         */
-        @Override
-        protected void exportDone(JComponent source, Transferable data,
-                                  int action) {
-            try {
-                CargoLabel label = (CargoLabel)data.getTransferData(DefaultTransferHandler.flavor);
-                if (source.getParent() instanceof CargoPanel) {
-                    TradeRouteInputPanel.this.cargoPanel.remove(label);
-                    int[] indices = TradeRouteInputPanel.this.stopList
-                        .getSelectedIndices();
-                    for (int stopIndex : indices) {
-                        TradeRouteStop stop = TradeRouteInputPanel.this
-                            .stopListModel.get(stopIndex);
-                        List<GoodsType> cargo = new ArrayList<>(stop.getCargo());
-                        for (int index = 0; index < cargo.size(); index++) {
-                            if (cargo.get(index) == label.getType()) {
-                                cargo.remove(index);
-                                break;
-                            }
-                        }
-                        stop.setCargo(cargo);
-                    }
-                    TradeRouteInputPanel.this.stopList.revalidate();
-                    TradeRouteInputPanel.this.stopList.repaint();
-                    TradeRouteInputPanel.this.cargoPanel.revalidate();
-                    TradeRouteInputPanel.this.cargoPanel.repaint();
-                }
-            } catch (IOException|UnsupportedFlavorException ex) {
-                logger.log(Level.WARNING, "CargoHandler export", ex);
-            }
-        }
-
-        @Override
-        public boolean canImport(JComponent c, DataFlavor[] flavors) {
-            return any(flavors, f -> f.equals(DefaultTransferHandler.flavor));
-        }
-    }
-
     private class DestinationCellRenderer extends JLabel
         implements ListCellRenderer<String> {
 
@@ -261,21 +120,26 @@ public final class TradeRouteInputPanel extends FreeColPanel
 
     /**
      * Panel for all types of goods that can be loaded onto a carrier.
+     * This should *not* change under drag/drop, but a drop indicates
+     * that the dropped goods type is no longer needed to be loaded
+     * at the currently selected stop.
      */
-    private class GoodsPanel extends JPanel {
+    private class AllGoodsTypesPanel extends GoodsTypePanel {
 
-        public GoodsPanel() {
-            super(new GridLayout(0, 4, MARGIN, MARGIN));
+        /** The currently selected goods types. */
+        private List<GoodsTypeLabel> labels;
+        
+        public AllGoodsTypesPanel(List<GoodsTypeLabel> labels) {
+            super(new GridLayout(0, 4, MARGIN, MARGIN), true);
 
-            for (GoodsType goodsType : getSpecification().getGoodsTypeList()) {
-                if (goodsType.isStorable()) {
-                    CargoLabel label = new CargoLabel(goodsType);
-                    add(label);
-                }
-            }
+            this.labels = labels;
             setOpaque(false);
             setBorder(Utility.localizedBorder("goods"));
-            addMouseListener(TradeRouteInputPanel.this.dropListener);
+            reset();
+        }
+
+        public void reset() {
+            setLabels(this.labels);
         }
 
         /**
@@ -285,13 +149,56 @@ public final class TradeRouteInputPanel extends FreeColPanel
         public void setEnabled(boolean enable) {
             super.setEnabled(enable);
             for (Component child : getComponents()) {
-                if (child instanceof CargoLabel) {
-                    child.setEnabled(enable);
-                }
+                if (child instanceof GoodsTypeLabel) child.setEnabled(enable);
             }
         }
+
+        // Interface DropTarget
+
+        /**
+         * {@inheritDoc}
+         */
+        @Override
+        public Component add(Component comp, boolean editState) {
+            Component ret = super.add(comp);
+            if (comp instanceof GoodsTypeLabel) {
+                GoodsTypeLabel gtl = (GoodsTypeLabel)comp;
+                cancelImport(gtl.getType());
+            }
+            return ret;
+        }
     }
 
+    /**
+     * Panel for the types of goods that are to be loaded onto the carrier
+     * at the current stop.  This changes under drag/drop, and a drop
+     * adds the dropped goods type to be loaded at the currently selected
+     * stop.
+     */
+    private class StopGoodsTypesPanel extends GoodsTypePanel {
+
+        public StopGoodsTypesPanel() {
+            super(new MigLayout("wrap 10"), false);
+        }
+
+        // Interface DropTarget
+
+        /**
+         * {@inheritDoc}
+         */
+        @Override
+        public Component add(Component comp, boolean editState) {
+            Component ret = super.add(comp);
+            if (ret instanceof GoodsTypeLabel) {
+                GoodsTypeLabel gtl = (GoodsTypeLabel)ret;
+                enableImport(gtl.getType());
+                revalidate();
+                repaint();
+            }
+            return ret;
+        }
+    }
+        
     private static class StopListTransferable implements Transferable {
 
         private final List<TradeRouteStop> stops;
@@ -342,7 +249,7 @@ public final class TradeRouteInputPanel extends FreeColPanel
          */
         @Override
         public boolean canImport(JComponent c, DataFlavor[] flavors) {
-            return any(flavors, f -> f.equals(STOP_FLAVOR));
+            return any(flavors, matchKeyEquals(STOP_FLAVOR));
         }
 
         /**
@@ -350,11 +257,12 @@ public final class TradeRouteInputPanel extends FreeColPanel
          */
         @Override
         protected Transferable createTransferable(JComponent c) {
-            JList list = (JList)c;
-            DefaultListModel model = (DefaultListModel)list.getModel();
-            List<TradeRouteStop> stops = new ArrayList<>();
-            for (int index : list.getSelectedIndices()) {
-                stops.add((TradeRouteStop)model.get(index));
+            final JList list = (JList) c;
+            final DefaultListModel model = (DefaultListModel)list.getModel();
+            final int[] indicies = list.getSelectedIndices();
+            final List<TradeRouteStop> stops = new ArrayList<>(indicies.length);
+            for (int i : indicies) {
+                stops.add((TradeRouteStop) model.get(i));
             }
             return new StopListTransferable(stops);
         }
@@ -372,47 +280,70 @@ public final class TradeRouteInputPanel extends FreeColPanel
          */
         @Override
         public boolean importData(JComponent target, Transferable data) {
-            JList<TradeRouteStop> stl = TradeRouteInputPanel.this.stopList;
-            if (canImport(target, data.getTransferDataFlavors())
-                && target == stl
-                && data instanceof StopListTransferable) {
-                List<TradeRouteStop> stops
-                    = ((StopListTransferable)data).getStops();
-                DefaultListModel<TradeRouteStop> model
-                    = new DefaultListModel<>();
-                int index = stl.getMaxSelectionIndex();
-                for (TradeRouteStop stop : stops) {
-                    if (index < 0) {
-                        model.addElement(stop);
+            final JList<TradeRouteStop> stl = TradeRouteInputPanel.this.stopList;
+            try {
+                if (target == stl && canImport(target, data.getTransferDataFlavors())) {
+                    @SuppressWarnings("unchecked")
+                    final List<TradeRouteStop> stops = (List<TradeRouteStop>) data.getTransferData(STOP_FLAVOR);
+                    
+                    final DefaultListModel<TradeRouteStop> newModel = new DefaultListModel<>();
+                    final int targetIndex = stl.getMaxSelectionIndex();
+                    
+                    final List<TradeRouteStop> oldList = getAllValues(stl);
+                    if (targetIndex < 0) {
+                        oldList.removeAll(stops);
+                        newModel.addAll(oldList);
+                        newModel.addAll(stops);
                     } else {
-                        index++;
-                        model.add(index, stop);
+                        final List<TradeRouteStop> beforeList = new ArrayList<>(oldList.subList(0, targetIndex));
+                        beforeList.removeAll(stops);
+                        
+                        final List<TradeRouteStop> afterList = new ArrayList<>(oldList.subList(targetIndex, oldList.size()));
+                        afterList.removeAll(stops);
+                        
+                        newModel.addAll(beforeList);
+                        newModel.addAll(stops);
+                        newModel.addAll(afterList);
                     }
+                    
+                    stopListModel = newModel;
+                    stl.setModel(newModel);
+                    if (targetIndex < 0) {
+                        stl.setSelectedIndex(newModel.getSize() - 1);
+                    } else {
+                        stl.setSelectedIndex(targetIndex);
+                    }
+                    stl.invalidate();
+                    stl.repaint();
+                    
+                    return true;
                 }
-                stl.setModel(model);
-                return true;
+            } catch (IOException e) {
+                logger.log(Level.WARNING, "Drag&drop failed.", e);
+                return false;
+            } catch (UnsupportedFlavorException e) {
+                logger.log(Level.WARNING, "Drag&drop failed.", e);
+                return false;
+            } catch (RuntimeException e) {
+                logger.log(Level.WARNING, "Drag&drop failed.", e);
+                throw e;
             }
             return false;
         }
 
+        private List<TradeRouteStop> getAllValues(final JList<TradeRouteStop> jList) {
+            final List<TradeRouteStop> oldList = new ArrayList<>();
+            for (int i=0; i<jList.getModel().getSize(); i++) {
+                oldList.add(jList.getModel().getElementAt(i));
+            }
+            return oldList;
+        }
+
         /**
          * {@inheritDoc}
          */
         @Override
-        protected void exportDone(JComponent source, Transferable data,
-                                  int action) {
-            try {
-                if (source instanceof JList && action == MOVE) {
-                    JList stopList = (JList)source;
-                    DefaultListModel listModel
-                        = (DefaultListModel)stopList.getModel();
-                    for (Object o : (List) data.getTransferData(STOP_FLAVOR)) {
-                        listModel.removeElement(o);
-                    }
-                }
-            } catch (Exception e) {
-                logger.warning(e.toString());
-            }
+        protected void exportDone(JComponent source, Transferable data, int action) {
         }
     }
 
@@ -422,9 +353,9 @@ public final class TradeRouteInputPanel extends FreeColPanel
         private final JPanel NORMAL_COMPONENT = new JPanel();
 
         public StopRenderer() {
-            NORMAL_COMPONENT.setLayout(new MigLayout("", "[80, center][]"));
+            NORMAL_COMPONENT.setLayout(new MigLayout("", "[center][]"));
             NORMAL_COMPONENT.setOpaque(false);
-            SELECTED_COMPONENT.setLayout(new MigLayout("", "[80, center][]"));
+            SELECTED_COMPONENT.setLayout(new MigLayout("", "[center][]"));
             SELECTED_COMPONENT.setOpaque(false);
             SELECTED_COMPONENT.setUI((PanelUI)FreeColSelectedPanelUI
                 .createUI(SELECTED_COMPONENT));
@@ -447,26 +378,39 @@ public final class TradeRouteInputPanel extends FreeColPanel
             panel.setFont(list.getFont());
             Location location = value.getLocation();
             ImageLibrary lib = getImageLibrary();
-            JLabel icon, name;
-            if (location instanceof Europe) {
-                Europe europe = (Europe) location;
-                Image image = lib.getSmallerMiscIconImage(
-                    europe.getOwner().getNation());
-                icon = new JLabel(new ImageIcon(image));
-                name = Utility.localizedLabel(europe);
-            } else if (location instanceof Colony) {
-                Colony colony = (Colony) location;
-                icon = new JLabel(new ImageIcon(ImageLibrary.getSettlementImage(
-                    colony, lib.getScaleFactor()* 0.5f)));
-                name = new JLabel(colony.getName());
+            ImageIcon ii;
+            JLabel name;
+
+            if (location instanceof TradeLocation) {
+                TradeLocation tl = (TradeLocation) location;
+                if (tl.canBeInput() == true) {
+                    name = tl.getNameAsJlabel();
+                } else   {
+                    throw new IllegalStateException("Bogus location: " + location);
+                }
             } else {
                 throw new IllegalStateException("Bogus location: " + location);
             }
-            panel.add(icon, "spany");
+
+            if (location instanceof Europe) {
+                Player owner = ((Europe)location).getOwner();
+                ii = new ImageIcon(lib.getSmallerNationImage(owner.getNation()));
+            } else if (location instanceof Colony) {
+                Colony colony = (Colony) location;
+                ii = new ImageIcon(lib.getSmallerSettlementImage(colony));
+            } else {
+                throw new IllegalStateException("Bogus location: " + location);
+            }
+            
+            name.setVerticalTextPosition(SwingConstants.CENTER);
+            
+            final int width = (int) (80 * getImageLibrary().getScaleFactor());
+            final int height = (int) (60 * getImageLibrary().getScaleFactor());
+            panel.add(new JLabel(ii), "spany, width " + width + ", height " + height);
             panel.add(name, "span, wrap");
             for (GoodsType cargo : value.getCargo()) {
-                panel.add(new JLabel(new ImageIcon(
-                    lib.getSmallerIconImage(cargo))));
+                ii = new ImageIcon(lib.getSmallerGoodsTypeImage(cargo));
+                panel.add(new JLabel(ii));
             }
             return panel;
         }
@@ -479,11 +423,11 @@ public final class TradeRouteInputPanel extends FreeColPanel
      */
     private final TradeRoute newRoute;
 
-    /** A TransferHandler for the cargo labels. */
-    private CargoHandler cargoHandler;
+    /** The TransferHandler for the cargo labels. */
+    private TransferHandler transferHandler;
 
-    /** Mouse listeners to use throughout. */
-    private MouseListener dragListener, dropListener;
+    /** Mouse listener to use throughout. */
+    private transient MouseListener dragListener;
 
     /** Model to contain the current stops. */
     private DefaultListModel<TradeRouteStop> stopListModel;
@@ -506,40 +450,39 @@ public final class TradeRouteInputPanel extends FreeColPanel
     /** A button to remove stops with. */
     private JButton removeStopButton;
 
-    /** The panel displaying the goods that could be transported. */
-    private final GoodsPanel goodsPanel;
+    /** The panel displaying all goods types that could be transported. */
+    private AllGoodsTypesPanel allGoodsTypesPanel;
 
-    /** The panel displaying the cargo at the selected stop. */
-    private CargoPanel cargoPanel;
+    /** The panel displaying the goods types to collect at the selected stop. */
+    private StopGoodsTypesPanel stopGoodsTypesPanel;
 
 
     /**
      * Create a panel to define trade route cargos.
      *
-     * @param freeColClient The <code>FreeColClient</code> for the game.
-     * @param newRoute The <code>TradeRoute</code> to operate on.
+     * @param freeColClient The {@code FreeColClient} for the game.
+     * @param newRoute The {@code TradeRoute} to operate on.
      */
     public TradeRouteInputPanel(FreeColClient freeColClient,
                                 TradeRoute newRoute) {
-        super(freeColClient, new MigLayout("wrap 4, fill", "[]20[fill]rel"));
+        super(freeColClient, null,
+              new MigLayout("wrap 4, fill", "[]20[fill]rel"));
 
         final Game game = freeColClient.getGame();
         final Player player = getMyPlayer();
-        final TradeRoute tradeRoute = newRoute.copy(game, TradeRoute.class);
+        final TradeRoute tradeRoute = newRoute.copy(game);
 
         this.newRoute = newRoute;
-        this.cargoHandler = new CargoHandler();
+        this.transferHandler = new DefaultTransferHandler(freeColClient, this);
         this.dragListener = new DragListener(freeColClient, this);
-        this.dropListener = new DropListener();
 
         this.stopListModel = new DefaultListModel<>();
-        for (TradeRouteStop stop : tradeRoute.getStops()) {
+        for (TradeRouteStop stop : tradeRoute.getStopList()) {
             this.stopListModel.addElement(stop);
         }
 
         this.stopList = new JList<>(this.stopListModel);
         this.stopList.setCellRenderer(new StopRenderer());
-        this.stopList.setFixedCellHeight(48);
         this.stopList.setDragEnabled(true);
         this.stopList.setTransferHandler(new StopListHandler());
         this.stopList.addKeyListener(new KeyListener() {
@@ -560,19 +503,21 @@ public final class TradeRouteInputPanel extends FreeColPanel
         this.stopList.addListSelectionListener(this);
         JScrollPane tradeRouteView = new JScrollPane(stopList);
 
-        JLabel nameLabel = Utility.localizedLabel("tradeRouteInputPanel.nameLabel");
+        JLabel nameLabel
+            = Utility.localizedLabel("tradeRouteInputPanel.nameLabel");
         this.tradeRouteName = new JTextField(tradeRoute.getName());
 
         JLabel destinationLabel
             = Utility.localizedLabel("tradeRouteInputPanel.destinationLabel");
         this.destinationSelector = new JComboBox<>();
         this.destinationSelector.setRenderer(new DestinationCellRenderer());
-        StringTemplate template = StringTemplate.template("tradeRouteInputPanel.allColonies");
+        StringTemplate template
+            = StringTemplate.template("tradeRouteInputPanel.allColonies");
         this.destinationSelector.addItem(Messages.message(template));
         if (player.getEurope() != null) {
             this.destinationSelector.addItem(player.getEurope().getId());
         }
-        for (Colony colony : freeColClient.getMySortedColonies()) {
+        for (Colony colony : player.getColonyList()) {
             this.destinationSelector.addItem(colony.getId());
         }
 
@@ -583,28 +528,34 @@ public final class TradeRouteInputPanel extends FreeColPanel
                 tradeRoute.setSilent(messagesBox.isSelected());
             });
 
-        this.addStopButton = Utility.localizedButton("tradeRouteInputPanel.addStop");
+        this.addStopButton
+            = Utility.localizedButton("tradeRouteInputPanel.addStop");
         this.addStopButton.addActionListener((ActionEvent ae) -> {
                 addSelectedStops();
             });
 
-        this.removeStopButton = Utility.localizedButton("tradeRouteInputPanel.removeStop");
+        this.removeStopButton
+            = Utility.localizedButton("tradeRouteInputPanel.removeStop");
         this.removeStopButton.addActionListener((ActionEvent ae) -> {
                 deleteCurrentlySelectedStops();
             });
 
-        this.goodsPanel = new GoodsPanel();
-        this.goodsPanel.setTransferHandler(this.cargoHandler);
-        this.goodsPanel.setEnabled(false);
-        this.cargoPanel = new CargoPanel();
-        this.cargoPanel.setTransferHandler(this.cargoHandler);
+        final List<GoodsType> gtl = getSpecification().getGoodsTypeList();
+        this.allGoodsTypesPanel = new AllGoodsTypesPanel(transform(gtl,
+                GoodsType::isStorable, gt -> buildCargoLabel(gt)));
+        this.allGoodsTypesPanel.setTransferHandler(this.transferHandler);
+        this.allGoodsTypesPanel.setEnabled(false);
+
+        this.stopGoodsTypesPanel = new StopGoodsTypesPanel();
+        this.stopGoodsTypesPanel.setTransferHandler(this.transferHandler);
 
         JButton cancelButton = Utility.localizedButton("cancel");
         cancelButton.setActionCommand(CANCEL);
         cancelButton.addActionListener(this);
         setCancelComponent(cancelButton);
 
-        add(Utility.localizedHeader("tradeRouteInputPanel.editRoute", false),
+        add(Utility.localizedHeader("tradeRouteInputPanel.editRoute",
+                                    Utility.FONTSPEC_TITLE),
             "span, align center");
         add(tradeRouteView, "span 1 5, grow");
         add(nameLabel);
@@ -614,24 +565,70 @@ public final class TradeRouteInputPanel extends FreeColPanel
         add(this.messagesBox);
         add(this.addStopButton);
         add(this.removeStopButton, "span");
-        add(this.goodsPanel, "span");
-        add(this.cargoPanel, "span, height 80:, growy");
+        add(this.allGoodsTypesPanel, "span");
+        
+        final int iconHeight = (int) (80 * getImageLibrary().getScaleFactor());
+        add(this.stopGoodsTypesPanel, "span, height " + iconHeight + "px:, growy");
         add(okButton, "newline 20, span, split 2, tag ok");
         add(cancelButton, "tag cancel");
 
         // update cargo panel if stop is selected
         if (this.stopListModel.getSize() > 0) {
             this.stopList.setSelectedIndex(0);
-            TradeRouteStop selectedStop = this.stopListModel.firstElement();
-            this.cargoPanel.initialize(selectedStop);
+            updateCargoPanel(this.stopListModel.firstElement());
         }
-
         // update buttons according to selection
         updateButtons();
 
         getGUI().restoreSavedSize(this, getPreferredSize());
     }
 
+    /**
+     * Update the cargo panel to show a given stop.
+     *
+     * @param stop The {@code TradeRouteStop} to select.
+     */
+    private void updateCargoPanel(TradeRouteStop stop) {
+        this.stopGoodsTypesPanel.setLabels((stop == null) ? null
+            : transform(stop.getCargo(), GoodsType::isStorable,
+                        gt -> buildCargoLabel(gt)));
+    }
+
+    /**
+     * Import new goods at the selected stops.
+     *
+     * @param gt The {@code GoodsType} to import.
+     */
+    private void enableImport(GoodsType gt) {
+        if (gt == null) return;
+        for (int stopIndex : this.stopList.getSelectedIndices()) {
+            TradeRouteStop stop = this.stopListModel.get(stopIndex);
+            stop.addCargo(gt);
+        }
+        this.stopList.revalidate();
+        this.stopList.repaint();
+    }
+
+    /**
+     * Cancel import of goods at the selected stops.
+     *
+     * @param gt The {@code GoodsType} to stop importing.
+     */
+    private void cancelImport(GoodsType gt) {
+        this.stopGoodsTypesPanel.removeGoodsType(gt);
+        for (int stopIndex : this.stopList.getSelectedIndices()) {
+            TradeRouteStop stop = this.stopListModel.get(stopIndex);
+            List<GoodsType> cargo = new ArrayList<>(stop.getCargo());
+            if (removeInPlace(cargo, matchKey(gt))) {
+                stop.setCargo(cargo);
+            }
+        }
+        this.stopList.revalidate();
+        this.stopList.repaint();
+        this.stopGoodsTypesPanel.revalidate();
+        this.stopGoodsTypesPanel.repaint();
+    }
+
     /**
      * Add any stops selected in the destination selector.
      */
@@ -646,11 +643,10 @@ public final class TradeRouteInputPanel extends FreeColPanel
             startIndex = sel;
             endIndex = startIndex+1;
         }
-        List<GoodsType> cargo = new ArrayList<>();
-        for (Component comp : cargoPanel.getComponents()) {
-            CargoLabel label = (CargoLabel)comp;
-            cargo.add(label.getType());
-        }
+        List<GoodsType> cargo
+            = transform(this.stopGoodsTypesPanel.getComponents(),
+                        c -> c instanceof GoodsTypeLabel,
+                        c -> ((GoodsTypeLabel)c).getType());
         int maxIndex = this.stopList.getMaxSelectionIndex();
         for (int i = startIndex; i < endIndex; i++) {
             String id = this.destinationSelector.getItemAt(i);
@@ -669,6 +665,19 @@ public final class TradeRouteInputPanel extends FreeColPanel
         }
     }
 
+    /**
+     * Convenience function to build a new {@code GoodsTypeLabel}.
+     *
+     * @param gt The {@code GoodsType} for the label.
+     * @return A {@code GoodsTypeLabel} for the goods type.
+     */
+    private GoodsTypeLabel buildCargoLabel(GoodsType gt) {
+        GoodsTypeLabel label = new GoodsTypeLabel(getFreeColClient(), gt);
+        label.setTransferHandler(this.transferHandler);
+        label.addMouseListener(this.dragListener);
+        return label;
+    }
+
     /**
      * Delete any stops currently selected in the stop list.
      */
@@ -701,14 +710,14 @@ public final class TradeRouteInputPanel extends FreeColPanel
      */
     public void cancelTradeRoute() {
         this.newRoute.setName(null);
-        getGUI().removeFromCanvas(this);
+        getGUI().removeComponent(this);
     }
 
     /**
      * Enables the remove stop button if a stop is selected and disables it
      * otherwise.
      */
-    public void updateButtons() {
+    private void updateButtons() {
         this.addStopButton.setEnabled(this.stopListModel.getSize() 
             < this.destinationSelector.getItemCount() - 1);
         this.removeStopButton.setEnabled(this.stopList.getSelectedIndices()
@@ -721,8 +730,6 @@ public final class TradeRouteInputPanel extends FreeColPanel
      * @return True if the trade route is valid.
      */
     private boolean verifyNewTradeRoute() {
-        final Player player = getFreeColClient().getMyPlayer();
-
         // Update the trade route with the current settings
         this.newRoute.setName(tradeRouteName.getText());
         this.newRoute.clearStops();
@@ -733,7 +740,7 @@ public final class TradeRouteInputPanel extends FreeColPanel
 
         StringTemplate err = this.newRoute.verify();
         if (err != null) {
-            getGUI().showInformationMessage(err);
+            getGUI().showInformationPanel(err);
             this.newRoute.setName(null); // Mark as unacceptable
             return false;
         }
@@ -777,11 +784,10 @@ public final class TradeRouteInputPanel extends FreeColPanel
         if (e.getValueIsAdjusting()) return;
         int[] idx = this.stopList.getSelectedIndices();
         if (idx.length > 0) {
-            TradeRouteStop stop = this.stopListModel.get(idx[0]);
-            this.cargoPanel.initialize(stop);
-            this.goodsPanel.setEnabled(true);
+            updateCargoPanel(this.stopListModel.get(idx[0]));
+            this.allGoodsTypesPanel.setEnabled(true);
         } else {
-            this.goodsPanel.setEnabled(false);
+            this.allGoodsTypesPanel.setEnabled(false);
         }
         updateButtons();
     }
@@ -794,9 +800,8 @@ public final class TradeRouteInputPanel extends FreeColPanel
      */
     @Override
     public void removeNotify() {
-        this.cargoHandler = null;
+        this.transferHandler = null;
         this.dragListener = null;
-        this.dropListener = null;
         this.stopListModel.clear();
         this.stopListModel = null;
         this.stopList = null;
@@ -805,7 +810,8 @@ public final class TradeRouteInputPanel extends FreeColPanel
         this.messagesBox = null;
         this.addStopButton = null;
         this.removeStopButton = null;
-        this.cargoPanel = null;
+        this.stopGoodsTypesPanel = null;
+        this.allGoodsTypesPanel = null;
 
         super.removeNotify();
     }
