@@ -1,5 +1,5 @@
 /**
- *  Copyright (C) 2002-2015   The FreeCol Team
+ *  Copyright (C) 2002-2022   The FreeCol Team
  *
  *  This file is part of FreeCol.
  *
@@ -19,40 +19,47 @@
 
 package net.sf.freecol.common.i18n;
 
+import static net.sf.freecol.common.util.CollectionUtils.find;
+import static net.sf.freecol.common.util.CollectionUtils.map;
+import static net.sf.freecol.common.util.CollectionUtils.matchKeyEquals;
+import static net.sf.freecol.common.util.StringUtils.downCase;
+
 import java.io.BufferedReader;
 import java.io.File;
-import java.io.FileInputStream;
+import java.io.IOException;
 import java.io.InputStream;
 import java.io.InputStreamReader;
-import java.io.IOException;
-import java.io.UnsupportedEncodingException;
-import java.util.ArrayList;
+import java.nio.charset.StandardCharsets;
+import java.nio.file.Files;
+import java.util.AbstractMap.SimpleEntry;
 import java.util.HashMap;
-import java.util.Locale;
 import java.util.List;
+import java.util.Locale;
 import java.util.Map;
 import java.util.StringTokenizer;
+import java.util.function.Predicate;
+import java.util.logging.Level;
 import java.util.logging.Logger;
 
 import javax.swing.UIManager;
+import javax.xml.stream.XMLStreamException;
 
 import net.sf.freecol.common.ObjectWithId;
 import net.sf.freecol.common.io.FreeColDirectories;
-import net.sf.freecol.common.io.FreeColDataFile;
 import net.sf.freecol.common.io.FreeColModFile;
-import net.sf.freecol.common.io.Mods;
+import net.sf.freecol.common.io.FreeColRules;
 import net.sf.freecol.common.model.Named;
 import net.sf.freecol.common.model.Role;
 import net.sf.freecol.common.model.StringTemplate;
 import net.sf.freecol.common.model.StringTemplate.TemplateType;
-import static net.sf.freecol.common.util.CollectionUtils.*;
+import net.sf.freecol.common.util.LogBuilder;
 
 
 /**
  * Represents a collection of messages in a particular locale.
  *
  * The individual messages are read from property files in the
- * <code>data/strings</code> directory. The property files are called
+ * {@code data/strings} directory. The property files are called
  * "FreeColMessages[_LANGUAGE[_COUNTRY[_VARIANT]]].properties", where
  * LANGUAGE should be an ISO 639-2 or ISO 639-3 language code, COUNTRY
  * should be an ISO 3166-2 country code, and VARIANT is an arbitrary
@@ -89,10 +96,6 @@ public class Messages {
 
     private static final Logger logger = Logger.getLogger(Messages.class.getName());
 
-    public static final String MESSAGE_FILE_PREFIX = "FreeColMessages";
-    public static final String MOD_MESSAGE_FILE_PREFIX = "ModMessages";
-    public static final String MESSAGE_FILE_SUFFIX = ".properties";
-
     public static final String DESCRIPTION_SUFFIX = ".description";
     public static final String SHORT_DESCRIPTION_SUFFIX = ".shortDescription";
     public static final String NAME_SUFFIX = ".name";
@@ -124,68 +127,39 @@ public class Messages {
      * Gets the Selector with the given tag.
      *
      * @param tag The tag to check.
-     * @return A suitable <code>Selector</code>.
+     * @return A suitable {@code Selector}.
      */
     private static Selector getSelector(String tag) {
-        return tagMap.get(tag.toLowerCase(Locale.US));
+        return tagMap.get(downCase(tag));
     }
 
     /**
      * Set the grammatical number rule.
      *
-     * @param number a <code>Number</code> value
+     * @param number a {@code Number} value
      */
     public static void setGrammaticalNumber(Number number) {
         tagMap.put("plural", number);
     }
 
-    /**
-     * Get a list of candidate message file names for a given locale.
-     *
-     * @param locale The <code>Locale</code> to generate file names for.
-     * @return A list of message file names.
-     */
-    private static List<String> getMessageFileNames(Locale locale) {
-        return FreeColDataFile.getFileNames(MESSAGE_FILE_PREFIX,
-            MESSAGE_FILE_SUFFIX, locale);
-    }
-
-    /**
-     * Get a list of candidate mod message file names for a given locale.
-     *
-     * @param locale The <code>Locale</code> to generate file names for.
-     * @return A list of mod message file names.
-     */
-    private static List<String> getModMessageFileNames(Locale locale) {
-        return FreeColDataFile.getFileNames(MOD_MESSAGE_FILE_PREFIX,
-            MESSAGE_FILE_SUFFIX, locale);
-    }
-        
     /**
      * Load the message bundle for the given locale
      *
      * Error messages have to go to System.err as this routine is called
      * before logging is enabled.
      *
-     * @param locale The <code>Locale</code> to set resources for.
+     * @param locale The {@code Locale} to set resources for.
      */
     public static void loadMessageBundle(Locale locale) {
         messageBundle.clear(); // Reset the message bundle.
 
-        if (!Locale.getDefault().equals(locale)) {
-            Locale.setDefault(locale);
-        }
-
-        File i18nDirectory = FreeColDirectories.getI18nDirectory();
         if (!NumberRules.isInitialized()) {
             // attempt to read grammatical rules
-            File cldr = new File(i18nDirectory, "plurals.xml");
+            File cldr = FreeColDirectories.getI18nPluralsFile();
             if (cldr.exists()) {
-                try {
-                    try (FileInputStream in = new FileInputStream(cldr)) {
-                        NumberRules.load(in);
-                    }
-                } catch (IOException e) {
+                try (InputStream in = Files.newInputStream(cldr.toPath())) {
+                    NumberRules.load(in);
+                } catch (IOException|XMLStreamException e) {
                     System.err.println("Failed to read CLDR rules: "
                         + e.getMessage());
                 }
@@ -199,14 +173,34 @@ public class Messages {
             ? Locale.getDefault() : locale;
         setGrammaticalNumber(NumberRules.getNumberForLanguage(loc.getLanguage()));
 
-        for (String name : getMessageFileNames(locale)) {
-            File file = new File(i18nDirectory, name);
-            if (!file.exists()) continue; // Expected
-            try {
-                loadMessages(new FileInputStream(file));
-            } catch (IOException e) {
-                System.err.println("Failed to load messages from " + name
-                    + ": " + e.getMessage());
+        for (File f : FreeColDirectories.getI18nMessageFileList(locale)) {
+            if (!f.canRead()) continue;
+            try (InputStream in = Files.newInputStream(f.toPath())) {
+                loadMessages(in);
+            } catch (IOException ioe) {
+                System.err.println("Failed to load messages from "
+                    + f.getPath() + ": " + ioe.getMessage());
+            }
+        }
+
+        List<String> filenames
+            = FreeColDirectories.getMessageFileNameList(locale);
+        for (FreeColModFile fctf : FreeColRules.getRulesList()) {
+            for (String fn : filenames) {
+                InputStream is = null;
+                try {
+                    is = fctf.getInputStream(fn);
+                } catch (IOException ioe) {
+                    continue; // Expecting failure
+                }
+                try {
+                    loadMessages(is);
+                } catch (IOException ioe) {
+                    System.err.println("Failed to load rules messages from "
+                        + fn + ": " + ioe.getMessage());
+                } finally {
+                    try { is.close(); } catch (IOException x) {}
+                }
             }
         }
     }
@@ -216,16 +210,12 @@ public class Messages {
      *
      * Public for the test suite.
      *
-     * @param is The <code>InputStream</code> to read from.
-     * @throws IOException
+     * @param is The {@code InputStream} to read from.
+     * @throws IOException on failure to read from the stream.
      */
     public static void loadMessages(InputStream is) throws IOException {
-        InputStreamReader inputReader;
-        try {
-            inputReader = new InputStreamReader(is, "UTF-8");
-        } catch (UnsupportedEncodingException uee) {
-            return; // We have big problems if UTF-8 is not supported.
-        }
+        InputStreamReader inputReader
+            = new InputStreamReader(is, StandardCharsets.UTF_8);
         BufferedReader in = new BufferedReader(inputReader);
 
         String line = null;
@@ -256,21 +246,24 @@ public class Messages {
      * loaded.  So this routine is called separately when the mods are
      * finally loaded.
      *
-     * @param locale The <code>Locale</code> to load resources for.
+     * @param locale The {@code Locale} to load resources for.
      */
     public static void loadModMessageBundle(Locale locale) {
-        List<FreeColModFile> allMods = new ArrayList<>();
-        allMods.addAll(Mods.getAllMods());
-        allMods.addAll(Mods.getRuleSets());
-
-        List<String> filenames = getMessageFileNames(locale);
-        for (FreeColModFile fcmf : allMods) {
+        List<String> filenames
+            = FreeColDirectories.getMessageFileNameList(locale);
+        LogBuilder lb = new LogBuilder(32);
+        lb.add("Failed to load mod messages:");
+        lb.mark();
+        for (FreeColModFile fcmf : FreeColModFile.getModsList()) {
             for (String name : filenames) {
                 try {
                     loadMessages(fcmf.getInputStream(name));
-                } catch (IOException e) {} // Failures expected
+                } catch (IOException e) { // Failures expected
+                    lb.add(' ', fcmf.getId(), '/', name);
+                }
             }
         }
+        if (lb.grew()) lb.log(logger, Level.FINE);
     }
 
     /**
@@ -278,28 +271,36 @@ public class Messages {
      *
      * Called when the spec is updated with the selected mods.
      *
-     * @param mods The list of <code>FreeColModFile</code> for the active mods.
-     * @param locale The <code>Locale</code> to load resources for.
+     * @param mods The list of {@code FreeColModFile} for the active mods.
+     * @param locale The {@code Locale} to load resources for.
      */
     public static void loadActiveModMessageBundle(List<FreeColModFile> mods,
                                                   Locale locale) {
+        List<String> filenames
+            = FreeColDirectories.getModMessageFileNames(locale);
+        LogBuilder lb = new LogBuilder(32);
+        lb.add("Failed to load active mod messages:");
+        lb.mark();
         for (FreeColModFile fcmf : mods) {
-            for (String name : getModMessageFileNames(locale)) {
+            for (String name : filenames) {
                 try {
                     loadMessages(fcmf.getInputStream(name));
-                } catch (IOException e) {} // Failures expected
+                } catch (IOException e) { // Failures expected
+                    lb.add(' ', fcmf.getId(), '/', name);
+                }
             }
         }
+        if (lb.grew()) lb.log(logger, Level.FINE);
     }
     
     /**
-     * Get the <code>Locale</code> corresponding to a given language name.
+     * Get the {@code Locale} corresponding to a given language name.
      *
      * Public as this is needed for language option processing and the
      * initial locale setting.
      *
      * @param languageID An underscore separated language/country/variant tuple.
-     * @return The <code>Locale</code> for the specified language.
+     * @return The {@code Locale} for the specified language.
      */
     public static Locale getLocale(String languageID) {
         String language, country = "", variant = "";
@@ -387,7 +388,7 @@ public class Messages {
     /**
      * Does the message bundle contain the given key?
      *
-     * @param key The key <code>String</code> to check.
+     * @param key The key {@code String} to check.
      * @return True if there is a message present with the given key.
      */
     public static boolean containsKey(String key) {
@@ -399,9 +400,9 @@ public class Messages {
      * bundle and the default key otherwise. This should be used to
      * select the most specific message key available.
      *
-     * @param preferredKey a <code>String</code> value
-     * @param defaultKey a <code>String</code> value
-     * @return a <code>String</code> value
+     * @param preferredKey a {@code String} value
+     * @param defaultKey a {@code String} value
+     * @return a {@code String} value
      */
     public static String getKey(String preferredKey, String defaultKey) {
         if (containsKey(preferredKey)) {
@@ -415,9 +416,9 @@ public class Messages {
         return getBestDescription(object.getId());
     }
 
-    public static String getBestDescription(String id) {
+    private static String getBestDescription(String id) {
         String key = find(map(DESCRIPTION_KEYS, s -> id + s),
-            k -> containsKey(k), null);
+                          k -> containsKey(k));
         return (key == null) ? id : message(key);
     }
 
@@ -427,8 +428,8 @@ public class Messages {
      * Favour the .name form, but degrade gracefully if it is not present.
      * If .name is present, also look for a description.
      *
-     * @param named The <code>Named</code> to look up.
-     * @return A 2-element array of name and description found.
+     * @param named The {@code Named} to look up.
+     * @return A 2-member array of name and description found.
      */
     public static String[] getBestNameAndDescription(Named named) {
         return getBestNameAndDescription(named.getNameKey());
@@ -441,7 +442,7 @@ public class Messages {
      * If .name is present, also look for a description.
      *
      * @param id The identifier to look up.
-     * @return A 2-element array of name and description found.
+     * @return A 2-member array of name and description found.
      */
     public static String[] getBestNameAndDescription(String id) {
         if (id != null && id.endsWith(NAME_SUFFIX)) { // Temporary hack
@@ -459,6 +460,26 @@ public class Messages {
         return new String[] { name, desc };
     }
 
+    /**
+     * A predicate maker to match by message name.
+     *
+     * @param key The name of the message.
+     * @return A suitable {@code Predicate}.
+     */
+    public static final Predicate<String> matchesName(String key) {
+        return matchKeyEquals(key, (String k) -> Messages.getName(k));
+    }
+
+    /**
+     * A predicate maker to match named types.
+     *
+     * @param key The name of the type.
+     * @return A suitable {@code Predicate}.
+     */
+    public static final Predicate<Named> matchesNamed(String key) {
+        return matchKeyEquals(key, (Named k) -> Messages.getName(k));
+    }
+    
 
     // Special purpose unit labelling
 
@@ -472,7 +493,7 @@ public class Messages {
      * @param nationId An optional nation identifier.
      * @param roleId The unit role identifier.
      * @param extra An optional extra annotation.
-     * @return A <code>StringTemplate</code> to describe the given unit.
+     * @return A {@code StringTemplate} to describe the given unit.
      */
     public static StringTemplate getUnitLabel(String name, String typeId,
                                               int number, String nationId,
@@ -483,7 +504,7 @@ public class Messages {
         // the role twice, e.g. "Seasoned Scout Scout".
         StringTemplate type;
         String roleKey;
-        String baseKey = typeId + "." + Role.getRoleSuffix(roleId);
+        String baseKey = typeId + "." + Role.getRoleIdSuffix(roleId);
         if (containsKey(baseKey)) {
             type = StringTemplate.template(baseKey)
                 .addAmount("%number%", number);
@@ -697,7 +718,7 @@ public class Messages {
     /**
      * Localizes a StringTemplate.
      *
-     * @param template The <code>StringTemplate</code> to localize.
+     * @param template The {@code StringTemplate} to localize.
      * @return The localized string.
      */
     public static String message(StringTemplate template) {
@@ -705,15 +726,16 @@ public class Messages {
         String result = "";
         switch (template.getTemplateType()) {
         case LABEL:
-            List<StringTemplate> replacements = template.getReplacements();
-            if (replacements.isEmpty()) {
+            if (template.isEmpty()) {
                 result = message(template.getId());
             } else {
-                for (StringTemplate other : replacements) {
-                    result += template.getId() + message(other);
+                StringBuilder sb = new StringBuilder(64);
+                for (SimpleEntry<String,StringTemplate> e
+                         : template.entryList()) {
+                    sb.append(template.getId()).append(message(e.getValue()));
                 }
-                if (result.length() >= template.getId().length()) {
-                    result = result.substring(template.getId().length());
+                if (sb.length() >= template.getId().length()) {
+                    result = sb.toString().substring(template.getId().length());
                 } else {
                     logger.warning("incorrect use of template " + template);
                 }
@@ -726,9 +748,11 @@ public class Messages {
                 result = messageBundle.get(template.getDefaultId());
             }
             result = replaceChoices(result, template);
-            for (String key : template.getKeys()) {
-                result = result.replace(key,
-                                        message(template.getReplacement(key)));
+            for (SimpleEntry<String, StringTemplate> e
+                     : template.entryList()) {
+                if (e.getKey() != null) {
+                    result = result.replace(e.getKey(), message(e.getValue()));
+                }
             }
             break;
         case KEY:
@@ -757,9 +781,9 @@ public class Messages {
      * separated by an assignment character. Example:
      * "{{tag:selector|key1=val1|key2=val2}}".
      *
-     * @param input a <code>String</code> value
-     * @param template a <code>StringTemplate</code> value
-     * @return a <code>String</code> value
+     * @param input a {@code String} value
+     * @param template a {@code StringTemplate} value
+     * @return a {@code String} value
      */
     private static String replaceChoices(String input, StringTemplate template) {
         int openChoice = 0;
@@ -793,9 +817,11 @@ public class Messages {
                 if (template == null) {
                     selector = "default";
                 } else {
-                    StringTemplate replacement = template.getReplacement(selector);
+                    StringTemplate replacement
+                        = template.getReplacement(selector);
                     if (replacement == null) {
-                        logger.warning("Failed to find replacement for " + selector);
+                        logger.warning("Failed to find replacement for "
+                            + selector);
                         continue;
                     } else {
                         selector = message(replacement);
@@ -815,11 +841,13 @@ public class Messages {
             if (keyIndex < 0 || keyIndex > closeChoice) {
                 // key not found, choice might be a key itself
                 String otherKey = input.substring(pipeIndex + 1, closeChoice);
-                if (otherKey.startsWith("%") && otherKey.endsWith("%")
-                    && template != null) {
-                    StringTemplate replacement = template.getReplacement(otherKey);
+                if (template != null
+                    && otherKey.startsWith("%") && otherKey.endsWith("%")) {
+                    StringTemplate replacement
+                        = template.getReplacement(otherKey);
                     if (replacement == null) {
-                        logger.warning("Failed to find replacement for " + otherKey);
+                        logger.warning("Failed to find replacement for "
+                            + selector + "/" + otherKey);
                         continue;
                     } else if (replacement.getTemplateType() == TemplateType.KEY) {
                         otherKey = messageBundle.get(replacement.getId());
@@ -830,25 +858,29 @@ public class Messages {
                         } else {
                             keyIndex = otherKey.indexOf(selector, keyIndex);
                             if (keyIndex < 0) {
-                                logger.warning("Failed to find key " + selector + " in replacement "
-                                               + replacement.getId());
+                                logger.warning("Failed to find key "
+                                    + selector + "/" + otherKey
+                                    + " in replacement " + replacement);
                                 continue;
                             } else {
                                 result.append(getChoice(otherKey, selector));
                             }
                         }
                     } else {
-                        logger.warning("Choice substitution attempted, but template type was "
-                                       + replacement.getTemplateType());
+                        logger.warning("Choice substitution for "
+                            + selector + "/" + otherKey
+                            + " attempted, but template was " + replacement
+                            + " for input " + input);
                         continue;
                     }
                 } else if (containsKey(otherKey)) {
                     otherKey = getChoice(messageBundle.get(otherKey), selector);
                     result.append(otherKey);
                 } else {
-                    logger.warning("Unknown key or untagged choice: '" + otherKey
-                                   + "', selector was '" + selector
-                                   + "', trying 'default' instead");
+                    logger.warning("Unknown key or untagged choice: '"
+                        + otherKey
+                        + "', selector was '" + selector
+                        + "', trying 'default' instead");
                     int defaultStart = otherKey.indexOf("default=");
                     if (defaultStart >= 0) {
                         defaultStart += 8;
@@ -892,9 +924,9 @@ public class Messages {
      * Return the choice tagged with the given key, or null, if the
      * given input string does not contain the key.
      *
-     * @param input a <code>String</code> value
-     * @param key a <code>String</code> value
-     * @return a <code>String</code> value
+     * @param input a {@code String} value
+     * @param key a {@code String} value
+     * @return a {@code String} value
      */
     private static String getChoice(String input, String key) {
         int keyIndex = input.indexOf(key);
@@ -919,15 +951,15 @@ public class Messages {
      * Return the index of the matching pair of brackets, or -1 if
      * none is found.
      *
-     * @param input a <code>String</code> value
-     * @param start an <code>int</code> value
-     * @return an <code>int</code> value
+     * @param input a {@code String} value
+     * @param start an {@code int} value
+     * @return an {@code int} value
      */
     private static int findMatchingBracket(String input, int start) {
         char last = 0;
         int level = 0;
         for (int index = start; index < input.length(); index++) {
-            switch(input.charAt(index)) {
+            switch (input.charAt(index)) {
             case '{':
                 if (last == '{') {
                     last = 0;
