@@ -1,5 +1,5 @@
 /**
- *  Copyright (C) 2002-2015   The FreeCol Team
+ *  Copyright (C) 2002-2022   The FreeCol Team
  *
  *  This file is part of FreeCol.
  *
@@ -21,13 +21,17 @@ package net.sf.freecol.common.model;
 
 import java.util.ArrayList;
 import java.util.List;
-import java.util.stream.Collectors;
 
+import javax.swing.JList;
+import javax.swing.ListModel;
 import javax.xml.stream.XMLStreamException;
 
 import net.sf.freecol.common.io.FreeColXMLReader;
 import net.sf.freecol.common.io.FreeColXMLWriter;
+import net.sf.freecol.common.model.Colony.NoBuildReason;
+import static net.sf.freecol.common.model.Constants.*;
 import net.sf.freecol.common.model.UnitLocation.NoAddReason;
+import static net.sf.freecol.common.util.CollectionUtils.*;
 
 
 /**
@@ -35,7 +39,10 @@ import net.sf.freecol.common.model.UnitLocation.NoAddReason;
  * {@link Building}, such as the number of workplaces, and the types
  * of goods it produces and consumes.
  */
-public final class BuildingType extends BuildableType {
+public final class BuildingType extends BuildableType
+        implements BaseProduction {
+
+    public static final String TAG = "building-type";
 
     /** The level of building. */
     private int level = 1;
@@ -54,6 +61,17 @@ public final class BuildingType extends BuildableType {
     /** Consumption order. */
     private int priority = Consumer.BUILDING_PRIORITY;
 
+    /** Maximum production from the "experts have connections" option. */
+    private int expertConnectionProduction = 0;
+
+    /**
+     * A multiplier for any unit goods-specific bonus for this building type.
+     */
+    private float competenceFactor = 1.0f;
+
+    /** The multiplier for the colony rebel bonus for this building type. */
+    private float rebelFactor = 1.0f;
+
     /** The building type this upgrades from. */
     private BuildingType upgradesFrom = null;
 
@@ -65,10 +83,10 @@ public final class BuildingType extends BuildableType {
 
 
     /**
-     * Creates a new <code>BuildingType</code> instance.
+     * Creates a new {@code BuildingType} instance.
      *
      * @param id The object identifier.
-     * @param specification The <code>Specification</code> to refer to.
+     * @param specification The {@code Specification} to refer to.
      */
     public BuildingType(String id, Specification specification) {
         super(id, specification);
@@ -94,6 +112,13 @@ public final class BuildingType extends BuildableType {
         return workPlaces;
     }
 
+    public int getMinimumSkill() {
+        return this.minSkill;
+    }
+    public int getMaximumSkill() {
+        return this.maxSkill;
+    }
+
     /**
      * Gets the amount of gold necessary to maintain a Building of
      * this type for one turn.
@@ -104,6 +129,35 @@ public final class BuildingType extends BuildableType {
         return upkeep;
     }
 
+    /**
+     * Get the maximum production for the Experts-With-Connections option.
+     *
+     * @return The production amount.
+     */
+    public int getExpertConnectionProduction() {
+        return this.expertConnectionProduction;
+    }
+
+    /**
+     * Get a work location specific factor to multiply any unit goods specific
+     * bonuses by.
+     *
+     * @return The competence factor.
+     */
+    public float getCompetenceFactor() {
+        return this.competenceFactor;
+    }
+
+    /**
+     * Get a work location specific factor to multiply the colony
+     * rebel bonus by.
+     *
+     * @return The rebel factor.
+     */
+    public float getRebelFactor() {
+        return this.rebelFactor;
+    }
+
     /**
      * The consumption priority of a Building of this type. The higher
      * the priority, the earlier will the Consumer be allowed to
@@ -119,21 +173,21 @@ public final class BuildingType extends BuildableType {
      * Gets the reason why a given unit type can not be added to a
      * building of this type.
      *
-     * @param unitType The <code>UnitType</code> to test.
+     * @param unitType The {@code UnitType} to test.
      * @return The reason why adding would fail.
      */
     public NoAddReason getNoAddReason(UnitType unitType) {
         return (workPlaces == 0) ? NoAddReason.CAPACITY_EXCEEDED
-            : (!unitType.hasSkill()) ? NoAddReason.MISSING_SKILL
-            : (unitType.getSkill() < minSkill) ? NoAddReason.MINIMUM_SKILL
-            : (unitType.getSkill() > maxSkill) ? NoAddReason.MAXIMUM_SKILL
-            : NoAddReason.NONE;
+                : (!unitType.hasSkill()) ? NoAddReason.MISSING_SKILL
+                : (unitType.getSkill() < minSkill) ? NoAddReason.MINIMUM_SKILL
+                : (unitType.getSkill() > maxSkill) ? NoAddReason.MAXIMUM_SKILL
+                : NoAddReason.NONE;
     }
 
     /**
      * Can a unit of a given type be added to a Building of this type?
      *
-     * @param unitType The <code>UnitType</code> to check.
+     * @param unitType The {@code UnitType} to check.
      * @return True if the unit type can be added.
      */
     public boolean canAdd(UnitType unitType) {
@@ -146,14 +200,14 @@ public final class BuildingType extends BuildableType {
      *
      * @return This.
      */
-    public FreeColGameObjectType getType() {
+    public FreeColSpecObjectType getType() {
         return this;
     }
 
     /**
      * Gets the BuildingType this BuildingType upgrades from.
      *
-     * @return The <code>BuildingType</code> that upgrades to this one.
+     * @return The {@code BuildingType} that upgrades to this one.
      */
     public BuildingType getUpgradesFrom() {
         return upgradesFrom;
@@ -162,7 +216,7 @@ public final class BuildingType extends BuildableType {
     /**
      * Get the BuildingType this BuildingType upgrades to.
      *
-     * @return The <code>BuildingType</code> to upgrade to from this one.
+     * @return The {@code BuildingType} to upgrade to from this one.
      */
     public BuildingType getUpgradesTo() {
         return upgradesTo;
@@ -171,7 +225,7 @@ public final class BuildingType extends BuildableType {
     /**
      * Gets the first level of this BuildingType.
      *
-     * @return The base <code>BuildingType</code>.
+     * @return The base {@code BuildingType}.
      */
     public BuildingType getFirstLevel() {
         BuildingType buildingType = this;
@@ -190,10 +244,29 @@ public final class BuildingType extends BuildableType {
         return !needsGoodsToBuild() && getUpgradesFrom() == null;
     }
 
+    /**
+     * Get the production type list.
+     *
+     * @return The {@code ProductionType} list.
+     */
+    protected List<ProductionType> getProductionTypes() {
+        return this.productionTypes;
+    }
+    
+    /**
+     * Set the production type list.
+     *
+     * @param productionTypes The new {@code ProductionType} list.
+     */
+    protected void setProductionTypes(List<ProductionType> productionTypes) {
+        this.productionTypes.clear();
+        this.productionTypes.addAll(productionTypes);
+    }
+
     /**
      * Add a production type to this building type.
      *
-     * @param productionType The <code>ProductionType</code> to add.
+     * @param productionType The {@code ProductionType} to add.
      */
     public void addProductionType(ProductionType productionType) {
         if (productionType != null) productionTypes.add(productionType);
@@ -204,7 +277,7 @@ public final class BuildingType extends BuildableType {
      * current difficulty level.
      *
      * @param unattended Whether the production is unattended.
-     * @return A list of <code>ProductionType</code>s.
+     * @return A list of {@code ProductionType}s.
      */
     public List<ProductionType> getAvailableProductionTypes(boolean unattended) {
         return getAvailableProductionTypes(unattended, null);
@@ -220,38 +293,24 @@ public final class BuildingType extends BuildableType {
      *
      * @param unattended Whether the production is unattended.
      * @param level The production level (NYI).
-     * @return A list of <code>ProductionType</code>s.
+     * @return A list of {@code ProductionType}s.
      */
     public List<ProductionType> getAvailableProductionTypes(boolean unattended,
                                                             String level) {
-        return productionTypes.stream()
-            .filter(pt -> pt.getUnattended() == unattended
-                && pt.appliesTo(level))
-            .collect(Collectors.toList());
+        return transform(productionTypes,
+                pt -> pt.getUnattended() == unattended
+                        && pt.appliesTo(level));
     }
 
-    // @compat 0.10.6
-    /**
-     * Get the type of goods consumed by this BuildingType.
-     *
-     * @return The consumed <code>GoodsType</code>.
-     */
-    private GoodsType getConsumedGoodsType() {
-        if (productionTypes.isEmpty()) return null;
-        List<AbstractGoods> inputs = productionTypes.get(0).getInputs();
-        return (inputs.isEmpty()) ? null : inputs.get(0).getType();
-    }
-    // end @compat
-
     /**
      * Gets the type of goods produced by this BuildingType.
      *
-     * @return The produced <code>GoodsType</code>.
+     * @return The produced {@code GoodsType}.
      */
     public GoodsType getProducedGoodsType() {
         if (productionTypes.isEmpty()) return null;
-        List<AbstractGoods> outputs = productionTypes.get(0).getOutputs();
-        return (outputs.isEmpty()) ? null : outputs.get(0).getType();
+        AbstractGoods ag = first(first(productionTypes).getOutputs());
+        return (ag == null) ? null : ag.getType();
     }
 
     /**
@@ -267,46 +326,23 @@ public final class BuildingType extends BuildableType {
     /**
      * Can a tile of this type produce a given goods type?
      *
-     * @param goodsType The <code>GoodsType</code> to produce.
-     * @param unitType An optional <code>UnitType</code> that is to do
+     * @param goodsType The {@code GoodsType} to produce.
+     * @param unitType An optional {@code UnitType} that is to do
      *     the work, if null the unattended production is considered.
      * @return True if this tile type produces the goods.
      */
     public boolean canProduce(GoodsType goodsType, UnitType unitType) {
         return goodsType != null
-            && ProductionType.canProduce(goodsType,
+                && ProductionType.canProduce(goodsType,
                 getAvailableProductionTypes(unitType == null));
     }
 
     /**
-     * Get the base production of a given goods type for an optional
-     * unit type.
-     *
-     * @param productionType An optional <code>ProductionType</code> to use,
-     *     if null the best available one is used.
-     * @param goodsType The <code>GoodsType</code> to produce.
-     * @param unitType An optional <code>UnitType</code> that is to do
-     *     the work, if null the unattended production is considered.
-     * @return The amount of goods produced.
-     */
-    public int getBaseProduction(ProductionType productionType,
-                                 GoodsType goodsType, UnitType unitType) {
-        if (goodsType == null) return 0;
-        if (productionType == null) {
-            productionType = ProductionType.getBestProductionType(goodsType,
-                getAvailableProductionTypes(unitType == null));
-        }
-        if (productionType == null) return 0;
-        AbstractGoods best = productionType.getOutput(goodsType);
-        return (best == null) ? 0 : best.getAmount();
-    }
-
-    /** 
      * Get the amount of goods of a given goods type the given unit
      * type could produce on a tile of this tile type.
      *
-     * @param goodsType The <code>GoodsType</code> to produce.
-     * @param unitType An optional <code>UnitType</code> that is to do
+     * @param goodsType The {@code GoodsType} to produce.
+     * @param unitType An optional {@code UnitType} that is to do
      *     the work, if null the unattended production is considered.
      * @return The amount of goods produced.
      */
@@ -314,11 +350,103 @@ public final class BuildingType extends BuildableType {
                                       UnitType unitType) {
         if (goodsType == null) return 0;
         int amount = getBaseProduction(null, goodsType, unitType);
-        amount = (int)applyModifiers(amount, null, goodsType.getId(),
-                                     unitType);
+        amount = (int)apply(amount, null, goodsType.getId(), unitType);
         return (amount < 0) ? 0 : amount;
     }
 
+    /**
+     * {@inheritDoc}
+     */
+    public NoBuildReason canBeBuiltInColony(Colony colony,
+        List<BuildableType> assumeBuilt) {
+        Building colonyBuilding = colony.getBuilding(this);
+        if (colonyBuilding == null) {
+            // the colony has no similar building yet
+            BuildingType from = this.getUpgradesFrom();
+            if (from != null && !assumeBuilt.contains(from)) {
+                // we are trying to build an advanced factory, we
+                // should build lower level shop first
+                return Colony.NoBuildReason.WRONG_UPGRADE;
+            }
+        } else {
+            // a building of the same family already exists
+            BuildingType from = colonyBuilding.getType().getUpgradesTo();
+            if (from != this && !assumeBuilt.contains(from)) {
+                // the existing building's next upgrade is not the
+                // new one we want to build
+                return Colony.NoBuildReason.WRONG_UPGRADE;
+            }
+        }
+        return Colony.NoBuildReason.NONE;
+    }
+
+    @Override
+    public int getMinimumIndex(Colony colony, JList<BuildableType> buildQueueList, int UNABLE_TO_BUILD) {
+        ListModel<BuildableType> buildQueue = buildQueueList.getModel();
+        BuildingType upgradesFrom = this.getUpgradesFrom();
+        if (upgradesFrom == null) return 0;
+        Building building = colony.getBuilding(this);
+        BuildingType buildingType = (building == null) ? null
+                : building.getType();
+        if (buildingType == upgradesFrom) return 0;
+        for (int index = 0; index < buildQueue.getSize(); index++) {
+            if (upgradesFrom.equals(buildQueue.getElementAt(index))) {
+                return index + 1;
+            }
+        }
+        return UNABLE_TO_BUILD;
+    }
+
+    @Override
+    public int getMaximumIndex(Colony colony, JList<BuildableType> buildQueueList, int UNABLE_TO_BUILD) {
+        ListModel<BuildableType> buildQueue = buildQueueList.getModel();
+        final int buildQueueLastPos = buildQueue.getSize();
+
+        boolean canBuild = false;
+        if (colony.canBuild(this)) {
+            canBuild = true;
+        }
+
+        BuildingType upgradesFrom = this.getUpgradesFrom();
+        BuildingType upgradesTo = this.getUpgradesTo();
+        // does not depend on nothing, but still cannot be built
+        if (!canBuild && upgradesFrom == null) {
+            return UNABLE_TO_BUILD;
+        }
+
+        // if can be built and does not have any upgrade,
+        // then it can be built at any time
+        if (canBuild && upgradesTo == null) {
+            return buildQueueLastPos;
+        }
+
+        // if can be built, does not depend on anything, mark
+        // upgrades from as found
+        boolean foundUpgradesFrom = canBuild;
+        for (int index = 0; index < buildQueue.getSize(); index++) {
+            BuildableType toBuild = buildQueue.getElementAt(index);
+
+            if (toBuild == this) continue;
+
+            if (!canBuild && !foundUpgradesFrom
+                    && upgradesFrom.equals(toBuild)) {
+                foundUpgradesFrom = true;
+                // nothing else to upgrade this building to
+                if (upgradesTo == null) return buildQueueLastPos;
+            }
+            // found a building it upgrades to, cannot go to or
+            // beyond this position
+            if (foundUpgradesFrom && upgradesTo != null
+                    && upgradesTo.equals(toBuild)) return index;
+
+            // Don't go past a unit this building can build.
+            if (this.hasAbility(Ability.BUILD, toBuild)) {
+                return index;
+            }
+        }
+        return buildQueueLastPos;
+    }
+
 
     // Override FreeColObject
 
@@ -338,21 +466,42 @@ public final class BuildingType extends BuildableType {
         return cmp;
     }
 
+    /**
+     * {@inheritDoc}
+     */
+    @Override
+    public <T extends FreeColObject> boolean copyIn(T other) {
+        BuildingType o = copyInCast(other, BuildingType.class);
+        if (o == null || !super.copyIn(o)) return false;
+        this.level = o.getLevel();
+        this.workPlaces = o.getWorkPlaces();
+        this.minSkill = o.getMinimumSkill();
+        this.maxSkill = o.getMaximumSkill();
+        this.upkeep = o.getUpkeep();
+        this.priority = o.getPriority();
+        this.expertConnectionProduction = o.getExpertConnectionProduction();
+        this.competenceFactor = o.getCompetenceFactor();
+        this.rebelFactor = o.getRebelFactor();
+        this.upgradesFrom = o.getUpgradesFrom();
+        this.upgradesTo = o.getUpgradesTo();
+        this.setProductionTypes(o.getProductionTypes());        
+        return true;
+    }
+
 
     // Serialization
 
+    private static final String COMPETENCE_FACTOR_TAG = "competence-factor";
+    private static final String EXPERTS_WITH_CONNECTION_PRODUCTION_TAG
+            = "experts-with-connections-production";
     private static final String MAXIMUM_SKILL_TAG = "maximum-skill";
     private static final String MINIMUM_SKILL_TAG = "minimum-skill";
     private static final String PRIORITY_TAG = "priority";
     private static final String PRODUCTION_TAG = "production";
+    private static final String REBEL_FACTOR_TAG = "rebel-factor";
     private static final String UPGRADES_FROM_TAG = "upgrades-from";
     private static final String UPKEEP_TAG = "upkeep";
     private static final String WORKPLACES_TAG = "workplaces";
-    // @compat 0.10.6
-    private static final String BASIC_PRODUCTION_TAG = "basicProduction";
-    private static final String CONSUMES_TAG = "consumes";
-    private static final String PRODUCES_TAG = "produces";
-    // end @compat 0.10.6
     // @compat 0.11.3
     private static final String OLD_MAX_SKILL_TAG = "maxSkill";
     private static final String OLD_MIN_SKILL_TAG = "minSkill";
@@ -364,7 +513,7 @@ public final class BuildingType extends BuildableType {
      * {@inheritDoc}
      */
     @Override
-        protected void writeAttributes(FreeColXMLWriter xw) throws XMLStreamException {
+    protected void writeAttributes(FreeColXMLWriter xw) throws XMLStreamException {
         super.writeAttributes(xw);
 
         if (upgradesFrom != null) {
@@ -389,6 +538,12 @@ public final class BuildingType extends BuildableType {
             xw.writeAttribute(PRIORITY_TAG, priority);
         }
 
+        xw.writeAttribute(EXPERTS_WITH_CONNECTION_PRODUCTION_TAG,
+                this.expertConnectionProduction);
+
+        xw.writeAttribute(COMPETENCE_FACTOR_TAG, this.competenceFactor);
+
+        xw.writeAttribute(REBEL_FACTOR_TAG, this.rebelFactor);
     }
 
     /**
@@ -414,16 +569,16 @@ public final class BuildingType extends BuildableType {
         final Specification spec = getSpecification();
 
         BuildingType parent = xr.getType(spec, EXTENDS_TAG,
-            BuildingType.class, this);
+                BuildingType.class, this);
 
         // @compat 0.11.3
         if (xr.hasAttribute(OLD_UPGRADES_FROM_TAG)) {
             upgradesFrom = xr.getType(spec, OLD_UPGRADES_FROM_TAG,
-                BuildingType.class, (BuildingType)null);
+                    BuildingType.class, (BuildingType)null);
         } else
         // end @compat 0.11.3
             upgradesFrom = xr.getType(spec, UPGRADES_FROM_TAG,
-                BuildingType.class, (BuildingType)null);
+                                      BuildingType.class, (BuildingType)null);
         if (upgradesFrom == null) {
             level = 1;
         } else {
@@ -451,17 +606,15 @@ public final class BuildingType extends BuildableType {
 
         priority = xr.getAttribute(PRIORITY_TAG, parent.priority);
 
-        // @compat 0.10.6
-        int basicProduction = xr.getAttribute(BASIC_PRODUCTION_TAG, -1);
-        if (basicProduction > 0) {
-            GoodsType consumes = xr.getType(spec, CONSUMES_TAG, GoodsType.class,
-                parent.getConsumedGoodsType());
-            GoodsType produces = xr.getType(spec, PRODUCES_TAG, GoodsType.class,
-                parent.getProducedGoodsType());
-            productionTypes.add(new ProductionType(consumes, produces,
-                    basicProduction));
-        }
-        // end @compat
+        this.expertConnectionProduction
+                = xr.getAttribute(EXPERTS_WITH_CONNECTION_PRODUCTION_TAG,
+                parent.expertConnectionProduction);
+
+        this.competenceFactor = xr.getAttribute(COMPETENCE_FACTOR_TAG,
+                parent.competenceFactor);
+
+        this.rebelFactor = xr.getAttribute(REBEL_FACTOR_TAG,
+                parent.rebelFactor);
 
         if (parent != this) { // Handle "extends" for super-type fields
             if (!xr.hasAttribute(REQUIRED_POPULATION_TAG)) {
@@ -495,20 +648,17 @@ public final class BuildingType extends BuildableType {
         } else {
             super.readChild(xr);
         }
+
+        // @compat 0.11.6
+        if (this.expertConnectionProduction == 0
+            && hasAbility(Ability.EXPERTS_USE_CONNECTIONS)) {
+            this.expertConnectionProduction = 4;
+        }
+        // end @compat 0.11.6
     }
 
     /**
      * {@inheritDoc}
      */
-    @Override
-    public String getXMLTagName() { return getXMLElementTagName(); }
-
-    /**
-     * Gets the tag name of the root element representing this object.
-     *
-     * @return "building-type".
-     */
-    public static String getXMLElementTagName() {
-        return "building-type";
-    }
+    public String getXMLTagName() { return TAG; }
 }
