@@ -1,5 +1,5 @@
 /**
- *  Copyright (C) 2002-2015   The FreeCol Team
+ *  Copyright (C) 2002-2022   The FreeCol Team
  *
  *  This file is part of FreeCol.
  *
@@ -19,16 +19,18 @@
 
 package net.sf.freecol.common.model;
 
-import java.util.ArrayList;
 import java.util.Collections;
 import java.util.List;
 import java.util.logging.Logger;
+import java.util.stream.Stream;
 
 import javax.xml.stream.XMLStreamException;
 
 import net.sf.freecol.common.io.FreeColXMLReader;
 import net.sf.freecol.common.io.FreeColXMLWriter;
 import net.sf.freecol.common.model.Player.NoClaimReason;
+import net.sf.freecol.common.option.GameOptions;
+import static net.sf.freecol.common.util.CollectionUtils.*;
 
 
 /**
@@ -43,11 +45,14 @@ public class ColonyTile extends WorkLocation {
     @SuppressWarnings("unused")
     private static final Logger logger = Logger.getLogger(ColonyTile.class.getName());
 
+    public static final String TAG = "colonyTile";
+
     public static final String UNIT_CHANGE = "UNIT_CHANGE";
 
     /** The maximum number of units a ColonyTile can hold. */
     public static final int UNIT_CAPACITY = 1;
 
+
     /**
      * The tile to work.  This is accessed through getWorkTile().
      * Beware!  Do not confuse this with getTile(), which returns
@@ -60,10 +65,10 @@ public class ColonyTile extends WorkLocation {
     /**
      * Constructor for ServerColonyTile.
      *
-     * @param game The enclosing <code>Game</code>.
-     * @param colony The <code>Colony</code> this object belongs to.
-     * @param workTile The tile in which this <code>ColonyTile</code>
-     *                 represents a <code>WorkLocation</code> for.
+     * @param game The enclosing {@code Game}.
+     * @param colony The {@code Colony} this object belongs to.
+     * @param workTile The tile in which this {@code ColonyTile}
+     *                 represents a {@code WorkLocation} for.
      */
     protected ColonyTile(Game game, Colony colony, Tile workTile) {
         super(game);
@@ -74,11 +79,11 @@ public class ColonyTile extends WorkLocation {
     }
 
     /**
-     * Create a new <code>ColonyTile</code> with the given identifier.
+     * Create a new {@code ColonyTile} with the given identifier.
      * The object should later be initialized by calling either
      * {@link #readFromXML(FreeColXMLReader)}.
      *
-     * @param game The enclosing <code>Game</code>.
+     * @param game The enclosing {@code Game}.
      * @param id The object identifier.
      */
     public ColonyTile(Game game, String id) {
@@ -87,20 +92,20 @@ public class ColonyTile extends WorkLocation {
 
 
     /**
-     * Is this the tile where the <code>Colony</code> is located?
+     * Is this the tile where the {@code Colony} is located?
      *
      * @return True if this is the colony center tile.
      */
     public boolean isColonyCenterTile() {
-        return getWorkTile() == getTile();
+        return this.workTile == getTile();
     }
 
     /**
      * Gets the work tile, that is the actual tile being worked.
      *
-     * @return The <code>Tile</code> in which this
-     *     <code>ColonyTile</code> represents a
-     *     <code>WorkLocation</code> for.
+     * @return The {@code Tile} in which this
+     *     {@code ColonyTile} represents a
+     *     {@code WorkLocation} for.
      */
     public Tile getWorkTile() {
         return workTile;
@@ -110,7 +115,7 @@ public class ColonyTile extends WorkLocation {
      * Sets the work tile.  Needed to fix copied colonies.  Do not use
      * otherwise!
      *
-     * @param workTile The new work <code>Tile</code>.
+     * @param workTile The new work {@code Tile}.
      */
     public void setWorkTile(Tile workTile) {
         this.workTile = workTile;
@@ -119,7 +124,7 @@ public class ColonyTile extends WorkLocation {
     /**
      * Gets a unit who is occupying the tile.
      *
-     * @return A <code>Unit</code> who is occupying the work tile, if any.
+     * @return A {@code Unit} who is occupying the work tile, if any.
      * @see #isOccupied()
      */
     public Unit getOccupyingUnit() {
@@ -155,30 +160,27 @@ public class ColonyTile extends WorkLocation {
         final Colony colony = getColony();
         ProductionInfo pi = new ProductionInfo();
         if (isColonyCenterTile()) {
-            for (AbstractGoods output : getOutputs()) {
-                boolean onlyNaturalImprovements = getSpecification()
-                    .getBoolean(GameOptions.ONLY_NATURAL_IMPROVEMENTS)
-                    && !output.getType().isFoodType();
-                int potential = output.getAmount();
-                if (workTile.getTileItemContainer() != null) {
-                    potential = workTile.getTileItemContainer()
-                        .getTotalBonusPotential(output.getType(), null, potential,
-                                                onlyNaturalImprovements);
-                }
-                potential += Math.max(0, colony.getProductionBonus());
-                AbstractGoods production
-                    = new AbstractGoods(output.getType(), potential);
-                pi.addProduction(production);
-            }
+            forEach(getOutputs(), output -> {
+                    boolean onlyNaturalImprovements = getSpecification()
+                        .getBoolean(GameOptions.ONLY_NATURAL_IMPROVEMENTS)
+                        && !output.getType().isFoodType();
+                    int potential = output.getAmount();
+                    if (workTile.getTileItemContainer() != null) {
+                        potential = workTile.getTileItemContainer()
+                            .getTotalBonusPotential(output.getType(), null,
+                                potential, onlyNaturalImprovements);
+                    }
+                    potential += Math.max(0, colony.getProductionBonus());
+                    AbstractGoods production
+                        = new AbstractGoods(output.getType(), potential);
+                    pi.addProduction(production);
+                });
         } else {
-            for (AbstractGoods output : getOutputs()) {
-                final GoodsType goodsType = output.getType();
-                int amount = getUnitList().stream()
-                    .mapToInt(u -> getUnitProduction(u, goodsType)).sum();
-                if (amount > 0) {
-                    pi.addProduction(new AbstractGoods(goodsType, amount));
-                }
-            }
+            forEach(map(getOutputs(), AbstractGoods::getType),
+                gt -> {
+                    int n = sum(getUnits(), u -> getUnitProduction(u, gt));
+                    if (n > 0) pi.addProduction(new AbstractGoods(gt, n));
+                });
         }
         return pi;
     }
@@ -186,7 +188,7 @@ public class ColonyTile extends WorkLocation {
     /**
      * Would a given tile improvement be beneficial to this colony tile?
      *
-     * @param ti The <code>TileImprovementType</code> to assess.
+     * @param ti The {@code TileImprovementType} to assess.
      * @return A measure of improvement (negative is a bad idea).
      */
     public int improvedBy(TileImprovementType ti) {
@@ -210,16 +212,15 @@ public class ColonyTile extends WorkLocation {
         if (productionType.getUnattended()) {
             if (newType == null) {
                 // Tile type stays the same, return the sum of any food bonues.
-                return getSpecification().getFoodGoodsTypeList().stream()
-                    .mapToInt(gt -> ti.getBonus(gt)).sum();
+                return sum(getSpecification().getFoodGoodsTypeList(),
+                           gt -> ti.getBonus(gt));
             }
 
             // Tile type change.
             final List<AbstractGoods> newProd
-                = newType.getPossibleProduction(true);
-            int food = newProd.stream()
-                .filter(ag -> ag.getType().isFoodType())
-                .mapToInt(AbstractGoods::getAmount).sum();
+                = toList(newType.getPossibleProduction(true));
+            int food = sum(newProd, AbstractGoods::isFoodType,
+                           AbstractGoods::getAmount);
             // Get the current food production.  Otherwise for goods
             // that are being passively produced and consumed, check
             // if production remains in surplus following a negative change.
@@ -229,7 +230,8 @@ public class ColonyTile extends WorkLocation {
                     food -= ag.getAmount();
                 } else if (colony.isConsuming(goodsType)) {
                     int change = -ag.getAmount()
-                        + AbstractGoods.getCount(goodsType, newProd);
+                        + sum(newProd, AbstractGoods.matches(goodsType),
+                              AbstractGoods::getAmount);
                     if (change < 0
                         && change + colony.getNetProductionOf(goodsType) < 0) {
                         // The change drives the net production (more?)
@@ -241,9 +243,13 @@ public class ColonyTile extends WorkLocation {
             return food;
         }
 
-        // Units are present, see what the change would do to their work.
+        // Units present?
+        final Unit unit = getFirstUnit();
+        if (unit == null) return 0;
+
+        // See what the change would do to their work.
         final GoodsType work = getCurrentWorkType();
-        final UnitType unitType = getFirstUnit().getType();
+        final UnitType unitType = unit.getType();
         return (work == null) // No work, improvement does nothing
             ? 0
             : (newType == null) // No tile change, but return the new bonus
@@ -260,14 +266,14 @@ public class ColonyTile extends WorkLocation {
     /**
      * Evaluate this work location for a given player.
      *
-     * @param player The <code>Player</code> to evaluate for.
+     * @param player The {@code Player} to evaluate for.
      * @return A value for the player.
      */
     @Override
     public int evaluateFor(Player player) {
         return super.evaluateFor(player)
-            + getProductionInfo().getProduction().stream()
-                .mapToInt(ag -> ag.evaluateFor(player)).sum();
+            + sum(getProductionInfo().getProduction(),
+                  ag -> ag.evaluateFor(player));
     }
 
 
@@ -280,7 +286,6 @@ public class ColonyTile extends WorkLocation {
     //   UnitLocation.canAdd
     //   WorkLocation.remove
     //   UnitLocation.getUnitCount
-    //   final UnitLocation.getUnitIterator
     //   UnitLocation.getUnitList
     //   UnitLocation.getGoodsContainer
     //   final WorkLocation getSettlement
@@ -342,6 +347,13 @@ public class ColonyTile extends WorkLocation {
 
     // Interface WorkLocation
 
+    @Override
+    protected boolean goodSuggestionCheck(UnitType unitType, Unit unit, GoodsType goodsType) {
+        final Tile tile = getWorkTile();
+        final Player owner = getOwner();
+        return owner.owns(tile) || owner.canClaimForSettlement(tile);
+    }
+
     /**
      * {@inheritDoc}
      */
@@ -403,6 +415,13 @@ public class ColonyTile extends WorkLocation {
             : NoAddReason.WRONG_TYPE;
     }
 
+    /**
+     * {@inheritDoc}
+     */
+    public int getLevel() {
+        return 0; // Level not meaningful for colony tiles
+    }
+
     /**
      * {@inheritDoc}
      */
@@ -435,11 +454,9 @@ public class ColonyTile extends WorkLocation {
      * {@inheritDoc}
      */
     @Override
-    public List<Modifier> getProductionModifiers(GoodsType goodsType,
-                                                 UnitType unitType) {
-        if (!canProduce(goodsType, unitType)) {
-            return Collections.<Modifier>emptyList();
-        }
+    public Stream<Modifier> getProductionModifiers(GoodsType goodsType,
+                                                   UnitType unitType) {
+        if (!canProduce(goodsType, unitType)) return Stream.<Modifier>empty();
 
         final Tile workTile = getWorkTile();
         final TileType type = workTile.getType();
@@ -447,27 +464,22 @@ public class ColonyTile extends WorkLocation {
         final Colony colony = getColony();
         final Player owner = colony.getOwner();
         final Turn turn = getGame().getTurn();
-
-        List<Modifier> mods = new ArrayList<>();
-        if (unitType == null) {
-            if (isColonyCenterTile()) { // Unattended only possible in center
-                mods.addAll(workTile.getProductionModifiers(goodsType, null));
-                mods.addAll(colony.getProductionModifiers(goodsType));
-                mods.addAll(colony.getModifiers(id, null, turn));
-                if (owner != null) {
-                    mods.addAll(owner.getModifiers(id, type, turn));
-                }
-            }
-
-        } else {
-            mods.addAll(workTile.getProductionModifiers(goodsType, unitType));
-            mods.addAll(colony.getProductionModifiers(goodsType));
-            mods.addAll(unitType.getModifiers(id, type, turn));
-            if (owner != null) {
-                mods.addAll(owner.getModifiers(id, unitType, turn));
-            }
-        }
-        return mods;
+        return (unitType != null)
+            // Unit modifiers apply
+            ? concat(workTile.getProductionModifiers(goodsType, unitType),
+                     colony.getProductionModifiers(goodsType, unitType, this),
+                     unitType.getModifiers(id, type, turn),
+                     ((owner == null) ? null
+                         : owner.getModifiers(id, unitType, turn)))
+            // Unattended only possible in center, colony modifiers apply
+            : (isColonyCenterTile())
+            ? concat(workTile.getProductionModifiers(goodsType, null),
+                     colony.getProductionModifiers(goodsType, null, this),
+                     colony.getModifiers(id, null, turn),
+                     ((owner == null) ? null
+                         : owner.getModifiers(id, type, turn)))
+            // Otherwise impossible
+            : Stream.<Modifier>empty();
     }
 
     /**
@@ -481,6 +493,22 @@ public class ColonyTile extends WorkLocation {
             : workTile.getType().getAvailableProductionTypes(unattended);
     }
 
+    /**
+     * {@inheritDoc}
+     */
+    @Override
+    public float getCompetenceFactor() {
+        return 1.0f;
+    }
+
+    /**
+     * {@inheritDoc}
+     */
+    @Override
+    public float getRebelFactor() {
+        return 1.0f;
+    }
+
     /**
      * {@inheritDoc}
      */
@@ -492,6 +520,29 @@ public class ColonyTile extends WorkLocation {
     }
 
 
+    // Override FreeColObject
+
+    /**
+     * {@inheritDoc}
+     */
+    @Override
+    public <T extends FreeColObject> boolean copyIn(T other) {
+        ColonyTile o = copyInCast(other, ColonyTile.class);
+        if (o == null || !super.copyIn(o)) return false;
+        final Game game = getGame();
+        this.workTile = game.updateRef(o.getWorkTile());
+        return true;
+    }
+
+    /**
+     * {@inheritDoc}
+     */
+    @Override
+    public FreeColObject getDisplayObject() {
+        return getTile().getDisplayObject();
+    }
+
+
     // Serialization
 
     private static final String WORK_TILE_TAG = "workTile";
@@ -514,35 +565,29 @@ public class ColonyTile extends WorkLocation {
     protected void readAttributes(FreeColXMLReader xr) throws XMLStreamException {
         super.readAttributes(xr);
 
-        workTile = xr.makeFreeColGameObject(getGame(), WORK_TILE_TAG,
-                                            Tile.class, true);
+        workTile = xr.makeFreeColObject(getGame(), WORK_TILE_TAG,
+                                        Tile.class, true);
     }
 
+    /**
+     * {@inheritDoc}
+     */
+    public String getXMLTagName() { return TAG; }
+
+
+    // Override Object
+
     /**
      * {@inheritDoc}
      */
     @Override
     public String toString() {
         StringBuilder sb = new StringBuilder(64);
-        sb.append("[").append(getId())
-            .append(" ").append(getWorkTile())
-            .append("/").append(getColony().getName())
-            .append("]");
+        Colony c = getColony();
+        sb.append('[').append(getId())
+            .append(' ').append(getWorkTile())
+            .append('/').append((c == null) ? "NO-COLONY" : c.getName())
+            .append(']');
         return sb.toString();
     }
-
-    /**
-     * {@inheritDoc}
-     */
-    @Override
-    public String getXMLTagName() { return getXMLElementTagName(); }
-
-    /**
-     * Gets the tag name of the root element representing this object.
-     *
-     * @return "colonyTile".
-     */
-    public static String getXMLElementTagName() {
-        return "colonyTile";
-    }
 }
