@@ -1,5 +1,5 @@
 /**
- *  Copyright (C) 2002-2015   The FreeCol Team
+ *  Copyright (C) 2002-2022   The FreeCol Team
  *
  *  This file is part of FreeCol.
  *
@@ -18,17 +18,13 @@
  */
 package net.sf.freecol.common.model;
 
-import java.util.ArrayList;
 import java.util.Collection;
-import java.util.Collections;
 import java.util.HashMap;
 import java.util.HashSet;
-import java.util.Iterator;
-import java.util.List;
 import java.util.Map;
-import java.util.Map.Entry;
 import java.util.Set;
 import java.util.logging.Logger;
+import java.util.stream.Stream;
 
 import static net.sf.freecol.common.util.CollectionUtils.*;
 
@@ -36,7 +32,7 @@ import static net.sf.freecol.common.util.CollectionUtils.*;
 /**
  * A container to hold abilities and modifiers for some FreeColObject-subclass.
  *
- * - FreeColGameObjectType, Europe, Player, Settlement are current
+ * - FreeColSpecObjectType, Europe, Player, Settlement are current
  *   implementors.
  *
  * - Building delegates some functionality to its type.
@@ -100,17 +96,40 @@ public final class FeatureContainer {
         }
     }
 
+    protected Collection<Ability> getAbilityValues() {
+        Set<Ability> ret = new HashSet<>();
+        synchronized (abilitiesLock) {
+            if (abilities != null) {
+                for (Set<Ability> as : this.abilities.values()) ret.addAll(as);
+            }
+        }
+        return ret;
+    }
+    protected Collection<Modifier> getModifierValues() {
+        Set<Modifier> ret = new HashSet<>();
+        synchronized (modifiersLock) {
+            if (modifiers != null) {
+                for (Set<Modifier> ms : this.modifiers.values()) ret.addAll(ms);
+            }
+        }
+        return ret;
+    }
+    
 
     /**
      * Is the given set of abilities non-empty and contains no
      * false-valued members?
      *
-     * @return True if the ability set is `satisfied'.
+     * @param abilities A stream of {@code Ability}s to check.
+     * @return True if the abilities are `satisfied'.
      */
-    public static boolean hasAbility(Set<Ability> abilitySet) {
-        return (abilitySet == null || abilitySet.isEmpty())
-            ? false
-            : all(abilitySet, Ability::getValue);
+    public static boolean allAbilities(Stream<Ability> abilities) {
+        boolean ret = false;
+        for (Ability ability : iterable(abilities)) {
+            if (!ability.getValue()) return false;
+            ret = true;
+        }
+        return ret;
     }
 
     /**
@@ -118,14 +137,14 @@ public final class FeatureContainer {
      * false-valued members?
      *
      * @param id The object identifier.
-     * @param fcgot An optional <code>FreeColGameObjectType</code> the
+     * @param fcgot An optional {@code FreeColSpecObjectType} the
      *     ability applies to.
-     * @param turn An optional applicable <code>Turn</code>.
+     * @param turn An optional applicable {@code Turn}.
      * @return True if the ability is present.
      */
-    public boolean hasAbility(String id, FreeColGameObjectType fcgot,
+    public boolean hasAbility(String id, FreeColSpecObjectType fcgot,
                               Turn turn) {
-        return FeatureContainer.hasAbility(getAbilities(id, fcgot, turn));
+        return FeatureContainer.allAbilities(getAbilities(id, fcgot, turn));
     }
 
     /**
@@ -135,7 +154,7 @@ public final class FeatureContainer {
      * @return True if the key is present.
      */
     public boolean containsAbilityKey(String key) {
-        return !getAbilities(key, null, null).isEmpty();
+        return first(getAbilities(key, null, null)) != null;
     }
 
     /**
@@ -143,13 +162,13 @@ public final class FeatureContainer {
      * container.
      *
      * @param id The object identifier (null matches all).
-     * @param fcgot An optional <code>FreeColGameObjectType</code> the
+     * @param fcgot An optional {@code FreeColSpecObjectType} the
      *     ability applies to.
-     * @param turn An optional applicable <code>Turn</code>.
-     * @return A set of abilities.
+     * @param turn An optional applicable {@code Turn}.
+     * @return A stream of abilities.
      */
-    public Set<Ability> getAbilities(String id, FreeColGameObjectType fcgot,
-                                     Turn turn) {
+    public Stream<Ability> getAbilities(String id, FreeColSpecObjectType fcgot,
+                                        Turn turn) {
         Set<Ability> result = new HashSet<>();
         if (abilitiesPresent()) {
             synchronized (abilitiesLock) {
@@ -162,20 +181,16 @@ public final class FeatureContainer {
                     if (aset != null) result.addAll(aset);
                 }
             }
-            Iterator<Ability> it = result.iterator();
-            while (it.hasNext()) {
-                Ability a = it.next();
-                if (!a.appliesTo(fcgot, turn)) it.remove();
-            }
+            removeInPlace(result, a -> !a.appliesTo(fcgot, turn));
         }
-        return result;
+        return result.stream();
     }
 
     /**
      * Add the given ability to a container.
      *
-     * @param ability An <code>Ability</code> to add.
-     * @return True if the <code>Ability</code> was added.
+     * @param ability An {@code Ability} to add.
+     * @return True if the {@code Ability} was added.
      */
     public boolean addAbility(Ability ability) {
         if (ability == null) return false;
@@ -194,7 +209,7 @@ public final class FeatureContainer {
     /**
      * Remove the given ability from a container.
      *
-     * @param ability An <code>Ability</code> to remove.
+     * @param ability An {@code Ability} to remove.
      * @return The ability removed or null on failure.
      */
     public Ability removeAbility(Ability ability) {
@@ -227,67 +242,74 @@ public final class FeatureContainer {
      * container.
      *
      * @param id The object identifier.
-     * @param fcgot An optional <code>FreeColGameObjectType</code> the
+     * @param fcgot An optional {@code FreeColSpecObjectType} the
      *     modifier applies to.
-     * @param turn An optional applicable <code>Turn</code>.
-     * @return A set of modifiers.
+     * @param turn An optional applicable {@code Turn}.
+     * @return A stream of {@code Modifier}s.
      */
-    public Set<Modifier> getModifiers(String id, FreeColGameObjectType fcgot,
-                                      Turn turn) {
-        Set<Modifier> result = new HashSet<>();
-        if (modifiersPresent()) {
-            synchronized (modifiersLock) {
-                if (id == null) {
-                    for (Set<Modifier> mset : modifiers.values()) {
-                        result.addAll(mset);
-                    }
-                } else {
-                    Set<Modifier> mset = modifiers.get(id);
-                    if (mset != null) result.addAll(mset);
-                }
-            }
-            Iterator<Modifier> it = result.iterator();
-            while (it.hasNext()) {
-                Modifier m = it.next();
-                if (!m.appliesTo(fcgot, turn)) it.remove();
+    public Stream<Modifier> getModifiers(String id,
+                                         FreeColSpecObjectType fcgot,
+                                         Turn turn) {
+        if (!modifiersPresent()) return Stream.<Modifier>empty();
+        Set<Modifier> mset = new HashSet<>();
+        synchronized (modifiersLock) {
+            if (id == null) {
+                for (Set<Modifier> ms : modifiers.values()) mset.addAll(ms);
+            } else {
+                Set<Modifier> ms = modifiers.get(id);
+                if (ms != null) mset.addAll(ms);
             }
         }
-        return result;
-    }
-
-    /**
-     * Applies this objects modifiers with the given identifier to the
-     * given number.
-     *
-     * @param number The number to modify.
-     * @param turn An optional applicable <code>Turn</code>.
-     * @param id The object identifier.
-     * @param fcgot An optional <code>FreeColGameObjectType</code> the
-     *     modifier applies to.
-     * @return The modified number.
-     */
-    public final float applyModifiers(float number, Turn turn,
-                                      String id, FreeColGameObjectType fcgot) {
-        return applyModifiers(number, turn, getModifiers(id, fcgot, turn));
+        removeInPlace(mset, m -> !m.appliesTo(fcgot, turn));
+        return (mset.isEmpty()) ? Stream.<Modifier>empty() : mset.stream();
     }
 
     /**
      * Applies a collection of modifiers to the given float value.
      *
      * @param number The number to modify.
-     * @param turn An optional applicable <code>Turn</code>.
-     * @param mods The <code>Modifier</code>s to apply.
+     * @param turn An optional applicable {@code Turn}.
+     * @param mods The {@code Modifier}s to apply.
      * @return The modified number.
      */
     public static float applyModifiers(float number, Turn turn,
                                        Collection<Modifier> mods) {
-        if (mods == null || mods.isEmpty()) return number;
-        List<Modifier> modifiers = new ArrayList<>(mods);
-        Collections.sort(modifiers);
+        return (mods == null || mods.isEmpty()) ? number
+            : applyModifiersInternal(number, turn,
+                sort(mods, Modifier.ascendingModifierIndexComparator));
+    }
+
+    /**
+     * Applies a stream of modifiers to the given float value.
+     *
+     * @param number The number to modify.
+     * @param turn An optional applicable {@code Turn}.
+     * @param mods The {@code Modifier}s to apply.
+     * @return The modified number.
+     */
+    public static float applyModifiers(float number, Turn turn,
+                                       Stream<Modifier> mods) {
+        return (mods == null) ? number
+            : applyModifiersInternal(number, turn,
+                sort(mods, Modifier.ascendingModifierIndexComparator));
+    }
+
+    /**
+     * Implement applyModifiers.
+     *
+     * @param number The number to modify.
+     * @param turn An optional applicable {@code Turn}.
+     * @param mods The {@code Modifier}s to apply.
+     * @return The modified number.
+     */
+    private static float applyModifiersInternal(float number, Turn turn,
+                                                Collection<Modifier> mods) {
         float result = number;
-        for (Modifier m : modifiers) {
+        for (Modifier m : mods) {
             float value = m.getValue(turn);
-            if (value == Modifier.UNKNOWN) return value;
+            if (Float.compare(value, Modifier.UNKNOWN) == 0) {
+                return value;
+            }
             result = m.apply(result, value);
         }
         return result;
@@ -296,7 +318,7 @@ public final class FeatureContainer {
     /**
      * Adds a modifier to a container.
      *
-     * @param modifier The <code>Modifier</code> to add.
+     * @param modifier The {@code Modifier} to add.
      * @return True if the modifier was added.
      */
     public boolean addModifier(Modifier modifier) {
@@ -316,7 +338,7 @@ public final class FeatureContainer {
     /**
      * Removes a modifier from a container.
      *
-     * @param modifier The <code>Modifier</code> to remove.
+     * @param modifier The {@code Modifier} to remove.
      * @return The modifier removed.
      */
     public Modifier removeModifier(Modifier modifier) {
@@ -346,7 +368,7 @@ public final class FeatureContainer {
     /**
      * Adds all the features in an object to this object.
      *
-     * @param fco The <code>FreeColObject</code> to add features from.
+     * @param fco The {@code FreeColObject} to add features from.
      */
     public void addFeatures(FreeColObject fco) {
         FeatureContainer c = fco.getFeatureContainer();
@@ -359,14 +381,14 @@ public final class FeatureContainer {
                 ca = new HashMap<>(c.abilities);
             }
             synchronized (abilitiesLock) {
-                for (Entry<String, Set<Ability>> e : ca.entrySet()) {
-                    Set<Ability> abilitySet = abilities.get(e.getKey());
-                    if (abilitySet == null) {
-                        abilitySet = new HashSet<>();
-                        abilities.put(e.getKey(), abilitySet);
-                    }
-                    abilitySet.addAll(e.getValue());
-                }
+                forEachMapEntry(ca, e -> {
+                        Set<Ability> abilitySet = abilities.get(e.getKey());
+                        if (abilitySet == null) {
+                            abilitySet = new HashSet<>();
+                            abilities.put(e.getKey(), abilitySet);
+                        }
+                        abilitySet.addAll(e.getValue());
+                    });
             }
         }
 
@@ -377,14 +399,14 @@ public final class FeatureContainer {
                 cm = new HashMap<>(c.modifiers);
             }
             synchronized (modifiersLock) {
-                for (Entry<String, Set<Modifier>> e : cm.entrySet()) {
-                    Set<Modifier> modifierSet = modifiers.get(e.getKey());
-                    if (modifierSet == null) {
-                        modifierSet = new HashSet<>();
-                        modifiers.put(e.getKey(), modifierSet);
-                    }
-                    modifierSet.addAll(e.getValue());
-                }
+                forEachMapEntry(cm, e -> {
+                        Set<Modifier> modifierSet = modifiers.get(e.getKey());
+                        if (modifierSet == null) {
+                            modifierSet = new HashSet<>();
+                            modifiers.put(e.getKey(), modifierSet);
+                        }
+                        modifierSet.addAll(e.getValue());
+                    });
             }
         }
     }
@@ -392,7 +414,7 @@ public final class FeatureContainer {
     /**
      * Removes all the features in an object from a container.
      *
-     * @param fco The <code>FreeColObject</code> to find features to remove
+     * @param fco The {@code FreeColObject} to find features to remove
      *     in.
      */
     public void removeFeatures(FreeColObject fco) {
@@ -432,6 +454,19 @@ public final class FeatureContainer {
         }
     }
 
+    /**
+     * Copy anther feature container.
+     *
+     * @param other The other {@code FeatureContainer}.
+     */
+    public void copy(FeatureContainer other) {
+        clear();
+        if (other != null) {
+            for (Ability a : other.getAbilityValues()) addAbility(a);
+            for (Modifier m : other.getModifierValues()) addModifier(m);
+        }
+    }
+                
     /**
      * Clear this feature container.
      */
@@ -452,30 +487,31 @@ public final class FeatureContainer {
      * Replaces the source field. This is necessary because objects
      * may inherit Features from other abstract objects.
      *
-     * @param oldSource The old source <code>FreeColGameObjectType</code>.
-     * @param newSource The new source <code>FreeColGameObjectType</code>.
+     * @param oldSource The old source {@code FreeColSpecObjectType}.
+     * @param newSource The new source {@code FreeColSpecObjectType}.
      */
-    public void replaceSource(FreeColGameObjectType oldSource,
-                              FreeColGameObjectType newSource) {
-        for (Ability ability : getAbilities(null, null, null)) {
-            if (oldSource == null || ability.getSource() == oldSource) {
-                removeAbility(ability);
-                Ability newAbility = new Ability(ability);
-                newAbility.setSource(newSource);
-                addAbility(newAbility);
-            }
+    public void replaceSource(FreeColSpecObjectType oldSource,
+                              FreeColSpecObjectType newSource) {
+        for (Ability ability : transform(getAbilities(null, null, null),
+                a -> oldSource == null || a.getSource() == oldSource)) {
+            removeAbility(ability);
+            Ability newAbility = new Ability(ability);
+            newAbility.setSource(newSource);
+            addAbility(newAbility);
         }
 
-        for (Modifier modifier : getModifiers(null, null, null)) {
-            if (oldSource == null || modifier.getSource() == oldSource) {
-                removeModifier(modifier);
-                Modifier newModifier = new Modifier(modifier);
-                newModifier.setSource(newSource);
-                addModifier(newModifier);
-            }
+        for (Modifier modifier : transform(getModifiers(null, null, null),
+                m -> oldSource == null || m.getSource() == oldSource)) {
+            removeModifier(modifier);
+            Modifier mod = Modifier.makeModifier(modifier);
+            mod.setSource(newSource);
+            addModifier(mod);
         }
     }
 
+
+    // Override Object
+
     /**
      * {@inheritDoc}
      */
@@ -483,23 +519,23 @@ public final class FeatureContainer {
     public String toString() {
         StringBuilder sb = new StringBuilder(256);
         sb.append("[FeatureContainer");
-        Set<Ability> abilities = getAbilities(null, null, null);
-        if (!abilities.isEmpty()) {
-            sb.append(" [abilities");
-            for (Ability ability : getAbilities(null, null, null)) {
-                sb.append(" ").append(ability);
-            }
-            sb.append("]");
+        int siz = sb.length();
+        for (Ability ability : iterable(getAbilities(null, null, null))) {
+            sb.append(' ').append(ability);
         }
-        Set<Modifier> modifiers = getModifiers(null, null, null);
-        if (!modifiers.isEmpty()) {
-            sb.append(" [modifiers");
-            for (Modifier modifier : getModifiers(null, null, null)) {
-                sb.append(" ").append(modifier);
-            }
-            sb.append("]");
+        if (sb.length() > siz) {
+            sb.insert(siz, " [abilities");
+            sb.append(']');
         }
-        sb.append("]");
+        siz = sb.length();
+        for (Modifier modifier : iterable(getModifiers(null, null, null))) {
+            sb.append(' ').append(modifier);
+        }
+        if (sb.length() > siz) {
+            sb.insert(siz, "[modifiers");
+            sb.append(']');
+        }
+        sb.append(']');
         return sb.toString();
     }
 }
