@@ -1,5 +1,5 @@
 /**
- *  Copyright (C) 2002-2015   The FreeCol Team
+ *  Copyright (C) 2002-2022   The FreeCol Team
  *
  *  This file is part of FreeCol.
  *
@@ -21,21 +21,19 @@ package net.sf.freecol.common.model;
 
 import java.util.ArrayList;
 import java.util.HashMap;
-import java.util.Iterator;
 import java.util.List;
 import java.util.Map;
 import java.util.Map.Entry;
+import java.util.function.Predicate;
 import java.util.logging.Logger;
 
-import javax.xml.stream.XMLStreamConstants;
 import javax.xml.stream.XMLStreamException;
 
 import net.sf.freecol.common.io.FreeColXMLReader;
 import net.sf.freecol.common.io.FreeColXMLWriter;
+import static net.sf.freecol.common.model.Constants.*;
 import static net.sf.freecol.common.util.CollectionUtils.*;
 
-import org.w3c.dom.Element;
-
 
 /**
  * Contains goods and can be used by a {@link Location} to make certain
@@ -44,7 +42,9 @@ import org.w3c.dom.Element;
 public class GoodsContainer extends FreeColGameObject implements Ownable {
 
     @SuppressWarnings("unused")
-    private static final Logger logger = Logger.getLogger(Location.class.getName());
+    private static final Logger logger = Logger.getLogger(GoodsContainer.class.getName());
+
+    public static final String TAG = "goodsContainer";
 
     /** The size of a standard `hold' of data. */
     public static final int CARGO_SIZE = 100;
@@ -59,7 +59,7 @@ public class GoodsContainer extends FreeColGameObject implements Ownable {
     public static final int HUGE_CARGO_SIZE = 100 * CARGO_SIZE;
 
     /**
-     * The list of Goods stored in this <code>GoodsContainer</code>.
+     * The list of Goods stored in this {@code GoodsContainer}.
      *
      * Always accessed synchronized (except I/O).
      */
@@ -67,7 +67,7 @@ public class GoodsContainer extends FreeColGameObject implements Ownable {
 
     /** 
      * The previous list of Goods stored in this
-     * <code>GoodsContainer</code>.
+     * {@code GoodsContainer}.
      *
      * Always accessed synchronized and *after synchronized(storedGoods)*.
      * This is only touched rarely so the extra lock is tolerable.
@@ -75,16 +75,16 @@ public class GoodsContainer extends FreeColGameObject implements Ownable {
      */
     private final Map<GoodsType, Integer> oldStoredGoods = new HashMap<>();
 
-    /** The location for this <code>GoodsContainer</code>. */
+    /** The location for this {@code GoodsContainer}. */
     private Location parent = null;
 
 
     /**
-     * Creates an empty <code>GoodsContainer</code>.
+     * Creates an empty {@code GoodsContainer}.
      *
-     * @param game The enclosing <code>Game</code>.
-     * @param parent The <code>Location</code> this
-     *     <code>GoodsContainer</code> contains goods for.
+     * @param game The enclosing {@code Game}.
+     * @param parent The {@code Location} this
+     *     {@code GoodsContainer} contains goods for.
      */
     public GoodsContainer(Game game, Location parent) {
         super(game);
@@ -93,95 +93,126 @@ public class GoodsContainer extends FreeColGameObject implements Ownable {
     }
 
     /**
-     * Create a new <code>GoodsContainer</code>.
+     * Create a new {@code GoodsContainer}.
      *
-     * @param game The enclosing <code>Game</code>.
+     * @param game The enclosing {@code Game}.
      * @param id The object identifier.
      */
     public GoodsContainer(Game game, String id) {
         super(game, id);
     }
 
+
     /**
-     * Create a new <code>GoodsContainer</code> from an
-     * <code>Element</code>.
+     * Get the parent location.
      *
-     * @param game The enclosing <code>Game</code>.
-     * @param parent The <code>Location</code> this
-     *     <code>GoodsContainer</code> contains goods for.
-     * @param e An XML-element that will be used to initialize
-     *     this object.
+     * @return The parent {@code Location}.
      */
-    public GoodsContainer(Game game, Location parent, Element e) {
-        super(game, null);
-
-        this.parent = parent;
-
-        readFromXMLElement(e);
+    protected Location getParent() {
+        return this.parent;
     }
 
-
     /**
      * Set the goods location.
      *
-     * @param location The <code>Location</code> to set.
+     * @param location The {@code Location} to set.
      */
     public void setLocation(Location location) {
         if (location == null) {
-            throw new IllegalArgumentException("Null GoodsContainer Location.");
+            throw new RuntimeException("Null GoodsContainer location: " + this);
         }
         this.parent = location;
     }
 
     /**
-     * Checks if the specified <code>Goods</code> is in this container.
+     * Get the stored goods.
      *
-     * @param g The <code>Goods</code> to test the presence of.
+     * @return A map of the stored goods.
+     */
+    protected Map<GoodsType, Integer> getStoredGoods() {
+        return this.storedGoods;
+    }
+
+    /**
+     * Set the stored goods.
+     *
+     * @param goods A map of the new stored goods.
+     */
+    protected void setStoredGoods(Map<GoodsType, Integer> goods) {
+        synchronized (this.storedGoods) {
+            this.storedGoods.clear();
+            this.storedGoods.putAll(goods);
+        }
+    }
+
+    /**
+     * Get the old stored goods.
+     *
+     * @return A map of the old stored goods.
+     */
+    protected Map<GoodsType, Integer> getOldStoredGoods() {
+        return this.oldStoredGoods;
+    }
+    
+    /**
+     * Set the old stored goods.
+     *
+     * @param goods A map of the new old stored goods.
+     */
+    protected void setOldStoredGoods(Map<GoodsType, Integer> goods) {
+        synchronized (this.oldStoredGoods) {
+            this.oldStoredGoods.clear();
+            this.oldStoredGoods.putAll(goods);
+        }
+    }
+
+    /**
+     * Checks if the specified {@code Goods} is in this container.
+     *
+     * @param <T> The base type of the goods.
+     * @param g The goods to test the presence of.
      * @return True if there is enough of the specified goods present that it
      *     can be removed without error.
      */
-    public boolean contains(Goods g) {
+    public <T extends AbstractGoods> boolean contains(T g) {
         return getGoodsCount(g.getType()) >= g.getAmount();
     }
 
     /**
      * Gets the amount of one type of goods in this container.
      *
-     * @param type The <code>GoodsType</code> being looked for.
+     * @param type The {@code GoodsType} being looked for.
      * @return The amount of this type of goods in this container.
      */
     public int getGoodsCount(GoodsType type) {
-        synchronized (storedGoods) {
-            return (storedGoods.containsKey(type)) 
-                ? storedGoods.get(type)
-                : 0;
+        synchronized (this.storedGoods) {
+            Integer val = this.storedGoods.get(type);
+            return (val == null) ? 0 : val;
         }
     }
 
     /**
      * Gets the amount of one type of goods at the beginning of the turn.
      *
-     * @param type The <code>GoodsType</code> being looked for.
+     * @param type The {@code GoodsType} being looked for.
      * @return The amount of this type of goods in this container at
      *     the beginning of the turn
      */
     public int getOldGoodsCount(GoodsType type) {
-        synchronized (storedGoods) {
-            synchronized (oldStoredGoods) {
-                return (oldStoredGoods.containsKey(type))
-                    ? oldStoredGoods.get(type)
-                    : 0;
-            }
+        synchronized (this.oldStoredGoods) {
+            Integer val = this.oldStoredGoods.get(type);
+            return (val == null) ? 0 : val;
         }
     }
 
     /**
      * Adds goods to this goods container.
      *
-     * @param goods The <code>Goods</code> to add.
+     * @param <T> The base type of the goods.
+     * @param goods The goods to add.
      * @return True if the addition succeeds.
      */
-    public boolean addGoods(AbstractGoods goods) {
+    public <T extends AbstractGoods> boolean addGoods(T goods) {
         return addGoods(goods.getType(), goods.getAmount());
     }
 
@@ -190,7 +221,7 @@ public class GoodsContainer extends FreeColGameObject implements Ownable {
      *
      * Note: negative amounts are allowed.
      *
-     * @param type The <code>GoodsType</code> to add.
+     * @param type The {@code GoodsType} to add.
      * @param amount The amount of goods to add.
      * @return True if the addition succeeds.
      */
@@ -203,12 +234,12 @@ public class GoodsContainer extends FreeColGameObject implements Ownable {
                 + newAmount + " goods of type " + type
                 + " in Location " + parent);
         } else if (newAmount == 0) {
-            synchronized (storedGoods) {
-                storedGoods.remove(type);
+            synchronized (this.storedGoods) {
+                this.storedGoods.remove(type);
             }
         } else {
-            synchronized (storedGoods) {
-                storedGoods.put(type, newAmount);
+            synchronized (this.storedGoods) {
+                this.storedGoods.put(type, newAmount);
             }
         }
         return true;
@@ -217,18 +248,19 @@ public class GoodsContainer extends FreeColGameObject implements Ownable {
     /**
      * Removes goods from this goods container.
      *
-     * @param goods The <code>Goods</code> to remove from this container.
-     * @return The <code>Goods</code> actually removed.
+     * @param <T> The base type of the goods.
+     * @param goods The goods to remove from this container.
+     * @return The {@code Goods} actually removed.
      */
-    public Goods removeGoods(AbstractGoods goods) {
+    public <T extends AbstractGoods> Goods removeGoods(T goods) {
         return removeGoods(goods.getType(), goods.getAmount());
     }
 
     /**
      * Removes all goods of a given type from this goods container.
      *
-     * @param type The <code>GoodsType</code> to remove.
-     * @return The <code>Goods</code> actually removed.
+     * @param type The {@code GoodsType} to remove.
+     * @return The {@code Goods} actually removed.
      */
     public Goods removeGoods(GoodsType type) {
         return removeGoods(type, INFINITY);
@@ -237,9 +269,9 @@ public class GoodsContainer extends FreeColGameObject implements Ownable {
     /**
      * Removes goods by type and amount from this goods container.
      *
-     * @param type The <code>GoodsType</code> to remove.
+     * @param type The {@code GoodsType} to remove.
      * @param amount The amount of goods to remove.
-     * @return The <code>Goods</code> actually removed, which may have a
+     * @return The {@code Goods} actually removed, which may have a
      *     lower actual amount, or null if nothing removed.
      */
     public Goods removeGoods(GoodsType type, int amount) {
@@ -250,13 +282,13 @@ public class GoodsContainer extends FreeColGameObject implements Ownable {
         Goods removedGoods;
         if (newAmount > 0) {
             removedGoods = new Goods(getGame(), null, type, amount);
-            synchronized (storedGoods) {
-                storedGoods.put(type, newAmount);
+            synchronized (this.storedGoods) {
+                this.storedGoods.put(type, newAmount);
             }
         } else {
             removedGoods = new Goods(getGame(), null, type, oldAmount);
-            synchronized (storedGoods) {
-                storedGoods.remove(type);
+            synchronized (this.storedGoods) {
+                this.storedGoods.remove(type);
             }
         }
         return removedGoods;
@@ -265,17 +297,17 @@ public class GoodsContainer extends FreeColGameObject implements Ownable {
     /**
      * Set the amount of goods in this container.
      *
-     * @param goodsType The <code>GoodsType</code> to set the amount of.
+     * @param goodsType The {@code GoodsType} to set the amount of.
      * @param newAmount The new amount.
      */
     public void setAmount(GoodsType goodsType, int newAmount) {
         if (newAmount == 0) {
-            synchronized (storedGoods) {
-                storedGoods.remove(goodsType);
+            synchronized (this.storedGoods) {
+                this.storedGoods.remove(goodsType);
             }
         } else {
-            synchronized (storedGoods) {
-                storedGoods.put(goodsType, newAmount);
+            synchronized (this.storedGoods) {
+                this.storedGoods.put(goodsType, newAmount);
             }
         }
     }
@@ -284,8 +316,8 @@ public class GoodsContainer extends FreeColGameObject implements Ownable {
      * Remove all goods.
      */
     public void removeAll() {
-        synchronized (storedGoods) {
-            storedGoods.clear();
+        synchronized (this.storedGoods) {
+            this.storedGoods.clear();
         }
     }
 
@@ -293,10 +325,10 @@ public class GoodsContainer extends FreeColGameObject implements Ownable {
      * Clear both containers.
      */
     private void clearContainers() {
-        synchronized (storedGoods) {
-            storedGoods.clear();
-            synchronized (oldStoredGoods) {
-                oldStoredGoods.clear();
+        synchronized (this.storedGoods) {
+            this.storedGoods.clear();
+            synchronized (this.oldStoredGoods) {
+                this.oldStoredGoods.clear();
             }
         }
     }
@@ -307,18 +339,19 @@ public class GoodsContainer extends FreeColGameObject implements Ownable {
      *
      * @param newAmount The threshold.
      */
-    public void removeAbove(int newAmount) {
-        synchronized (storedGoods) {
+    public void removeAbove(final int newAmount) {
+        synchronized (this.storedGoods) {
             if (newAmount <= 0) {
-                storedGoods.clear();
+                this.storedGoods.clear();
                 return;
             }
-            for (GoodsType goodsType : storedGoods.keySet()) {
-                if (goodsType.isStorable() && !goodsType.limitIgnored()
-                    && storedGoods.get(goodsType) > newAmount) {
-                    setAmount(goodsType, newAmount);
-                }
-            }
+            final Predicate<Entry<GoodsType, Integer>> hiPred = e -> {
+                final GoodsType gt = e.getKey();
+                return gt.isStorable() && !gt.limitIgnored()
+                    && e.getValue() > newAmount;
+            };
+            forEachMapEntry(this.storedGoods, hiPred,
+                            e -> this.storedGoods.put(e.getKey(), newAmount));
         }
     }
 
@@ -330,9 +363,11 @@ public class GoodsContainer extends FreeColGameObject implements Ownable {
      *     given amount.
      */
     public boolean hasReachedCapacity(int amount) {
-        synchronized (storedGoods) {
-            return any(storedGoods.keySet(), gt -> gt.isStorable()
-                && !gt.limitIgnored() && storedGoods.get(gt) > amount);
+        synchronized (this.storedGoods) {
+            return any(this.storedGoods.entrySet(), e ->
+                       e.getKey().isStorable()
+                           && !e.getKey().limitIgnored()
+                           && e.getValue() > amount);
         }
     }
 
@@ -343,65 +378,49 @@ public class GoodsContainer extends FreeColGameObject implements Ownable {
      * @return The amount of space taken by this containers goods.
      */
     public int getSpaceTaken() {
-        synchronized (storedGoods) {
-            return storedGoods.values().stream()
-                .mapToInt(amount -> (amount % CARGO_SIZE == 0)
-                    ? amount/CARGO_SIZE
-                    : amount/CARGO_SIZE + 1).sum();
+        synchronized (this.storedGoods) {
+            return sum(this.storedGoods.values(),
+                       amount -> ((amount % CARGO_SIZE == 0)
+                           ? amount/CARGO_SIZE
+                           : amount/CARGO_SIZE + 1));
         }
     }
 
-    /**
-     * Gets an iterator over all holds of goods in this goods container.
-     * Each <code>Goods</code> returned has a maximum amount of CARGO_SIZE.
-     *
-     * @return The <code>Iterator</code>.
-     * @see #getCompactGoods
-     */
-    public Iterator<Goods> getGoodsIterator() {
-        return getGoods().iterator();
-    }
-
     /**
      * Gets a list containing all holds of goods in this goods container.
-     * Each <code>Goods</code> returned has a maximum amount of CARGO_SIZE.
+     * Each {@code Goods} returned has a maximum amount of CARGO_SIZE.
      *
-     * @return A list of <code>Goods</code>.
-     * @see #getGoodsIterator
+     * @return A list of {@code Goods}.
      */
-    public List<Goods> getGoods() {
-        List<Goods> totalGoods = new ArrayList<>();
-        synchronized (storedGoods) {
-            for (GoodsType goodsType : storedGoods.keySet()) {
-                int amount = storedGoods.get(goodsType);
-                while (amount > 0) {
-                    totalGoods.add(new Goods(getGame(), parent, goodsType,
-                            ((amount >= CARGO_SIZE) ? CARGO_SIZE : amount)));
-                    amount -= CARGO_SIZE;
-                }
-            }
+    public List<Goods> getGoodsList() {
+        final Game game = getGame();
+        List<Goods> result = new ArrayList<>();
+        synchronized (this.storedGoods) {
+            forEachMapEntry(this.storedGoods, e -> {
+                    int amount = e.getValue();
+                    while (amount > 0) {
+                        result.add(new Goods(game, parent, e.getKey(),
+                                ((amount >= CARGO_SIZE) ? CARGO_SIZE : amount)));
+                        amount -= CARGO_SIZE;
+                    }
+                });
         }
-        return totalGoods;
+        return result;
     }
 
     /**
-     * Gets a list of all goods in this goods container.
-     * There is only one <code>Goods</code> for each distinct
-     * <code>GoodsType</code>.
+     * Gets a list of all goods in this goods container.  There is
+     * only one {@code Goods} for each distinct {@code GoodsType}.
      *
-     * @return A list of <code>Goods</code>.
+     * @return A list of {@code Goods}.
      */
-    public List<Goods> getCompactGoods() {
-        List<Goods> totalGoods = new ArrayList<>();
-        synchronized (storedGoods) {
-            for (Entry<GoodsType, Integer> entry : storedGoods.entrySet()) {
-                if (entry.getValue() > 0) {
-                    totalGoods.add(new Goods(getGame(), parent, entry.getKey(),
-                                             entry.getValue()));
-                }
-            }
+    public List<Goods> getCompactGoodsList() {
+        final Game game = getGame();
+        synchronized (this.storedGoods) {
+            return transform(this.storedGoods.entrySet(),
+                             e -> e.getValue() > 0,
+                             e -> new Goods(game, parent, e.getKey(), e.getValue()));
         }
-        return totalGoods;
     }
 
     /**
@@ -409,10 +428,10 @@ public class GoodsContainer extends FreeColGameObject implements Ownable {
      * stored goods.
      */
     public void saveState() {
-        synchronized (storedGoods) {
-            synchronized (oldStoredGoods) {
-                oldStoredGoods.clear();
-                oldStoredGoods.putAll(storedGoods);
+        synchronized (this.storedGoods) {
+            synchronized (this.oldStoredGoods) {
+                this.oldStoredGoods.clear();
+                this.oldStoredGoods.putAll(this.storedGoods);
             }
         }
     }
@@ -422,10 +441,10 @@ public class GoodsContainer extends FreeColGameObject implements Ownable {
      * old state.
      */
     public void restoreState() {
-        synchronized (storedGoods) {
-            synchronized (oldStoredGoods) {
-                storedGoods.clear();
-                storedGoods.putAll(oldStoredGoods);
+        synchronized (this.storedGoods) {
+            synchronized (this.oldStoredGoods) {
+                this.storedGoods.clear();
+                this.storedGoods.putAll(this.oldStoredGoods);
             }
         }
     }
@@ -438,7 +457,7 @@ public class GoodsContainer extends FreeColGameObject implements Ownable {
      */
     public boolean hasChanged() {
         return any(getSpecification().getGoodsTypeList(),
-            gt -> getOldGoodsCount(gt) != getGoodsCount(gt));
+                   gt -> getOldGoodsCount(gt) != getGoodsCount(gt));
     }
 
     /**
@@ -460,6 +479,28 @@ public class GoodsContainer extends FreeColGameObject implements Ownable {
         return ret;
     }
 
+    /**
+     * Utility to move goods from one location to another.
+     *
+     * @param src The source {@code GoodsContainer}.
+     * @param goodsType The {@code GoodsType} to move.
+     * @param amount The amount of goods to move.
+     * @param dst The new {@code GoodsContainer}.
+     */
+    public static void moveGoods(GoodsContainer src,
+                                 GoodsType goodsType, int amount,
+                                 GoodsContainer dst) {
+        if (src != null) {
+            src.saveState();
+            src.removeGoods(goodsType, amount);
+        }
+        if (dst != null) {
+            dst.saveState();
+            dst.addGoods(goodsType, amount);
+        }
+    }
+
+    
     // Interface Ownable
 
     /**
@@ -491,6 +532,22 @@ public class GoodsContainer extends FreeColGameObject implements Ownable {
     }
 
 
+    // Override FreeColObject
+
+    /**
+     * {@inheritDoc}
+     */
+    @Override
+    public <T extends FreeColObject> boolean copyIn(T other) {
+        GoodsContainer o = copyInCast(other, GoodsContainer.class);
+        if (o == null || !super.copyIn(o)) return false;
+        // Parent is fixed, and not serialized
+        this.setStoredGoods(o.getStoredGoods());
+        this.setOldStoredGoods(o.getOldStoredGoods());
+        return true;
+    }
+
+
     // Serialization
 
     public static final String AMOUNT_TAG = "amount";
@@ -508,16 +565,19 @@ public class GoodsContainer extends FreeColGameObject implements Ownable {
 
         if (xw.validFor(getOwner())) {
 
-            writeStorage(xw, STORED_GOODS_TAG, storedGoods);
-
-            writeStorage(xw, OLD_STORED_GOODS_TAG, oldStoredGoods);
+            synchronized (this.storedGoods) {
+                writeStorage(xw, STORED_GOODS_TAG, this.storedGoods);
+                synchronized (this.oldStoredGoods) {
+                    writeStorage(xw, OLD_STORED_GOODS_TAG, this.oldStoredGoods);
+                }
+            }
         }
     }
 
     /**
      * Write a storage container to a stream.
      *
-     * @param xw The <code>FreeColXMLWriter</code> to write to.
+     * @param xw The {@code FreeColXMLWriter} to write to.
      * @param tag The element tag.
      * @param storage The storage container.
      * @exception XMLStreamException if there is a problem writing to
@@ -529,9 +589,9 @@ public class GoodsContainer extends FreeColGameObject implements Ownable {
 
         xw.writeStartElement(tag);
 
-        for (GoodsType goodsType : getSortedCopy(storage.keySet())) {
-
-            xw.writeStartElement(Goods.getXMLElementTagName());
+        for (GoodsType goodsType : sort(storage.keySet())) {
+            
+            xw.writeStartElement(Goods.TAG);
 
             xw.writeAttribute(TYPE_TAG, goodsType);
 
@@ -562,10 +622,14 @@ public class GoodsContainer extends FreeColGameObject implements Ownable {
         final String tag = xr.getLocalName();
 
         if (OLD_STORED_GOODS_TAG.equals(tag)) {
-            readStorage(xr, oldStoredGoods);
+            synchronized (this.oldStoredGoods) {
+                readStorage(xr, this.oldStoredGoods);
+            }
 
         } else if (STORED_GOODS_TAG.equals(tag)) {
-            readStorage(xr, storedGoods);
+            synchronized (this.storedGoods) {
+                readStorage(xr, this.storedGoods);
+            }
 
         } else {
             super.readChild(xr);
@@ -575,7 +639,7 @@ public class GoodsContainer extends FreeColGameObject implements Ownable {
     /**
      * Read a storage container from a stream.
      *
-     * @param xr The <code>FreeColXMLReader</code> to read from.
+     * @param xr The {@code FreeColXMLReader} to read from.
      * @param storage The storage container.
      * @exception XMLStreamException if there is a problem reading from
      *     the stream.
@@ -584,10 +648,10 @@ public class GoodsContainer extends FreeColGameObject implements Ownable {
         Map<GoodsType, Integer> storage) throws XMLStreamException {
         final Specification spec = getGame().getSpecification();
 
-        while (xr.nextTag() != XMLStreamConstants.END_ELEMENT) {
+        while (xr.moreTags()) {
             String tag = xr.getLocalName();
 
-            if (Goods.getXMLElementTagName().equals(tag)) {
+            if (Goods.TAG.equals(tag)) {
                 GoodsType goodsType = xr.getType(spec, TYPE_TAG,
                     GoodsType.class, (GoodsType)null);
 
@@ -606,38 +670,28 @@ public class GoodsContainer extends FreeColGameObject implements Ownable {
     /**
      * {@inheritDoc}
      */
-    @Override
-    public String toString() {
-        StringBuilder sb = new StringBuilder(128);
-        sb.append("[").append(getId()).append(" [");
-        // Do not bother to synchronize containers for display
-        for (Map.Entry<GoodsType, Integer> entry : storedGoods.entrySet()) {
-            sb.append(entry.getKey()).append("=").append(entry.getValue())
-                .append(", ");
-        }
-        sb.setLength(sb.length() - ", ".length());
-        sb.append("][");
-        for (Map.Entry<GoodsType, Integer> entry : oldStoredGoods.entrySet()) {
-            sb.append(entry.getKey()).append("=").append(entry.getValue())
-                .append(", ");
-        }
-        sb.setLength(sb.length() - ", ".length());
-        sb.append("]]");
-        return sb.toString();
-    }
+    public String getXMLTagName() { return TAG; }
+
+
+    // Override Object
 
     /**
      * {@inheritDoc}
      */
     @Override
-    public String getXMLTagName() { return getXMLElementTagName(); }
-
-    /**
-     * Gets the tag name of the root element representing this object.
-     *
-     * @return "goodsContainer".
-     */
-    public static String getXMLElementTagName() {
-        return "goodsContainer";
+    public String toString() {
+        final String sep = ", ";
+        StringBuilder sb = new StringBuilder(128);
+        sb.append('[').append(getId()).append(" [");
+        // Do not bother to synchronize containers for display
+        forEachMapEntry(storedGoods, e ->
+            sb.append(e.getKey()).append('=').append(e.getValue()).append(sep));
+        sb.setLength(sb.length() - sep.length());
+        sb.append("][");
+        forEachMapEntry(oldStoredGoods, e ->
+            sb.append(e.getKey()).append('=').append(e.getValue()).append(sep));
+        sb.setLength(sb.length() - sep.length());
+        sb.append("]]");
+        return sb.toString();
     }
 }
