@@ -1,5 +1,5 @@
 /**
- *  Copyright (C) 2002-2015   The FreeCol Team
+ *  Copyright (C) 2002-2022   The FreeCol Team
  *
  *  This file is part of FreeCol.
  *
@@ -21,20 +21,25 @@ package net.sf.freecol.common.model;
 
 import java.util.ArrayList;
 import java.util.Comparator;
-import java.util.HashSet;
 import java.util.List;
 import java.util.Set;
+import java.util.function.Function;
+import java.util.stream.Collectors;
 
 import javax.xml.stream.XMLStreamException;
 
 import net.sf.freecol.common.io.FreeColXMLReader;
 import net.sf.freecol.common.io.FreeColXMLWriter;
+import static net.sf.freecol.common.model.Constants.*;
+import static net.sf.freecol.common.util.CollectionUtils.*;
 
 
 /**
  * A type of goods, derived from the specification.
  */
-public final class GoodsType extends FreeColGameObjectType {
+public final class GoodsType extends FreeColSpecObjectType {
+
+    public static final String TAG = "goods-type";
 
     private static final float DEFAULT_PRODUCTION_WEIGHT = 1.0f;
     private static final float DEFAULT_LOW_PRODUCTION_THRESHOLD = 0.0f;
@@ -42,26 +47,8 @@ public final class GoodsType extends FreeColGameObjectType {
 
     /** A comparator to impose a useful order on goods types. */
     public static final Comparator<GoodsType> goodsTypeComparator
-        = new Comparator<GoodsType>() {
-            private int rank(GoodsType g) {
-                return (!g.isStorable() || g.isTradeGoods()) ? -1
-                    : (g.isFoodType()) ? 1
-                    : (g.isNewWorldGoodsType()) ? 2
-                    : (g.isFarmed()) ? 3
-                    : (g.isRawMaterial()) ? 4
-                    : (g.isNewWorldLuxuryType()) ? 5
-                    : (g.isRefined()) ? 6
-                    : -1;
-            }
-
-            @Override
-            public int compare(GoodsType g1, GoodsType g2) {
-                int r1 = rank(g1);
-                int r2 = rank(g2);
-                return (r1 != r2) ? r1 - r2
-                : g1.getId().compareTo(g2.getId());
-            }
-        };
+        = Comparator.comparingInt(GoodsType::getRank)
+            .thenComparing(Function.<FreeColObject>identity());
 
     /** Is this a farmed goods type. */
     private boolean isFarmed;
@@ -147,7 +134,7 @@ public final class GoodsType extends FreeColGameObjectType {
      * Create a new goods type.
      *
      * @param id The object identifier.
-     * @param specification The <code>Specification</code> to refer to.
+     * @param specification The {@code Specification} to refer to.
      */
     public GoodsType(String id, Specification specification) {
         super(id, specification);
@@ -177,16 +164,16 @@ public final class GoodsType extends FreeColGameObjectType {
      *
      * @return True if this is a military goods type.
      */
-    public boolean isMilitaryGoods() {
-        return isMilitary;
+    public boolean getMilitary() {
+        return this.isMilitary;
     }
 
-    // @compat 0.10.x
-    // Needed by Specification fixup010x()
+    // @compat 0.10.x/0.11.x
+    // Needed by Specification.fixSpec()
     public void setMilitary() {
         this.isMilitary = true;
     }
-    // end @compat 0.10.x
+    // end @compat 0.10.x/0.11.x
 
     /**
      * Do warehouse storage limits not apply to this goods type?
@@ -273,7 +260,7 @@ public final class GoodsType extends FreeColGameObjectType {
     /**
      * What type of goods is this goods type stored as?
      *
-     * @return The <code>GoodsType</code> this is stored as, usually itself.
+     * @return The {@code GoodsType} this is stored as, usually itself.
      */
     public GoodsType getStoredAs() {
         return (storedAs == null) ? this : storedAs;
@@ -282,7 +269,7 @@ public final class GoodsType extends FreeColGameObjectType {
     /**
      * Can this goods type be made into somthing?
      *
-     * @return True if this <code>GoodsType</code> can be made into something.
+     * @return True if this {@code GoodsType} can be made into something.
      */
     public boolean isRawMaterial() {
         return makes != null;
@@ -291,7 +278,7 @@ public final class GoodsType extends FreeColGameObjectType {
     /**
      * Is this goods type made from somthing?
      *
-     * @return True if this <code>GoodsType</code> is made from something.
+     * @return True if this {@code GoodsType} is made from something.
      */
     public boolean isRefined() {
         return madeFrom != null;
@@ -300,7 +287,7 @@ public final class GoodsType extends FreeColGameObjectType {
     /**
      * What is this goods made into?
      *
-     * @return The <code>GoodsType</code> this is made into, if anything.
+     * @return The {@code GoodsType} this is made into, if anything.
      */
     public GoodsType getOutputType() {
         return makes;
@@ -309,7 +296,7 @@ public final class GoodsType extends FreeColGameObjectType {
     /**
      * What is this goods type made from?
      *
-     * @return The <code>GoodsType</code> this is made from, if anything.
+     * @return The {@code GoodsType} this is made from, if anything.
      */
     public GoodsType getInputType() {
         return madeFrom;
@@ -368,7 +355,7 @@ public final class GoodsType extends FreeColGameObjectType {
     /**
      * Is this type of goods breedable?
      *
-     * @return True if this <code>GoodsType</code> is breedable.
+     * @return True if this {@code GoodsType} is breedable.
      */
     public boolean isBreedable() {
         return breedingNumber != INFINITY;
@@ -418,7 +405,7 @@ public final class GoodsType extends FreeColGameObjectType {
     /**
      * Gets the i18n-ed name for this goods type.
      *
-     * @return The name of this <code>GoodsType</code>.
+     * @return The name of this {@code GoodsType}.
      */
     public StringTemplate getLabel() {
         return StringTemplate.key(getNameKey());
@@ -436,13 +423,13 @@ public final class GoodsType extends FreeColGameObjectType {
     /**
      * Gets the production chain of the goods type, beginning with a
      * raw material that can not be produced from any other.  The last
-     * element of the production chain is the goods type itself.
+     * member of the production chain is the goods type itself.
      *
      * Note that this does not really handle goods that are stored as
      * something else as is the case for grain etc.
      * FIXME: fix or retire this routine?
      *
-     * @return The production chain of this <code>GoodsType</code> as
+     * @return The production chain of this {@code GoodsType} as
      *     a list.
      */
     public List<GoodsType> getProductionChain() {
@@ -477,16 +464,13 @@ public final class GoodsType extends FreeColGameObjectType {
      * Get all the equivalent goods types, in the sense that they are
      * stored as this type.
      *
-     * @return A set of equivalent <code>GoodsType</code>s, which
+     * @return A set of equivalent {@code GoodsType}s, which
      *     must include this one.
      */
     public Set<GoodsType> getEquivalentTypes() {
-        Set<GoodsType> result = new HashSet<>();
-        for (GoodsType type : getSpecification().getGoodsTypeList()) {
-            if (type == this
-                || type.getStoredAs() == this) result.add(type);
-        }
-        return result;
+        return transform(getSpecification().getGoodsTypeList(),
+                         gt -> gt == this || gt.getStoredAs() == this,
+                         Function.<GoodsType>identity(), Collectors.toSet());
     }
         
     /**
@@ -505,7 +489,7 @@ public final class GoodsType extends FreeColGameObjectType {
      * in the classic ruleset and made-from grain in the freecol
      * ruleset.
      *
-     * @param spec The <code>Specification<code> to operate on.
+     * @param spec The {@code Specification} to operate on.
      */
     public static void setDerivedAttributes(Specification spec) {
         // Reset to default state
@@ -518,12 +502,9 @@ public final class GoodsType extends FreeColGameObjectType {
         List<BuildableType> buildableTypes = new ArrayList<>();
         buildableTypes.addAll(spec.getBuildingTypeList());
         buildableTypes.addAll(spec.getUnitTypeList());
-        buildableTypes.addAll(spec.getRoles());
-        for (BuildableType b : buildableTypes) {
-            for (AbstractGoods ag : b.getRequiredGoods()) {
-                ag.getType().buildingMaterial = true;
-            }
-        }
+        buildableTypes.addAll(spec.getRolesList());
+        forEach(flatten(buildableTypes, BuildableType::getRequiredGoods),
+                ag -> ag.getType().buildingMaterial = true);
 
         // Set makes attribute
         for (GoodsType g : spec.getGoodsTypeList()) {
@@ -531,6 +512,54 @@ public final class GoodsType extends FreeColGameObjectType {
         }
     }
 
+    /**
+     * Impose a general order on the goods types.
+     *
+     * @return An integer useful in comparators.
+     */
+    private int getRank() {
+        return (!isStorable() || isTradeGoods()) ? -1
+            : (isFoodType()) ? 1
+            : (isNewWorldGoodsType()) ? 2
+            : (isFarmed()) ? 3
+            : (isRawMaterial()) ? 4
+            : (isNewWorldLuxuryType()) ? 5
+            : (isRefined()) ? 6
+            : -1;
+    }
+
+
+    // Override FreeColObject
+
+    /**
+     * {@inheritDoc}
+     */
+    @Override
+    public <T extends FreeColObject> boolean copyIn(T other) {
+        GoodsType o = copyInCast(other, GoodsType.class);
+        if (o == null || !super.copyIn(o)) return false;
+        this.isFarmed = o.isFarmed();
+        this.isFood = o.isFoodType();
+        this.ignoreLimit = o.limitIgnored();
+        this.newWorldGoods = o.isNewWorldGoodsType();
+        this.isMilitary = o.getMilitary();
+        this.buildingMaterial = o.isBuildingMaterial();
+        this.tradeGoods = o.isTradeGoods();
+        this.storable = o.isStorable();
+        this.storedAs = o.getStoredAs();
+        this.madeFrom = o.getInputType();
+        this.makes = o.getOutputType();
+        this.initialAmount = o.getInitialAmount();
+        this.initialPrice = o.getInitialSellPrice();
+        this.priceDiff = o.getPriceDifference();
+        this.breedingNumber = o.getBreedingNumber();
+        this.price = o.getPrice();
+        this.productionWeight = o.getProductionWeight();
+        this.lowProductionThreshold = o.getLowProductionThreshold();
+        this.zeroProductionFactor = o.getZeroProductionFactor();
+        return true;
+    }
+
 
     // Serialization
 
@@ -659,12 +688,6 @@ public final class GoodsType extends FreeColGameObjectType {
 
         productionWeight = xr.getAttribute(PRODUCTION_WEIGHT_TAG,
             DEFAULT_PRODUCTION_WEIGHT);
-        // @compat 0.10.7
-        if (!xr.hasAttribute(PRODUCTION_WEIGHT_TAG) && isFarmed) {
-            // Use something a bit more than the default for old games
-            productionWeight = 3.0f;
-        }
-        // end @compat 0.10.7
 
         lowProductionThreshold = xr.getAttribute(LOW_PRODUCTION_THRESHOLD_TAG,
             DEFAULT_LOW_PRODUCTION_THRESHOLD);
@@ -697,15 +720,5 @@ public final class GoodsType extends FreeColGameObjectType {
     /**
      * {@inheritDoc}
      */
-    @Override
-    public String getXMLTagName() { return getXMLElementTagName(); }
-
-    /**
-     * Gets the tag name of the root element representing this object.
-     *
-     * @return "goods-type".
-     */
-    public static String getXMLElementTagName() {
-        return "goods-type";
-    }
+    public String getXMLTagName() { return TAG; }
 }
