@@ -1,5 +1,5 @@
 /**
- *  Copyright (C) 2002-2015   The FreeCol Team
+ *  Copyright (C) 2002-2022   The FreeCol Team
  *
  *  This file is part of FreeCol.
  *
@@ -19,6 +19,7 @@
 
 package net.sf.freecol.common.model;
 
+import java.util.Comparator;
 import java.util.List;
 
 import javax.xml.stream.XMLStreamException;
@@ -27,17 +28,27 @@ import net.sf.freecol.common.io.FreeColXMLReader;
 import net.sf.freecol.common.io.FreeColXMLWriter;
 import net.sf.freecol.common.util.Utils;
 
-import org.w3c.dom.Element;
-
 
 /**
- * The <code>Modifier</code> class encapsulates a bonus or penalty
+ * The {@code Modifier} class encapsulates a bonus or penalty
  * that can be applied to any action within the game, most obviously
  * combat. The Modifier may be applicable only to certain Objects
- * specified by means of <code>Scope</code> objects.
+ * specified by means of {@code Scope} objects.
  */
 public class Modifier extends Feature {
 
+    public static final String TAG = "modifier";
+    
+    /**
+     * Comparator to sort by ascending modifier index, then type, then
+     * source, then FCO order.
+     */
+    public static final Comparator<Modifier> ascendingModifierIndexComparator
+        = Comparator.<Modifier>comparingInt(Modifier::getModifierIndex)
+            .thenComparingInt(m -> m.getType().ordinal())
+            .thenComparing(FreeColObject.fcoComparator)
+            .thenComparing(Modifier::getSource, FreeColObject.fcoComparator);
+    
     public static final String AMPHIBIOUS_ATTACK
         = "model.modifier.amphibiousAttack";
     public static final String ARTILLERY_AGAINST_RAID
@@ -125,9 +136,11 @@ public class Modifier extends Feature {
 
     public static final int DEFAULT_MODIFIER_INDEX = 0;
 
-    // @compat 0.10.x
-    // These are now attached to modifiers in the spec, but
-    // Specification.fixup010x() still needs them for now.
+    // These are now attached to modifiers in the spec and mostly
+    // should not be needed any more.  However the dynamic modifiers
+    // (e.g. disaster, goods party related) do still need to be built
+    // explicitly with the right index.  So we might as well leave
+    // them all visible here for clarity.
     public static final int RESOURCE_PRODUCTION_INDEX = 10;
     public static final int COLONY_PRODUCTION_INDEX = 20;
     public static final int EXPERT_PRODUCTION_INDEX = 30;
@@ -138,7 +151,6 @@ public class Modifier extends Feature {
     public static final int NATION_PRODUCTION_INDEX = 80;
     public static final int PARTY_PRODUCTION_INDEX = 90;
     public static final int DISASTER_PRODUCTION_INDEX = 100;
-    // end @compat 0.10.x
     public static final int DEFAULT_PRODUCTION_INDEX = 100;
 
     // Specific combat indicies
@@ -175,110 +187,84 @@ public class Modifier extends Feature {
 
     /**
      * Deliberately empty constructor.
-     */
-    protected Modifier() {}
-
-    /**
-     * Use this only for reading from DOM Elements.
      *
-     * @param specification The <code>Specification</code> to refer to.
+     * @param specification The defining {@code Specification}.
      */
     public Modifier(Specification specification) {
-        setSpecification(specification);
+        super(specification);
     }
 
     /**
-     * Creates a new <code>Modifier</code> instance.
+     * Creates a new {@code Modifier} instance.
+     *
+     * @param id The object identifier.
+     * @param value The modifier value.
+     * @param type The type of the modifier.
+     * @param source The source {@code FreeColObject}.
+     * @param modifierIndex The modifier index.
+     */
+    public Modifier(String id, float value, ModifierType type,
+                    FreeColObject source, int modifierIndex) {
+        this((source == null) ? null : source.getSpecification());
+
+        setId(id);
+        setValue(value);
+        setType(type);
+        setSource(source);
+        setModifierIndex(modifierIndex);
+    }
+
+    /**
+     * Creates a new {@code Modifier} instance.
+     *
+     * @param id The object identifier.
+     * @param value The modifier value.
+     * @param type The type of the modifier.
+     * @param source The source {@code FreeColObject}.
+     */
+    public Modifier(String id, float value, ModifierType type,
+                    FreeColObject source) {
+        this(id, value, type, source, DEFAULT_MODIFIER_INDEX);
+    }
+
+    /**
+     * Creates a new {@code Modifier} instance.
      *
      * @param id The object identifier.
      * @param value The modifier value.
      * @param type The type of the modifier.
      */
     public Modifier(String id, float value, ModifierType type) {
-        setId(id);
-        setType(type);
-        setValue(value);
+        this(id, value, type, null);
     }
 
     /**
-     * Creates a new <code>Modifier</code> instance.
+     * Creates a new {@code Modifier} instance.
      *
-     * @param id The object identifier.
-     * @param value The modifier value.
-     * @param type The type of the modifier.
-     * @param source The source <code>FreeColObject</code>.
-     */
-    public Modifier(String id, float value, ModifierType type,
-                    FreeColObject source) {
-        this(id, value, type);
-        setSource(source);
-    }
-
-    /**
-     * Creates a new <code>Modifier</code> instance.
-     *
-     * @param id The object identifier.
-     * @param value The modifier value.
-     * @param type The type of the modifier.
-     * @param source The source <code>FreeColObject</code>.
-     * @param modifierIndex The modifier index.
-     */
-    public Modifier(String id, float value, ModifierType type,
-                    FreeColObject source, int modifierIndex) {
-        this(id, value, type, source);
-        setModifierIndex(modifierIndex);
-    }
-
-    /**
-     * Creates a new <code>Modifier</code> instance from another.
-     *
-     * @param template A <code>Modifier</code> to copy.
-     */
-    public Modifier(Modifier template) {
-        copyFrom(template);
-        setType(template.getType());
-        setValue(template.getValue());
-        setIncrementType(template.getIncrementType());
-        setIncrement(template.getIncrement());
-        setModifierIndex(template.getModifierIndex());
-    }
-
-    /**
-     * Creates a new <code>Modifier</code> instance from another with
-     * an identifier override.
-     *
-     * @param id The object identifier.
-     * @param template A <code>Modifier</code> to copy.
-     */
-    public Modifier(String id, Modifier template) {
-        this(template);
-        setId(id);
-    }
-
-    /**
-     * Creates a new <code>Modifier</code> instance.
-     *
-     * @param xr The <code>FreeColXMLReader</code> to read from.
-     * @param specification The <code>Specification</code> to refer to.
+     * @param xr The {@code FreeColXMLReader} to read from.
+     * @param specification The {@code Specification} to refer to.
      * @exception XMLStreamException if there is an error reading the
      *     stream.
      */
     public Modifier(FreeColXMLReader xr, Specification specification) throws XMLStreamException {
-        setSpecification(specification);
+        this(specification);
+
         readFromXML(xr);
     }
 
-    /**
-     * Creates a new <code>Modifier</code> instance.
-     *
-     * @param e The <code>Element</code> to read from.
-     * @param specification The <code>Specification</code> to refer to.
-     */
-    public Modifier(Element e, Specification specification) {
-        setSpecification(specification);
-        readFromXMLElement(e);
-    }
 
+    /**
+     * Make a copy of a modifier.
+     *
+     * @param modifier The {@code Modifier} to copy.
+     * @return A copy of the modifier.
+     */
+    public static Modifier makeModifier(Modifier modifier) {
+        Modifier ret = new Modifier(modifier.getSpecification());
+        ret.setId(modifier.getId());
+        ret.copyIn(modifier);
+        return ret;
+    }
 
     /**
      * Makes a timed modifier (one with start/end turn and increment)
@@ -289,13 +275,14 @@ public class Modifier extends Feature {
      * "model.modifier.colonyGoodsParty".
      *
      * @param id The id for the new modifier.
-     * @param template A template <code>Modifier</code> with increment.
-     * @param start The starting <code>Turn</code>.
+     * @param template A template {@code Modifier} with increment.
+     * @param start The starting {@code Turn}.
      * @return A new timed modifier.
      */
     public static Modifier makeTimedModifier(String id, Modifier template,
                                              Turn start) {
-        Modifier modifier = new Modifier(id, template);
+        Modifier modifier = makeModifier(template);
+        modifier.setId(id);
         float inc = template.getIncrement();
         int duration = template.getDuration();
         modifier.setTemporary(template.isTemporary());
@@ -311,7 +298,7 @@ public class Modifier extends Feature {
     /**
      * Get the modifier type.
      *
-     * @return The <code>ModifierType</code>.
+     * @return The {@code ModifierType}.
      */
     public final ModifierType getType() {
         return modifierType;
@@ -320,10 +307,33 @@ public class Modifier extends Feature {
     /**
      * Set the modifier type.
      *
-     * @param modifierType The new <code>ModifierType</code> value.
+     * @param modifierType The new {@code ModifierType} value.
+     * @return This modifier.
      */
-    public final void setType(final ModifierType modifierType) {
+    public final Modifier setType(final ModifierType modifierType) {
         this.modifierType = modifierType;
+        return this;
+    }
+
+    /**
+     * Is this a "known" value.
+     *
+     * @param value The value to test.
+     * @return True if the value is known.
+     */
+    public static boolean isFloatKnown(float value) {
+        return Float.compare(value, UNKNOWN) != 0;
+    }
+
+    /**
+     * Does this modifier have a known value?
+     *
+     * That is, is it not set to the UNKNOWN value.
+     *
+     * @return True if the value is known.
+     */
+    public boolean isKnown() {
+        return Modifier.isFloatKnown(this.value);
     }
 
     /**
@@ -332,13 +342,13 @@ public class Modifier extends Feature {
      * @return The modifier value.
      */
     public final float getValue() {
-        return value;
+        return this.value;
     }
 
     /**
      * Get the value the modifier during the given Turn.
      *
-     * @param turn The <code>Turn</code> to check.
+     * @param turn The {@code Turn} to check.
      * @return The turn-dependent modifier value.
      */
     public final float getValue(Turn turn) {
@@ -346,9 +356,9 @@ public class Modifier extends Feature {
             if (hasIncrement()) {
                 float f = (turn.getNumber() - getFirstTurn().getNumber())
                     * increment;
-                return apply(value, f, incrementType);
+                return apply(this.value, f, incrementType);
             } else {
-                return value;
+                return this.value;
             }
         } else {
             return 0;
@@ -359,9 +369,11 @@ public class Modifier extends Feature {
      * Set the modifier value.
      *
      * @param value The new value.
+     * @return This modifier.
      */
-    public final void setValue(final float value) {
+    public final Modifier setValue(final float value) {
         this.value = value;
+        return this;
     }
 
     /**
@@ -376,7 +388,7 @@ public class Modifier extends Feature {
     /**
      * Get the increment type.
      *
-     * @return The increment <code>ModifierType</code>.
+     * @return The increment {@code ModifierType}.
      */
     public final ModifierType getIncrementType() {
         return incrementType;
@@ -385,10 +397,12 @@ public class Modifier extends Feature {
     /**
      * Set the increment type.
      *
-     * @param incrementType The new increment <code>ModifierType</code>.
+     * @param incrementType The new increment {@code ModifierType}.
+     * @return This modifier.
      */
-    public final void setIncrementType(final ModifierType incrementType) {
+    public final Modifier setIncrementType(final ModifierType incrementType) {
         this.incrementType = incrementType;
+        return this;
     }
 
     /**
@@ -404,31 +418,35 @@ public class Modifier extends Feature {
      * Set the increment value.
      *
      * @param increment The new value.
+     * @return This modifier.
      */
-    public final void setIncrement(final float increment) {
+    public final Modifier setIncrement(final float increment) {
         this.increment = increment;
+        return this;
     }
 
     /**
      * Set the whole increment.
      *
-     * @param incrementType The new <code>ModifierType</code>.
+     * @param incrementType The new {@code ModifierType}.
      * @param increment The new increment value.
-     * @param firstTurn The first <code>Turn</code> the increment is
+     * @param firstTurn The first {@code Turn} the increment is
      *     active.
-     * @param lastTurn The last <code>Turn</code> the increment is
+     * @param lastTurn The last {@code Turn} the increment is
      *     active.
+     * @return This modifier.
      */
-    public final void setIncrement(final ModifierType incrementType,
-                                   final float increment,
-                                   Turn firstTurn, Turn lastTurn) {
+    public final Modifier setIncrement(final ModifierType incrementType,
+                                       final float increment,
+                                       Turn firstTurn, Turn lastTurn) {
         if (firstTurn == null) {
-            throw new IllegalArgumentException("Null firstTurn");
+            throw new RuntimeException("Null firstTurn: " + this);
         }
         this.incrementType = incrementType;
         this.increment = increment;
         setFirstTurn(firstTurn);
         setLastTurn(lastTurn);
+        return this;
     }
 
     /**
@@ -444,18 +462,20 @@ public class Modifier extends Feature {
      * Set the modifier index.
      *
      * @param modifierIndex The new modifier index value.
+     * @return This modifier.
      */
-    public final void setModifierIndex(final int modifierIndex) {
+    public final Modifier setModifierIndex(final int modifierIndex) {
         this.modifierIndex = modifierIndex;
+        return this;
     }
 
     /**
      * Applies the given value to the given base value, depending on
      * the type of this Modifier.
      *
-     * @param base a <code>float</code> value
-     * @param value a <code>float</code> value
-     * @return a <code>float</code> value
+     * @param base a {@code float} value
+     * @param value a {@code float} value
+     * @return a {@code float} value
      */
     public float apply(float base, float value) {
         return apply(base, value, getType());
@@ -467,7 +487,7 @@ public class Modifier extends Feature {
      *
      * @param base The base value.
      * @param value The modifier value.
-     * @param type The <code>ModifierType</code>.
+     * @param type The {@code ModifierType}.
      * @return The result of applying the value to the base.
      */
     private float apply(float base, float value, ModifierType type) {
@@ -487,8 +507,8 @@ public class Modifier extends Feature {
      * Applies this Modifier to a number. This method does not take
      * scopes, increments or time limits into account.
      *
-     * @param number a <code>float</code> value
-     * @return a <code>float</code> value
+     * @param number a {@code float} value
+     * @return a {@code float} value
      */
     public float applyTo(float number) {
         return apply(number, value);
@@ -499,7 +519,7 @@ public class Modifier extends Feature {
      * increments into account.
      *
      * @param number The number to modify.
-     * @param turn The <code>Turn</code> to evaluate increments in.
+     * @param turn The {@code Turn} to evaluate increments in.
      * @return The modified number.
      */
     public float applyTo(float number, Turn turn) {
@@ -507,19 +527,6 @@ public class Modifier extends Feature {
             : apply(number, getValue(turn), getType());
     }
 
-    // @compat 0.10.7
-    /**
-     * Helper for scope fixups.
-     *
-     * @return True if a negated person scope was added.
-     */
-    public boolean requireNegatedPersonScope() {
-        if (hasScope()) return false;
-        addScope(Scope.makeNegatedPersonScope());
-        return true;
-    }
-    // end @compat 0.10.7
-
 
     // Override FreeColObject
 
@@ -527,55 +534,15 @@ public class Modifier extends Feature {
      * {@inheritDoc}
      */
     @Override
-    public int compareTo(FreeColObject other) {
-        int cmp = 0;
-        if (other instanceof Modifier) {
-            Modifier modifier = (Modifier)other;
-            cmp = modifierIndex - modifier.modifierIndex;
-            if (cmp == 0) {
-                cmp = modifierType.ordinal() - modifier.modifierType.ordinal();
-            }
-            if (cmp == 0) {
-                cmp = FreeColObject.compareIds(getSource(), 
-                                               modifier.getSource());
-            }
-        }
-        if (cmp == 0) cmp = super.compareTo(other);
-        return cmp;
-    }
-
-
-    // Override Object
-
-    /**
-     * {@inheritDoc}
-     */
-    @Override
-    public boolean equals(Object o) {
-        if (o == this) return true;
-        if (o instanceof Modifier) { 
-            Modifier m = (Modifier)o;
-            return Utils.equals(this.modifierType, m.modifierType)
-                && this.value == m.value
-                && this.increment == m.increment
-                && Utils.equals(this.incrementType, m.incrementType)
-                && this.modifierIndex == m.modifierIndex
-                && super.equals(o);
-        }
-        return false;
-    }
-
-    /**
-     * {@inheritDoc}
-     */
-    @Override
-    public int hashCode() {
-        int hash = super.hashCode();
-        hash = 31 * hash + Float.floatToIntBits(value);
-        hash = 31 * hash + Float.floatToIntBits(increment);
-        hash = 31 * hash + Utils.hashCode(modifierType);
-        hash = 31 * hash + Utils.hashCode(incrementType);
-        return 31 * hash + modifierIndex;
+    public <T extends FreeColObject> boolean copyIn(T other) {
+        Modifier o = copyInCast(other, Modifier.class);
+        if (o == null || !super.copyIn(o)) return false;
+        this.setType(o.getType());
+        this.setValue(o.getValue());
+        this.setIncrement(o.getIncrement());
+        this.setIncrementType(o.getIncrementType());
+        this.setModifierIndex(o.getModifierIndex());
+        return true;
     }
 
 
@@ -645,43 +612,67 @@ public class Modifier extends Feature {
         modifierIndex = xr.getAttribute(INDEX_TAG, DEFAULT_MODIFIER_INDEX);
     }
 
+    /**
+     * {@inheritDoc}
+     */
+    public String getXMLTagName() { return TAG; }
+
+
+    // Override Object
+
     /**
      * {@inheritDoc}
      */
     @Override
-    public String toString() {
-        StringBuilder sb = new StringBuilder(64);
-        sb.append("[Modifier ").append(getId());
-        if (getSource() != null) {
-            sb.append(" (" + getSource().getId() + ")");
+    public boolean equals(Object o) {
+        if (o == this) return true;
+        if (o instanceof Modifier) { 
+            Modifier other = (Modifier)o;
+            return Float.compare(this.value, other.value) == 0
+                && Float.compare(this.increment, other.increment) == 0
+                && this.modifierIndex == other.modifierIndex
+                && Utils.equals(this.modifierType, other.modifierType)
+                && Utils.equals(this.incrementType, other.incrementType)
+                && super.equals(other);
         }
-        sb.append(" ").append(modifierType)
-            .append(" ").append(value);
-        if (modifierIndex >= DEFAULT_MODIFIER_INDEX) {
-            sb.append(" index=").append(modifierIndex);
-        }
-        List<Scope> scopes = getScopes();
-        if (!scopes.isEmpty()) {
-            sb.append(" [");
-            for (Scope s : scopes) sb.append(" ").append(s);
-            sb.append(" ]");
-        }
-        sb.append("]");
-        return sb.toString();
+        return false;
     }
 
     /**
      * {@inheritDoc}
      */
     @Override
-    public String getXMLTagName() { return getXMLElementTagName(); }
+    public int hashCode() {
+        int hash = super.hashCode();
+        hash = 31 * hash + Float.floatToIntBits(value);
+        hash = 31 * hash + Float.floatToIntBits(increment);
+        hash = 31 * hash + Utils.hashCode(modifierType);
+        hash = 31 * hash + Utils.hashCode(incrementType);
+        return 31 * hash + modifierIndex;
+    }
 
     /**
-     * Gets the XML tag name for this element.
-     *
-     * @return "modifier".
+     * {@inheritDoc}
      */
-    public static String getXMLElementTagName() {
-        return "modifier";
+    @Override
+    public String toString() {
+        StringBuilder sb = new StringBuilder(64);
+        sb.append("[Modifier ").append(getId());
+        if (getSource() != null) {
+            sb.append(" (").append(getSource().getId()).append(')');
+        }
+        sb.append(' ').append(modifierType)
+            .append(' ').append(value);
+        if (modifierIndex >= DEFAULT_MODIFIER_INDEX) {
+            sb.append(" index=").append(modifierIndex);
+        }
+        List<Scope> scopes = getScopeList();
+        if (!scopes.isEmpty()) {
+            sb.append(" [");
+            for (Scope s : scopes) sb.append(' ').append(s);
+            sb.append(" ]");
+        }
+        sb.append(']');
+        return sb.toString();
     }
 }
