@@ -1,5 +1,5 @@
 /**
- *  Copyright (C) 2002-2015   The FreeCol Team
+ *  Copyright (C) 2002-2022   The FreeCol Team
  *
  *  This file is part of FreeCol.
  *
@@ -19,18 +19,31 @@
 
 package net.sf.freecol.common.model;
 
+import static net.sf.freecol.common.model.Constants.INFINITY;
+import static net.sf.freecol.common.util.CollectionUtils.all;
+import static net.sf.freecol.common.util.CollectionUtils.any;
+import static net.sf.freecol.common.util.CollectionUtils.find;
+import static net.sf.freecol.common.util.CollectionUtils.minimize;
+import static net.sf.freecol.common.util.CollectionUtils.sum;
+
 import java.util.ArrayList;
+import java.util.Collection;
+import java.util.Collections;
 import java.util.HashSet;
 import java.util.List;
 import java.util.Random;
 import java.util.Set;
+import java.util.function.Predicate;
 import java.util.logging.Logger;
 
 import javax.xml.stream.XMLStreamException;
 
+import net.sf.freecol.client.gui.ImageLibrary;
+import net.sf.freecol.common.FreeColException;
 import net.sf.freecol.common.io.FreeColXMLReader;
 import net.sf.freecol.common.io.FreeColXMLWriter;
-import static net.sf.freecol.common.util.CollectionUtils.*;
+import net.sf.freecol.common.model.Constants.IntegrityType;
+import net.sf.freecol.common.util.LogBuilder;
 
 
 /**
@@ -44,18 +57,19 @@ public abstract class Settlement extends GoodsLocation
 
     public static final int FOOD_PER_COLONIST = 200;
 
-    /** The <code>Player</code> owning this <code>Settlement</code>. */
+
+    /** The type of settlement. */
+    private SettlementType type = null;
+
+    /** The {@code Player} owning this {@code Settlement}. */
     protected Player owner;
 
     /** The name of the Settlement. */
     private String name;
 
-    /** The <code>Tile</code> where this <code>Settlement</code> is located. */
+    /** The {@code Tile} where this {@code Settlement} is located. */
     protected Tile tile;
 
-    /** The type of settlement. */
-    private SettlementType type = null;
-
     /** The tiles this settlement owns. */
     private final Set<Tile> ownedTiles = new HashSet<>();
 
@@ -64,14 +78,14 @@ public abstract class Settlement extends GoodsLocation
 
 
     /**
-     * Create a new <code>Settlement</code>.
+     * Create a new {@code Settlement}.
      *
-     * @param game The enclosing <code>Game</code>.
-     * @param owner The owning <code>Player</code>.
+     * @param game The enclosing {@code Game}.
+     * @param owner The owning {@code Player}.
      * @param name The settlement name.
-     * @param tile The containing <code>Tile</code>.
+     * @param tile The containing {@code Tile}.
      */
-    public Settlement(Game game, Player owner, String name, Tile tile) {
+    protected Settlement(Game game, Player owner, String name, Tile tile) {
         super(game);
 
         this.owner = owner;
@@ -81,12 +95,11 @@ public abstract class Settlement extends GoodsLocation
     }
 
     /**
-     * Initiates a new <code>Settlement</code> with the given identifier.
-     * The object should later be initialized by calling either
-     * {@link #readFromXML(FreeColXMLReader)} or
-     * {@link #readFromXMLElement(Element)}.
+     * Initiates a new {@code Settlement} with the given identifier.
      *
-     * @param game The enclosing <code>Game</code>.
+     * The object should be initialized later.
+     *
+     * @param game The enclosing {@code Game}.
      * @param id The object identifier.
      */
     public Settlement(Game game, String id) {
@@ -106,12 +119,30 @@ public abstract class Settlement extends GoodsLocation
     /**
      * Set the settlement type.
      *
-     * @param newType The new <code>SettlementType</code>.
+     * @param newType The new {@code SettlementType}.
      */
     public void setType(final SettlementType newType) {
         this.type = newType;
     }
 
+    /**
+     * Gets the immigration points.
+     *
+     * @return The current immigration.
+     */
+    public int getImmigration() {
+        return 0;
+    }
+
+    /**
+     * Gets the liberty points.
+     *
+     * @return The current liberty.
+     */
+    public int getLiberty() {
+        return 0;
+    }
+
     /**
      * Is this settlement landlocked?
      *
@@ -124,7 +155,7 @@ public abstract class Settlement extends GoodsLocation
     /**
      * Change the settlement type, setting the consequent features.
      *
-     * @param newType The new <code>SettlementType</code>.
+     * @param newType The new {@code SettlementType}.
      */
     private final void changeType(final SettlementType newType) {
         if (type != null) removeFeatures(type);
@@ -150,12 +181,22 @@ public abstract class Settlement extends GoodsLocation
     /**
      * Get the tiles this settlement owns.
      *
-     * @return A list of tiles.
+     * @return A set of tiles.
      */
     public Set<Tile> getOwnedTiles() {
         return new HashSet<>(ownedTiles);
     }
 
+    /**
+     * Set the owned tiles set.
+     *
+     * @param ownedTiles The new set of owned {@code Tile}s.
+     */
+    protected void setOwnedTiles(Collection<Tile> ownedTiles) {
+        this.ownedTiles.clear();
+        this.ownedTiles.addAll(ownedTiles);
+    }
+
     /**
      * Adds a tile to this settlement.
      *
@@ -165,23 +206,23 @@ public abstract class Settlement extends GoodsLocation
      * the map definition.  So we just accumulate and defend against
      * duplicates.
      *
-     * @param tile The <code>Tile</code> to add.
+     * @param tile The {@code Tile} to add.
      */
     public void addTile(Tile tile) {
-        if (!ownedTiles.contains(tile)) ownedTiles.add(tile);
+        ownedTiles.add(tile);
     }
 
     /**
      * Removes a tile from this settlement.
      *
-     * @param tile The <code>Tile</code> to remove.
+     * @param tile The {@code Tile} to remove.
      */
     public void removeTile(Tile tile) {
         ownedTiles.remove(tile);
     }
 
     /**
-     * Gets the radius of what the <code>Settlement</code> considers
+     * Gets the radius of what the {@code Settlement} considers
      * as it's own land.
      *
      * @return Settlement radius
@@ -196,14 +237,14 @@ public abstract class Settlement extends GoodsLocation
      * @return The line of sight value.
      */
     public int getLineOfSight() {
-        return (int)applyModifiers((float)getType().getVisibleRadius(),
+        return (int)apply((float)getType().getVisibleRadius(),
             getGame().getTurn(), Modifier.LINE_OF_SIGHT_BONUS);
     }
 
     /**
      * Gets an amount of plunder when this settlement is taken.
      *
-     * @param attacker The <code>Unit</code> that takes the settlement.
+     * @param attacker The {@code Unit} that takes the settlement.
      * @param random A pseudo-random number source.
      * @return An amount of gold plundered.
      */
@@ -213,6 +254,17 @@ public abstract class Settlement extends GoodsLocation
             : range.getAmount("Plunder " + getName(), random, false);
     }
 
+    /**
+     * Get the tiles visible from this settlement.
+     *
+     * @return A set of visible tiles.
+     */
+    public Set<Tile> getVisibleTileSet() {
+        final Tile tile = getTile();
+        return (tile == null) ? Collections.<Tile>emptySet()
+            : new HashSet<Tile>(tile.getSurroundingTiles(0, getLineOfSight()));
+    }
+
     /**
      * Put a prepared settlement onto the map.
      *
@@ -258,28 +310,30 @@ public abstract class Settlement extends GoodsLocation
         settlementTile.setSettlement(null);//-vis(owner),-til
         settlementTile.changeOwnership(null, null);//-til
         TileImprovement road = settlementTile.getRoad();
-        if (road != null && road.isVirtual()) {
+        if (road != null && road.getVirtual()) {
             settlementTile.removeRoad();//-til
         }
     }
 
     /**
-     * Change the owner of this <code>Settlement</code>.
+     * Change the owner of this {@code Settlement}.
      *
      * Does not fix up the units!  That is handled in the server.
      *
      * -vis: Changes visibility.
      * -til: Changes tile appearance.
      *
-     * @param newOwner The <code>Player</code> that shall own this
-     *            <code>Settlement</code>.
+     * @param newOwner The {@code Player} that shall own this
+     *            {@code Settlement}.
      * @see #getOwner
      */
     public void changeOwner(Player newOwner) {
         final Player oldOwner = this.owner;
         if (newOwner.isIndian() != oldOwner.isIndian()) {
-            throw new IllegalArgumentException("Can not transfer settlements between native and European players.");
+            throw new RuntimeException("Can not transfer settlements between native and European players: " + oldOwner + " -> " + newOwner);
         }
+        getGame().notifyOwnerChanged(this, oldOwner, newOwner);
+
         setOwner(newOwner);//-til,-vis
 
         getGame().checkOwners(this, oldOwner);
@@ -287,8 +341,6 @@ public abstract class Settlement extends GoodsLocation
         for (Tile t : getOwnedTiles()) {
             t.changeOwnership(newOwner, this);//-til
         }
-
-        getGame().notifyOwnerChanged(this, oldOwner, newOwner);
     }
 
     /**
@@ -300,7 +352,7 @@ public abstract class Settlement extends GoodsLocation
      */
     public boolean isConnectedPort() {
         return any(getTile().getSurroundingTiles(1, 1),
-            t -> !t.isLand() && t.isHighSeasConnected());
+                   t -> !t.isLand() && t.isHighSeasConnected());
     }
 
     /**
@@ -310,39 +362,33 @@ public abstract class Settlement extends GoodsLocation
      * @return A high seas count, INFINITY if not connected.
      */
     public int getHighSeasCount() {
-        int best = INFINITY;
-        for (Tile t : getTile().getSurroundingTiles(1)) {
-            if (t.isLand() || t.getHighSeasCount() < 0) continue;
-            if (best > t.getHighSeasCount()) {
-                best = t.getHighSeasCount();
-            }
-        }
-        return best;
+        Tile best = minimize(getTile().getSurroundingTiles(1, 1),
+                             Tile.isSeaTile, Tile.highSeasComparator);
+        return (best == null) ? INFINITY : best.getHighSeasCount();
     }
 
     /**
      * Returns the number of goods of a given type used by the settlement
      * each turn.
      *
-     * @param goodsType a <code>GoodsType</code> value
-     * @return an <code>int</code> value
+     * @param goodsType a {@code GoodsType} value
+     * @return an {@code int} value
      */
-    public int getConsumptionOf(GoodsType goodsType) {
-        return Math.max(0, getUnitList().stream()
-                .mapToInt(u -> u.getType().getConsumptionOf(goodsType)).sum());
+    protected int getConsumptionOf(GoodsType goodsType) {
+        return Math.max(0, sum(getUnits(),
+                               u -> u.getType().getConsumptionOf(goodsType)));
     }
 
     /**
      * Returns the number of goods of all given types used by the
      * settlement each turn.
      *
-     * @param goodsTypes <code>GoodsType</code> values
-     * @return an <code>int</code> value
+     * @param goodsTypes {@code GoodsType} values
+     * @return an {@code int} value
      */
-    public int getConsumptionOf(List<GoodsType> goodsTypes) {
+    protected int getConsumptionOf(List<GoodsType> goodsTypes) {
         return (goodsTypes == null) ? 0
-            : goodsTypes.stream()
-                .mapToInt(gt -> getConsumptionOf(gt)).sum();
+            : sum(goodsTypes, gt -> getConsumptionOf(gt));
     }
 
     /**
@@ -359,16 +405,14 @@ public abstract class Settlement extends GoodsLocation
      * equipment.  Unlike priceGoods, this takes goods "reserved" for
      * other purposes into account (e.g. breeding).
      *
-     * @param goods A list of <code>AbstractGoods</code>
+     * @param goods A list of {@code AbstractGoods}
      * @return True if the settlement can provide the equipment.
      */
-    public boolean canProvideGoods(List<AbstractGoods> goods) {
+    protected boolean canProvideGoods(List<AbstractGoods> goods) {
         return all(goods, ag -> {
                 int available = getGoodsCount(ag.getType());
                 int breedingNumber = ag.getType().getBreedingNumber();
-                if (breedingNumber != GoodsType.INFINITY) {
-                    available -= breedingNumber;
-                }
+                if (breedingNumber != INFINITY) available -= breedingNumber;
                 return available >= ag.getAmount();
             });
     }
@@ -399,22 +443,35 @@ public abstract class Settlement extends GoodsLocation
      * Can this settlement provide the goods to improve a given unit's
      * role?
      *
-     * @param unit The <code>Unit</code> to check.
-     * @return The <code>Role</code> that this settlement could provide.
+     * @param unit The {@code Unit} to check.
+     * @return The {@code Role} that this settlement could provide.
      */
     public Role canImproveUnitMilitaryRole(Unit unit) {
         final Specification spec = getSpecification();
         final Role role = unit.getRole();
 
         // Get the military roles that are superior to the current role
-        List<Role> military = spec.getMilitaryRoles();
+        List<Role> military = spec.getMilitaryRolesList();
         int index = military.indexOf(role);
         if (index >= 0) military = military.subList(0, index);
 
         // To succeed, there must exist an available role for the unit
         // where the extra equipment for the role is present.
         return find(unit.getAvailableRoles(military),
-            r -> canProvideGoods(unit.getGoodsDifference(r, 1)));
+                    r -> canProvideGoods(unit.getGoodsDifference(r, 1)));
+    }
+
+    /**
+     * Get all the units present here.  That is, not just the units in the
+     * settlement but also the units on the tile.
+     *
+     * @return A list of {@code Unit}s.
+     */
+    public List<Unit> getAllUnitsList() {
+        List<Unit> units = getUnitList();
+        if (units.isEmpty()) return getTile().getUnitList();
+        units.addAll(getTile().getUnitList());
+        return units;
     }
 
 
@@ -435,7 +492,7 @@ public abstract class Settlement extends GoodsLocation
      */
     @Override
     public FeatureContainer getFeatureContainer() {
-        return featureContainer;
+        return this.featureContainer;
     }
 
     /**
@@ -459,7 +516,7 @@ public abstract class Settlement extends GoodsLocation
     // Interface Nameable
 
     /**
-     * Get the name of this <code>Settlement</code>.
+     * Get the name of this {@code Settlement}.
      *
      * @return The settlement name.
      */
@@ -469,7 +526,7 @@ public abstract class Settlement extends GoodsLocation
     }
 
     /**
-     * Sets the name of this <code>Settlement</code>.
+     * Sets the name of this {@code Settlement}.
      *
      * -til: Changes the tile appearance.
      *
@@ -543,7 +600,15 @@ public abstract class Settlement extends GoodsLocation
      */
     @Override
     public final int getRank() {
-        return Location.getRank(getTile());
+        return Location.rankOf(getTile());
+    }
+
+    /**
+     * {@inheritDoc}
+     */
+    @Override
+    public String getLocationImageKey() {
+        return ImageLibrary.getSettlementKey(this);
     }
 
 
@@ -577,10 +642,14 @@ public abstract class Settlement extends GoodsLocation
      * {@inheritDoc}
      */
     @Override
-    public int priceGoods(List<AbstractGoods> goods) {
-        return (any(goods,
-                ag -> getGoodsCount(ag.getType()) < ag.getAmount())) ? -1
-            : 0;
+    public int priceGoods(List<AbstractGoods> goods) throws FreeColException {
+        final Predicate<AbstractGoods> pred = ag ->
+            getGoodsCount(ag.getType()) < ag.getAmount();
+        AbstractGoods missing = find(goods, pred);
+        if (missing != null) {
+            throw new FreeColException("Goods missing: " + missing);
+        }
+        return 0;
     }
 
     /**
@@ -591,13 +660,17 @@ public abstract class Settlement extends GoodsLocation
         if (!unit.roleIsAvailable(role)) return false;
 
         // Get the change in goods
-        List<AbstractGoods> required = unit.getGoodsDifference(role, roleCount);
+        List<AbstractGoods> req = unit.getGoodsDifference(role, roleCount);
 
         // Check if the required goods are available
-        if (priceGoods(required) < 0) return false;
+        try {
+            priceGoods(req);
+        } catch (FreeColException fce) {
+            return false;
+        }
 
         // Make the change
-        for (AbstractGoods ag : required) {
+        for (AbstractGoods ag : req) {
             addGoods(ag.getType(), -ag.getAmount());
         }
 
@@ -617,20 +690,13 @@ public abstract class Settlement extends GoodsLocation
     // Settlement routines to be implemented by subclasses.
 
     /**
-     * Gets an image key for this settlement.
+     * Gets the {@code Unit} that is currently defending this
+     * {@code Settlement}.
      *
-     * @return An image key
-     */
-    public abstract String getImageKey();
-
-    /**
-     * Gets the <code>Unit</code> that is currently defending this
-     * <code>Settlement</code>.
-     *
-     * @param attacker The <code>Unit</code> that is attacking this
-     *     <code>Settlement</code>.
-     * @return The <code>Unit</code> that has been chosen to defend
-     *     this <code>Settlement</code>.
+     * @param attacker The {@code Unit} that is attacking this
+     *     {@code Settlement}.
+     * @return The {@code Unit} that has been chosen to defend
+     *     this {@code Settlement}.
      */
     public abstract Unit getDefendingUnit(Unit attacker);
 
@@ -652,8 +718,8 @@ public abstract class Settlement extends GoodsLocation
     /**
      * Gets the range of gold plunderable when this settlement is captured.
      *
-     * @param attacker The <code>Unit</code> that takes the settlement.
-     * @return A <code>RandomRange</code> encapsulating the range of plunder
+     * @param attacker The {@code Unit} that takes the settlement.
+     * @return A {@code RandomRange} encapsulating the range of plunder
      *     available.
      */
     public abstract RandomRange getPlunderRange(Unit attacker);
@@ -687,7 +753,7 @@ public abstract class Settlement extends GoodsLocation
      * Allow player == null as this is true in the map editor where
      * the user player is moot.
      *
-     * @param player The other <code>Player</code> to check.
+     * @param player The other {@code Player} to check.
      * @return True if the settlement has contacted the player.
      */
     public abstract boolean hasContacted(Player player);
@@ -696,12 +762,64 @@ public abstract class Settlement extends GoodsLocation
      * Gets a label indicating the alarm level at this settlement with
      * respect to another player.
      *
-     * @param player The other <code>Player</code>.
-     * @return A <code>StringTemplate</code> describing the alarm.
+     * @param player The other {@code Player}.
+     * @return A {@code StringTemplate} describing the alarm.
      */
     public abstract StringTemplate getAlarmLevelLabel(Player player);
 
-    
+
+    /**
+     * Get the value of attacking a {@code Settlement}
+     *
+     * @param value The previously calculated input value from
+     *          {@link net.sf.freecol.server.ai.mission.UnitSeekAndDestroyMission
+     *                  #scoreSettlementPath(AIUnit, PathNode, Settlement)}
+     * @param unit The Unit doing the attacking.
+     * @return The calculated value
+     */
+    public abstract int calculateSettlementValue(int value, Unit unit);
+
+
+    // Override FreeColGameObject
+
+    /**
+     * {@inheritDoc}
+     */
+    @Override
+    public IntegrityType checkIntegrity(boolean fix, LogBuilder lb) {
+        IntegrityType result = super.checkIntegrity(fix, lb);
+        final Player owner = getOwner();
+        if (owner == null) {
+            lb.add("\n  Settlement without owner: ", getId());
+            result = result.fail();
+        }
+        return result;
+    }
+
+
+    // Override FreeColObject
+
+    /**
+     * {@inheritDoc}
+     */
+    @Override
+    public <T extends FreeColObject> boolean copyIn(T other) {
+        Settlement o = copyInCast(other, Settlement.class);
+        if (o == null || !super.copyIn(o)) return false;
+        final Game game = getGame();
+        this.type = o.getType();
+        this.owner = game.updateRef(o.getOwner());
+        this.name = o.getName();
+        this.tile = game.updateRef(o.getTile());
+        this.setOwnedTiles(game.updateRef(o.getOwnedTiles()));
+        this.featureContainer.copy(o.getFeatureContainer());
+
+        // Owner can be null when creating a Settlement known only by
+        // its tile ownership.
+        if (this.owner != null) this.owner.addSettlement(this);
+        return true;
+    }
+
 
     // Serialization
 
@@ -765,13 +883,14 @@ public abstract class Settlement extends GoodsLocation
         Player oldOwner = owner;
         owner = xr.findFreeColGameObject(game, OWNER_TAG,
                                          Player.class, (Player)null, true);
-        if (xr.shouldIntern()) game.checkOwners(this, oldOwner);
 
         tile = xr.findFreeColGameObject(game, TILE_TAG,
                                         Tile.class, (Tile)null, true);
 
         String newType = xr.getAttribute(SETTLEMENT_TYPE_TAG, (String)null);
         type = owner.getNationType().getSettlementType(newType);
+
+        if (xr.shouldIntern()) game.checkOwners(this, oldOwner);
     }
 
     /**
@@ -796,11 +915,11 @@ public abstract class Settlement extends GoodsLocation
         final Specification spec = getSpecification();
         final String tag = xr.getLocalName();
 
-        if (Ability.getXMLElementTagName().equals(tag)) {
+        if (Ability.TAG.equals(tag)) {
             Ability ability = new Ability(xr, spec);
             if (ability.isIndependent()) addAbility(ability);
 
-        } else if (Modifier.getXMLElementTagName().equals(tag)) {
+        } else if (Modifier.TAG.equals(tag)) {
             Modifier modifier = new Modifier(xr, spec);
             if (modifier.isIndependent()) addModifier(modifier);
 
@@ -808,4 +927,6 @@ public abstract class Settlement extends GoodsLocation
             super.readChild(xr);
         }
     }
+
+    // getXMLTagName left to subclasses
 }
