@@ -1,5 +1,5 @@
 /**
- *  Copyright (C) 2002-2015   The FreeCol Team
+ *  Copyright (C) 2002-2022   The FreeCol Team
  *
  *  This file is part of FreeCol.
  *
@@ -26,10 +26,8 @@ import java.util.Random;
 import java.util.Set;
 import java.util.logging.Level;
 import java.util.logging.Logger;
-import java.util.stream.Collectors;
 
 import net.sf.freecol.common.model.Modifier.ModifierType;
-import net.sf.freecol.common.model.UnitTypeChange.ChangeType;
 import static net.sf.freecol.common.util.CollectionUtils.*;
 import net.sf.freecol.common.util.LogBuilder;
 import static net.sf.freecol.common.util.RandomUtils.*;
@@ -83,7 +81,7 @@ public class SimpleCombatModel extends CombatModel {
      *
      * @param attacker The attacker.
      * @param defender The defender.
-     * @param lb An optional <code>LogBuilder</code> to log to.
+     * @param lb An optional {@code LogBuilder} to log to.
      * @return The combat odds.
      */
     private CombatOdds calculateCombatOdds(FreeColGameObject attacker,
@@ -123,12 +121,12 @@ public class SimpleCombatModel extends CombatModel {
     /**
      * Helper to log modifiers with.
      *
-     * @param lb The <code>LogBuilder</code> to log to.
-     * @param modSet A set of <code>Modifiers</code> to log.
+     * @param lb The {@code LogBuilder} to log to.
+     * @param modSet A set of {@code Modifiers} to log.
      */   
     private void logModifiers(LogBuilder lb, Set<Modifier> modSet) {
-        lb.addCollection(" ", modSet.stream()
-            .sorted().collect(Collectors.toList()));
+        lb.addCollection(" ",
+            sort(modSet, Modifier.ascendingModifierIndexComparator));
     }
 
     /**
@@ -136,7 +134,7 @@ public class SimpleCombatModel extends CombatModel {
      *
      * @param attacker The attacker.
      * @param defender The defender.
-     * @param lb An optional <code>LogBuilder</code> to log to.
+     * @param lb An optional {@code LogBuilder} to log to.
      * @return The offensive power.
      */
     private double getOffencePower(FreeColGameObject attacker,
@@ -144,7 +142,7 @@ public class SimpleCombatModel extends CombatModel {
                                    LogBuilder lb) {
         double result = 0.0;
         if (attacker == null) {
-            throw new IllegalStateException("Null attacker");
+            throw new RuntimeException("Null attacker: " + this);
 
         } else if (combatIsAttackMeasurement(attacker, defender)
             || combatIsAttack(attacker, defender)
@@ -160,15 +158,16 @@ public class SimpleCombatModel extends CombatModel {
         } else if (combatIsBombard(attacker, defender)) {
             Settlement attackerSettlement = (Settlement) attacker;
             if (attackerSettlement.hasAbility(Ability.BOMBARD_SHIPS)) {
-                result += attackerSettlement.getTile().getUnitList().stream()
-                    .filter(u -> u.hasAbility(Ability.BOMBARD))
-                    .mapToDouble(u -> u.getType().getOffence()).sum();
+                result += sumDouble(attackerSettlement.getTile().getUnits(),
+                                    u -> u.hasAbility(Ability.BOMBARD),
+                                    u -> u.getType().getOffence());
             }
             if (result > MAXIMUM_BOMBARD_POWER) result = MAXIMUM_BOMBARD_POWER;
             if (lb != null) lb.add(" bombard=", result);
 
         } else {
-            throw new IllegalArgumentException("Bogus combat");
+            throw new RuntimeException("Bogus combat: " + attacker
+                + " v " + defender);
         }
         return result;
     }
@@ -191,7 +190,7 @@ public class SimpleCombatModel extends CombatModel {
      *
      * @param attacker The attacker.
      * @param defender The defender.
-     * @param lb An optional <code>LogBuilder</code> to log to.
+     * @param lb An optional {@code LogBuilder} to log to.
      * @return The defensive power.
      */
     public double getDefencePower(FreeColGameObject attacker,
@@ -227,9 +226,8 @@ public class SimpleCombatModel extends CombatModel {
     public Set<Modifier> getOffensiveModifiers(FreeColGameObject attacker,
                                                FreeColGameObject defender) {
         Set<Modifier> result = new HashSet<>();
-        Modifier m;
         if (attacker == null) {
-            throw new IllegalStateException("Null attacker");
+            throw new RuntimeException("Null attacker: " + this);
         } else if (combatIsAttackMeasurement(attacker, defender)
             || combatIsAttack(attacker, defender)
             || combatIsSettlementAttack(attacker, defender)) {
@@ -254,9 +252,9 @@ public class SimpleCombatModel extends CombatModel {
             // Special bonuses against certain nation types
             if (defender instanceof Ownable) {
                 Player owner = ((Ownable)defender).getOwner();
-                result.addAll(attackerUnit
-                    .getModifiers(Modifier.OFFENCE_AGAINST,
-                                  owner.getNationType()));
+                result.addAll(toList(attackerUnit
+                        .getModifiers(Modifier.OFFENCE_AGAINST,
+                                      owner.getNationType())));
             }
 
             // Land/naval specific
@@ -270,15 +268,16 @@ public class SimpleCombatModel extends CombatModel {
             ; // Bombard strength handled by getOffensePower
 
         } else {
-            throw new IllegalArgumentException("Bogus combat");
+            throw new RuntimeException("Bogus combat: " + attacker
+                + " v " + defender);
         }
 
         // @compat 0.11.0
         // Any modifier with the default modifier index needs to be fixed
-        for (Modifier r : result) {
-            if (r.getModifierIndex() == Modifier.DEFAULT_MODIFIER_INDEX) {
-                r.setModifierIndex(Modifier.GENERAL_COMBAT_INDEX);
-            }
+        for (Modifier m : transform(result,
+                matchKeyEquals(Modifier.DEFAULT_MODIFIER_INDEX,
+                               Modifier::getModifierIndex))) {
+            m.setModifierIndex(Modifier.GENERAL_COMBAT_INDEX);
         }
         // end @compat 0.11.0
 
@@ -295,24 +294,22 @@ public class SimpleCombatModel extends CombatModel {
                                             Set<Modifier> result) {
         // Attack bonus
         final Specification spec = attacker.getSpecification();
-        result.addAll(spec.getModifiers(Modifier.ATTACK_BONUS));
+        result.addAll(toList(spec.getModifiers(Modifier.ATTACK_BONUS)));
 
         // Goods penalty always applies
         int goodsCount = attacker.getGoodsSpaceTaken();
         if (goodsCount > 0) {
-            for (Modifier m : spec.getModifiers(Modifier.CARGO_PENALTY)) {
-                Modifier c = new Modifier(m);
-                c.setValue(c.getValue() * goodsCount);
-                result.add(c);
-            }
+            result.addAll(transform(spec.getModifiers(Modifier.CARGO_PENALTY),
+                    alwaysTrue(),
+                    m -> Modifier.makeModifier(m).setValue(m.getValue() * goodsCount)));
         }
     }
 
     /**
      * Add the popular support bonus to the result set if applicable.
      *
-     * @param colony The <code>Colony</code> under attack.
-     * @param attacker The attacking <code>Unit</code>.
+     * @param colony The {@code Colony} under attack.
+     * @param attacker The attacking {@code Unit}.
      * @param result The set of modifiers to add to.
      */
     private void addPopularSupportBonus(Colony colony, Unit attacker,
@@ -331,7 +328,7 @@ public class SimpleCombatModel extends CombatModel {
     /**
      * Add all the offensive modifiers that apply to a land attack.
      *
-     * @param attacker The attacker <code>Unit</code>.
+     * @param attacker The attacker {@code Unit}.
      * @param defender The defender.
      * @param result The set of modifiers to add to.
      */
@@ -341,15 +338,15 @@ public class SimpleCombatModel extends CombatModel {
         final Specification spec = attacker.getSpecification();
 
         // Attack bonus
-        result.addAll(spec.getModifiers(Modifier.ATTACK_BONUS));
+        result.addAll(toList(spec.getModifiers(Modifier.ATTACK_BONUS)));
 
         // Movement penalty
         switch (attacker.getMovesLeft()) {
         case 1:
-            result.addAll(spec.getModifiers(Modifier.BIG_MOVEMENT_PENALTY));
+            result.addAll(toList(spec.getModifiers(Modifier.BIG_MOVEMENT_PENALTY)));
             break;
         case 2:
-            result.addAll(spec.getModifiers(Modifier.SMALL_MOVEMENT_PENALTY));
+            result.addAll(toList(spec.getModifiers(Modifier.SMALL_MOVEMENT_PENALTY)));
             break;
         default:
             break;
@@ -357,7 +354,7 @@ public class SimpleCombatModel extends CombatModel {
 
         // Amphibious attack?
         if (combatIsAmphibious(attacker, defender)) {
-            result.addAll(spec.getModifiers(Modifier.AMPHIBIOUS_ATTACK));
+            result.addAll(toList(spec.getModifiers(Modifier.AMPHIBIOUS_ATTACK)));
         }
 
         if (combatIsAttackMeasurement(attacker, defender)) {
@@ -365,7 +362,7 @@ public class SimpleCombatModel extends CombatModel {
 
         } else if (combatIsSettlementAttack(attacker, defender)) {
             // Settlement present, apply bombardment bonus
-            result.addAll(attacker.getModifiers(Modifier.BOMBARD_BONUS));
+            result.addAll(toList(attacker.getModifiers(Modifier.BOMBARD_BONUS)));
 
             // Popular support bonus
             if (combatIsWarOfIndependence(attacker, defender)) {
@@ -378,8 +375,7 @@ public class SimpleCombatModel extends CombatModel {
             if (tile != null) {
                 if (tile.hasSettlement()) {
                     // Bombard bonus applies to settlement defence
-                    result.addAll(attacker
-                                  .getModifiers(Modifier.BOMBARD_BONUS));
+                    result.addAll(toList(attacker.getModifiers(Modifier.BOMBARD_BONUS)));
 
                     // Popular support bonus
                     if (combatIsWarOfIndependence(attacker, defender)) {
@@ -391,7 +387,8 @@ public class SimpleCombatModel extends CombatModel {
                     // bonus, if defender is REF, or attacker is indian.
                     if (isAmbush(attacker, defender)) {
                         for (Modifier m : tile.getDefenceModifiers()) {
-                            Modifier mod = new Modifier(Modifier.OFFENCE, m);
+                            Modifier mod = Modifier.makeModifier(m);
+                            mod.setId(Modifier.OFFENCE);
                             mod.setSource(Specification.AMBUSH_BONUS_SOURCE);
                             result.add(mod);
                         }
@@ -402,14 +399,15 @@ public class SimpleCombatModel extends CombatModel {
             // Artillery in the open penalty, attacker must be on a
             // tile and neither unit can be in a settlement.
             if (attacker.hasAbility(Ability.BOMBARD)
-                && attacker.getLocation() instanceof Tile
+                && attacker.isOnTile()
                 && attacker.getSettlement() == null
                 && attacker.getState() != Unit.UnitState.FORTIFIED
                 && defenderUnit.getSettlement() == null) {
-                result.addAll(spec.getModifiers(Modifier.ARTILLERY_IN_THE_OPEN));
+                result.addAll(toList(spec.getModifiers(Modifier.ARTILLERY_IN_THE_OPEN)));
             }
         } else {
-            throw new IllegalStateException("Bogus combat");
+            throw new RuntimeException("Bogus combat: " + attacker
+                + " v " + defender);
         }
     }
 
@@ -466,15 +464,16 @@ public class SimpleCombatModel extends CombatModel {
             result.add(UNKNOWN_DEFENCE_MODIFIER);
 
         } else {
-            throw new IllegalArgumentException("Bogus combat");
+            throw new RuntimeException("Bogus combat: " + attacker
+                + " v " + defender);
         }
 
         // @compat 0.11.0
         // Any modifier with the default modifier index needs to be fixed
-        for (Modifier r : result) {
-            if (r.getModifierIndex() == Modifier.DEFAULT_MODIFIER_INDEX) {
-                r.setModifierIndex(Modifier.GENERAL_COMBAT_INDEX);
-            }
+        for (Modifier m : transform(result,
+                matchKeyEquals(Modifier.DEFAULT_MODIFIER_INDEX,
+                               Modifier::getModifierIndex))) {
+            m.setModifierIndex(Modifier.GENERAL_COMBAT_INDEX);
         }
         // end @compat 0.11.0
 
@@ -484,7 +483,7 @@ public class SimpleCombatModel extends CombatModel {
     /**
      * Add all the defensive modifiers that apply to a naval attack.
      *
-     * @param defender The defender <code>Unit</code>.
+     * @param defender The defender {@code Unit}.
      * @param result The set of modifiers to add to.
      */
     private void addNavalDefensiveModifiers(Unit defender,
@@ -494,31 +493,29 @@ public class SimpleCombatModel extends CombatModel {
         // Cargo penalty always applies
         int goodsCount = defender.getVisibleGoodsCount();
         if (goodsCount > 0) {
-            for (Modifier m : spec.getModifiers(Modifier.CARGO_PENALTY)) {
-                Modifier c = new Modifier(m);
-                c.setValue(c.getValue() * goodsCount);
-                result.add(c);
-            }
+            result.addAll(transform(spec.getModifiers(Modifier.CARGO_PENALTY),
+                    alwaysTrue(),
+                    m -> Modifier.makeModifier(m).setValue(m.getValue() * goodsCount)));
         }
     }
 
     /**
      * Does a given object provide a strong defence bonus?
      *
-     * @param fco The <code>FreeColObject</code> to check.
+     * @param fco The {@code FreeColObject} to check.
      * @return True if a strong defence bonus is present.
      */
     private boolean hasStrongDefenceModifier(FreeColObject fco) {
         return any(fco.getDefenceModifiers(),
-            m -> m.getType() == ModifierType.PERCENTAGE
-                && m.getValue() >= STRONG_DEFENCE_THRESHOLD);
+                   m -> m.getType() == ModifierType.PERCENTAGE
+                       && m.getValue() >= STRONG_DEFENCE_THRESHOLD);
     }
 
     /**
      * Add all the defensive modifiers that apply to a land attack.
      *
      * @param attacker The attacker.
-     * @param defender The defender <code>Unit</code>.
+     * @param defender The defender {@code Unit}.
      * @param result The set of modifiers to add to.
      */
     private void addLandDefensiveModifiers(FreeColGameObject attacker,
@@ -543,7 +540,7 @@ public class SimpleCombatModel extends CombatModel {
                 // Artillery in the Open penalty
                 if (defender.hasAbility(Ability.BOMBARD)
                     && defender.getState() != Unit.UnitState.FORTIFIED) {
-                    result.addAll(spec.getModifiers(Modifier.ARTILLERY_IN_THE_OPEN));
+                    result.addAll(toList(spec.getModifiers(Modifier.ARTILLERY_IN_THE_OPEN)));
                 }
 
             } else { // In settlement
@@ -551,10 +548,9 @@ public class SimpleCombatModel extends CombatModel {
                 result.addAll(settlement.getDefenceModifiers());
 
                 // Artillery defence bonus against an Indian raid
-                if (defender.hasAbility(Ability.BOMBARD)
-                    && attacker != null
-                    && ((Unit)attacker).getOwner().isIndian()) {
-                    result.addAll(spec.getModifiers(Modifier.ARTILLERY_AGAINST_RAID));
+                if (attacker != null && ((Unit)attacker).getOwner().isIndian()
+                    && defender.hasAbility(Ability.BOMBARD)) {
+                    result.addAll(toList(spec.getModifiers(Modifier.ARTILLERY_AGAINST_RAID)));
                 }
 
                 // Automatic defensive role (e.g. Revere)
@@ -572,9 +568,9 @@ public class SimpleCombatModel extends CombatModel {
             }
 
             // Fortify bonus
-            if (defender.getState() == Unit.UnitState.FORTIFIED
-                && !disableFortified) {
-                result.addAll(spec.getModifiers(Modifier.FORTIFIED));
+            if (!disableFortified
+                && defender.getState() == Unit.UnitState.FORTIFIED) {
+                result.addAll(toList(spec.getModifiers(Modifier.FORTIFIED)));
             }
         }
     }
@@ -644,7 +640,8 @@ public class SimpleCombatModel extends CombatModel {
             if (!defenderUnit.isNaval()) {
                 // One day we might want:
                 //   crs.add(CombatResult.SLAUGHTER_UNIT_BOMBARD);
-                throw new IllegalStateException("Bombard of non-naval");
+                throw new RuntimeException("Bombard of non-naval: " + attacker
+                    + " v " + defender);
             }
             action = "Bombard";
 
@@ -676,7 +673,8 @@ public class SimpleCombatModel extends CombatModel {
             }
 
         } else {
-            throw new IllegalStateException("Bogus combat");
+            throw new RuntimeException("Bogus combat: " + attacker
+                + " v " + defender);
         }
 
         // Log the results so that we have a solid record of combat
@@ -692,11 +690,11 @@ public class SimpleCombatModel extends CombatModel {
     /**
      * Resolve all the consequences of a normal attack.
      *
-     * @param winner The winning <code>Unit</code>.
-     * @param loser The losing <code>Unit</code>.
+     * @param winner The winning {@code Unit}.
+     * @param loser The losing {@code Unit}.
      * @param great True if this is a great win/loss.
      * @param r A "residual" random value (for convert/burn mission).
-     * @param crs A list of <code>CombatResult</code>s to add to.
+     * @param crs A list of {@code CombatResult}s to add to.
      */
     private void resolveAttack(Unit winner, Unit loser, boolean great,
                                double r, List<CombatResult> crs) {
@@ -705,7 +703,8 @@ public class SimpleCombatModel extends CombatModel {
         Player winnerPlayer = winner.getOwner();
         boolean attackerWon = crs.get(0) == CombatResult.WIN;
         boolean loserMustDie = loser.hasAbility(Ability.DISPOSE_ON_COMBAT_LOSS);
-
+        UnitTypeChange uc;
+        
         if (loser.isNaval()) {
             // Naval victors get to loot the defenders hold.  Sink the
             // loser on great win/loss, lack of repair location, or
@@ -738,7 +737,7 @@ public class SimpleCombatModel extends CombatModel {
                 // until the colony falls for lack of survivors.
                 // Ships in a falling colony will be damaged or sunk
                 // if they have no repair location.
-                if (!loser.isDefensiveUnit() && autoRole == null) {
+                if (autoRole == null && !loser.isDefensiveUnit()) {
                     List<Unit> ships = colony.getTile().getNavalUnits();
                     final CombatResult shipResult = (ships.isEmpty()) ? null
                         : (ships.get(0).getRepairLocation() == null)
@@ -799,8 +798,8 @@ public class SimpleCombatModel extends CombatModel {
                             lose++;
                         }
                     } else if (r >= 1.0 - winner.getBurnProbability()) {
-                        if (any(loserPlayer.getIndianSettlements(),
-                                s -> s.hasMissionary(winnerPlayer))) {
+                        if (any(transform(loserPlayer.getIndianSettlements(),
+                                    s -> s.hasMissionary(winnerPlayer)))) {
                             crs.add(CombatResult.BURN_MISSIONS);
                         }
                     }
@@ -838,8 +837,7 @@ public class SimpleCombatModel extends CombatModel {
                     crs.add((winner.canCaptureEquipment(loserRole) != null)
                         ? CombatResult.CAPTURE_EQUIP
                         : CombatResult.LOSE_EQUIP);
-                    if (loserMustDie
-                        || loser.losingEquipmentKillsUnit()) {
+                    if (loser.losingEquipmentKillsUnit()) {
                         crs.add(CombatResult.SLAUGHTER_UNIT);
                     } else if (loser.losingEquipmentDemotesUnit()) {
                         crs.add(CombatResult.DEMOTE_UNIT);
@@ -853,8 +851,7 @@ public class SimpleCombatModel extends CombatModel {
                     crs.add(CombatResult.CAPTURE_UNIT);
 
                 // Or losing just causes a demotion.
-                } else if (loser.getTypeChange(ChangeType.DEMOTION,
-                                               loserPlayer) != null) {
+                } else if (loser.getUnitChange(UnitChangeType.DEMOTION) != null) {
                     crs.add(CombatResult.DEMOTE_UNIT);
 
                 // But finally, the default is to kill them.
@@ -865,13 +862,11 @@ public class SimpleCombatModel extends CombatModel {
         }
 
         // Promote great winners or with automatic promotion, if possible.
-        UnitTypeChange promotion = winner.getType()
-            .getUnitTypeChange(ChangeType.PROMOTION, winnerPlayer);
-        if (promotion != null
+        uc = winner.getUnitChange(UnitChangeType.PROMOTION);
+        if (uc != null
             && (winner.hasAbility(Ability.AUTOMATIC_PROMOTION)
                 || (great
-                    && (100 * (r - Math.floor(r))
-                        <= promotion.getProbability(ChangeType.PROMOTION))))) {
+                    && (100 * (r - Math.floor(r)) <= uc.probability)))) {
             crs.add(CombatResult.PROMOTE_UNIT);
         }
     }
@@ -885,19 +880,7 @@ public class SimpleCombatModel extends CombatModel {
      */
     private boolean isAmbush(FreeColGameObject attacker,
                              FreeColGameObject defender) {
-        if (attacker instanceof Unit && defender instanceof Unit) {
-            Unit attackerUnit = (Unit)attacker;
-            Unit defenderUnit = (Unit)defender;
-            return attackerUnit.getSettlement() == null
-                && attackerUnit.hasTile()
-                && defenderUnit.getSettlement() == null
-                && defenderUnit.getState() != Unit.UnitState.FORTIFIED
-                && defenderUnit.hasTile()
-                && (attackerUnit.hasAbility(Ability.AMBUSH_BONUS)
-                    || defenderUnit.hasAbility(Ability.AMBUSH_PENALTY))
-                && (attackerUnit.getTile().hasAbility(Ability.AMBUSH_TERRAIN)
-                    || defenderUnit.getTile().hasAbility(Ability.AMBUSH_TERRAIN));
-        }
-        return false;
+        return attacker instanceof Unit && defender instanceof Unit
+            && ((Unit)attacker).canAmbush((Unit)defender);
     }
 }
