@@ -1,5 +1,5 @@
 /**
- *  Copyright (C) 2002-2015   The FreeCol Team
+ *  Copyright (C) 2002-2022   The FreeCol Team
  *
  *  This file is part of FreeCol.
  *
@@ -20,13 +20,13 @@
 package net.sf.freecol.common.model;
 
 import java.util.ArrayList;
-import java.util.Collections;
 import java.util.HashMap;
 import java.util.HashSet;
 import java.util.List;
 import java.util.Map;
 import java.util.Map.Entry;
 import java.util.Set;
+import java.util.stream.Stream;
 
 import javax.xml.stream.XMLStreamException;
 
@@ -39,7 +39,9 @@ import net.sf.freecol.common.util.RandomChoice;
 /**
  * An improvement to make to a tile.
  */
-public final class TileImprovementType extends FreeColGameObjectType {
+public final class TileImprovementType extends FreeColSpecObjectType {
+
+    public static final String TAG = "tile-improvement-type";
 
     /** Is this improvement natural or man-made? */
     private boolean natural;
@@ -59,24 +61,9 @@ public final class TileImprovementType extends FreeColGameObjectType {
     /** The amount of the equipment expended in making this improvement. */
     private int expendedAmount;
 
-    // @compat 0.10.4
-    /** The type of goods delivered by making this improvement. */
-    private GoodsType deliverGoodsType = null;
-
-    /** The amount of goods delivered by making this improvement. */
-    private int deliverAmount;
-    // end @compat
-
     /** The change to the movement cost due to this tile improvement. */
     private int movementCost = -1;
 
-    /**
-     * The layer a TileItem belongs to. Items with higher zIndex
-     * will be displayed above items with a lower zIndex. E.g. the
-     * LostCityRumour would be displayed above the Plow improvement.
-     */
-    private int zIndex;
-
     /**
      * Can this improvement expose a resource when completed? This
      * should only apply to improvement types that change the
@@ -84,6 +71,13 @@ public final class TileImprovementType extends FreeColGameObjectType {
      */
     private int exposeResourcePercent;
 
+    /**
+     * The layer a TileItem belongs to. Items with higher zIndex
+     * will be displayed above items with a lower zIndex. E.g. the
+     * LostCityRumour would be displayed above the Plow improvement.
+     */
+    private int zIndex;
+
     /** The workers that can make this improvement. */
     private Set<String> allowedWorkers = null;
 
@@ -93,18 +87,12 @@ public final class TileImprovementType extends FreeColGameObjectType {
     /** The disasters that may strike this type of tile improvement. */
     private List<RandomChoice<Disaster>> disasters = null;
 
-    /**
-     * The scopes define which TileTypes support this improvement.
-     * An eligible TileType must match all scopes.
-     */
-    private List<Scope> scopes = null;
-
 
     /**
      * Create a new tile improvement type.
      *
      * @param id The object identifier.
-     * @param specification The <code>Specification</code> to refer to.
+     * @param specification The {@code Specification} to refer to.
      */
     public TileImprovementType(String id, Specification specification) {
         super(id, specification);
@@ -150,7 +138,7 @@ public final class TileImprovementType extends FreeColGameObjectType {
     /**
      * Get the role required to perform this improvement, if any.
      *
-     * @return The required <code>Role</code>.
+     * @return The required {@code Role}.
      */
     public Role getRequiredRole() {
         return requiredRole;
@@ -165,6 +153,32 @@ public final class TileImprovementType extends FreeColGameObjectType {
         return expendedAmount;
     }
 
+    /**
+     * Get the standard movement cost of the tile improvement type.
+     *
+     * @return The movement cost.
+     */
+    public int getMoveCost() {
+        return this.movementCost;
+    }
+
+    /**
+     * Possibly reduce the cost of moving due to this tile improvement
+     * type.
+     *
+     * Only applies if movementCost is positive (see spec).  Do not
+     * return zero from a movement costing routine or units get free
+     * moves!
+     *
+     * @param originalCost The original movement cost.
+     * @return The movement cost after any change.
+     */
+    public int getMoveCost(int originalCost) {
+        return (movementCost > 0 && movementCost < originalCost)
+            ? movementCost
+            : originalCost;
+    }
+
     /**
      * Gets the percent chance that this tile improvement can expose a
      * resource on the tile. This only applies to TileImprovementTypes
@@ -176,49 +190,6 @@ public final class TileImprovementType extends FreeColGameObjectType {
         return exposeResourcePercent;
     }
 
-    /**
-     * Get the scopes applicable to this improvement.
-     *
-     * @return A list of <code>Scope</code>s.
-     */
-    public List<Scope> getScopes() {
-        return (scopes == null) ? Collections.<Scope>emptyList()
-            : scopes;
-    }
-
-    /**
-     * Add a scope.
-     *
-     * @param scope The <code>Scope</code> to add.
-     */
-    private void addScope(Scope scope) {
-        if (scopes == null) scopes = new ArrayList<>();
-        scopes.add(scope);
-    }
-
-    /**
-     * Get a weighted list of natural disasters than can strike this
-     * tile improvement type.
-     *
-     * @return A random choice list of <code>Disaster</code>s.
-     */
-    public List<RandomChoice<Disaster>> getDisasters() {
-        return (disasters == null)
-            ? Collections.<RandomChoice<Disaster>>emptyList()
-            : disasters;
-    }
-
-    /**
-     * Add a disaster.
-     *
-     * @param disaster The <code>Disaster</code> to add.
-     * @param probability The probability of the disaster.
-     */
-    private void addDisaster(Disaster disaster, int probability) {
-        if (disasters == null) disasters = new ArrayList<>();
-        disasters.add(new RandomChoice<>(disaster, probability));
-    }
-
     /**
      * Get the layer.
      *
@@ -237,25 +208,46 @@ public final class TileImprovementType extends FreeColGameObjectType {
         this.zIndex = newZIndex;
     }
 
+    /**
+     * Get the set of allowed worker identifiers.
+     *
+     * @return The set of allowed worker identifiers.
+     */
+    protected Set<String> getAllowedWorkers() {
+        if (this.allowedWorkers == null) this.allowedWorkers = new HashSet<>();
+        return this.allowedWorkers;
+    }
+
+    /**
+     * Set the set of allowed worker identifiers.
+     *
+     * @param allowed The new set of allowed worker identifiers.
+     */
+    protected void setAllowedWorkers(Set<String> allowed) {
+        if (this.allowedWorkers == null) {
+            this.allowedWorkers = new HashSet<>();
+        } else {
+            this.allowedWorkers.clear();
+        }
+        this.allowedWorkers.addAll(allowed);
+    }
+
     /**
      * Add an allowed worker identifier.
      *
      * @param id The worker identifier to add.
      */
     private void addAllowedWorker(String id) {
-        if (allowedWorkers == null) {
-            allowedWorkers = new HashSet<>();
-        }
-        allowedWorkers.add(id);
+        getAllowedWorkers().add(id);
     }
 
     /**
      * Is a particular unit type allowed to build this improvement?
      *
-     * @param unitType The <code>UnitType</code> to check.
-     * @return True if the <code>UnitType</code> can build this improvement.
+     * @param unitType The {@code UnitType} to check.
+     * @return True if the {@code UnitType} can build this improvement.
      */
-    public boolean isWorkerTypeAllowed(UnitType unitType) {
+    private boolean isWorkerTypeAllowed(UnitType unitType) {
         return allowedWorkers == null || allowedWorkers.isEmpty()
             || allowedWorkers.contains(unitType.getId());
     }
@@ -265,8 +257,8 @@ public final class TileImprovementType extends FreeColGameObjectType {
      *
      * Checks both the unit type and the available equipment.
      *
-     * @param unit The <code>Unit</code> to check.
-     * @return True if the <code>Unit</code> can build this improvement.
+     * @param unit The {@code Unit} to check.
+     * @return True if the {@code Unit} can build this improvement.
      */
     public boolean isWorkerAllowed(Unit unit) {
         return isWorkerTypeAllowed(unit.getType())
@@ -281,7 +273,7 @@ public final class TileImprovementType extends FreeColGameObjectType {
      * If you want to find out if an improvement is allowed for a tile, call
      * {@link Tile#isImprovementAllowed(TileImprovement)}.
      *
-     * @param tileType The <code>TileType</code> to check.
+     * @param tileType The {@code TileType} to check.
      * @return True if improvement is possible.
      */
     public boolean isTileTypeAllowed(TileType tileType) {
@@ -298,16 +290,31 @@ public final class TileImprovementType extends FreeColGameObjectType {
     }
 
     public Modifier getProductionModifier(GoodsType goodsType) {
-        Set<Modifier> modifierSet = getModifiers(goodsType.getId());
-        if (modifierSet == null || modifierSet.isEmpty()) {
-            return null;
+        return first(getModifiers(goodsType.getId()));
+    }
+
+    /**
+     * Get the tile type change map.
+     *
+     * @return The map of tile type to changes.
+     */
+    protected Map<TileType, TileTypeChange> getTileTypeChanges() {
+        if (this.tileTypeChanges == null) this.tileTypeChanges = new HashMap<>();
+        return this.tileTypeChanges;
+    }
+
+    /**
+     * Set the tile type change map.
+     *
+     * @param changes A new map of tile type to changes.
+     */
+    protected void setTileTypeChanges(Map<TileType, TileTypeChange> changes) {
+        if (this.tileTypeChanges == null) {
+            this.tileTypeChanges = new HashMap<>();
         } else {
-            if (modifierSet.size() > 1) {
-                logger.warning("Only one Modifier for " + goodsType.getId()
-                    + " expected!");
-            }
-            return modifierSet.iterator().next();
+            this.tileTypeChanges.clear();
         }
+        this.tileTypeChanges.putAll(changes);
     }
 
     /**
@@ -323,8 +330,8 @@ public final class TileImprovementType extends FreeColGameObjectType {
      * Gets the goods produced by applying this TileImprovementType
      * to a Tile with the given TileType.
      *
-     * @param from The original <code>TileType</code>.
-     * @return The <code>AbstractGoods</code> produced.
+     * @param from The original {@code TileType}.
+     * @return The {@code AbstractGoods} produced.
      */
     public AbstractGoods getProduction(TileType from) {
         if (tileTypeChanges == null) return null;
@@ -335,8 +342,8 @@ public final class TileImprovementType extends FreeColGameObjectType {
     /**
      * Gets the destination type of a tile type change (or null).
      *
-     * @param tileType The <code>TileType</code> that is to change.
-     * @return The resulting <code>TileType</code>.
+     * @param tileType The {@code TileType} that is to change.
+     * @return The resulting {@code TileType}.
      */
     public TileType getChange(TileType tileType) {
         if (tileTypeChanges == null) return null;
@@ -348,48 +355,31 @@ public final class TileImprovementType extends FreeColGameObjectType {
      * Can this tile improvement type change a tile type to the given
      * tile type.
      *
-     * @param tileType The required <code>TileType</code>.
-     * @return True if the required <code>TileType</code> can be changed to.
+     * @param tileType The required {@code TileType}.
+     * @return True if the required {@code TileType} can be changed to.
      */
     public boolean changeContainsTarget(TileType tileType) {
         return (tileTypeChanges == null) ? false
             : any(tileTypeChanges.values(),
-                change -> change.getTo() == tileType);
+                  matchKey(tileType, TileTypeChange::getTo));
     }
 
     /**
      * Add a tile type change.
      *
-     * @param change The <code>TileTypeChange</code> to add.
+     * @param change The {@code TileTypeChange} to add.
      */
     private void addChange(TileTypeChange change) {
         if (tileTypeChanges == null) tileTypeChanges = new HashMap<>();
         tileTypeChanges.put(change.getFrom(), change);
     }
 
-    /**
-     * Possibly reduce the cost of moving due to this tile improvement
-     * type.
-     *
-     * Only applies if movementCost is positive (see spec).  Do not
-     * return zero from a movement costing routine or units get free
-     * moves!
-     *
-     * @param originalCost The original movement cost.
-     * @return The movement cost after any change.
-     */
-    public int getMoveCost(int originalCost) {
-        return (movementCost > 0 && movementCost < originalCost)
-            ? movementCost
-            : originalCost;
-    }
-
     /**
      * Gets the increase in production of the given GoodsType
      * this tile improvement type would yield at a specified tile.
      *
-     * @param tile The <code>Tile</code> to be considered.
-     * @param goodsType An optional preferred <code>GoodsType</code>.
+     * @param tile The {@code Tile} to be considered.
+     * @param goodsType An optional preferred {@code GoodsType}.
      * @return The increase in production
      */
     public int getImprovementValue(Tile tile, GoodsType goodsType) {
@@ -403,8 +393,7 @@ public final class TileImprovementType extends FreeColGameObjectType {
             if (tt == null) { // simple bonus
                 int production = tile.getPotentialProduction(goodsType, colonistType);
                 if (production > 0) {
-                    float chg = applyModifiers(production, null,
-                                               goodsType.getId());
+                    float chg = apply(production, null, goodsType.getId());
                     value = (int)(chg - production);
                 }
             } else { // tile type change
@@ -416,13 +405,82 @@ public final class TileImprovementType extends FreeColGameObjectType {
         return value;
     }
 
+    /**
+     * Get the list of disasters for this tile type.
+     *
+     * @return The disaster list.
+     */
+    protected List<RandomChoice<Disaster>> getDisasters() {
+        if (this.disasters == null) this.disasters = new ArrayList<>();
+        return this.disasters;
+    }
+
+    /**
+     * Set the list of disasters for this tile type.
+     *
+     * @param disasters The new list of {@code Disaster}s.
+     */
+    protected void setDisasters(List<RandomChoice<Disaster>> disasters) {
+        if (this.disasters == null) {
+            this.disasters = new ArrayList<>();
+        } else {
+            this.disasters.clear();
+        }
+        this.disasters.addAll(disasters);
+    }
+
+    /**
+     * Get a weighted list of natural disasters than can strike this
+     * tile improvement type.
+     *
+     * @return A stream of {@code Disaster} choices.
+     */
+    public Stream<RandomChoice<Disaster>> getDisasterChoices() {
+        return (disasters == null)
+            ? Stream.<RandomChoice<Disaster>>empty()
+            : disasters.stream();
+    }
+
+    /**
+     * Add a disaster.
+     *
+     * @param disaster The {@code Disaster} to add.
+     * @param probability The probability of the disaster.
+     */
+    private void addDisaster(Disaster disaster, int probability) {
+        if (disasters == null) disasters = new ArrayList<>();
+        disasters.add(new RandomChoice<>(disaster, probability));
+    }
+
+
+    // Override FreeColObject
+
+    /**
+     * {@inheritDoc}
+     */
+    @Override
+    public <T extends FreeColObject> boolean copyIn(T other) {
+        TileImprovementType o = copyInCast(other, TileImprovementType.class);
+        if (o == null || !super.copyIn(o)) return false;
+        this.natural = o.isNatural();
+        this.magnitude = o.getMagnitude();
+        this.addWorkTurns = o.getAddWorkTurns();
+        this.requiredImprovementType = o.getRequiredImprovementType();
+        this.requiredRole = o.getRequiredRole();
+        this.expendedAmount = o.getExpendedAmount();
+        this.movementCost = o.getMoveCost();
+        this.exposeResourcePercent = o.getExposeResourcePercent();
+        this.zIndex = o.getZIndex();
+        this.setAllowedWorkers(o.getAllowedWorkers());
+        this.setTileTypeChanges(o.getTileTypeChanges());
+        this.setDisasters(o.getDisasters());
+        return true;
+    }
+
 
     // Serialization
 
     private static final String ADD_WORK_TURNS_TAG = "add-work-turns";
-    private static final String CHANGE_TAG = "change";
-    private static final String DELIVER_AMOUNT_TAG = "deliver-amount";
-    private static final String DELIVER_GOODS_TYPE_TAG = "deliver-goods-type";
     private static final String DISASTER_TAG = "disaster";
     private static final String EXPENDED_AMOUNT_TAG = "expended-amount";
     private static final String EXPOSE_RESOURCE_PERCENT_TAG = "expose-resource-percent";
@@ -433,16 +491,19 @@ public final class TileImprovementType extends FreeColGameObjectType {
     private static final String PROBABILITY_TAG = "probability";
     private static final String REQUIRED_IMPROVEMENT_TAG = "required-improvement";
     private static final String REQUIRED_ROLE_TAG = "required-role";
+    private static final String TILE_TYPE_CHANGE_TAG = "tile-type-change";
     private static final String TO_TAG = "to";
     private static final String WORKER_TAG = "worker";
     private static final String ZINDEX_TAG = "zIndex";
-    // @compat 0.10.x
-    private static final String EXPENDED_EQUIPMENT_TYPE_TAG = "expended-equipment-type";
-    // end @compat 0.10.x
+    // @compat 0.11.0
+    private static final String OLD_EXPENDED_EQUIPMENT_TYPE_TAG = "expended-equipment-type";
+    // end @compat 0.11.0
     // @compat 0.11.3
     private static final String OLD_EXPOSE_RESOURCE_PERCENT_TAG = "exposeResourcePercent";
-
     // end @compat 0.11.3
+    // @compat 0.11.6
+    private static final String OLD_CHANGE_TAG = "change";
+    // end @compat 0.11.6
 
 
     /**
@@ -485,8 +546,6 @@ public final class TileImprovementType extends FreeColGameObjectType {
     protected void writeChildren(FreeColXMLWriter xw) throws XMLStreamException {
         super.writeChildren(xw);
 
-        for (Scope scope : getScopes()) scope.toXML(xw);
-
         if (allowedWorkers != null) {
             for (String id : allowedWorkers) {
                 xw.writeStartElement(WORKER_TAG);
@@ -504,7 +563,7 @@ public final class TileImprovementType extends FreeColGameObjectType {
             }
         }
 
-        for (RandomChoice<Disaster> choice : getDisasters()) {
+        for (RandomChoice<Disaster> choice : iterable(getDisasterChoices())) {
             xw.writeStartElement(DISASTER_TAG);
 
             xw.writeAttribute(ID_ATTRIBUTE_TAG, choice.getObject().getId());
@@ -535,21 +594,14 @@ public final class TileImprovementType extends FreeColGameObjectType {
 
         requiredRole = xr.getType(spec, REQUIRED_ROLE_TAG,
             Role.class, (Role)null);
-        // @compat 0.10.x
-        if (xr.hasAttribute(EXPENDED_EQUIPMENT_TYPE_TAG)) {
+        // @compat 0.11.x
+        if (xr.hasAttribute(OLD_EXPENDED_EQUIPMENT_TYPE_TAG)) {
             requiredRole = spec.getRole("model.role.pioneer");
         }
-        // end @compat 0.10.x
+        // end @compat 0.11.x
 
         expendedAmount = xr.getAttribute(EXPENDED_AMOUNT_TAG, 0);
 
-        // @compat 0.10.4
-        deliverGoodsType = xr.getType(spec, DELIVER_GOODS_TYPE_TAG,
-                                      GoodsType.class, (GoodsType)null);
-
-        deliverAmount = xr.getAttribute(DELIVER_AMOUNT_TAG, 0);
-        // end @compat
-
         movementCost = xr.getAttribute(MOVEMENT_COST_TAG, 0);
 
         zIndex = xr.getAttribute(ZINDEX_TAG, 0);
@@ -569,7 +621,6 @@ public final class TileImprovementType extends FreeColGameObjectType {
     protected void readChildren(FreeColXMLReader xr) throws XMLStreamException {
         // Clear containers.
         if (xr.shouldClearContainers()) {
-            scopes = null;
             allowedWorkers = null;
             tileTypeChanges = null;
             disasters = null;
@@ -586,24 +637,12 @@ public final class TileImprovementType extends FreeColGameObjectType {
         final Specification spec = getSpecification();
         final String tag = xr.getLocalName();
 
-        if (CHANGE_TAG.equals(tag)) {
-            TileTypeChange change = new TileTypeChange();
-            if (deliverGoodsType == null) {
-                change.readFromXML(xr, spec);
-            } else {
-                // @compat 0.10.4
-                TileType from = xr.getType(spec, FROM_TAG,
-                                           TileType.class, (TileType)null);
-                TileType to = xr.getType(spec, TO_TAG,
-                                         TileType.class, (TileType)null);
-                change.setFrom(from);
-                change.setTo(to);
-                change.setProduction(new AbstractGoods(deliverGoodsType,
-                                                       deliverAmount));
-                xr.closeTag(CHANGE_TAG);
-                // end @compat
-            }
-            addChange(change);
+        if (TILE_TYPE_CHANGE_TAG.equals(tag)
+            // @compat 0.11.6
+            || OLD_CHANGE_TAG.equals(tag)
+            // end @compat 0.11.6
+            ) {
+            addChange(new TileTypeChange(xr, spec));
 
         } else if (DISASTER_TAG.equals(tag)) {
             Disaster disaster = xr.getType(spec, ID_ATTRIBUTE_TAG,
@@ -616,9 +655,6 @@ public final class TileImprovementType extends FreeColGameObjectType {
             addAllowedWorker(xr.readId());
             xr.closeTag(WORKER_TAG);
 
-        } else if (Scope.getXMLElementTagName().equals(tag)) {
-            addScope(new Scope(xr));
-
         } else {
             super.readChild(xr);
         }
@@ -627,15 +663,5 @@ public final class TileImprovementType extends FreeColGameObjectType {
     /**
      * {@inheritDoc}
      */
-    @Override
-    public String getXMLTagName() { return getXMLElementTagName(); }
-
-    /**
-     * Gets the tag name of the root element representing this object.
-     *
-     * @return "tile-improvement-type".
-     */
-    public static String getXMLElementTagName() {
-        return "tile-improvement-type";
-    }
+    public String getXMLTagName() { return TAG; }
 }
