@@ -1,5 +1,5 @@
 /**
- *  Copyright (C) 2002-2015   The FreeCol Team
+ *  Copyright (C) 2002-2022   The FreeCol Team
  *
  *  This file is part of FreeCol.
  *
@@ -19,21 +19,31 @@
 
 package net.sf.freecol.common.model;
 
+import static net.sf.freecol.common.util.CollectionUtils.alwaysTrue;
+import static net.sf.freecol.common.util.CollectionUtils.flatten;
+import static net.sf.freecol.common.util.CollectionUtils.iterable;
+import static net.sf.freecol.common.util.CollectionUtils.matchKey;
+import static net.sf.freecol.common.util.CollectionUtils.transform;
+
 import java.util.ArrayList;
-import java.util.Collections;
 import java.util.List;
+import java.util.stream.Stream;
 
 import javax.xml.stream.XMLStreamException;
 
 import net.sf.freecol.common.io.FreeColXMLReader;
 import net.sf.freecol.common.io.FreeColXMLWriter;
+import net.sf.freecol.common.option.GameOptions;
 import net.sf.freecol.common.util.RandomChoice;
 
 
 /**
  * The types of tiles.
  */
-public final class TileType extends FreeColGameObjectType {
+public final class TileType extends FreeColSpecObjectType
+                            implements BaseProduction {
+
+    public static final String TAG = "tile-type";
 
     public static enum RangeType { HUMIDITY, TEMPERATURE, ALTITUDE };
 
@@ -44,6 +54,7 @@ public final class TileType extends FreeColGameObjectType {
     public static final TileType WATER = new TileType("WATER", true);
     public static final TileType LAND  = new TileType("LAND", false);
 
+
     /** Is this a forested tile? */
     private boolean forest;
 
@@ -90,16 +101,16 @@ public final class TileType extends FreeColGameObjectType {
      * Create a new tile type.
      *
      * @param id The object identifier.
-     * @param specification The <code>Specification</code> to refer to.
+     * @param specification The {@code Specification} to refer to.
      */
     public TileType(String id, Specification specification) {
         super(id, specification);
     }
 
     /**
-     * Creates a new <code>TileType</code> instance. This constructor
-     * is used to create the "virtual" tile types <code>LAND</code>
-     * and <code>WATER</code>, which are intended to simplify map
+     * Creates a new {@code TileType} instance. This constructor
+     * is used to create the "virtual" tile types {@code LAND}
+     * and {@code WATER}, which are intended to simplify map
      * loading.
      *
      * @param id The object identifier.
@@ -187,10 +198,29 @@ public final class TileType extends FreeColGameObjectType {
         return basicWorkTurns;
     }
 
+    protected int getHumidity(int i) {
+        return this.humidity[i];
+    }
+    protected void setHumidity(int i, int value) {
+        this.humidity[i] = value;
+    }
+    protected int getTemperature(int i) {
+        return this.temperature[i];
+    }
+    protected void setTemperature(int i, int value) {
+        this.temperature[i] = value;
+    }
+    protected int getAltitude(int i) {
+        return this.altitude[i];
+    }
+    protected void setAltitude(int i, int value) {
+        this.altitude[i] = value;
+    }
+    
     /**
      * Is this tile type suitable for a given range type value.
      *
-     * @param rangeType The <code>RangeType</code> to test.
+     * @param rangeType The {@code RangeType} to test.
      * @param value The value to check.
      * @return True if the tile type meets the range limits.
      */
@@ -213,64 +243,95 @@ public final class TileType extends FreeColGameObjectType {
      *
      * @return A weighted list of resource types.
      */
-    public List<RandomChoice<ResourceType>> getWeightedResources() {
-        return (resourceTypes == null)
-            ? Collections.<RandomChoice<ResourceType>>emptyList()
-            : resourceTypes;
+    public List<RandomChoice<ResourceType>> getResourceTypes() {
+        if (this.resourceTypes == null) this.resourceTypes = new ArrayList<>();
+        return this.resourceTypes;
+    }
+
+    /**
+     * Set the resource types list.
+     *
+     * @param resourceTypes The list of resource type choices.
+     */
+    protected void setResourceTypes(List<RandomChoice<ResourceType>> resourceTypes) {
+        if (this.resourceTypes == null) {
+            this.resourceTypes = new ArrayList<>();
+        } else {
+            this.resourceTypes.clear();
+        }
+        this.resourceTypes.addAll(resourceTypes);
     }
 
     /**
      * Gets the resource types that can be found on this tile type.
      *
-     * @return A list of <code>ResourceType</code>s.
+     * @return A list of {@code ResourceType}s.
      */
-    public List<ResourceType> getResourceTypes() {
-        List<ResourceType> result = new ArrayList<>();
-        if (resourceTypes != null) {
-            for (RandomChoice<ResourceType> resource : resourceTypes) {
-                result.add(resource.getObject());
-            }
-        }
-        return result;
+    public List<ResourceType> getResourceTypeValues() {
+        return transform(getResourceTypes(), alwaysTrue(),
+                         RandomChoice::getObject);
     }
 
     /**
      * Add a resource type.
      *
-     * @param type The <code>ResourceType</code> to add.
+     * @param type The {@code ResourceType} to add.
      * @param prob The percentage probability of the resource being
      *     present.
      */
     private void addResourceType(ResourceType type, int prob) {
-        if (resourceTypes == null) resourceTypes = new ArrayList<>();
-        resourceTypes.add(new RandomChoice<>(type, prob));
+        if (this.resourceTypes == null) this.resourceTypes = new ArrayList<>();
+        this.resourceTypes.add(new RandomChoice<>(type, prob));
     }
 
     /**
      * Can this tile type contain a specified resource type?
      *
-     * @param resourceType The <code>ResourceType</code> to test.
-     * @return True if the <code>ResourceType</code> is compatible.
+     * @param resourceType The {@code ResourceType} to test.
+     * @return True if the {@code ResourceType} is compatible.
      */
     public boolean canHaveResourceType(ResourceType resourceType) {
-        return getResourceTypes().contains(resourceType);
+        return getResourceTypeValues().contains(resourceType);
+    }
+
+    /**
+     * Get the disaster choices.
+     *
+     * @return A list of {@code Disaster} choices.
+     */
+    protected List<RandomChoice<Disaster>> getDisasters() {
+        return this.disasters;
+    }
+
+    /**
+     * Set the disaster choices.
+     *
+     * @param disasters The new {@code Disaster} choice list.
+     */
+    protected void setDisasters(List<RandomChoice<Disaster>> disasters) {
+        if (this.disasters == null) {
+            this.disasters = new ArrayList<>();
+        } else {
+            this.disasters.clear();
+        }
+        this.disasters.addAll(disasters);
     }
 
     /**
      * Gets the natural disasters than can strike this tile type.
      *
-     * @return a <code>List<RandomChoice<Disaster>></code> value
+     * @return A stream of {@code Disaster} choices.
      */
-    public List<RandomChoice<Disaster>> getDisasters() {
-        return (disasters == null)
-            ? Collections.<RandomChoice<Disaster>>emptyList()
-            : disasters;
+    public Stream<RandomChoice<Disaster>> getDisasterChoices() {
+        return (this.disasters == null)
+            ? Stream.<RandomChoice<Disaster>>empty()
+            : disasters.stream();
     }
 
     /**
      * Add a disaster.
      *
-     * @param disaster The <code>Disaster</code> to add.
+     * @param disaster The {@code Disaster} to add.
      * @param probability The probability of the disaster.
      */
     private void addDisaster(Disaster disaster, int probability) {
@@ -278,12 +339,31 @@ public final class TileType extends FreeColGameObjectType {
         disasters.add(new RandomChoice<>(disaster, probability));
     }
 
+    /**
+     * Get the production type list.
+     *
+     * @return The {@code ProductionType} list.
+     */
+    protected List<ProductionType> getProductionTypes() {
+        return this.productionTypes;
+    }
+    
+    /**
+     * Set the production type list.
+     *
+     * @param productionTypes The new {@code ProductionType} list.
+     */
+    protected void setProductionTypes(List<ProductionType> productionTypes) {
+        this.productionTypes.clear();
+        this.productionTypes.addAll(productionTypes);
+    }
+    
     /**
      * Gets the production types available at the current difficulty
      * level.
      *
      * @param unattended Whether the production is unattended.
-     * @return A list of <code>ProductionType</code>s.
+     * @return A list of {@code ProductionType}s.
      */
     public List<ProductionType> getAvailableProductionTypes(boolean unattended) {
         return getAvailableProductionTypes(unattended,
@@ -295,16 +375,18 @@ public final class TileType extends FreeColGameObjectType {
      * of colony center tile and production level.  If the production
      * level is null, all production levels will be returned.
      *
+     * Public for the test suite.
+     *
      * @param unattended Whether the production is unattended.
      * @param level The production level.
-     * @return A list of <code>ProductionType</code>s.
+     * @return A list of {@code ProductionType}s.
      */
     public List<ProductionType> getAvailableProductionTypes(boolean unattended,
                                                             String level) {
         List<ProductionType> good = new ArrayList<>(),
             better = new ArrayList<>();
-        for (ProductionType productionType : productionTypes) {
-            if (productionType.getUnattended() != unattended) continue;
+        for (ProductionType productionType : transform(productionTypes,
+                matchKey(unattended, ProductionType::getUnattended))) {
             if (productionType.appliesExactly(level)) {
                 better.add(productionType);
             } else if (productionType.appliesTo(level)) {
@@ -320,8 +402,8 @@ public final class TileType extends FreeColGameObjectType {
     /**
      * Can a tile of this type produce a given goods type?
      *
-     * @param goodsType The <code>GoodsType</code> to produce.
-     * @param unitType An optional <code>UnitType</code> that is to do
+     * @param goodsType The {@code GoodsType} to produce.
+     * @param unitType An optional {@code UnitType} that is to do
      *     the work, if null the unattended production is considered.
      * @return True if this tile type produces the goods.
      */
@@ -331,35 +413,12 @@ public final class TileType extends FreeColGameObjectType {
                 getAvailableProductionTypes(unitType == null));
     }
 
-    /**
-     * Get the base production of a given goods type for an optional
-     * unit type.
-     * 
-     * @param productionType An optional <code>ProductionType</code> to use,
-     *     if null the best available one is used.
-     * @param goodsType The <code>GoodsType</code> to produce.
-     * @param unitType An optional <code>UnitType</code> that is to do
-     *     the work, if null the unattended production is considered.
-     * @return The amount of goods produced.
-     */
-    public int getBaseProduction(ProductionType productionType,
-                                 GoodsType goodsType, UnitType unitType) {
-        if (goodsType == null) return 0;
-        if (productionType == null) {
-            productionType = ProductionType.getBestProductionType(goodsType,
-                getAvailableProductionTypes(unitType == null));
-        }
-        if (productionType == null) return 0;
-        AbstractGoods best = productionType.getOutput(goodsType);
-        return (best == null) ? 0 : best.getAmount();
-    }
-
     /**
      * Get the amount of goods of given goods type the given unit type
      * could produce on a tile of this tile type.
      *
-     * @param goodsType The <code>GoodsType</code> to produce.
-     * @param unitType An optional <code>UnitType</code> that is to do
+     * @param goodsType The {@code GoodsType} to produce.
+     * @param unitType An optional {@code UnitType} that is to do
      *     the work, if null the unattended production is considered.
      * @return The amount of goods produced.
      */
@@ -367,8 +426,7 @@ public final class TileType extends FreeColGameObjectType {
                                       UnitType unitType) {
         if (goodsType == null) return 0;
         int amount = getBaseProduction(null, goodsType, unitType);
-        amount = (int)applyModifiers(amount, null, goodsType.getId(),
-                                     unitType);
+        amount = (int)apply(amount, null, goodsType.getId(), unitType);
         return (amount < 0) ? 0 : amount;
     }
 
@@ -380,25 +438,43 @@ public final class TileType extends FreeColGameObjectType {
      * {@link #getPotentialProduction(GoodsType, UnitType)}
      *
      * @param unattended Select unattended production.
-     * @return A list of produced <code>AbstractGoods</code>.
+     * @return A stream of produced {@code AbstractGoods}.
      */
-    public List<AbstractGoods> getPossibleProduction(boolean unattended) {
-        List<AbstractGoods> production = new ArrayList<>();
-        for (ProductionType pt : getAvailableProductionTypes(unattended)) {
-            List<AbstractGoods> outputs = pt.getOutputs();
-            if (!outputs.isEmpty()) production.addAll(outputs);
-        }
-        return production;
+    public Stream<AbstractGoods> getPossibleProduction(boolean unattended) {
+        return flatten(getAvailableProductionTypes(unattended),
+                       ProductionType::getOutputs);
     }
 
+
+    // Override FreeColObject
+
     /**
      * {@inheritDoc}
-     *
-     * Kludge to make this public so that MapViewer can see it.
      */
     @Override
-    public int getIndex() {
-        return super.getIndex();
+    public <T extends FreeColObject> boolean copyIn(T other) {
+        TileType o = copyInCast(other, TileType.class);
+        if (o == null || !super.copyIn(o)) return false;
+        this.forest = o.isForested();
+        this.water = o.isWater();
+        this.canSettle = o.canSettle();
+        this.connected = o.isHighSeasConnected();
+        this.elevation = o.isElevation();
+        this.basicMoveCost = o.getBasicMoveCost();
+        this.basicWorkTurns = o.getBasicWorkTurns();
+        for (int i = 0; i < this.humidity.length; i++) {
+            this.setHumidity(i, o.getHumidity(i));
+        }
+        for (int i = 0; i < this.temperature.length; i++) {
+            this.setTemperature(i, o.getTemperature(i));
+        }
+        for (int i = 0; i < this.altitude.length; i++) {
+            this.setAltitude(i, o.getAltitude(i));
+        }
+        this.setResourceTypes(o.getResourceTypes());
+        this.setDisasters(o.getDisasters());
+        this.setProductionTypes(o.getProductionTypes());
+        return true;
     }
 
 
@@ -424,11 +500,10 @@ public final class TileType extends FreeColGameObjectType {
     private static final String TEMPERATURE_MIN_TAG = "temperature-minimum";
     private static final String TEMPERATURE_MAX_TAG = "temperature-maximum";
     private static final String TYPE_TAG = "type";
-    // @compat 0.10.x
+    // @compat 0.11.x
     private static final String PRIMARY_PRODUCTION_TAG = "primary-production";
     private static final String SECONDARY_PRODUCTION_TAG = "secondary-production";
-    private static final String TILE_PRODUCTION_TAG = "tile-production";
-    // end @compat 0.10.x
+    // end @compat 0.11.x
     // @compat 0.11.3
     private static final String OLD_ALTITUDE_MIN_TAG = "altitudeMin";
     private static final String OLD_ALTITUDE_MAX_TAG = "altitudeMax";
@@ -446,19 +521,19 @@ public final class TileType extends FreeColGameObjectType {
     protected void writeAttributes(FreeColXMLWriter xw) throws XMLStreamException {
         super.writeAttributes(xw);
 
-        xw.writeAttribute(BASIC_MOVE_COST_TAG, basicMoveCost);
+        xw.writeAttribute(BASIC_MOVE_COST_TAG, this.basicMoveCost);
 
-        xw.writeAttribute(BASIC_WORK_TURNS_TAG, basicWorkTurns);
+        xw.writeAttribute(BASIC_WORK_TURNS_TAG, this.basicWorkTurns);
 
-        xw.writeAttribute(IS_FOREST_TAG, forest);
+        xw.writeAttribute(IS_FOREST_TAG, this.forest);
 
-        xw.writeAttribute(IS_WATER_TAG, water);
+        xw.writeAttribute(IS_WATER_TAG, this.water);
 
-        xw.writeAttribute(IS_ELEVATION_TAG, elevation);
+        xw.writeAttribute(IS_ELEVATION_TAG, this.elevation);
 
-        xw.writeAttribute(IS_CONNECTED_TAG, connected);
+        xw.writeAttribute(IS_CONNECTED_TAG, this.connected);
 
-        xw.writeAttribute(CAN_SETTLE_TAG, canSettle);
+        xw.writeAttribute(CAN_SETTLE_TAG, this.canSettle);
     }
 
     /**
@@ -470,35 +545,37 @@ public final class TileType extends FreeColGameObjectType {
 
         xw.writeStartElement(GEN_TAG);
 
-        xw.writeAttribute(HUMIDITY_MIN_TAG, humidity[0]);
+        xw.writeAttribute(HUMIDITY_MIN_TAG, this.humidity[0]);
 
-        xw.writeAttribute(HUMIDITY_MAX_TAG, humidity[1]);
+        xw.writeAttribute(HUMIDITY_MAX_TAG, this.humidity[1]);
 
-        xw.writeAttribute(TEMPERATURE_MIN_TAG, temperature[0]);
+        xw.writeAttribute(TEMPERATURE_MIN_TAG, this.temperature[0]);
 
-        xw.writeAttribute(TEMPERATURE_MAX_TAG, temperature[1]);
+        xw.writeAttribute(TEMPERATURE_MAX_TAG, this.temperature[1]);
 
-        xw.writeAttribute(ALTITUDE_MIN_TAG, altitude[0]);
+        xw.writeAttribute(ALTITUDE_MIN_TAG, this.altitude[0]);
 
-        xw.writeAttribute(ALTITUDE_MAX_TAG, altitude[1]);
+        xw.writeAttribute(ALTITUDE_MAX_TAG, this.altitude[1]);
 
         xw.writeEndElement();
 
-        for (ProductionType productionType : productionTypes) {
+        for (ProductionType productionType : this.productionTypes) {
             productionType.toXML(xw);
         }
 
-        for (RandomChoice<ResourceType> choice : getWeightedResources()) {
-            xw.writeStartElement(RESOURCE_TAG);
+        if (this.resourceTypes != null) {
+            for (RandomChoice<ResourceType> choice : this.resourceTypes) {
+                xw.writeStartElement(RESOURCE_TAG);
 
-            xw.writeAttribute(TYPE_TAG, choice.getObject());
+                xw.writeAttribute(TYPE_TAG, choice.getObject());
 
-            xw.writeAttribute(PROBABILITY_TAG, choice.getProbability());
+                xw.writeAttribute(PROBABILITY_TAG, choice.getProbability());
 
-            xw.writeEndElement();
+                xw.writeEndElement();
+            }
         }
 
-        for (RandomChoice<Disaster> choice : getDisasters()) {
+        for (RandomChoice<Disaster> choice : iterable(getDisasterChoices())) {
             xw.writeStartElement(DISASTER_TAG);
 
             xw.writeAttribute(ID_ATTRIBUTE_TAG, choice.getObject());
@@ -516,19 +593,19 @@ public final class TileType extends FreeColGameObjectType {
     protected void readAttributes(FreeColXMLReader xr) throws XMLStreamException {
         super.readAttributes(xr);
 
-        basicMoveCost = xr.getAttribute(BASIC_MOVE_COST_TAG, 1);
+        this.basicMoveCost = xr.getAttribute(BASIC_MOVE_COST_TAG, 1);
 
-        basicWorkTurns = xr.getAttribute(BASIC_WORK_TURNS_TAG, 1);
+        this.basicWorkTurns = xr.getAttribute(BASIC_WORK_TURNS_TAG, 1);
 
-        forest = xr.getAttribute(IS_FOREST_TAG, false);
+        this.forest = xr.getAttribute(IS_FOREST_TAG, false);
 
-        water = xr.getAttribute(IS_WATER_TAG, false);
+        this.water = xr.getAttribute(IS_WATER_TAG, false);
 
-        elevation = xr.getAttribute(IS_ELEVATION_TAG, false);
+        this.elevation = xr.getAttribute(IS_ELEVATION_TAG, false);
 
-        canSettle = xr.getAttribute(CAN_SETTLE_TAG, !water);
+        this.canSettle = xr.getAttribute(CAN_SETTLE_TAG, !water);
 
-        connected = xr.getAttribute(IS_CONNECTED_TAG, false);
+        this.connected = xr.getAttribute(IS_CONNECTED_TAG, false);
 
     }
 
@@ -539,9 +616,9 @@ public final class TileType extends FreeColGameObjectType {
     protected void readChildren(FreeColXMLReader xr) throws XMLStreamException {
         // Clear containers.
         if (xr.shouldClearContainers()) {
-            disasters = null;
-            resourceTypes = null;
-            productionTypes.clear();
+            this.disasters = null;
+            this.resourceTypes = null;
+            this.productionTypes.clear();
         }
 
         super.readChildren(xr);
@@ -564,77 +641,41 @@ public final class TileType extends FreeColGameObjectType {
             xr.closeTag(DISASTER_TAG);
 
         } else if (GEN_TAG.equals(tag)) {
-            humidity[0] = xr.getAttribute(HUMIDITY_MIN_TAG, 0);
-            humidity[1] = xr.getAttribute(HUMIDITY_MAX_TAG, 100);
-            temperature[0] = xr.getAttribute(TEMPERATURE_MIN_TAG, -20);
-            temperature[1] = xr.getAttribute(TEMPERATURE_MAX_TAG, 40);
-            altitude[0] = xr.getAttribute(ALTITUDE_MIN_TAG, 0);
-            altitude[1] = xr.getAttribute(ALTITUDE_MAX_TAG, 0);
+            this.humidity[0] = xr.getAttribute(HUMIDITY_MIN_TAG, 0);
+            this.humidity[1] = xr.getAttribute(HUMIDITY_MAX_TAG, 100);
+            this.temperature[0] = xr.getAttribute(TEMPERATURE_MIN_TAG, -20);
+            this.temperature[1] = xr.getAttribute(TEMPERATURE_MAX_TAG, 40);
+            this.altitude[0] = xr.getAttribute(ALTITUDE_MIN_TAG, 0);
+            this.altitude[1] = xr.getAttribute(ALTITUDE_MAX_TAG, 0);
             // @compat 0.11.3
             if (xr.hasAttribute(OLD_HUMIDITY_MIN_TAG)) {
-                humidity[0] = xr.getAttribute(OLD_HUMIDITY_MIN_TAG, 0);
+                this.humidity[0] = xr.getAttribute(OLD_HUMIDITY_MIN_TAG, 0);
             }
             if (xr.hasAttribute(OLD_HUMIDITY_MAX_TAG)) {
-                humidity[1] = xr.getAttribute(OLD_HUMIDITY_MAX_TAG, 100);
+                this.humidity[1] = xr.getAttribute(OLD_HUMIDITY_MAX_TAG, 100);
             }
             if (xr.hasAttribute(OLD_TEMPERATURE_MIN_TAG)) {
-                temperature[0] = xr.getAttribute(OLD_TEMPERATURE_MIN_TAG, -20);
+                this.temperature[0] = xr.getAttribute(OLD_TEMPERATURE_MIN_TAG, -20);
             }
             if (xr.hasAttribute(OLD_TEMPERATURE_MAX_TAG)) {
-                temperature[1] = xr.getAttribute(OLD_TEMPERATURE_MAX_TAG, 40);
+                this.temperature[1] = xr.getAttribute(OLD_TEMPERATURE_MAX_TAG, 40);
             }
             if (xr.hasAttribute(OLD_ALTITUDE_MIN_TAG)) {
-                altitude[0] = xr.getAttribute(OLD_ALTITUDE_MIN_TAG, 0);
+                this.altitude[0] = xr.getAttribute(OLD_ALTITUDE_MIN_TAG, 0);
             }
             if (xr.hasAttribute(OLD_ALTITUDE_MAX_TAG)) {
-                altitude[1] = xr.getAttribute(OLD_ALTITUDE_MAX_TAG, 0);
+                this.altitude[1] = xr.getAttribute(OLD_ALTITUDE_MAX_TAG, 0);
             }
             // end @compat 0.11.3
             xr.closeTag(GEN_TAG);
 
         } else if (PRODUCTION_TAG.equals(tag)
             && xr.getAttribute(DELETE_TAG, false)) {
-            productionTypes.clear();
+            this.productionTypes.clear();
             xr.closeTag(PRODUCTION_TAG);
 
-        } else if (PRODUCTION_TAG.equals(tag)
-                   // @compat 0.10.6
-                   && xr.getAttribute(GOODS_TYPE_TAG, (String)null) == null
-                   // end @compat 0.10.6
-                   ) {
-            productionTypes.add(new ProductionType(xr, spec));
-
-        // @compat 0.10.6
-        } else if (PRODUCTION_TAG.equals(tag)
-            || PRIMARY_PRODUCTION_TAG.equals(tag)
-            || SECONDARY_PRODUCTION_TAG.equals(tag)) {
-            GoodsType type = xr.getType(spec, GOODS_TYPE_TAG,
-                                        GoodsType.class, (GoodsType)null);
-            int amount = xr.getAttribute(VALUE_TAG, 0);
-            AbstractGoods goods = new AbstractGoods(type, amount);
-            String tileProduction = xr.getAttribute(TILE_PRODUCTION_TAG,
-                                                    (String)null);
-            // CAUTION: this only works if the primary production is
-            // defined before the secondary production
-            if (PRIMARY_PRODUCTION_TAG.equals(tag)) {
-                productionTypes.add(new ProductionType(goods, true,
-                                                       tileProduction));
-            } else if (SECONDARY_PRODUCTION_TAG.equals(tag)) {
-                String level = spec.getDifficultyLevel();
-                List<ProductionType> unattendedTypes
-                    = getAvailableProductionTypes(true, level);
-                for (ProductionType productionType : unattendedTypes) {
-                    if (tileProduction == null
-                        || tileProduction.equals(productionType.getProductionLevel())) {
-                        productionType.getOutputs().add(goods);
-                    }
-                }
-            } else {
-                productionTypes.add(new ProductionType(goods, false,
-                                                       tileProduction));
-            }
-            xr.closeTag(tag);
-            // end @compat 0.10.6
+        } else if (PRODUCTION_TAG.equals(tag)) {
+            this.productionTypes.add(new ProductionType(xr, spec));
 
         } else if (RESOURCE_TAG.equals(tag)) {
             addResourceType(xr.getType(spec, TYPE_TAG, ResourceType.class,
@@ -642,16 +683,20 @@ public final class TileType extends FreeColGameObjectType {
                             xr.getAttribute(PROBABILITY_TAG, 100));
             xr.closeTag(RESOURCE_TAG);
 
-        } else if (Modifier.getXMLElementTagName().equals(tag)) {
-            // @compat 0.10.7
-            // the tile type no longer contains the base production modifier
-            String id = xr.getAttribute(ID_ATTRIBUTE_TAG, null);
-            if (id.startsWith("model.goods.")) {
-                xr.closeTag(Modifier.getXMLElementTagName());
-            } else {
-                super.readChild(xr);
-            }
-            // end @compat
+        // @compat 0.11.x
+        // Primary and secondary production was dropped at 0.11.0, but
+        // some saved games slipped through.
+        } else if (PRIMARY_PRODUCTION_TAG.equals(tag)
+            || SECONDARY_PRODUCTION_TAG.equals(tag)) {
+            GoodsType type = xr.getType(spec, GOODS_TYPE_TAG,
+                                        GoodsType.class, (GoodsType)null);
+            int amount = xr.getAttribute(VALUE_TAG, 0);
+            ProductionType pt = new ProductionType(null, type, amount);
+            pt.setUnattended(true);
+            this.productionTypes.add(pt);
+            xr.closeTag(tag);
+        // @end compat 0.11.x
+
         } else {
             super.readChild(xr);
         }
@@ -660,15 +705,5 @@ public final class TileType extends FreeColGameObjectType {
     /**
      * {@inheritDoc}
      */
-    @Override
-    public String getXMLTagName() { return getXMLElementTagName(); }
-
-    /**
-     * Gets the tag name of the root element representing this object.
-     *
-     * @return "tile-type".
-     */
-    public static String getXMLElementTagName() {
-        return "tile-type";
-    }
+    public String getXMLTagName() { return TAG; }
 }
