@@ -1,5 +1,5 @@
 /**
- *  Copyright (C) 2002-2015   The FreeCol Team
+ *  Copyright (C) 2002-2022   The FreeCol Team
  *
  *  This file is part of FreeCol.
  *
@@ -21,15 +21,18 @@ package net.sf.freecol.common.model;
 
 import java.util.ArrayList;
 import java.util.Collections;
-import java.util.Iterator;
+import java.util.Comparator;
 import java.util.List;
-import java.util.stream.Collectors;
+import java.util.stream.Stream;
 
+import javax.swing.JList;
+import javax.swing.ListModel;
 import javax.xml.stream.XMLStreamException;
 
 import net.sf.freecol.common.io.FreeColXMLReader;
 import net.sf.freecol.common.io.FreeColXMLWriter;
-import net.sf.freecol.common.model.UnitTypeChange.ChangeType;
+import net.sf.freecol.common.model.Colony.NoBuildReason;
+import static net.sf.freecol.common.model.Constants.*;
 import static net.sf.freecol.common.util.CollectionUtils.*;
 
 
@@ -38,26 +41,33 @@ import static net.sf.freecol.common.util.CollectionUtils.*;
  */
 public final class UnitType extends BuildableType implements Consumer {
 
+    public static final String TAG = "unit-type";
+
+    /** Comparator for defence ability. */
+    public static final Comparator<UnitType> defenceComparator
+            = Comparator.comparingDouble(UnitType::getDefence);
+
     /** The default offence value. */
     public static final int DEFAULT_OFFENCE = 0;
 
     /** The default offence value. */
     public static final int DEFAULT_DEFENCE = 1;
 
+
     /**
      * The offence of this UnitType. Only Units with an offence value
      * greater than zero can attack.
      */
-    private int offence = DEFAULT_OFFENCE;
+    private int baseOffence = DEFAULT_OFFENCE;
 
     /** The defence of this UnitType. */
-    private int defence = DEFAULT_DEFENCE;
+    private int baseDefence = DEFAULT_DEFENCE;
 
     /** The capacity of this UnitType. */
     private int space = 0;
 
     /** Is this the default unit type? */
-    private boolean defaultUnit = false;
+    private boolean defaultUnitType = false;
 
     /**
      * The number of hit points this UnitType has. At the moment, this
@@ -75,6 +85,9 @@ public final class UnitType extends BuildableType implements Consumer {
     /** The price of this UnitType. */
     private int price = UNDEFINED;
 
+    /** The optional price of this UnitType for use by mercenary forces. */
+    private int mercenaryPrice = UNDEFINED;
+
     /** The initial moves of this UnitType. */
     private int movement = 3;
 
@@ -108,18 +121,15 @@ public final class UnitType extends BuildableType implements Consumer {
     /** The default role for a unit of this type. */
     private Role defaultRole = null;
 
-    /** The possible type changes for this unit type. */
-    private List<UnitTypeChange> typeChanges = null;
-
     /** The goods consumed per turn when in a settlement. */
     private TypeCountMap<GoodsType> consumption = null;
 
 
     /**
-     * Creates a new <code>UnitType</code> instance.
+     * Creates a new {@code UnitType} instance.
      *
      * @param id The object identifier.
-     * @param specification The <code>Specification</code> to refer to.
+     * @param specification The {@code Specification} to refer to.
      */
     public UnitType(String id, Specification specification) {
         super(id, specification);
@@ -170,7 +180,7 @@ public final class UnitType extends BuildableType implements Consumer {
      * @return The base offence value.
      */
     public int getBaseOffence() {
-        return offence;
+        return this.baseOffence;
     }
 
     /**
@@ -179,7 +189,7 @@ public final class UnitType extends BuildableType implements Consumer {
      * @return The offence value.
      */
     public double getOffence() {
-        return applyModifiers(offence, null, Modifier.OFFENCE);
+        return apply(this.baseOffence, null, Modifier.OFFENCE);
     }
 
     /**
@@ -197,7 +207,7 @@ public final class UnitType extends BuildableType implements Consumer {
      * @return The defence value.
      */
     public int getBaseDefence() {
-        return defence;
+        return this.baseDefence;
     }
 
     /**
@@ -206,7 +216,7 @@ public final class UnitType extends BuildableType implements Consumer {
      * @return The defence value.
      */
     public double getDefence() {
-        return applyModifiers(defence, null, Modifier.DEFENCE);
+        return apply(this.baseDefence, null, Modifier.DEFENCE);
     }
 
     /**
@@ -228,7 +238,7 @@ public final class UnitType extends BuildableType implements Consumer {
      * @return True if this is the default unit type.
      */
     public boolean isDefaultUnitType() {
-        return defaultUnit;
+        return this.defaultUnitType;
     }
 
     /**
@@ -335,6 +345,28 @@ public final class UnitType extends BuildableType implements Consumer {
         return price;
     }
 
+    /**
+     * Get the special price used by mercenary forces when there is no
+     * standard base price (e.g. ManOWar).
+     *
+     * @return The mercenary price.
+     */
+    public int getMercenaryPrice() {
+        return mercenaryPrice;
+    }
+
+    // @compat 0.11.6
+    // Only used in spec fixups
+    /**
+     * Set the mercenary price.
+     *
+     * @param price The mercenary price.
+     */
+    public void setMercenaryPrice(int price) {
+        this.mercenaryPrice = price;
+    }
+    // end @compat 0.11.6
+    
     /**
      * Get the base movement of this unit type.
      *
@@ -353,6 +385,15 @@ public final class UnitType extends BuildableType implements Consumer {
         return maximumExperience;
     }
 
+    /**
+     * Is this unit type subject to attrition?
+     *
+     * @return True if attrition can happen for this unit type.
+     */
+    public boolean hasMaximumAttrition() {
+        return maximumAttrition != INFINITY;
+    }
+
     /**
      * Get the maximum attrition for this unit type (greater attrition than
      * this destroys the unit).
@@ -366,7 +407,7 @@ public final class UnitType extends BuildableType implements Consumer {
     /**
      * Get the type of goods this unit type has expert ability to produce.
      *
-     * @return The expert production <code>GoodsType</code>.
+     * @return The expert production {@code GoodsType}.
      */
     public GoodsType getExpertProduction() {
         return expertProduction;
@@ -384,7 +425,7 @@ public final class UnitType extends BuildableType implements Consumer {
     /**
      * Gets the default role of this unit type, mostly model.role.default.
      *
-     * @return The default <code>Role</code>.
+     * @return The default {@code Role}.
      */
     public Role getDefaultRole() {
         return defaultRole;
@@ -397,9 +438,8 @@ public final class UnitType extends BuildableType implements Consumer {
      * @return a list of expert roles
      */
     public List<Role> getExpertRoles() {
-        return getSpecification().getRoles().stream()
-            .filter(r -> r.getExpertUnit() == this)
-            .collect(Collectors.toList());
+        return transform(getSpecification().getRoles(),
+                matchKey(this, Role::getExpertUnit));
     }
 
     /**
@@ -408,138 +448,39 @@ public final class UnitType extends BuildableType implements Consumer {
      * @return A suitable role identifier for display purposes.
      */
     public String getDisplayRoleId() {
-        for (Role r : getExpertRoles()) return r.getId();
-        return Specification.DEFAULT_ROLE_ID;
+        Role r = first(getExpertRoles());
+        return (r != null) ? r.getId() : Specification.DEFAULT_ROLE_ID;
     }
 
     /**
-     * Gets the list of all type changes associated with this unit type.
+     * Get the type that this unit type can be educated to by a
+     * teacher unit type, if any.
      *
-     * @return The list of type changes.
+     * @param teacherType The {@code UnitType} of the teacher.
+     * @return The {@code UnitType} that this unit type can be educated
+     *     to by the teacher unit type, or null if education is not possible.
      */
-    public List<UnitTypeChange> getTypeChanges() {
-        return (typeChanges == null) ? Collections.<UnitTypeChange>emptyList()
-            : typeChanges;
-    }
+    public UnitType getTeachingType(UnitType teacherType) {
+        final Specification spec = getSpecification();
+        final UnitType taught = teacherType.getSkillTaught();
+        final int taughtLevel = taught.getSkill();
+        if (getSkill() >= taughtLevel) return null; // Fail fast
 
-    /**
-     * Sets the list of all type changes associated with this unit type.
-     * Public for the test suite.
-     *
-     * @param typeChanges The new list of type changes.
-     */
-    public void setTypeChanges(List<UnitTypeChange> typeChanges) {
-        this.typeChanges = typeChanges;
-    }
-
-    /**
-     * Add a unit type change.
-     *
-     * @param change The <code>UnitTypeChange</code> to add.
-     */
-    private void addTypeChange(UnitTypeChange change) {
-        if (typeChanges == null) typeChanges = new ArrayList<>();
-        typeChanges.add(change);
-    }
-
-    /**
-     * Gets a unit type resulting from a given change type and player.
-     *
-     * @param changeType A <code>ChangeType</code> to match.
-     * @param player A <code>Player</code> to check.
-     * @return Any changed <code>UnitType</code> found, or null if none.
-     */
-    public UnitType getTargetType(ChangeType changeType, Player player) {
-        UnitTypeChange change = getUnitTypeChange(changeType, player);
-        return (change == null) ? null : change.getNewUnitType();
-    }
-
-    /**
-     * Gets any suitable unit type changes applicable to a given change type
-     * and player.
-     *
-     * @param changeType The <code>ChangeType</code> to match.
-     * @param player The <code>Player</code> to check.
-     * @return Any <code>UnitTypeChange</code> found, or null if none.
-     */
-    public UnitTypeChange getUnitTypeChange(ChangeType changeType,
-                                            Player player) {
-        return find(getTypeChanges(),
-            c -> c.asResultOf(changeType) && c.appliesTo(player)
-                && c.getNewUnitType().isAvailableTo(player));
-    }
-
-    /**
-     * Gets the type change required to become another given unit type.
-     *
-     * @param newType The target <code>UnitType</code>.
-     * @return The type change, or null if impossible.
-     */
-    public UnitTypeChange getUnitTypeChange(UnitType newType) {
-        return find(getTypeChanges(), c -> c.getNewUnitType() == newType);
-    }
-
-    /**
-     * Can this type of unit be upgraded to another given type by a given
-     * educational change type?
-     *
-     * If the target type is null, return true if the UnitType can be
-     * upgraded to any other type by the given means of education.
-     *
-     * @param newType The <code>UnitType</code> to learn (may be null
-     *     in the case of attempting to move to a native settlement
-     *     when the skill taught there is still unknown).
-     * @param changeType The educational <code>ChangeType</code>.
-     * @return True if this unit type can learn.
-     */
-    public boolean canBeUpgraded(UnitType newType, ChangeType changeType) {
-        return any(getTypeChanges(),
-            c -> (newType == null || newType == c.getNewUnitType())
-                && c.getProbability(changeType) > 0);
-    }
-
-    /**
-     * Gets the unit types which can be learned from a lost city rumour.
-     *
-     * @return A list of unit types.
-     */
-    public List<UnitType> getUnitTypesLearntInLostCity() {
-        List<UnitType> unitTypes = new ArrayList<>();
-        for (UnitTypeChange change : getTypeChanges()) {
-            if (change.asResultOf(ChangeType.LOST_CITY)) {
-                unitTypes.add(change.getNewUnitType());
-            }
+        // Is there an education change that gets this unit type to the
+        // type taught by the teacher type?  If so, the taught type is valid
+        // and should be returned at once.  Accumulate other intermediate
+        // changes that do not reach the taught type level.
+        List<UnitType> todo = new ArrayList<>();
+        for (UnitTypeChange uc : spec.getUnitChanges(UnitChangeType.EDUCATION, this)) {
+            if (uc.to == taught) return taught;
+            if (uc.to.getSkill() < taughtLevel) todo.add(uc.to);
         }
-        return unitTypes;
-    }
-
-    /**
-     * Get a UnitType to learn with a level skill less or equal than
-     * given level.
-     *
-     * @param maximumSkill The maximum level skill which we are searching for.
-     * @return A <code>UnitType</code> with a skill equal or less than given
-     *     maximum.
-     */
-    public UnitType getEducationUnit(int maximumSkill) {
-        return find(getTypeChanges().stream()
-                .filter(c -> c.canBeTaught())
-                .map(UnitTypeChange::getNewUnitType),
-            ut -> ut.hasSkill() && ut.getSkill() <= maximumSkill, null);
-    }
-
-    /**
-     * Get the number of turns to educate this unit type to become
-     * another type.
-     *
-     * @param unitType The <code>UnitType</code> to teach.
-     * @return The number of turns, or UNDEFINED if impossible.
-     */
-    public int getEducationTurns(UnitType unitType) {
-        UnitTypeChange utc = find(getTypeChanges(),
-            c -> c.asResultOf(UnitTypeChange.ChangeType.EDUCATION)
-                && unitType == c.getNewUnitType());
-        return (utc == null) ? UNDEFINED : utc.getTurnsToLearn();
+        // Can the teacher teach any of the intermediate changes?  If so,
+        // that change is valid.  Otherwise, education is not possible.
+        for (UnitType ut : todo) {
+            if (ut.getTeachingType(teacherType) != null) return ut;
+        }
+        return null;
     }
 
     /**
@@ -560,6 +501,15 @@ public final class UnitType extends BuildableType implements Consumer {
         return hasAbility(Ability.PERSON);
     }
 
+    /**
+     * Is this unit type able to build a colony?
+     *
+     * @return True if this unit type can build colonies.
+     */
+    public boolean canBuildColony() {
+        return hasAbility(Ability.FOUND_COLONY);
+    }
+
     /**
      * Can this unit type move to the High Seas?
      *
@@ -590,10 +540,29 @@ public final class UnitType extends BuildableType implements Consumer {
         return price != UNDEFINED;
     }
 
+    /**
+     * Get the consumption map.
+     *
+     * @return The map of the consumed goods.
+     */
+    protected TypeCountMap<GoodsType> getConsumption() {
+        return this.consumption;
+    }
+
+    /**
+     * Set the consumption map.
+     *
+     * @param consumption The new map of the consumed goods.
+     */
+    protected void setConsumption(TypeCountMap<GoodsType> consumption) {
+        this.consumption = consumption;
+    }
+
     /**
      * Gets the number of units of the given GoodsType this UnitType
      * consumes per turn (when in a settlement).
      *
+     * @param goodsType The {@code GoodsType} to consume.
      * @return The amount of goods consumed per turn.
      */
     public int getConsumptionOf(GoodsType goodsType) {
@@ -603,7 +572,7 @@ public final class UnitType extends BuildableType implements Consumer {
     /**
      * Add consumption.
      *
-     * @param type The <code>GoodsType</code> to consume.
+     * @param type The {@code GoodsType} to consume.
      * @param amount The amount of goods to consume.
      */
     private void addConsumption(GoodsType type, int amount) {
@@ -613,49 +582,70 @@ public final class UnitType extends BuildableType implements Consumer {
         consumption.incrementCount(type, amount);
     }
 
-
-    // Override FreeColObject
-
     /**
      * {@inheritDoc}
      */
-    @Override
-    public int compareTo(FreeColObject other) {
-        int cmp = 0;
-        if (other instanceof UnitType) {
-            UnitType ut = (UnitType)other;
-            cmp = getIndex() - ut.getIndex();
+    public NoBuildReason canBeBuiltInColony(Colony colony,
+                                            List<BuildableType> assumeBuilt) {
+        // Non-person units need a BUILD ability, present or assumed.
+        if (!hasAbility(Ability.PERSON)
+            && !colony.hasAbility(Ability.BUILD, this)
+            && none(assumeBuilt, bt -> bt.hasAbility(Ability.BUILD, this))) {
+            return Colony.NoBuildReason.MISSING_BUILD_ABILITY;
         }
-        if (cmp == 0) cmp = super.compareTo(other);
-        return cmp;
+        return Colony.NoBuildReason.NONE;
+    }
+
+    @Override
+    public int getMinimumIndex(Colony colony, JList<BuildableType> buildQueueList, int UNABLE_TO_BUILD) {
+        ListModel<BuildableType> buildQueue = buildQueueList.getModel();
+        if (colony.canBuild(this)) return 0;
+        for (int index = 0; index < buildQueue.getSize(); index++) {
+            if (buildQueue.getElementAt(index).hasAbility(Ability.BUILD, this)) return index + 1;
+        }
+        return UNABLE_TO_BUILD;
+    }
+
+    @Override
+    public int getMaximumIndex(Colony colony, JList<BuildableType> buildQueueList, int UNABLE_TO_BUILD) {
+        ListModel<BuildableType> buildQueue = buildQueueList.getModel();
+        final int buildQueueLastPos = buildQueue.getSize();
+
+        boolean canBuild = false;
+        if (colony.canBuild(this)) {
+            canBuild = true;
+        }
+
+        // does not depend on anything, nothing depends on it
+        // can be built at any time
+        if (canBuild) return buildQueueLastPos;
+        // check for building in queue that allows builting this unit
+        for (int index = 0; index < buildQueue.getSize(); index++) {
+            BuildableType toBuild = buildQueue.getElementAt(index);
+            if (toBuild == this) continue;
+            if (toBuild.hasAbility(Ability.BUILD, this)) {
+                return buildQueueLastPos;
+            }
+        }
+        return UNABLE_TO_BUILD;
     }
 
 
     // Interface Consumer
 
     /**
-     * Gets a list of goods this Consumer consumes.
-     *
-     * @return The goods consumed by this unit type.
+     * {@inheritDoc}
      */
     @Override
     public List<AbstractGoods> getConsumedGoods() {
-        List<AbstractGoods> result = new ArrayList<>();
-        if (consumption != null) {
-            for (GoodsType goodsType : consumption.keySet()) {
-                result.add(new AbstractGoods(goodsType,
-                        consumption.getCount(goodsType)));
-            }
-        }
-        return result;
+        return (consumption == null) ? Collections.<AbstractGoods>emptyList()
+                : transform(consumption.keySet(),
+                gt -> consumption.getCount(gt) != 0,
+                gt -> new AbstractGoods(gt, consumption.getCount(gt)));
     }
 
     /**
-     * Gets the priority of this Consumer.  The higher the priority,
-     * the earlier will the Consumer be allowed to consume the goods
-     * it requires.
-     *
-     * @return The priority of this unit type.
+     * {@inheritDoc}
      */
     @Override
     public int getPriority() {
@@ -663,27 +653,56 @@ public final class UnitType extends BuildableType implements Consumer {
     }
 
     /**
-     * Is this unit type able to build a colony?
-     *
-     * @return True if this unit type can build colonies.
+     * {@inheritDoc}
      */
-    public boolean canBuildColony() {
-        return hasAbility(Ability.FOUND_COLONY);
+    @Override
+    public Stream<Modifier> getConsumptionModifiers(String id) {
+        return getModifiers(id);
+    }
+
+
+    // Override FreeColObject
+
+    /**
+     * {@inheritDoc}
+     */
+    @Override
+    public <T extends FreeColObject> boolean copyIn(T other) {
+        UnitType o = copyInCast(other, UnitType.class);
+        if (o == null || !super.copyIn(o)) return false;
+        this.baseOffence = o.getBaseOffence();
+        this.baseDefence = o.getBaseDefence();
+        this.space = o.getSpace();
+        this.defaultUnitType = o.isDefaultUnitType();
+        this.hitPoints = o.getHitPoints();
+        this.spaceTaken = o.getSpaceTaken();
+        this.skill = o.getSkill();
+        this.price = o.getPrice();
+        this.movement = o.getMovement();
+        this.lineOfSight = o.getLineOfSight();
+        this.recruitProbability = o.getRecruitProbability();
+        this.expertProduction = o.getExpertProduction();
+        this.scoreValue = o.getScoreValue();
+        this.maximumExperience = o.getMaximumExperience();
+        this.maximumAttrition = o.getMaximumAttrition();
+        this.priority = o.getPriority();
+        this.skillTaught = o.getSkillTaught();
+        this.defaultRole = o.getDefaultRole();
+        this.consumption = o.getConsumption();
+        return true;
     }
 
 
     // Serialization
 
     private static final String CONSUMES_TAG = "consumes";
-    // @compat 0.10.7
-    private static final String DEFAULT_EQUIPMENT_TAG = "default-equipment";
-    // end @compat
     private static final String DEFAULT_ROLE_TAG = "default-role";
     private static final String DEFAULT_UNIT_TAG = "default-unit";
     private static final String DEFENCE_TAG = "defence";
     private static final String EXPERT_PRODUCTION_TAG = "expert-production";
     private static final String HIT_POINTS_TAG = "hit-points";
     private static final String LINE_OF_SIGHT_TAG = "line-of-sight";
+    private static final String MERCENARY_PRICE_TAG = "mercenary-price";
     private static final String MOVEMENT_TAG = "movement";
     private static final String MAXIMUM_EXPERIENCE_TAG = "maximum-experience";
     private static final String MAXIMUM_ATTRITION_TAG = "maximum-attrition";
@@ -696,9 +715,10 @@ public final class UnitType extends BuildableType implements Consumer {
     private static final String SKILL_TAUGHT_TAG = "skill-taught";
     private static final String SPACE_TAG = "space";
     private static final String SPACE_TAKEN_TAG = "space-taken";
-    private static final String DOWNGRADE_TAG = "downgrade";
     private static final String UNIT_TAG = "unit";
-    private static final String UPGRADE_TAG = "upgrade";
+    // @compat 0.11.0
+    private static final String OLD_DEFAULT_EQUIPMENT_TAG = "default-equipment";
+    // end @compat 0.11.0
     // @compat 0.11.3
     private static final String OLD_DEFAULT_UNIT_TAG = "defaultUnit";
     private static final String OLD_HIT_POINTS_TAG = "hitPoints";
@@ -710,6 +730,10 @@ public final class UnitType extends BuildableType implements Consumer {
     private static final String OLD_SKILL_TAUGHT_TAG = "skillTaught";
     private static final String OLD_SPACE_TAKEN_TAG = "spaceTaken";
     // end @compat 0.11.3
+    // @compat 0.11.6
+    private static final String DOWNGRADE_TAG = "downgrade";
+    private static final String UPGRADE_TAG = "upgrade";
+    // end @compat 0.11.6
 
     /**
      * {@inheritDoc}
@@ -718,11 +742,11 @@ public final class UnitType extends BuildableType implements Consumer {
     protected void writeAttributes(FreeColXMLWriter xw) throws XMLStreamException {
         super.writeAttributes(xw);
 
-        xw.writeAttribute(OFFENCE_TAG, offence);
+        xw.writeAttribute(OFFENCE_TAG, baseOffence);
 
-        xw.writeAttribute(DEFENCE_TAG, defence);
+        xw.writeAttribute(DEFENCE_TAG, baseDefence);
 
-        xw.writeAttribute(DEFAULT_UNIT_TAG, defaultUnit);
+        xw.writeAttribute(DEFAULT_UNIT_TAG, this.defaultUnitType);
 
         xw.writeAttribute(MOVEMENT_TAG, movement);
 
@@ -738,7 +762,7 @@ public final class UnitType extends BuildableType implements Consumer {
 
         xw.writeAttribute(MAXIMUM_EXPERIENCE_TAG, maximumExperience);
 
-        if (maximumAttrition < INFINITY) {
+        if (hasMaximumAttrition()) {
             xw.writeAttribute(MAXIMUM_ATTRITION_TAG, maximumAttrition);
         }
 
@@ -752,6 +776,10 @@ public final class UnitType extends BuildableType implements Consumer {
             xw.writeAttribute(PRICE_TAG, price);
         }
 
+        if (mercenaryPrice != UNDEFINED) {
+            xw.writeAttribute(MERCENARY_PRICE_TAG, mercenaryPrice);
+        }
+
         xw.writeAttribute(SKILL_TAUGHT_TAG, skillTaught);
 
         if (expertProduction != null) {
@@ -778,8 +806,6 @@ public final class UnitType extends BuildableType implements Consumer {
             xw.writeEndElement();
         }
 
-        for (UnitTypeChange change : getTypeChanges()) change.toXML(xw);
-
         if (consumption != null) {
             for (GoodsType goodsType : consumption.keySet()) {
                 xw.writeStartElement(CONSUMES_TAG);
@@ -804,16 +830,16 @@ public final class UnitType extends BuildableType implements Consumer {
 
         UnitType parent = xr.getType(spec, EXTENDS_TAG, UnitType.class, this);
 
-        offence = xr.getAttribute(OFFENCE_TAG, parent.offence);
+        this.baseOffence = xr.getAttribute(OFFENCE_TAG, parent.baseOffence);
 
-        defence = xr.getAttribute(DEFENCE_TAG, parent.defence);
+        this.baseDefence = xr.getAttribute(DEFENCE_TAG, parent.baseDefence);
 
         // @compat 0.11.3
         if (xr.hasAttribute(OLD_DEFAULT_UNIT_TAG)) {
-            defaultUnit = xr.getAttribute(OLD_DEFAULT_UNIT_TAG, false);
+            this.defaultUnitType = xr.getAttribute(OLD_DEFAULT_UNIT_TAG, false);
         } else
-        // end @compat 0.11.3
-            defaultUnit = xr.getAttribute(DEFAULT_UNIT_TAG, false);
+            // end @compat 0.11.3
+            this.defaultUnitType = xr.getAttribute(DEFAULT_UNIT_TAG, false);
 
         movement = xr.getAttribute(MOVEMENT_TAG, parent.movement);
 
@@ -821,14 +847,14 @@ public final class UnitType extends BuildableType implements Consumer {
         if (xr.hasAttribute(OLD_LINE_OF_SIGHT_TAG)) {
             lineOfSight = xr.getAttribute(OLD_LINE_OF_SIGHT_TAG, parent.lineOfSight);
         } else
-        // end @compat 0.11.3
+            // end @compat 0.11.3
             lineOfSight = xr.getAttribute(LINE_OF_SIGHT_TAG, parent.lineOfSight);
 
         // @compat 0.11.3
         if (xr.hasAttribute(OLD_SCORE_VALUE_TAG)) {
             scoreValue = xr.getAttribute(OLD_SCORE_VALUE_TAG, parent.scoreValue);
         } else
-        // end @compat 0.11.3
+            // end @compat 0.11.3
             scoreValue = xr.getAttribute(SCORE_VALUE_TAG, parent.scoreValue);
 
         space = xr.getAttribute(SPACE_TAG, parent.space);
@@ -837,49 +863,49 @@ public final class UnitType extends BuildableType implements Consumer {
         if (xr.hasAttribute(OLD_HIT_POINTS_TAG)) {
             hitPoints = xr.getAttribute(OLD_HIT_POINTS_TAG, parent.hitPoints);
         } else
-        // end @compat 0.11.3
+            // end @compat 0.11.3
             hitPoints = xr.getAttribute(HIT_POINTS_TAG, parent.hitPoints);
 
         // @compat 0.11.3
         if (xr.hasAttribute(OLD_SPACE_TAKEN_TAG)) {
             spaceTaken = xr.getAttribute(OLD_SPACE_TAKEN_TAG, parent.spaceTaken);
         } else
-        // end @compat 0.11.3
+            // end @compat 0.11.3
             spaceTaken = xr.getAttribute(SPACE_TAKEN_TAG, parent.spaceTaken);
 
         // @compat 0.11.3
         if (xr.hasAttribute(OLD_MAXIMUM_EXPERIENCE_TAG)) {
             maximumExperience = xr.getAttribute(OLD_MAXIMUM_EXPERIENCE_TAG,
-                                                parent.maximumExperience);
+                    parent.maximumExperience);
         } else
-        // end @compat 0.11.3
+            // end @compat 0.11.3
             maximumExperience = xr.getAttribute(MAXIMUM_EXPERIENCE_TAG,
-                                                parent.maximumExperience);
+                    parent.maximumExperience);
 
         // @compat 0.11.3
         if (xr.hasAttribute(OLD_MAXIMUM_ATTRITION_TAG)) {
             maximumAttrition = xr.getAttribute(OLD_MAXIMUM_ATTRITION_TAG,
-                                               parent.maximumAttrition);
+                    parent.maximumAttrition);
         } else
-        // end @compat 0.11.3
+            // end @compat 0.11.3
             maximumAttrition = xr.getAttribute(MAXIMUM_ATTRITION_TAG,
-                                               parent.maximumAttrition);
+                    parent.maximumAttrition);
 
         // @compat 0.11.3
         if (xr.hasAttribute(OLD_SKILL_TAUGHT_TAG)) {
             skillTaught = xr.getType(spec, OLD_SKILL_TAUGHT_TAG, UnitType.class, this);
         } else
-        // end @compat 0.11.3
+            // end @compat 0.11.3
             skillTaught = xr.getType(spec, SKILL_TAUGHT_TAG, UnitType.class, this);
 
         // @compat 0.11.3
         if (xr.hasAttribute(OLD_RECRUIT_PROBABILITY_TAG)) {
             recruitProbability = xr.getAttribute(OLD_RECRUIT_PROBABILITY_TAG,
-                                                 parent.recruitProbability);
+                    parent.recruitProbability);
         } else
-        // end @compat 0.11.3
+            // end @compat 0.11.3
             recruitProbability = xr.getAttribute(RECRUIT_PROBABILITY_TAG,
-                                                 parent.recruitProbability);
+                    parent.recruitProbability);
 
         // New in 0.11.4, but default is backward compatible.
         priority = xr.getAttribute(PRIORITY_TAG, Consumer.UNIT_PRIORITY);
@@ -888,8 +914,11 @@ public final class UnitType extends BuildableType implements Consumer {
 
         price = xr.getAttribute(PRICE_TAG, parent.price);
 
+        mercenaryPrice = xr.getAttribute(MERCENARY_PRICE_TAG,
+                                         parent.mercenaryPrice);
+        
         expertProduction = xr.getType(spec, EXPERT_PRODUCTION_TAG,
-                                      GoodsType.class, parent.expertProduction);
+                GoodsType.class, parent.expertProduction);
 
         if (parent != this) { // Handle "extends" for super-type fields
             if (!xr.hasAttribute(REQUIRED_POPULATION_TAG)) {
@@ -908,7 +937,6 @@ public final class UnitType extends BuildableType implements Consumer {
         // Clear containers.
         if (xr.shouldClearContainers()) {
             consumption = null;
-            typeChanges = null;
         }
         defaultRole = spec.getDefaultRole();
 
@@ -916,11 +944,6 @@ public final class UnitType extends BuildableType implements Consumer {
         if (parent != this) {
             defaultRole = parent.defaultRole;
 
-            if (parent.typeChanges != null) {
-                if (typeChanges == null) typeChanges = new ArrayList<>();
-                typeChanges.addAll(parent.typeChanges);
-            }
-
             if (parent.consumption != null) {
                 if (consumption == null) consumption = new TypeCountMap<>();
                 consumption.putAll(parent.consumption);
@@ -933,28 +956,6 @@ public final class UnitType extends BuildableType implements Consumer {
         }
 
         super.readChildren(xr);
-
-        // @compat 0.10.6
-        if (hasAbility(Ability.PERSON)) {
-            Modifier m;
-            if (!containsModifierKey(Modifier.CONVERSION_SKILL)) {
-                m = new Modifier(Modifier.CONVERSION_SKILL, 8.0f, 
-                                 Modifier.ModifierType.ADDITIVE);
-                addModifier(m);
-
-                if (hasAbility(Ability.EXPERT_MISSIONARY)) {
-                    m = new Modifier(Modifier.CONVERSION_SKILL, 5.0f,
-                                     Modifier.ModifierType.ADDITIVE);
-                    addModifier(m);
-                }
-            }
-            if (!containsModifierKey(Modifier.CONVERSION_ALARM_RATE)) {
-                m = new Modifier(Modifier.CONVERSION_ALARM_RATE, 2.0f,
-                                 Modifier.ModifierType.PERCENTAGE);
-                addModifier(m);
-            }
-        }
-        // end @compat
     }
 
     /**
@@ -967,55 +968,24 @@ public final class UnitType extends BuildableType implements Consumer {
 
         if (CONSUMES_TAG.equals(tag)) {
             addConsumption(xr.getType(spec, ID_ATTRIBUTE_TAG,
-                                      GoodsType.class, (GoodsType)null),
-                           xr.getAttribute(VALUE_TAG, UNDEFINED));
+                    GoodsType.class, (GoodsType)null),
+                    xr.getAttribute(VALUE_TAG, UNDEFINED));
             xr.closeTag(CONSUMES_TAG);
 
-        // @compat 0.10.7
-        } else if (DEFAULT_EQUIPMENT_TAG.equals(tag)) {
-            String id = xr.getAttribute(ID_ATTRIBUTE_TAG, null);
-            String roleId = ("model.equipment.horses".equals(id))
-                ? "model.role.scout"
-                : ("model.equipment.muskets".equals(id))
-                ? "model.role.soldier"
-                : ("model.equipment.tools".equals(id))
-                ? "model.role.pioneer"
-                : ("model.equipment.missionary".equals(id))
-                ? "model.role.missionary"
-                : Specification.DEFAULT_ROLE_ID;
-            defaultRole = spec.getRole(roleId);
-            xr.closeTag(DEFAULT_EQUIPMENT_TAG);
-        // end @compat
+        // @compat 0.11.0
+        } else if (OLD_DEFAULT_EQUIPMENT_TAG.equals(tag)) {
+            xr.swallowTag(OLD_DEFAULT_EQUIPMENT_TAG);
+        // end @compat 0.11.0
 
         } else if (DEFAULT_ROLE_TAG.equals(tag)) {
             defaultRole = xr.getType(spec, ID_ATTRIBUTE_TAG,
-                                     Role.class, spec.getDefaultRole());
+                    Role.class, spec.getDefaultRole());
             xr.closeTag(DEFAULT_ROLE_TAG);
 
+        // @compat 0.11.6
         } else if (DOWNGRADE_TAG.equals(tag) || UPGRADE_TAG.equals(tag)) {
-            if (xr.getAttribute(DELETE_TAG, false)) {
-                if (typeChanges != null) {
-                    String unitId = xr.getAttribute(UNIT_TAG, (String)null);
-                    Iterator<UnitTypeChange> it = typeChanges.iterator();
-                    while (it.hasNext()) {
-                        if (unitId.equals(it.next().getNewUnitType().getId())) {
-                            it.remove();
-                            break;
-                        }
-                    }
-                }
-                xr.closeTag(tag);
-
-            } else {
-                UnitTypeChange change
-                    = new UnitTypeChange(xr, spec);// Closes tag
-                if (DOWNGRADE_TAG.equals(tag)
-                    && change.getChangeTypes().isEmpty()) {
-                    // add default downgrade type
-                    change.getChangeTypes().put(ChangeType.CLEAR_SKILL, 100);
-                }
-                addTypeChange(change);
-            }
+            xr.closeTag(tag, Scope.TAG);
+        // end @compat 0.11.6
 
         } else {
             super.readChild(xr);
@@ -1025,23 +995,16 @@ public final class UnitType extends BuildableType implements Consumer {
     /**
      * {@inheritDoc}
      */
-    @Override
-    public String toString() {
-        return getId();
-    }
+    public String getXMLTagName() { return TAG; }
+
+
+    // Override Object
 
     /**
      * {@inheritDoc}
      */
     @Override
-    public String getXMLTagName() { return getXMLElementTagName(); }
-
-    /**
-     * Gets the tag name of the root element representing this object.
-     *
-     * @return "unit-type".
-     */
-    public static String getXMLElementTagName() {
-        return "unit-type";
+    public String toString() {
+        return getId();
     }
 }
