@@ -1,5 +1,5 @@
 /**
- *  Copyright (C) 2002-2015   The FreeCol Team
+ *  Copyright (C) 2002-2022   The FreeCol Team
  *
  *  This file is part of FreeCol.
  *
@@ -24,22 +24,23 @@ import java.util.Collections;
 import java.util.Comparator;
 import java.util.HashMap;
 import java.util.List;
+import java.util.function.ToIntFunction;
 import java.util.logging.Level;
 import java.util.logging.Logger;
+import java.util.stream.Stream;
 
 import javax.xml.stream.XMLStreamException;
 
 import net.sf.freecol.common.io.FreeColXMLReader;
 import net.sf.freecol.common.io.FreeColXMLWriter;
-import static net.sf.freecol.common.util.CollectionUtils.*;
 import net.sf.freecol.common.util.LogBuilder;
 import static net.sf.freecol.common.util.CollectionUtils.*;
 import net.sf.freecol.common.util.Utils;
 
 
 /**
- * The <code>WorkLocation</code> is a place in a {@link Colony} where
- * <code>Units</code> can work.  The unit capacity of a WorkLocation
+ * The {@code WorkLocation} is a place in a {@link Colony} where
+ * {@code Units} can work.  The unit capacity of a WorkLocation
  * is likely to be limited.  ColonyTiles can only hold a single
  * worker, and Buildings can hold no more than three workers, for
  * example.  WorkLocations do not store any Goods.  They take any
@@ -60,20 +61,10 @@ public abstract class WorkLocation extends UnitLocation
     public static class Suggestion {
 
         public static final Comparator<Suggestion> descendingAmountComparator
-            = new Comparator<Suggestion>() {
-                    @Override
-                    public int compare(Suggestion s1, Suggestion s2) {
-                        int cmp = s2.amount - s1.amount;
-                        if (cmp == 0) {
-                            cmp = GoodsType.goodsTypeComparator
-                                .compare(s1.goodsType, s2.goodsType);
-                        }
-                        if (cmp == 0) {
-                            cmp = s2.newType.getId().compareTo(s1.newType.getId());
-                        }
-                        return cmp;
-                    }
-                };
+            = Comparator.comparingInt(Suggestion::getAmount).reversed()
+                .thenComparing(Suggestion::getGoodsType,
+                               GoodsType.goodsTypeComparator)
+                .thenComparing(Suggestion::getNewUnitType);
                     
         public final WorkLocation workLocation;
         public final UnitType oldType;
@@ -83,17 +74,17 @@ public abstract class WorkLocation extends UnitLocation
 
 
         /**
-         * Suggest that work done by (optional) <code>oldType</code>
-         * would be better done by <code>newType</code> because it
-         * could produce <code>amount</code> more
-         * <code>goodsType</code>.
+         * Suggest that work done by (optional) {@code oldType}
+         * would be better done by {@code newType} because it
+         * could produce {@code amount} more
+         * {@code goodsType}.
          *
-         * @param workLocation The <code>WorkLocation</code> to add
+         * @param workLocation The {@code WorkLocation} to add
          *     a unit to.
-         * @param oldType The optional <code>UnitType</code> currently
+         * @param oldType The optional {@code UnitType} currently
          *     doing the work.
-         * @param newType A new <code>UnitType</code> to do the work.
-         * @param goodsType The <code>GoodsType</code> to produce.
+         * @param newType A new {@code UnitType} to do the work.
+         * @param goodsType The {@code GoodsType} to produce.
          * @param amount The extra goods that would be produced if the
          *     suggestion is taken.
          */
@@ -105,6 +96,18 @@ public abstract class WorkLocation extends UnitLocation
             this.goodsType = goodsType;
             this.amount = amount;
         }
+
+        public UnitType getNewUnitType() {
+            return this.newType;
+        }
+        
+        public GoodsType getGoodsType() {
+            return this.goodsType;
+        }
+        
+        public int getAmount() {
+            return this.amount;
+        }
     };
 
     /** The colony that contains this work location. */
@@ -117,19 +120,18 @@ public abstract class WorkLocation extends UnitLocation
     /**
      * Constructor for ServerWorkLocation.
      *
-     * @param game The enclosing <code>Game</code>.
+     * @param game The enclosing {@code Game}.
      */
     protected WorkLocation(Game game) {
         super(game);
     }
 
     /**
-     * Creates a new <code>WorkLocation</code> with the given identifier.
-     * The object should later be initialized by calling either
-     * {@link #readFromXML(FreeColXMLReader)} or
-     * {@link #readFromXMLElement(Element)}.
+     * Creates a new {@code WorkLocation} with the given identifier.
      *
-     * @param game The enclosing <code>Game</code>.
+     * The object should be initialized later.
+     *
+     * @param game The enclosing {@code Game}.
      * @param id The object identifier.
      */
     public WorkLocation(Game game, String id) {
@@ -137,33 +139,28 @@ public abstract class WorkLocation extends UnitLocation
     }
 
 
-    /**
-     * Gets the owning settlement for this work location.
-     *
-     * @return The owning settlement for this work location.
-     */
-    public Settlement getOwningSettlement() {
-        return colony;
+    public void setColony(Colony colony) {
+        this.colony = colony;
     }
 
     /**
      * Get the production type.
      *
-     * @return The <code>ProductionType</code> for this work location.
+     * @return The {@code ProductionType} for this work location.
      */
     public final ProductionType getProductionType() {
-        return productionType;
+        return this.productionType;
     }
 
     /**
      * Set the prodution type.
      *
-     * @param newProductionType The new <code>ProductionType</code> value.
+     * @param newProductionType The new {@code ProductionType} value.
      */
     public final void setProductionType(final ProductionType newProductionType) {
-        if (!Utils.equals(newProductionType, productionType)) {
-            productionType = newProductionType;
-            colony.invalidateCache();
+        if (!Utils.equals(newProductionType, this.productionType)) {
+            this.productionType = newProductionType;
+            getColony().invalidateCache();
             logger.fine("Production type at " + this
                 + " is now: " + newProductionType);
         }
@@ -175,7 +172,7 @@ public abstract class WorkLocation extends UnitLocation
      * This assumes that all units in a work location are doing the same
      * work, which is true for now.
      *
-     * @return The current <code>GoodsType</code> being produced, or
+     * @return The current {@code GoodsType} being produced, or
      *     null if none.
      */
     public GoodsType getCurrentWorkType() {
@@ -199,43 +196,33 @@ public abstract class WorkLocation extends UnitLocation
      *
      * @param unattended Whether to require unattended production.
      * @param workType An optional work type to require.
-     * @return The best available <code>ProductionType</code> given the
+     * @return The best available {@code ProductionType} given the
      *     argument constraints.
      */
     public ProductionType getBestProductionType(boolean unattended,
                                                 GoodsType workType) {
-        ProductionType best = null;
-        int amount = -1;
-        for (ProductionType pt : getAvailableProductionTypes(unattended)) {
-            for (AbstractGoods output : pt.getOutputs()) {
-                if (workType != null && workType != output.getType()) continue;
-                if (amount < output.getAmount()) {
-                    amount = output.getAmount();
-                    best = pt;
-                }
-            }
-        }
-        return best;
+        return ProductionType.getBestProductionType(workType,
+            getAvailableProductionTypes(unattended));
     }
 
     /**
      * Gets the best occupation for a given unit at this work location.
      *
-     * @param unit The <code>Unit</code> to find an
-     *     <code>Occupation</code> for.
+     * @param unit The {@code Unit} to find an
+     *     {@code Occupation} for.
      * @param userMode If a user requested this, favour the current
      *     work type, if not favour goods that the unit requires.
-     * @return An <code>Occupation</code> for the given unit, or
+     * @return An {@code Occupation} for the given unit, or
      *     null if none found.
      */
     public Occupation getOccupation(Unit unit, boolean userMode) {
-        LogBuilder lb = new LogBuilder((colony.getOccupationTrace()) ? 64 : 0);
-        lb.add(colony.getName(), "/", this, ".getOccupation(", unit, ")");
+        LogBuilder lb = new LogBuilder((getColony().getOccupationTrace()) ? 64 : 0);
+        lb.add(getColony().getName(), "/", this, ".getOccupation(", unit, ")");
 
         Occupation best = new Occupation(null, null, null);
         int bestAmount = 0;
         for (Collection<GoodsType> types
-                 : colony.getWorkTypeChoices(unit, userMode)) {
+                 : getColony().getWorkTypeChoices(unit, userMode)) {
             lb.add("\n  ");
             logFreeColObjects(types, lb);
             bestAmount = best.improve(unit, this, bestAmount, types, lb);
@@ -252,9 +239,9 @@ public abstract class WorkLocation extends UnitLocation
     /**
      * Gets the best occupation for a given unit type at this work location.
      *
-     * @param unitType An optional <code>UnitType</code> to find an
-     *     <code>Occupation</code> for.  If null, use the default unit type.
-     * @return An <code>Occupation</code> for the given unit, or
+     * @param unitType An optional {@code UnitType} to find an
+     *     {@code Occupation} for.  If null, use the default unit type.
+     * @return An {@code Occupation} for the given unit, or
      *     null if none found.
      */
     public Occupation getOccupation(UnitType unitType) {
@@ -263,8 +250,8 @@ public abstract class WorkLocation extends UnitLocation
             unitType = spec.getDefaultUnitType(getOwner().getNationType());
         }
         
-        LogBuilder lb = new LogBuilder((colony.getOccupationTrace()) ? 64 : 0);
-        lb.add(colony.getName(), "/", this, ".getOccupation(",
+        LogBuilder lb = new LogBuilder((getColony().getOccupationTrace()) ? 64 : 0);
+        lb.add(getColony().getName(), "/", this, ".getOccupation(",
                unitType.getSuffix(), ")");
 
         Collection<GoodsType> types = spec.getGoodsTypeList();
@@ -273,7 +260,7 @@ public abstract class WorkLocation extends UnitLocation
         logFreeColObjects(types, lb);
         int bestAmount = best.improve(unitType, this, 0, types, lb);
         if (best.workType != null) {
-            lb.add("\n  => ", best);
+            lb.add("\n  => ", best, "/", bestAmount);
         } else {
             lb.add("\n  FAILED");
         }
@@ -285,8 +272,8 @@ public abstract class WorkLocation extends UnitLocation
      * Get the best work type for a unit at this work location, favouring
      * the existing work.
      *
-     * @param unit The <code>Unit</code> to find a work type for.
-     * @return The best work <code>GoodsType</code> for the unit, or null
+     * @param unit The {@code Unit} to find a work type for.
+     * @return The best work {@code GoodsType} for the unit, or null
      *     if none found.
      */
     public GoodsType getWorkFor(Unit unit) {
@@ -297,7 +284,7 @@ public abstract class WorkLocation extends UnitLocation
     /**
      * Install a unit at the best occupation for it at this work location.
      *
-     * @param unit The <code>Unit</code> to install.
+     * @param unit The {@code Unit} to install.
      * @return True if the installation succeeds.
      */
     public boolean setWorkFor(Unit unit) {
@@ -309,19 +296,19 @@ public abstract class WorkLocation extends UnitLocation
      * Is it a good idea to produce a goods type at this work location
      * using a better unit type?
      *
-     * @param unit The <code>Unit</code> that is doing the job at
+     * @param unit The {@code Unit} that is doing the job at
      *     present, which may be null if none is at work.
-     * @param productionType The <code>ProductionType</code> to use.
-     * @param goodsType The <code>GoodsType</code> to produce.
-     * @return A <code>Suggestion</code> for a better worker, or null if
+     * @param productionType The {@code ProductionType} to use.
+     * @param goodsType The {@code GoodsType} to produce.
+     * @return A {@code Suggestion} for a better worker, or null if
      *     improvement is not worthwhile.
      */
     private Suggestion getSuggestion(Unit unit, ProductionType productionType,
                                      GoodsType goodsType) {
         // Check first if there is space.
-        if (((unit == null || !contains(unit)) && isFull())
-            || productionType == null
-            || goodsType == null) return null;
+        if (productionType == null
+            || goodsType == null
+            || ((unit == null || !contains(unit)) && isFull())) return null;
 
         final Specification spec = getSpecification();
         final Player owner = getOwner();
@@ -337,54 +324,60 @@ public abstract class WorkLocation extends UnitLocation
             delta -= getPotentialProduction(goodsType, unit.getType());
         }
         // Do we have a chance of satisfying the inputs?
-        for (AbstractGoods in : productionType.getInputs()) {
-            // TODO: should really consider in.getAmount
-            delta = Math.min(delta, colony.getNetProductionOf(in.getType()));
-        }
+        final ToIntFunction<AbstractGoods> prod = ag ->
+            getColony().getNetProductionOf(ag.getType());
+        delta = Math.min(delta, min(productionType.getInputs(), prod));
         if (delta <= 0) return null;
 
         // Is the production actually a good idea?  Not if we are independent
         // and have maximized liberty, or for immigration.
         if (owner.getPlayerType() == Player.PlayerType.INDEPENDENT
-            && ((goodsType.isLibertyType() && colony.getSoL() >= 100)
+            && ((goodsType.isLibertyType() && getColony().getSoL() >= 100)
                 || goodsType.isImmigrationType())) 
             return null;
         
-        // FIXME: OO
-        boolean ok = false;
-        if (this instanceof ColonyTile) {
-            // Assume the work is worth doing for owned or trivially
-            // claimable colony tiles.
-            Tile tile = ((ColonyTile)this).getWorkTile();
-            ok = owner.owns(tile) || owner.canClaimForSettlement(tile);
-        } else if (this instanceof Building) {
-            Building bu = (Building)this;
-            // Make sure the type can be added.
-            if (bu.canAddType(better)) {
-                Colony colony = getColony();
-                BuildableType bt;
-                // Assume work is worth doing if a unit is already
-                // there, or if the building has been upgraded, or if
-                // the goods are required for the current building job.
-                if (bu.getLevel() > 1 || unit != null) {
-                    ok = true;
-                } else if (colony.getTotalProductionOf(goodsType) == 0
-                    && (bt = colony.getCurrentlyBuilding()) != null
-                    && AbstractGoods.containsType(goodsType, bt.getRequiredGoods())) {
-                    ok = true;
-                }
-            }
-        }
+        final Boolean ok = goodSuggestionCheck(better,unit, goodsType);
         return (!ok) ? null
             : new Suggestion(this, (unit == null) ? null : unit.getType(),
                              better, goodsType, delta);
     }
 
+
+    /**
+     * Determines whether a {@code WorkLocation} is a good suggestion
+     *
+     * @param unitType The type of {@code Unit} to check.
+     * @param unit The Unit itself.
+     * @param goodsType The {@code GoodsType} to check.
+     *
+     * @return True or false, depending on whether the suggestion is valid.
+     */
+    protected boolean goodSuggestionCheck(UnitType unitType, Unit unit, GoodsType goodsType) {
+        return goodSuggestionCheck(unitType, unit, goodsType, this);
+    }
+
+    /**
+     * Determines whether a {@code WorkLocation} is a good suggestion
+     *
+     * @param unitType The type of {@code Unit} to check.
+     * @param unit The Unit itself.
+     * @param goodsType The {@code GoodsType} to check.
+     * @param workLocation The {@code WorkLocation} to check. This can
+     *     be overloaded with a default parameter of "this" to call
+     *     the WorkLocation instance currently in in reference.
+     * @return True or false, depending on whether the suggestion is valid.
+     */
+    protected boolean goodSuggestionCheck(UnitType unitType, Unit unit,
+                                          GoodsType goodsType,
+                                          WorkLocation workLocation) {
+        return false;
+    }
+
     /**
      * Get a map of suggestions for better or additional units.
      *
      * @return A mapping of either existing units or null (denoting
-     *     adding a unit) to a <code>Suggestion</code>.
+     *     adding a unit) to a {@code Suggestion}.
      */
     public java.util.Map<Unit, Suggestion> getSuggestions() {
         java.util.Map<Unit, Suggestion> result = new HashMap<>();
@@ -393,9 +386,8 @@ public abstract class WorkLocation extends UnitLocation
         Occupation occ = getOccupation(null);
         GoodsType work;
         Suggestion sug;
-        // Check if the existing units can be improved.
-        for (Unit u : getUnitList()) {
-            if (u.getTeacher() != null) continue; // Students assumed temporary
+        // Check if the existing non-student units can be improved.
+        for (Unit u : transform(getUnits(), isNull(Unit::getTeacher))) {
             if ((work = u.getWorkType()) == null) {
                 if (occ != null) work = occ.workType;
             }
@@ -404,8 +396,8 @@ public abstract class WorkLocation extends UnitLocation
             }
         }
         // Check for a suggestion for an extra worker if there is space.
-        if (!isFull() && occ != null
-            && (work = occ.workType) != null
+        if (occ != null && (work = occ.workType) != null
+            && !isFull()
             && (sug = getSuggestion(null, occ.productionType, work)) != null) {
             result.put(null, sug);
         }
@@ -413,34 +405,34 @@ public abstract class WorkLocation extends UnitLocation
     }
             
     /**
-     * Get the <code>AbstractGoods</code> consumed by this work location.
+     * Get the {@code AbstractGoods} consumed by this work location.
      *
-     * @return A list of <code>AbstractGoods</code> consumed.
+     * @return A stream of {@code AbstractGoods} consumed.
      */
-    public List<AbstractGoods> getInputs() {
-        return (productionType == null) ? EMPTY_LIST
+    public Stream<AbstractGoods> getInputs() {
+        return (productionType == null) ? Stream.<AbstractGoods>empty()
             : productionType.getInputs();
     }
 
     /**
-     * Get the <code>AbstractGoods</code> produced by this work location.
+     * Get the {@code AbstractGoods} produced by this work location.
      *
-     * @return A list of <code>AbstractGoods</code> produced.
+     * @return A stream of {@code AbstractGoods} produced.
      */
-    public List<AbstractGoods> getOutputs() {
-        return (productionType == null) ? EMPTY_LIST
+    public Stream<AbstractGoods> getOutputs() {
+        return (productionType == null) ? Stream.<AbstractGoods>empty()
             : productionType.getOutputs();
     }
 
     /**
      * Does this work location produce a given type of goods?
      *
-     * @param goodsType The <code>GoodsType</code> to check.
-     * @return True if this <code>WorkLocation</code> produces the
-     *     given <code>GoodsType</code>.
+     * @param goodsType The {@code GoodsType} to check.
+     * @return True if this {@code WorkLocation} produces the
+     *     given {@code GoodsType}.
      */
     public boolean produces(GoodsType goodsType) {
-        return AbstractGoods.containsType(goodsType, getOutputs());
+        return any(getOutputs(), AbstractGoods.matches(goodsType));
     }
 
     /**
@@ -449,8 +441,7 @@ public abstract class WorkLocation extends UnitLocation
      * @return True if there are any inputs.
      */
     public boolean hasInputs() {
-        return productionType != null
-            && !productionType.getInputs().isEmpty();
+        return any(getInputs());
     }
 
     /**
@@ -459,8 +450,8 @@ public abstract class WorkLocation extends UnitLocation
      * @return True if there are any outputs.
      */
     public boolean hasOutputs() {
-        return productionType != null
-            && !productionType.getOutputs().isEmpty();
+        return this.productionType != null
+            && any(this.productionType.getOutputs());
     }
 
     /**
@@ -486,7 +477,7 @@ public abstract class WorkLocation extends UnitLocation
      * Gets the ProductionInfo for this WorkLocation from the Colony's
      * cache.
      *
-     * @return The work location <code>ProductionInfo</code>.
+     * @return The work location {@code ProductionInfo}.
      */
     public ProductionInfo getProductionInfo() {
         return getColony().getProductionInfo(this);
@@ -506,12 +497,12 @@ public abstract class WorkLocation extends UnitLocation
      * Gets the total production of a specified goods type at this
      * work location.
      *
-     * @param goodsType The <code>GoodsType</code> to check.
+     * @param goodsType The {@code GoodsType} to check.
      * @return The amount of production.
      */
     public int getTotalProductionOf(GoodsType goodsType) {
         if (goodsType == null) {
-            throw new IllegalArgumentException("Null GoodsType.");
+            throw new RuntimeException("Null GoodsType: " + this);
         }
         return AbstractGoods.getCount(goodsType, getProduction());
     }
@@ -520,7 +511,7 @@ public abstract class WorkLocation extends UnitLocation
      * Gets the maximum production of this work location for a given
      * goods type, assuming the current workers and input goods.
      *
-     * @param goodsType The <code>GoodsType</code> to check.
+     * @param goodsType The {@code GoodsType} to check.
      * @return The maximum production of the goods at this work location.
      */
     public int getMaximumProductionOf(GoodsType goodsType) {
@@ -528,7 +519,7 @@ public abstract class WorkLocation extends UnitLocation
         if (info == null) return 0;
         List<AbstractGoods> production = info.getMaximumProduction();
         if (production != null) {
-            AbstractGoods ag = AbstractGoods.findByType(goodsType, production);
+            AbstractGoods ag = find(production, AbstractGoods.matches(goodsType));
             if (ag != null) return ag.getAmount();
         }
         return getTotalProductionOf(goodsType);
@@ -538,22 +529,22 @@ public abstract class WorkLocation extends UnitLocation
      * Gets the unit type that is the expert for this work location
      * using its first output for which an expert type can be found.
      *
-     * @return The expert <code>UnitType</code>.
+     * @return The expert {@code UnitType} or null if none found.
      */
     public UnitType getExpertUnitType() {
         final Specification spec = getSpecification();
         ProductionType pt = getBestProductionType(false, null);
         return (pt == null) ? null
             : find(map(pt.getOutputs(),
-                    ag -> spec.getExpertForProducing(ag.getType())),
-                ut -> ut != null, null);
+                       ag -> spec.getExpertForProducing(ag.getType())),
+                   isNotNull());
     }
 
     /**
      * Get the potential production of a given goods type using the
      * default unit.  This is useful for planning.
      *
-     * @param goodsType The <code>GoodsType</code> to check.
+     * @param goodsType The {@code GoodsType} to check.
      * @return The potential production.
      */
     public int getGenericPotential(GoodsType goodsType) {
@@ -568,22 +559,18 @@ public abstract class WorkLocation extends UnitLocation
      *
      * Used below, only public for the test suite.
      *
-     * @param unit The <code>Unit</code> to check.
+     * @param unit The {@code Unit} to check.
+     * @param goodsType The {@code GoodsType} to check the production of.
      * @return The maximum return from this unit.
      */
     public int getUnitProduction(Unit unit, GoodsType goodsType) {
         if (unit == null || unit.getWorkType() != goodsType) return 0;
         final UnitType unitType = unit.getType();
         final Turn turn = getGame().getTurn();
-        int bestAmount = 0;
-        for (AbstractGoods output : getOutputs()) {
-            if (output.getType() != goodsType) continue;
-            int amount = (int)applyModifiers(getBaseProduction(getProductionType(),
-                    goodsType, unitType),
-                turn, getProductionModifiers(goodsType, unitType));
-            if (bestAmount < amount) bestAmount = amount;
-        }
-        return bestAmount;
+        return Math.max(0,
+                        (int)applyModifiers(getBaseProduction(getProductionType(), goodsType, unitType),
+                                            turn,
+                                            getProductionModifiers(goodsType, unitType)));
     }
 
     /**
@@ -594,7 +581,7 @@ public abstract class WorkLocation extends UnitLocation
      * @return The production of the given type of goods.
      */
     public int getProductionOf(Unit unit, GoodsType goodsType) {
-        if (unit == null) throw new IllegalArgumentException("Null unit.");
+        if (unit == null) throw new RuntimeException("Null unit: " + this);
         return (!produces(goodsType)) ? 0
             : Math.max(0, getPotentialProduction(goodsType, unit.getType()));
     }
@@ -617,8 +604,8 @@ public abstract class WorkLocation extends UnitLocation
      * willing to change (a case could be made for including the
      * OCCUPIED_BY_ENEMY condition).
      *
-     * @param goodsType The <code>GoodsType</code> to produce.
-     * @param unitType The optional <code>UnitType</code> to do the work.
+     * @param goodsType The {@code GoodsType} to produce.
+     * @param unitType The optional {@code UnitType} to do the work.
      * @return The potential production with the given goods type and
      *     unit type.
      */
@@ -646,11 +633,24 @@ public abstract class WorkLocation extends UnitLocation
 
         int amount = getBaseProduction(null, goodsType, unitType);
         amount = (int)applyModifiers(amount, getGame().getTurn(),
-            getProductionModifiers(goodsType, unitType));
+                                     getProductionModifiers(goodsType, unitType));
         return (amount < 0) ? 0 : amount;
     }
 
-
+    /**
+     * Helper to get the production deficit for a type of goods.
+     *
+     * @param goodsType The {@code GoodsType} to check.
+     * @return The production deficit as {@code AbstractGoods}, or null
+     *     if not in deficit.
+     */
+    public AbstractGoods getProductionDeficit(GoodsType goodsType) {
+        ProductionInfo pi = getProductionInfo();
+        return (pi == null) ? null
+            : find(pi.getProductionDeficit(), AbstractGoods.matches(goodsType));
+    }
+ 
+   
     // Interface Location
     // Inherits:
     //   FreeColObject.getId
@@ -658,7 +658,6 @@ public abstract class WorkLocation extends UnitLocation
     //   UnitLocation.contains
     //   UnitLocation.canAdd
     //   UnitLocation.getUnitCount
-    //   final UnitLocation.getUnitIterator
     //   UnitLocation.getGoodsContainer
 
     /**
@@ -675,7 +674,7 @@ public abstract class WorkLocation extends UnitLocation
      */
     @Override
     public final Tile getTile() {
-        return colony.getTile();
+        return getSettlement().getTile();
     }
 
     /**
@@ -733,7 +732,15 @@ public abstract class WorkLocation extends UnitLocation
      */
     @Override
     public final Settlement getSettlement() {
-        return colony;
+        return this.colony;
+    }
+
+    /**
+     * {@inheritDoc}
+     */
+    @Override
+    public final Colony getColony() {
+        return this.colony;
     }
 
     /**
@@ -741,7 +748,7 @@ public abstract class WorkLocation extends UnitLocation
      */
     @Override
     public final int getRank() {
-        return Location.getRank(getTile());
+        return Location.rankOf(getTile());
     }
 
     
@@ -771,7 +778,7 @@ public abstract class WorkLocation extends UnitLocation
      * Get a description of the work location, with any expected extra
      * detail.
      *
-     * @return A label <code>StringTemplate</code> for this work location.
+     * @return A label {@code StringTemplate} for this work location.
      */
     public abstract StringTemplate getLabel();
 
@@ -797,6 +804,22 @@ public abstract class WorkLocation extends UnitLocation
      */
     public abstract NoAddReason getNoWorkReason();
 
+    /**
+     * Get an associated tile, if any.
+     *
+     * @return The underlying {@code Tile} which is worked, if any.
+     */
+    public abstract Tile getWorkTile();
+    
+    /**
+     * Get the "level" of this work location.
+     *
+     * Note: only meaningful for buildings at present.
+     *
+     * @return The work location level.
+     */
+    public abstract int getLevel();
+        
     /**
      * Can this work location can produce goods without workers?
      *
@@ -809,8 +832,8 @@ public abstract class WorkLocation extends UnitLocation
      * Can this work location produce a given goods type with
      * an optional unit.
      *
-     * @param goodsType The <code>GoodsType</code> to produce.
-     * @param unitType An optional <code>UnitType</code>, if null the
+     * @param goodsType The {@code GoodsType} to produce.
+     * @param unitType An optional {@code UnitType}, if null the
      *     unattended production is considered.
      * @return True if this location can produce the goods.
      */
@@ -823,10 +846,10 @@ public abstract class WorkLocation extends UnitLocation
      * production exclusive of any modifiers.  If no unit type is
      * specified, the unattended production is calculated.
      *
-     * @param productionType An optional <code>ProductionType</code> to use,
+     * @param productionType An optional {@code ProductionType} to use,
      *     if null the best available one is used.
-     * @param goodsType The <code>GoodsType</code> to produce.
-     * @param unitType An optional <code>UnitType</code> to produce
+     * @param goodsType The {@code GoodsType} to produce.
+     * @param unitType An optional {@code UnitType} to produce
      *     the goods.
      * @return The amount of goods potentially produced.
      */
@@ -839,31 +862,52 @@ public abstract class WorkLocation extends UnitLocation
      * unit type.  If no unit is specified the unattended production
      * is calculated.
      *
-     * @param goodsType The <code>GoodsType</code> to produce.
-     * @param unitType The optional <code>UnitType</code> to produce them.
-     * @return A list of the applicable modifiers.
+     * @param goodsType The {@code GoodsType} to produce.
+     * @param unitType The optional {@code UnitType} to produce them.
+     * @return A stream of the applicable modifiers.
      */
-    public abstract List<Modifier> getProductionModifiers(GoodsType goodsType,
-                                                          UnitType unitType);
+    public abstract Stream<Modifier> getProductionModifiers(GoodsType goodsType,
+                                                            UnitType unitType);
 
     /**
      * Get the production types available for this work location.
      *
      * @param unattended If true, get unattended production types.
-     * @return A list of suitable <code>ProductionType</code>s.
+     * @return A list of suitable {@code ProductionType}s.
      */
     public abstract List<ProductionType> getAvailableProductionTypes(boolean unattended);
 
+    /**
+     * Get a work location specific factor to multiply any goods-specific
+     * unit production bonuses by.
+     *
+     * @return The competence factor.
+     */
+    public abstract float getCompetenceFactor();
+
+    /**
+     * Get a work location specific factor to multiply the colony
+     * rebel bonus by.
+     *
+     * @return The rebel factor.
+     */
+    public abstract float getRebelFactor();
+      
     /**
      * Evaluate this work location for a given player.
      * To be overridden by subclasses.
      *
-     * @param player The <code>Player</code> to evaluate for.
+     * @param player The {@code Player} to evaluate for.
      * @return A value for the player.
      */
     public int evaluateFor(Player player) {
-        return getUnitList().stream()
-            .mapToInt(u -> u.evaluateFor(player)).sum();
+        int result = 0;
+        for (Unit u : getUnitList()) {
+            int v = u.evaluateFor(player);
+            if (v == Integer.MIN_VALUE) return Integer.MIN_VALUE;
+            result += v;
+        }
+        return result;
     }
 
     /**
@@ -882,22 +926,23 @@ public abstract class WorkLocation extends UnitLocation
     // Interface Ownable
 
     /**
-     * Gets the owner of this <code>Ownable</code>.
+     * Gets the owner of this {@code Ownable}.
      *
-     * @return The <code>Player</code> controlling this
+     * @return The {@code Player} controlling this
      *         {@link Ownable}.
      */
     @Override
     public Player getOwner() {
-        return colony.getOwner();
+        final Colony colony = getColony(); // Defend against init-time NPE
+        return (colony == null) ? null : colony.getOwner();
     }
 
     /**
-     * Sets the owner of this <code>Ownable</code>.  Do not call this
+     * Sets the owner of this {@code Ownable}.  Do not call this
      * method, ever.  The owner of this WorkLocation is the owner
      * of the Colony, you must set the owner of the Colony instead.
      *
-     * @param p The <code>Player</code> that should take ownership
+     * @param p The {@code Player} that should take ownership
      *     of this {@link Ownable}.
      * @exception UnsupportedOperationException is always thrown by
      *     this method.
@@ -908,6 +953,21 @@ public abstract class WorkLocation extends UnitLocation
     }
 
 
+    // Override FreeColObject
+
+    /**
+     * {@inheritDoc}
+     */
+    @Override
+    public <T extends FreeColObject> boolean copyIn(T other) {
+        WorkLocation o = copyInCast(other, WorkLocation.class);
+        if (o == null || !super.copyIn(o)) return false;
+        this.colony = getGame().updateRef(o.getColony());
+        this.productionType = o.getProductionType();
+        return true;
+    }
+
+
     // Serialization
 
     private static final String COLONY_TAG = "colony";
@@ -920,7 +980,7 @@ public abstract class WorkLocation extends UnitLocation
     protected void writeAttributes(FreeColXMLWriter xw) throws XMLStreamException {
         super.writeAttributes(xw);
 
-        xw.writeAttribute(COLONY_TAG, colony);
+        xw.writeAttribute(COLONY_TAG, this.colony);
     }
 
     /**
@@ -930,7 +990,7 @@ public abstract class WorkLocation extends UnitLocation
     protected void writeChildren(FreeColXMLWriter xw) throws XMLStreamException {
         super.writeChildren(xw);
 
-        if (productionType != null) productionType.toXML(xw);
+        if (this.productionType != null) this.productionType.toXML(xw);
     }
 
     /**
@@ -940,19 +1000,8 @@ public abstract class WorkLocation extends UnitLocation
     protected void readAttributes(FreeColXMLReader xr) throws XMLStreamException {
         super.readAttributes(xr);
 
-        colony = xr.findFreeColGameObject(getGame(), COLONY_TAG,
-                                          Colony.class, (Colony)null, true);
-    }
-
-    /**
-     * {@inheritDoc}
-     */
-    @Override
-    public void readChildren(FreeColXMLReader xr) throws XMLStreamException {
-
-        super.readChildren(xr);
-
-        updateProductionType();
+        this.colony = xr.findFreeColGameObject(getGame(), COLONY_TAG,
+            Colony.class, (Colony)null, true);
     }
 
     /**
@@ -963,11 +1012,13 @@ public abstract class WorkLocation extends UnitLocation
         final Specification spec = getSpecification();
         final String tag = xr.getLocalName();
 
-        if (ProductionType.getXMLElementTagName().equals(tag)) {
-            productionType = new ProductionType(xr, spec);
+        if (ProductionType.TAG.equals(tag)) {
+            this.productionType = new ProductionType(xr, spec);
 
         } else {
             super.readChild(xr);
         }
     }
+
+    // getXMLTagName left to subclasses
 }
