@@ -1,5 +1,5 @@
 /**
- *  Copyright (C) 2002-2015   The FreeCol Team
+ *  Copyright (C) 2002-2022   The FreeCol Team
  *
  *  This file is part of FreeCol.
  *
@@ -19,29 +19,36 @@
 
 package net.sf.freecol.common.option;
 
+import java.io.File;
+import java.io.IOException;
 import java.util.ArrayList;
 import java.util.HashMap;
-import java.util.Iterator;
 import java.util.List;
 import java.util.Map;
+import java.util.logging.Level;
 import java.util.logging.Logger;
 
 import javax.xml.stream.XMLStreamException;
 
+import net.sf.freecol.client.ClientOptions;
 import net.sf.freecol.common.i18n.Messages;
 import net.sf.freecol.common.io.FreeColXMLReader;
 import net.sf.freecol.common.io.FreeColXMLWriter;
 import net.sf.freecol.common.model.Specification;
 import static net.sf.freecol.common.util.CollectionUtils.*;
+import net.sf.freecol.common.util.LogBuilder;
 
 
 /**
  * Used for grouping {@link Option}s.
  */
-public class OptionGroup extends AbstractOption<OptionGroup> {
+public class OptionGroup extends AbstractOption<OptionGroup>
+    implements OptionContainer {
 
     private static final Logger logger = Logger.getLogger(OptionGroup.class.getName());
 
+    public static final String TAG = "optionGroup";
+
     /** The options in this group. */
     private final List<Option> options = new ArrayList<>();
 
@@ -55,9 +62,12 @@ public class OptionGroup extends AbstractOption<OptionGroup> {
     /** Is this option group user editable? */
     private boolean editable = true;
 
+    /** Should this option group be visible? */
+    private boolean visible = true;
+    
 
     /**
-     * Creates a new <code>OptionGroup</code>.
+     * Creates a new {@code OptionGroup}.
      *
      * @param id The object identifier.
      */
@@ -66,33 +76,33 @@ public class OptionGroup extends AbstractOption<OptionGroup> {
     }
 
     /**
-     * Creates a new <code>OptionGroup</code>.
+     * Creates a new {@code OptionGroup}.
      *
-     * @param specification The <code>Specification</code> to refer to.
+     * @param specification The {@code Specification} to refer to.
      */
     public OptionGroup(Specification specification) {
         super(specification);
     }
 
     /**
-     * Creates a new <code>OptionGroup</code>.
+     * Creates a new {@code OptionGroup}.
      *
      * @param id The object identifier.
-     * @param specification The <code>Specification</code> to refer to.
+     * @param specification The {@code Specification} to refer to.
      */
     public OptionGroup(String id, Specification specification) {
         super(id, specification);
     }
 
     /**
-     * Creates a new <code>OptionGroup</code>.
+     * Creates a new {@code OptionGroup}.
      *
-     * @param xr The <code>FreeColXMLReader</code> to read from.
-     * @param specification The <code>Specification</code> to refer to.
+     * @param xr The {@code FreeColXMLReader} to read from.
+     * @param specification The {@code Specification} to refer to.
      * @exception XMLStreamException if there is a problem reading the stream.
      */
     public OptionGroup(FreeColXMLReader xr,
-                       Specification specification) throws XMLStreamException {
+        Specification specification) throws XMLStreamException {
         super(specification);
 
         readFromXML(xr);
@@ -118,7 +128,16 @@ public class OptionGroup extends AbstractOption<OptionGroup> {
     }
 
     /**
-     * Gets the i18n-name of this <code>Option</code>.
+     * Should this option group be visible?
+     *
+     * @return True if the option group should be visible.
+     */
+    public boolean isVisible() {
+        return visible;
+    }
+
+    /**
+     * Gets the i18n-name of this {@code Option}.
      *
      * @return The name as provided in the constructor.
      */
@@ -127,10 +146,10 @@ public class OptionGroup extends AbstractOption<OptionGroup> {
     }
 
     /**
-     * Gets the i18n short description of this <code>Option</code>.
+     * Gets the i18n short description of this {@code Option}.
      * Should be suitable for use as a tooltip text.
      *
-     * @return A short description of this <code>Option</code>.
+     * @return A short description of this {@code Option}.
      */
     public String getShortDescription() {
         return Messages.getShortDescription(getId());
@@ -139,7 +158,7 @@ public class OptionGroup extends AbstractOption<OptionGroup> {
     /**
      * Get the options in this group.
      *
-     * @return The list of <code>Option</code>s.
+     * @return The list of {@code Option}s.
      */
     public List<Option> getOptions() {
         return options;
@@ -158,7 +177,7 @@ public class OptionGroup extends AbstractOption<OptionGroup> {
     /**
      * Does this option group contain any subgroups?
      *
-     * @return True if there are any child <code>OptionGroup</code>s present.
+     * @return True if there are any child {@code OptionGroup}s present.
      */
     public boolean hasOptionGroup() {
         return any(options, o -> o instanceof OptionGroup);
@@ -168,7 +187,7 @@ public class OptionGroup extends AbstractOption<OptionGroup> {
      * Adds the given option to this group.  The option is assumed to
      * be correct.
      *
-     * @param option The <code>Option</code> to add.
+     * @param option The {@code Option} to add.
      */
     public void add(Option option) {
         String id = option.getId();
@@ -192,7 +211,7 @@ public class OptionGroup extends AbstractOption<OptionGroup> {
     }
 
     /**
-     * Merges the given <code>Option</code> into this group.
+     * Merges the given {@code Option} into this group.
      *
      * - Option groups are *not* merged, but their leaf options are.
      * - If an option is not already present it is ignored.
@@ -202,11 +221,13 @@ public class OptionGroup extends AbstractOption<OptionGroup> {
      * The intent is that the option group structure is never subject to
      * merging.
      *
-     * @param option The <code>Option</code> to merge.
+     * @param option The {@code Option} to merge.
+     * @param lb A {@code LogBuilder} to log to.
      * @return True if the merge was accepted.
      */
-    public boolean merge(Option option) {
+    public boolean merge(Option option, LogBuilder lb) {
         final String id = option.getId();
+        OptionGroup etc = null;
 
         // Check first, it is valid to merge an option group onto
         // one at the same level, for which it will not contain a key.
@@ -214,21 +235,30 @@ public class OptionGroup extends AbstractOption<OptionGroup> {
             OptionGroup optionGroup = (OptionGroup)option;
             boolean result = true;
             for (Option o : optionGroup.getOptions()) {
+                // @compat 0.11.6
+                // Placement options move to their own group
+                if (o.getId().startsWith("net.sf.freecol.client.gui.panel.")) {
+                    if (etc == null) etc = getOptionGroup(ClientOptions.ETC);
+                    etc.add(o);
+                    lb.add(o.getId(), " moved to ", ClientOptions.ETC, ".\n");
+                    continue;
+                }
+                // end @compat 0.11.6
+                
                 // Merge from the top level, so that the new
                 // option will end up in the group inherited
                 // from the standard client-options.xml.
-                result = result && this.merge(o);
+                result = result && this.merge(o, lb);
             }
             if (result) {
                 optionGroup.setEditable(editable && optionGroup.isEditable());
             }
-            logger.finest("Merged option group " + id
-                + " contents into " + this.getId());
+            lb.add(id, " group merged contents into ", this.getId(), ".\n");
             return result;
         }
                 
         if (!optionMap.containsKey(id)) {
-            logger.warning("Ignoring unknown option " + id);
+            lb.add(id, " unknown/ignored.\n");
             return false;
         }
 
@@ -238,19 +268,19 @@ public class OptionGroup extends AbstractOption<OptionGroup> {
                 options.remove(index);
                 options.add(index, option);
                 optionMap.put(id, option);
-                logger.finest("Merged option " + id + " into " + this.getId()
-                    + ": " + option.toString() + "/");
+                lb.add(id, " merged into ", this.getId(),
+                    ": ", option.toString() + ".\n");
                 return true;
             }
             if (o instanceof OptionGroup) {
                 OptionGroup og = (OptionGroup)o;
-                if (og.optionMap.containsKey(id) && og.merge(option)) {
+                if (og.optionMap.containsKey(id) && og.merge(option, lb)) {
                     optionMap.put(id, option);
                     return true;
                 }
             }
         }
-        logger.warning("Option " + id + " registered but not found!");
+        lb.add(id, " registered but not found!\n");
         return false;
     }
         
@@ -258,7 +288,7 @@ public class OptionGroup extends AbstractOption<OptionGroup> {
      * Helper function to recursively add option group members to the
      * optionMap.
      *
-     * @param group The initial <code>OptionGroup</code> to add.
+     * @param group The initial {@code OptionGroup} to add.
      */
     private void addOptionGroup(OptionGroup group) {
         for (Option option : group.getOptions()) {
@@ -273,7 +303,7 @@ public class OptionGroup extends AbstractOption<OptionGroup> {
      * Remove an option.
      *
      * @param id The identifier of the option to remove.
-     * @return The <code>Option</code> removed if any.
+     * @return The {@code Option} removed if any.
      */
     public Option remove(String id) {
         Option op = optionMap.remove(id);
@@ -282,8 +312,8 @@ public class OptionGroup extends AbstractOption<OptionGroup> {
     }
 
     /**
-     * Removes all of the <code>Option</code>s from this
-     * <code>OptionGroup</code>.
+     * Removes all of the {@code Option}s from this
+     * {@code OptionGroup}.
      */
     public void removeAll() {
         options.clear();
@@ -291,196 +321,106 @@ public class OptionGroup extends AbstractOption<OptionGroup> {
     }
 
     /**
-     * Gets an <code>Iterator</code> for the <code>Option</code>s.
+     * Loads the options from the given reader.
      *
-     * @return The <code>Iterator</code>.
+     * @param xr The {@code FreeColXMLReader} to read from.
+     * @return True if the options were loaded without error.
+     * @exception XMLStreamException if there is an error reading the stream.
      */
-    public Iterator<Option> iterator() {
-        return options.iterator();
+    protected boolean load(FreeColXMLReader xr) throws XMLStreamException {
+        if (xr == null) return false;
+        xr.nextTag();
+        readFromXML(xr);
+        return true;
     }
-
-
-    // Convenience accessors.
-
+    
     /**
-     * Gets the value of an option as an option group.
+     * Loads the options from the given file.
      *
-     * @param id The object identifier.
-     * @return The <code>OptionGroup</code> value.
-     * @exception IllegalArgumentException If there is no option group
-     *     value associated with the specified option.
-     * @exception NullPointerException if the given
-     *     <code>Option</code> does not exist.
+     * @param file The {@code File} to read the options from.
+     * @return True if the options were loaded without error.
      */
-    public OptionGroup getOptionGroup(String id) {
-        try {
-            return ((OptionGroup)getOption(id)).getValue();
-        } catch (ClassCastException e) {
-            throw new IllegalArgumentException("No option group"
-                + " value associated with the specified option: " + id, e);
+    public boolean load(File file) {
+        if (file == null) return false;
+        boolean ret = false;
+        try (
+             FreeColXMLReader xr = new FreeColXMLReader(file);
+             ) {
+            ret = load(xr);
+        } catch (IOException|XMLStreamException xse) {
+            logger.log(Level.WARNING, "Load OptionGroup(" + getId()
+                + ") from file " + file.getPath() + " crashed", xse);
+            return false;
         }
+        logger.info("Load OptionGroup(" + getId() + ") from " + file.getPath()
+            + ((ret) ? " succeeded" : " failed"));
+        return ret;
     }
 
     /**
-     * Gets the integer value of an option.
+     * Load a specific option group from a file by identifier.
      *
-     * @param id The object identifier.
-     * @return The integer value.
-     * @exception IllegalArgumentException If there is no integer
-     *     value associated with the specified option.
-     * @exception NullPointerException if the given
-     *     <code>Option</code> does not exist.
+     * @param file The {@code File} to load from.
+     * @param optionId The option identifier to look for.
+     * @param spec A base {@code Specification}.
+     * @return The {@code OptionGroup} found, or null on error or
+     *     not found.
      */
-    public int getInteger(String id) {
-        try {
-            return ((IntegerOption)getOption(id)).getValue();
-        } catch (ClassCastException e) {
-            throw new IllegalArgumentException("No integer"
-                + " value associated with the specified option: " + id, e);
+    public static OptionGroup loadOptionGroup(File file, String optionId,
+                                              Specification spec) {
+        OptionGroup ret = new OptionGroup(spec);
+        if (ret.load(file)) {
+            if (!optionId.equals(ret.getId())) {
+                Option op = ret.getOption(optionId);
+                ret = (op instanceof OptionGroup) ? (OptionGroup)op : null;
+            }
+        } else {
+            ret = null;
         }
+        return ret;
+    }
+
+
+    // Interface OptionContainer
+
+    /**
+     * {@inheritDoc}
+     */    
+    public <T extends Option> boolean hasOption(String id,
+                                                Class<T> returnClass) {
+        if (id == null) return false;
+        Option val = this.optionMap.get(id);
+        return (val == null) ? false
+            : returnClass.isAssignableFrom(val.getClass());
     }
 
     /**
-     * Sets the integer value of an option.
-     *
-     * @param id The object identifier.
-     * @param value The new integer value of the option.
-     * @exception IllegalArgumentException If there is no integer
-     *     value associated with the specified option.
-     * @exception NullPointerException if the given
-     *     <code>Option</code> does not exist.
+     * {@inheritDoc}
      */
-    public void setInteger(String id, int value) {
-        try {
-            ((IntegerOption)getOption(id)).setValue(value);
-        } catch (ClassCastException e) {
-            throw new IllegalArgumentException("No integer"
-                + " value associated with the specified option: " + id, e);
+    public <T extends Option> T getOption(String id,
+                                          Class<T> returnClass) {
+        if (id == null) {
+            throw new RuntimeException("Null id: " + this);
+        } else if (!this.optionMap.containsKey(id)) {
+            throw new RuntimeException("Missing option: " + id);
+        } else {
+            try {
+                return returnClass.cast(optionMap.get(id));
+            } catch (ClassCastException cce) {
+                throw new RuntimeException("Not a " + returnClass.getName()
+                    + ": " + id, cce);
+            }
         }
     }
 
-    /**
-     * Gets the boolean value of an option.
-     *
-     * @param id The object identifier.
-     * @return The boolean value.
-     * @exception IllegalArgumentException If there is no boolean
-     *     value associated with the specified option.
-     * @exception NullPointerException if the given
-     *     <code>Option</code> does not exist.
-     */
-    public boolean getBoolean(String id) {
-        try {
-            return ((BooleanOption)getOption(id)).getValue();
-        } catch (ClassCastException e) {
-            throw new IllegalArgumentException("No boolean"
-                + " value associated with the specified option: " + id, e);
-        }
-    }
-
-    /**
-     * Sets the boolean value of an option.
-     *
-     * @param id The object identifier.
-     * @param value The new boolean value of the option.
-     * @exception IllegalArgumentException If there is no boolean
-     *     value associated with the specified option.
-     * @exception NullPointerException if the given
-     *     <code>Option</code> does not exist.
-     */
-    public void setBoolean(String id, boolean value) {
-        try {
-            ((BooleanOption)getOption(id)).setValue(value);
-        } catch (ClassCastException e) {
-            throw new IllegalArgumentException("No boolean"
-                + " value associated with the specified option: " + id, e);
-        }
-    }
-
-    /**
-     * Gets the string value of an option.
-     *
-     * @param id The object identifier.
-     * @return The string value.
-     * @exception IllegalArgumentException If there is no string value
-     *     associated with the specified option.
-     * @exception NullPointerException if the given
-     *     <code>Option</code> does not exist.
-     */
-    public String getString(String id) {
-        try {
-            return ((StringOption)getOption(id)).getValue();
-        } catch (ClassCastException e) {
-            throw new IllegalArgumentException("No String"
-                + " value associated with the specified option: " + id, e);
-        }
-    }
-
-    /**
-     * Sets the string value of an option.
-     *
-     * @param id The object identifier.
-     * @param value The new string value.
-     * @exception IllegalArgumentException If there is no string value
-     *     associated with the specified option.
-     * @exception NullPointerException if the given
-     *     <code>Option</code> does not exist.
-     */
-    public void setString(String id, String value) {
-        try {
-            ((StringOption)getOption(id)).setValue(value);
-        } catch (ClassCastException e) {
-            throw new IllegalArgumentException("No String"
-                + " value associated with the specified option: " + id, e);
-        }
-    }
-
-    /**
-     * Gets the string value of an option.
-     *
-     * @param id The object identifier.
-     * @return The string value.
-     * @exception IllegalArgumentException If there is no string value
-     *     associated with the specified option.
-     * @exception NullPointerException if the given
-     *     <code>Option</code> does not exist.
-     */
-    public String getText(String id) {
-        try {
-            return ((TextOption)getOption(id)).getValue();
-        } catch (ClassCastException e) {
-            throw new IllegalArgumentException("No String"
-                + " value associated with the specified option: " + id, e);
-        }
-    }
-
-    /**
-     * Sets the string value of an option.
-     *
-     * @param id The object identifier.
-     * @param value The new string value.
-     * @exception IllegalArgumentException If there is no string value
-     *     associated with the specified option.
-     * @exception NullPointerException if the given
-     *     <code>Option</code> does not exist.
-     */
-    public void setText(String id, String value) {
-        try {
-            ((TextOption)getOption(id)).setValue(value);
-        } catch (ClassCastException e) {
-            throw new IllegalArgumentException("No String"
-                + " value associated with the specified option: " + id, e);
-        }
-    }
-
-
+    
     // Interface Option
 
     /**
      * {@inheritDoc}
      */
     @Override
-    public OptionGroup clone() {
+    public OptionGroup cloneOption() {
         OptionGroup result = new OptionGroup(this.getId(), getSpecification());
         result.editable = this.editable;
         result.setValues(this);
@@ -526,6 +466,20 @@ public class OptionGroup extends AbstractOption<OptionGroup> {
 
     // Override AbstractOption
 
+    /**
+     * {@inheritDoc}
+     */
+    @Override
+    public void generateChoices() {
+        for (Option o : options) {
+            if (o instanceof OptionGroup) {
+                ((OptionGroup)o).generateChoices();
+            } else if (o instanceof AbstractOption) {
+                ((AbstractOption)o).generateChoices();
+            }
+        }
+    }
+
     /**
      * {@inheritDoc}
      */
@@ -535,9 +489,31 @@ public class OptionGroup extends AbstractOption<OptionGroup> {
     }
 
 
+    // Override FreeColObject
+
+    /**
+     * {@inheritDoc}
+     */
+    @Override
+    public boolean save(File file) {
+        boolean ret = false;
+        try {
+            ret = this.save(file, null, true);
+        } catch (Exception e) {
+            logger.log(Level.WARNING, "Save OptionGroup(" + getId()
+                + ") to " + file.getPath() + " crashed", e);
+            return false;
+        }
+        logger.info("Save OptionGroup(" + getId() + ") to " + file.getPath()
+            + ((ret) ? " succeeded" : " failed"));
+        return ret;
+    }
+    
+
     // Serialization
 
     private static final String EDITABLE_TAG = "editable";
+    private static final String VISIBLE_TAG = "visible";
 
 
     /**
@@ -548,6 +524,8 @@ public class OptionGroup extends AbstractOption<OptionGroup> {
         super.writeAttributes(xw);
 
         xw.writeAttribute(EDITABLE_TAG, editable);
+
+        xw.writeAttribute(VISIBLE_TAG, visible);
     }
 
     /**
@@ -568,6 +546,8 @@ public class OptionGroup extends AbstractOption<OptionGroup> {
         super.readAttributes(xr);
 
         editable = xr.getAttribute(EDITABLE_TAG, true);
+
+        visible = xr.getAttribute(VISIBLE_TAG, true);
     }
 
     /**
@@ -603,29 +583,22 @@ public class OptionGroup extends AbstractOption<OptionGroup> {
     /**
      * {@inheritDoc}
      */
-    @Override
-    public String toString() {
-        StringBuilder sb = new StringBuilder();
-        sb.append("[").append(getId()).append(" <");
-        for (Option o : getOptions()) {
-            sb.append(" ").append(o.toString());
-        }
-        sb.append(" >]");
-        return sb.toString();
-    }
+    public String getXMLTagName() { return TAG; }
 
+
+    // Override Object
+    
     /**
      * {@inheritDoc}
      */
     @Override
-    public String getXMLTagName() { return getXMLElementTagName(); }
-
-    /**
-     * Gets the tag name of the root element representing this object.
-     *
-     * @return "optionGroup".
-     */
-    public static String getXMLElementTagName() {
-        return "optionGroup";
+    public String toString() {
+        StringBuilder sb = new StringBuilder();
+        sb.append('[').append(getId()).append(" <");
+        for (Option o : getOptions()) {
+            sb.append(' ').append(o.toString());
+        }
+        sb.append(" >]");
+        return sb.toString();
     }
 }
