@@ -1,5 +1,5 @@
 /**
- *  Copyright (C) 2002-2015   The FreeCol Team
+ *  Copyright (C) 2002-2022   The FreeCol Team
  *
  *  This file is part of FreeCol.
  *
@@ -27,24 +27,26 @@ import java.util.logging.Logger;
 
 import javax.xml.stream.XMLStreamException;
 
+import net.sf.freecol.common.FreeColException;
 import net.sf.freecol.common.io.FreeColXMLReader;
 import net.sf.freecol.common.io.FreeColXMLWriter;
+import net.sf.freecol.common.model.Ability;
 import net.sf.freecol.common.model.AbstractGoods;
 import net.sf.freecol.common.model.Colony;
+import static net.sf.freecol.common.model.Constants.*;
+import net.sf.freecol.common.model.Direction;
 import net.sf.freecol.common.model.Locatable;
 import net.sf.freecol.common.model.Location;
 import net.sf.freecol.common.model.Map;
-import net.sf.freecol.common.model.Direction;
 import net.sf.freecol.common.model.PathNode;
 import net.sf.freecol.common.model.Player;
+import net.sf.freecol.common.model.Role;
 import net.sf.freecol.common.model.Settlement;
-import net.sf.freecol.common.model.Specification;
 import net.sf.freecol.common.model.Tile;
 import net.sf.freecol.common.model.Unit;
 import net.sf.freecol.common.model.UnitLocation;
-import net.sf.freecol.common.model.Role;
 import net.sf.freecol.common.util.LogBuilder;
-import net.sf.freecol.server.ai.goal.Goal;
+import net.sf.freecol.common.util.Utils;
 import net.sf.freecol.server.ai.mission.BuildColonyMission;
 import net.sf.freecol.server.ai.mission.CashInTreasureTrainMission;
 import net.sf.freecol.server.ai.mission.DefendSettlementMission;
@@ -63,35 +65,35 @@ import net.sf.freecol.server.ai.mission.UnitWanderMission;
 import net.sf.freecol.server.ai.mission.WishRealizationMission;
 import net.sf.freecol.server.ai.mission.WorkInsideColonyMission;
 
-import org.w3c.dom.Element;
-
 
 /**
  * Objects of this class contains AI-information for a single {@link Unit}.
  *
+ * AIUnits are successfully initialized by a call to
+ * {@link #setUnit(Unit)} with a non-null actual {@link Unit}.
+
  * The method {@link #doMission(LogBuilder)} is called once each turn,
  * by {@link AIPlayer#startWorking()}, to perform the assigned
- * <code>Mission</code>.  Most of the methods in this class just
+ * {@code Mission}.  Most of the methods in this class just
  * delegates the call to that mission.
  *
  * @see Mission
  */
-public class AIUnit extends TransportableAIObject {
+public final class AIUnit extends TransportableAIObject {
 
     private static final Logger logger = Logger.getLogger(AIUnit.class.getName());
 
+    public static final String TAG = "aiUnit";
+
     /** The Unit this AIObject contains AI-information for. */
     private Unit unit;
 
     /** The mission to which this AI unit has been assigned. */
     private Mission mission;
 
-    /** The goal.  Currently unused. */
-    private Goal goal = null;
-
 
     /**
-     * Creates a new uninitialized <code>AIUnit</code>.
+     * Creates a new uninitialized {@code AIUnit}.
      *
      * @param aiMain The main AI-object.
      * @param id The object identifier.
@@ -101,11 +103,11 @@ public class AIUnit extends TransportableAIObject {
 
         this.unit = null;
         this.mission = null;
-        this.goal = null;
+        this.initialized = false;
     }
 
     /**
-     * Creates a new <code>AIUnit</code>.
+     * Creates a new {@code AIUnit}.
      *
      * @param aiMain The main AI-object.
      * @param unit The unit to make an {@link AIObject} for.
@@ -115,27 +117,11 @@ public class AIUnit extends TransportableAIObject {
 
         this.unit = unit;
         this.mission = null;
-        this.goal = null;
-
-        uninitialized = unit == null;
+        setInitialized();
     }
 
     /**
-     * Creates a new <code>AIUnit</code> from the given
-     * XML-representation.
-     *
-     * @param aiMain The main AI-object.
-     * @param element The root element for the XML-representation
-     *       of a <code>Wish</code>.
-     */
-    public AIUnit(AIMain aiMain, Element element) {
-        super(aiMain, element);
-
-        uninitialized = getUnit() == null;
-    }
-
-    /**
-     * Creates a new <code>AIUnit</code> from the given
+     * Creates a new {@code AIUnit} from the given
      * XML-representation.
      *
      * @param aiMain The main AI-object.
@@ -147,19 +133,42 @@ public class AIUnit extends TransportableAIObject {
                   FreeColXMLReader xr) throws XMLStreamException {
         super(aiMain, xr);
 
-        uninitialized = getUnit() == null;
+        setInitialized();
     }
 
 
     /**
-     * Gets the <code>Unit</code> this <code>AIUnit</code> controls.
+     * {@inheritDoc}
+     */
+    public void setInitialized() {
+        this.initialized = getUnit() != null;
+    }
+
+    /**
+     * Gets the {@code Unit} this {@code AIUnit} controls.
      *
-     * @return The <code>Unit</code>.
+     * @return The {@code Unit}.
      */
     public final Unit getUnit() {
         return this.unit;
     }
 
+    /**
+     * Set the {@code Unit} this {@code AIUnit} controls.
+     *
+     * Note: it is an error to change the unit once it is set.
+     *
+     * @param unit The new {@code Unit}.
+     */
+    private final void setUnit(Unit unit) {
+        if (this.unit != null && unit != this.unit) {
+            throw new RuntimeException("Attempt to change AI Unit " + getId()
+                + " to " + unit
+                + "\n" + net.sf.freecol.common.debug.FreeColDebugger.stackTraceToString());
+        }
+        this.unit = unit;
+    }
+
     /**
      * Checks if this unit has been assigned a mission.
      *
@@ -172,7 +181,7 @@ public class AIUnit extends TransportableAIObject {
     /**
      * Gets the mission this unit has been assigned.
      *
-     * @return The <code>Mission</code>.
+     * @return The {@code Mission}.
      */
     public final Mission getMission() {
         return this.mission;
@@ -181,30 +190,12 @@ public class AIUnit extends TransportableAIObject {
     /**
      * Assigns a mission to unit. 
      *
-     * @param mission The new <code>Mission</code>.
+     * @param mission The new {@code Mission}.
      */
     public final void setMission(Mission mission) {
         this.mission = mission;
     }
 
-    /**
-     * Gets the goal of this AI unit.
-     *
-     * @return The goal of this AI unit.
-     */
-    public final Goal getGoal() {
-        return this.goal;
-    }
-
-    /**
-     * Sets the goal of this AI unit.
-     *
-     * @param goal The new <code>Goal</code>.
-     */
-    public final void setGoal(Goal goal) {
-        this.goal = goal;
-    }
-
 
     // Internal
 
@@ -215,6 +206,7 @@ public class AIUnit extends TransportableAIObject {
         Location loc;
         Colony colony;
         AIColony aiColony;
+        final Unit unit = getUnit();
         if (unit != null
             && (loc = unit.getLocation()) != null
             && (colony = loc.getColony()) != null
@@ -244,24 +236,34 @@ public class AIUnit extends TransportableAIObject {
 
     // Public routines
 
+    /**
+     * Get the {@code Player} that owns this AIUnit.
+     *
+     * @return The owning {@code Player}.
+     */
+    public Player getOwner() {
+        final Unit unit = getUnit();
+        return (unit == null) ? null : unit.getOwner();
+    }
+
     /**
      * Gets the AIPlayer that owns this AIUnit.
      *
      * @return The owning AIPlayer.
      */
     public AIPlayer getAIOwner() {
-        return (unit == null) ? null
-            : (unit.getOwner() == null) ? null
-            : getAIMain().getAIPlayer(unit.getOwner());
+        final Player owner = getOwner();
+        return (owner == null) ? null : getAIMain().getAIPlayer(owner);
     }
 
     /**
      * Gets the PRNG to use with this unit.
      *
-     * @return A <code>Random</code> instance.
+     * @return A {@code Random} instance.
      */
     public Random getAIRandom() {
-        return (unit == null) ? null : getAIOwner().getAIRandom();
+        AIPlayer aiOwner = getAIOwner();
+        return (aiOwner == null) ? null : aiOwner.getAIRandom();
     }
 
     /**
@@ -270,8 +272,10 @@ public class AIUnit extends TransportableAIObject {
      * @return A trivial target, or null if none found.
      */
     public Location getTrivialTarget() {
-        PathNode path = unit.getTrivialPath();
-        return (path == null) ? null
+        final Unit unit = getUnit();
+        PathNode path;
+        return (unit == null
+            || (path = unit.getTrivialPath()) == null) ? null
             : Location.upLoc(path.getLastNode().getLocation());
     }
 
@@ -281,13 +285,15 @@ public class AIUnit extends TransportableAIObject {
      * @return True if the unit has cargo aboard.
      */
     public final boolean hasCargo() {
+        final Unit unit = getUnit();
         return (unit == null) ? false : unit.hasCargo();
     }
 
     /**
      * Does this unit have a particular class of mission?
      *
-     * @param returnClass The <code>Class</code> of mission to check.
+     * @param <T> The type of the mission.
+     * @param returnClass The {@code Class} of mission to check.
      * @return True if the mission is of the given class.
      */
     public <T extends Mission> boolean hasMission(Class<T> returnClass) {
@@ -297,8 +303,9 @@ public class AIUnit extends TransportableAIObject {
     /**
      * Get the unit mission if it is of a given class.
      *
-     * @param returnClass The <code>Class</code> of the mission.
-     * @return The <code>Mission</code>, or null if it is not of the
+     * @param <T> The type of the mission.
+     * @param returnClass The {@code Class} of the mission.
+     * @return The {@code Mission}, or null if it is not of the
      *     given class.
      */
     public <T extends Mission> T getMission(Class<T> returnClass) {
@@ -315,7 +322,7 @@ public class AIUnit extends TransportableAIObject {
      * Do *not* check mission validity.  The mission itself does that,
      * and has special case error handling.
      *
-     * @param lb A <code>LogBuilder</code> to log to.
+     * @param lb A {@code LogBuilder} to log to.
      */
     public void doMission(LogBuilder lb) {
         if (this.mission != null) this.mission.doMission(lb);
@@ -324,17 +331,13 @@ public class AIUnit extends TransportableAIObject {
     /**
      * Change the mission of a unit.
      *
-     * @param mission The new <code>Mission</code>.
-     * @return The new current <code>Mission</code>.
+     * @param mission The new {@code Mission}.
+     * @return The new current {@code Mission}.
      */
     public Mission changeMission(Mission mission) {
         if (this.mission == mission) return this.mission;
 
-        if (this.mission != null) {
-            this.mission.dispose();
-            this.mission = null;
-        }
-       
+        removeMission();
         setMission(mission);
         if (mission != null) {
             setTransportPriority(mission.getBaseTransportPriority());
@@ -342,13 +345,121 @@ public class AIUnit extends TransportableAIObject {
         return this.mission;
     }
 
+    public void removeMission() {
+        if (this.mission != null) {
+            this.mission.dispose();
+            this.mission = null;
+        }
+    }
+
+
+    // Helper methods for AIColony
+
+    public boolean hasDefendSettlementMission() {
+        return hasMission(DefendSettlementMission.class);
+    }
+
+    public boolean isCompleteWishRealizationMission(Colony colony) {
+        WishRealizationMission wm
+            = getMission(WishRealizationMission.class);
+        return (wm != null && Map.isSameLocation(wm.getTarget(), colony));
+    }
+
+    public boolean isAvailableForWork(Colony colony) {
+        Mission m = getMission();
+        return (m == null
+            || (m instanceof BuildColonyMission
+                && (Map.isSameLocation(m.getTarget(), colony.getTile())
+                    || colony.getUnitCount() <= 1))
+            // FIXME: drop this when the AI stops building excessive armies
+            || m instanceof DefendSettlementMission
+            || m instanceof IdleAtSettlementMission
+            || m instanceof WorkInsideColonyMission
+            );
+    }
+
+    public boolean tryWorkInsideColonyMission(AIColony aiColony, LogBuilder lb) {
+        WorkInsideColonyMission wic
+            = getMission(WorkInsideColonyMission.class);
+        if (wic == null) {
+            AIPlayer aiPlayer = getAIOwner();
+            if (!(aiPlayer instanceof EuropeanAIPlayer) ||
+                    ((EuropeanAIPlayer)aiPlayer)
+                    .getWorkInsideColonyMission(this, aiColony) == null) {
+                return false; 
+            }
+            lb.add(", ", getMission());
+            dropTransport();
+        }
+        return true;
+    }
+
+    public boolean tryPioneeringMission(LogBuilder lb) {
+        Mission m = getMission();
+        Location oldTarget = (m == null) ? null : m.getTarget();
+        AIPlayer aiPlayer = getAIOwner();
+
+        if (aiPlayer instanceof EuropeanAIPlayer) {
+            EuropeanAIPlayer euaiPlayer = (EuropeanAIPlayer)aiPlayer;
+            if (euaiPlayer.getPioneeringMission(this, null) != null) {
+                lb.add(", ", getMission());
+                euaiPlayer.updateTransport(this, oldTarget, lb);
+                return true;
+            }
+        }
+        return false;
+    }
+
+    public boolean trySomeUsefulMission(Colony colony, LogBuilder lb) {
+        Mission m = getMission();
+        if (m instanceof BuildColonyMission
+            || m instanceof DefendSettlementMission
+            || m instanceof MissionaryMission
+            || m instanceof PioneeringMission
+            || m instanceof ScoutingMission
+            || m instanceof UnitSeekAndDestroyMission)
+            return true;
+        Location oldTarget = (m == null) ? null : m.getTarget();
+        AIPlayer aiPlayer = getAIOwner();
+
+        final Unit unit = getUnit();
+        if(aiPlayer instanceof EuropeanAIPlayer) {
+            EuropeanAIPlayer euaiPlayer = (EuropeanAIPlayer)getAIOwner();
+
+            if (unit.hasAbility(Ability.SPEAK_WITH_CHIEF)
+                && (m = euaiPlayer.getScoutingMission(this)) != null) {
+                lb.add(", ", m);
+                euaiPlayer.updateTransport(this, oldTarget, lb);
+                return true;
+            } else if (unit.isDefensiveUnit()
+                && (m = euaiPlayer.getDefendSettlementMission(this, colony)) != null) {
+                lb.add(", ", m);
+                euaiPlayer.updateTransport(this, oldTarget, lb);
+                return true;
+            } else if (unit.hasAbility(Ability.ESTABLISH_MISSION)
+                && (m = euaiPlayer.getMissionaryMission(this)) != null) {
+                lb.add(", ", m);
+                euaiPlayer.updateTransport(this, oldTarget, lb);
+                return true;
+            }
+        }
+        return false;
+    }
+
+    public void removeTransportable(AIGoods ag) {
+        TransportMission tm = getMission(TransportMission.class);
+        if (tm != null) {
+            tm.removeTransportable(ag);
+        }
+    }
+
     /**
      * Moves a unit to the new world.
      *
      * @return True if there was no c-s problem.
      */
     public boolean moveToAmerica() {
-        return AIMessage.askMoveTo(this, unit.getOwner().getGame().getMap());
+        return AIMessage.askMoveTo(this, getOwner().getGame().getMap());
     }
 
     /**
@@ -357,16 +468,17 @@ public class AIUnit extends TransportableAIObject {
      * @return True if there was no c-s problem.
      */
     public boolean moveToEurope() {
-        return AIMessage.askMoveTo(this, unit.getOwner().getEurope());
+        return AIMessage.askMoveTo(this, getOwner().getEurope());
     }
 
     /**
      * Moves this AI unit.
      *
-     * @param direction The <code>Direction</code> to move.
+     * @param direction The {@code Direction} to move.
      * @return True if the move succeeded.
      */
     public boolean move(Direction direction) {
+        final Unit unit = getUnit();
         Tile start = unit.getTile();
         return unit.getMoveType(direction).isProgress()
             && AIMessage.askMove(this, direction)
@@ -380,29 +492,72 @@ public class AIUnit extends TransportableAIObject {
      * (possibly requiring a purchase, which may fail due to lack of gold
      * or boycotts in effect).
      *
-     * @param role The <code>Role</code> to equip for identifier.
+     * @param role The {@code Role} to equip for identifier.
      * @return True if the role change was successful.
      */
     public boolean equipForRole(Role role) {
-        final Specification spec = getSpecification();
-        final Player player = unit.getOwner();
+        final Player player = getOwner();
+        final Unit unit = getUnit();
         Location loc = Location.upLoc(unit.getLocation());
         if (!(loc instanceof UnitLocation)) return false;
         int count = role.getMaximumCount();
         if (count > 0) {
             for (; count > 0; count--) {
                 List<AbstractGoods> req = unit.getGoodsDifference(role, count);
-                int price = ((UnitLocation)loc).priceGoods(req);
-                if (price < 0) continue;
-                if (player.checkGold(price)) break;
+                try {
+                    int price = ((UnitLocation)loc).priceGoods(req);
+                    if (player.checkGold(price)) break;
+                } catch (FreeColException fce) {
+                    continue;
+                }
             }
             if (count <= 0) return false;
         }
         return AIMessage.askEquipForRole(this, role, count)
-            && unit.getRole() == role && unit.getRoleCount() == count;
+            && unit.getRole() == role
+            && unit.getRoleCount() == count;
     }
 
-        
+    /**
+     * Score this AI unit with its suitability for building.
+     *
+     * Favour unequipped freeColonists, and other unskilled over experts.
+     * Also slightly favour units on the map.
+     *
+     * @return An integer score.
+     */
+    public int getBuilderScore() {
+        final Unit unit = getUnit();
+        if (unit == null || BuildColonyMission.invalidMissionReason(this) != null)
+            return -1000;
+        int ret = (!unit.hasDefaultRole()) ? 0
+            : (unit.getSkillLevel() > 0) ? 100
+            : 500 + 100 * unit.getSkillLevel();
+        if (unit.hasTile()) ret += 50;
+        return ret;
+    }        
+
+    /**
+     * Score this AI unit with its suitability for pioneering.
+     *
+     * @return An integer score.
+     */
+    public int getPioneerScore() {
+        final Unit unit = getUnit();
+        return (unit == null) ? -1000 : unit.getPioneerScore();
+    }
+
+    /**
+     * Score this AI unit with its suitability for scouting.
+     *
+     * @return An integer score.
+     */
+    public int getScoutScore() {
+        final Unit unit = getUnit();
+        return (unit == null) ? -1000 : unit.getScoutScore();
+    }
+
+    
     // Implement TransportableAIObject
 
     /**
@@ -418,7 +573,7 @@ public class AIUnit extends TransportableAIObject {
      */
     @Override
     public Locatable getTransportLocatable() {
-        return unit;
+        return getUnit();
     }
 
     /**
@@ -426,8 +581,9 @@ public class AIUnit extends TransportableAIObject {
      */
     @Override
     public Location getTransportSource() {
-        return (getUnit() == null || getUnit().isDisposed()) ? null
-            : getUnit().getLocation();
+        final Unit unit = getUnit();
+        return (unit == null || unit.isDisposed()) ? null
+            : unit.getLocation();
     }
 
     /**
@@ -435,9 +591,10 @@ public class AIUnit extends TransportableAIObject {
      */
     @Override
     public Location getTransportDestination() {
-        return (getUnit() == null || getUnit().isDisposed() || !hasMission())
+        final Unit unit = getUnit();
+        return (unit == null || unit.isDisposed() || !hasMission())
             ? null
-            : mission.getTransportDestination();
+            : this.mission.getTransportDestination();
     }
 
     /**
@@ -452,8 +609,9 @@ public class AIUnit extends TransportableAIObject {
         dst = Location.upLoc(dst);
 
         PathNode path;
+        final Unit unit = getUnit();
         if (unit.getLocation() == carrier) {
-            path = unit.findPath(carrier.getLocation(), dst, carrier, null);
+            path = unit.findPath(carrier.getLocation(), dst, carrier);
             if (path == null && dst.getTile() != null) {
                 path = unit.findPathToNeighbour(carrier.getLocation(),
                     dst.getTile(), carrier, null);
@@ -461,7 +619,7 @@ public class AIUnit extends TransportableAIObject {
         } else if (unit.getLocation() instanceof Unit) {
             return null;
         } else {
-            path = unit.findPath(unit.getLocation(), dst, carrier, null);
+            path = unit.findPath(unit.getLocation(), dst, carrier);
             if (path == null && dst.getTile() != null) {
                 path = unit.findPathToNeighbour(unit.getLocation(),
                     dst.getTile(), carrier, null);
@@ -484,7 +642,7 @@ public class AIUnit extends TransportableAIObject {
      */
     @Override
     public void setTransportDestination(Location destination) {
-        throw new RuntimeException("AI unit transport destination set by mission.");
+        throw new RuntimeException("AI unit transport destination set by mission:" + destination);
     }
 
     /**
@@ -560,7 +718,7 @@ public class AIUnit extends TransportableAIObject {
 
         // Pick the available tile with the shortest path to one of our
         // settlements, or the tile with the highest defence value.
-        final Player player = unit.getOwner();
+        final Player player = getOwner();
         Tile safe = tiles.get(0);
         Tile best = null;
         int bestTurns = Unit.MANY_TURNS;
@@ -581,20 +739,20 @@ public class AIUnit extends TransportableAIObject {
             }
         }
         return leaveTransport(tile.getDirection((best != null) ? best : safe));
-    }               
+    }
 
     /**
      * {@inheritDoc}
      */
     @Override
     public boolean leaveTransport(Direction direction) {
+        final Unit unit = getUnit();
         if (!unit.isOnCarrier()) return false;
         final Unit carrier = unit.getCarrier();
         boolean result = (direction == null)
             ? (AIMessage.askDisembark(this)
                 && unit.getLocation() == carrier.getLocation())
             : move(direction);
-
         if (result) {
             requestLocalRearrange();
             dropTransport();
@@ -609,6 +767,7 @@ public class AIUnit extends TransportableAIObject {
     public boolean joinTransport(Unit carrier, Direction direction) {
         AIUnit aiCarrier = getAIMain().getAIUnit(carrier);
         if (aiCarrier == null) return false;
+        final Unit unit = getUnit();
         boolean result = AIMessage.askEmbark(aiCarrier, unit, direction)
             && unit.getLocation() == carrier;
 
@@ -640,14 +799,10 @@ public class AIUnit extends TransportableAIObject {
     public void dispose() {
         dropTransport();
         AIPlayer aiOwner = getAIOwner();
-        if (aiOwner != null) {
-            aiOwner.removeAIUnit(this);
-        } else {
-            // This happens with missionaries, and legitimately when a unit
-            // changes owner.  FIXME: cleanup.
-            logger.warning("Disposing of " + getId() + " but owner is null!");
-        }
-        if (mission != null) {
+        // Might not be owned by an AI
+        if (aiOwner != null) aiOwner.removeAIObject(this);
+
+        if (this.mission != null) {
             this.mission.dispose();
             this.mission = null;
         }
@@ -655,17 +810,18 @@ public class AIUnit extends TransportableAIObject {
     }
 
     /**
-     * Checks the integrity of this AIUnit.
-     *
-     * @param fix Fix problems if possible.
-     * @return Negative if there are problems remaining, zero if
-     *     problems were fixed, positive if no problems found at all.
+     * {@inheritDoc}
      */
     @Override
-    public int checkIntegrity(boolean fix) {
-        int result = super.checkIntegrity(fix);
-        if (unit == null || unit.isDisposed()) {
-            result = -1;
+    public IntegrityType checkIntegrity(boolean fix, LogBuilder lb) {
+        IntegrityType result = super.checkIntegrity(fix, lb);
+        final Unit unit = getUnit();
+        if (unit == null) {
+            lb.add("\n  AIUnit with null unit: ", getId());
+            result = result.fail();
+        } else if (unit.isDisposed()) {
+            lb.add("\n  AIUnit with disposed unit: ", getId());
+            result = result.fail();
         }
         return result;
     }
@@ -673,13 +829,6 @@ public class AIUnit extends TransportableAIObject {
 
     // Serialization
 
-    // @compat 0.10.3
-    private static final String TILE_IMPROVEMENT_PLAN_MISSION_TAG = "tileImprovementPlanMission";
-    // end @compat
-    // @compat 0.10.5
-    private static final String IDLE_AT_COLONY_MISSION_TAG = "idleAtColonyMission";
-    // end @compat
-
 
     /**
      * {@inheritDoc}
@@ -702,25 +851,17 @@ public class AIUnit extends TransportableAIObject {
 
         final AIMain aiMain = getAIMain();
 
-        unit = xr.findFreeColGameObject(aiMain.getGame(), ID_ATTRIBUTE_TAG,
-                                        Unit.class, (Unit)null, true);
-        if (unit.isUninitialized()) {
+        Unit u = xr.findFreeColGameObject(aiMain.getGame(), ID_ATTRIBUTE_TAG,
+                                          Unit.class, (Unit)null, true);
+        if (u.isInitialized()) {
+            setUnit(u);
+        } else {
             xr.nextTag(); // Move off the opening <AIUnit> tag
             throw new XMLStreamException("AIUnit for uninitialized Unit: "
-                + unit.getId());
+                + u.getId());
         }
     }
 
-    /**
-     * {@inheritDoc}
-     */
-    @Override
-    protected void readChildren(FreeColXMLReader xr) throws XMLStreamException {
-        super.readChildren(xr);
-
-        if (getUnit() != null) uninitialized = false;
-    }
-
     /**
      * {@inheritDoc}
      */
@@ -730,60 +871,52 @@ public class AIUnit extends TransportableAIObject {
         final String tag = xr.getLocalName();
 
         mission = null;
-        if (BuildColonyMission.getXMLElementTagName().equals(tag)) {
+        if (BuildColonyMission.TAG.equals(tag)) {
             mission = new BuildColonyMission(aiMain, this, xr);
 
-        } else if (CashInTreasureTrainMission.getXMLElementTagName().equals(tag)) {
+        } else if (CashInTreasureTrainMission.TAG.equals(tag)) {
             mission = new CashInTreasureTrainMission(aiMain, this, xr);
 
-        } else if (DefendSettlementMission.getXMLElementTagName().equals(tag)) {
+        } else if (DefendSettlementMission.TAG.equals(tag)) {
             mission = new DefendSettlementMission(aiMain, this, xr);
 
-        } else if (IdleAtSettlementMission.getXMLElementTagName().equals(tag)
-            // @compat 0.10.5
-            || IDLE_AT_COLONY_MISSION_TAG.equals(tag)
-            // end @compat
-                   ) {
+        } else if (IdleAtSettlementMission.TAG.equals(tag)) {
             mission = new IdleAtSettlementMission(aiMain, this, xr);
 
-        } else if (IndianBringGiftMission.getXMLElementTagName().equals(tag)) {
+        } else if (IndianBringGiftMission.TAG.equals(tag)) {
             mission = new IndianBringGiftMission(aiMain, this, xr);
 
-        } else if (IndianDemandMission.getXMLElementTagName().equals(tag)) {
+        } else if (IndianDemandMission.TAG.equals(tag)) {
             mission = new IndianDemandMission(aiMain, this, xr);
 
-        } else if (MissionaryMission.getXMLElementTagName().equals(tag)) {
+        } else if (MissionaryMission.TAG.equals(tag)) {
             mission = new MissionaryMission(aiMain, this, xr);
 
-        } else if (PioneeringMission.getXMLElementTagName().equals(tag)
-            // @compat 0.10.3
-            || TILE_IMPROVEMENT_PLAN_MISSION_TAG.equals(tag)
-            // end @compat
-                   ) {
+        } else if (PioneeringMission.TAG.equals(tag)) {
             mission = new PioneeringMission(aiMain, this, xr);
 
-        } else if (PrivateerMission.getXMLElementTagName().equals(tag)) {
+        } else if (PrivateerMission.TAG.equals(tag)) {
             mission = new PrivateerMission(aiMain, this, xr);
 
-        } else if (ScoutingMission.getXMLElementTagName().equals(tag)) {
+        } else if (ScoutingMission.TAG.equals(tag)) {
             mission = new ScoutingMission(aiMain, this, xr);
 
-        } else if (TransportMission.getXMLElementTagName().equals(tag)) {
+        } else if (TransportMission.TAG.equals(tag)) {
             mission = new TransportMission(aiMain, this, xr);
 
-        } else if (UnitSeekAndDestroyMission.getXMLElementTagName().equals(tag)) {
+        } else if (UnitSeekAndDestroyMission.TAG.equals(tag)) {
             mission = new UnitSeekAndDestroyMission(aiMain, this, xr);
 
-        } else if (UnitWanderHostileMission.getXMLElementTagName().equals(tag)) {
+        } else if (UnitWanderHostileMission.TAG.equals(tag)) {
             mission = new UnitWanderHostileMission(aiMain, this, xr);
 
-        } else if (UnitWanderMission.getXMLElementTagName().equals(tag)) {
+        } else if (UnitWanderMission.TAG.equals(tag)) {
             mission = new UnitWanderMission(aiMain, this, xr);
 
-        } else if (WishRealizationMission.getXMLElementTagName().equals(tag)) {
+        } else if (WishRealizationMission.TAG.equals(tag)) {
             mission = new WishRealizationMission(aiMain, this, xr);
 
-        } else if (WorkInsideColonyMission.getXMLElementTagName().equals(tag)) {
+        } else if (WorkInsideColonyMission.TAG.equals(tag)) {
             mission = new WorkInsideColonyMission(aiMain, this, xr);
 
         } else {
@@ -795,22 +928,39 @@ public class AIUnit extends TransportableAIObject {
      * {@inheritDoc}
      */
     @Override
-    public String toString() {
-        return (unit == null) ? "AIUnit-null" : unit.toString("AIUnit ");
+    public String getXMLTagName() { return TAG; }
+
+
+    // Override Object
+
+    /**
+     * {@inheritDoc}
+     */
+    @Override
+    public boolean equals(Object o) {
+        if (!(o instanceof AIUnit)) return false;
+        AIUnit other = (AIUnit)o;
+        return Utils.equals(this.unit, other.unit)
+            && Utils.equals(this.mission, other.mission)
+            && super.equals(other);
     }
 
     /**
      * {@inheritDoc}
      */
     @Override
-    public String getXMLTagName() { return getXMLElementTagName(); }
+    public int hashCode() {
+        int hash = super.hashCode();
+        hash = 37 * hash + Utils.hashCode(this.unit);
+        return 37 * hash + Utils.hashCode(this.mission);
+    }
 
     /**
-     * Gets the tag name of the root element representing this object.
-     *
-     * @return "aiUnit"
+     * {@inheritDoc}
      */
-    public static String getXMLElementTagName() {
-        return "aiUnit";
+    @Override
+    public String toString() {
+        return (this.unit == null) ? "AIUnit-null"
+            : this.unit.toString("AIUnit ");
     }
 }
