@@ -1,5 +1,5 @@
 /**
- *  Copyright (C) 2002-2015   The FreeCol Team
+ *  Copyright (C) 2002-2022   The FreeCol Team
  *
  *  This file is part of FreeCol.
  *
@@ -20,13 +20,15 @@
 package net.sf.freecol.server.ai;
 
 import java.util.ArrayList;
-import java.util.Collections;
 import java.util.Comparator;
 import java.util.HashMap;
 import java.util.HashSet;
 import java.util.List;
 import java.util.Random;
 import java.util.Set;
+import java.util.function.Function;
+import java.util.function.Predicate;
+import java.util.function.ToIntFunction;
 import java.util.logging.Level;
 import java.util.logging.Logger;
 
@@ -35,12 +37,16 @@ import javax.xml.stream.XMLStreamException;
 import net.sf.freecol.common.io.FreeColXMLReader;
 import net.sf.freecol.common.model.Colony;
 import net.sf.freecol.common.model.CombatModel;
+import net.sf.freecol.common.model.Constants.IndianDemandAction;
 import net.sf.freecol.common.model.FeatureContainer;
-import net.sf.freecol.common.model.GameOptions;
 import net.sf.freecol.common.model.Goods;
+import net.sf.freecol.common.model.GoodsType;
 import net.sf.freecol.common.model.IndianSettlement;
 import net.sf.freecol.common.model.Location;
 import net.sf.freecol.common.model.Modifier;
+import net.sf.freecol.common.model.NativeTrade;
+import net.sf.freecol.common.model.NativeTrade.NativeTradeAction;
+import net.sf.freecol.common.model.NativeTradeItem;
 import net.sf.freecol.common.model.Ownable;
 import net.sf.freecol.common.model.PathNode;
 import net.sf.freecol.common.model.Player;
@@ -54,7 +60,8 @@ import net.sf.freecol.common.model.Unit;
 import net.sf.freecol.common.model.Role;
 import net.sf.freecol.common.model.pathfinding.CostDecider;
 import net.sf.freecol.common.model.pathfinding.CostDeciders;
-import net.sf.freecol.common.networking.NetworkConstants;
+import net.sf.freecol.common.option.GameOptions;
+import static net.sf.freecol.common.util.CollectionUtils.*;
 import net.sf.freecol.common.util.LogBuilder;
 import net.sf.freecol.common.util.RandomChoice;
 import static net.sf.freecol.common.util.RandomUtils.*;
@@ -64,7 +71,6 @@ import net.sf.freecol.server.ai.mission.IndianDemandMission;
 import net.sf.freecol.server.ai.mission.Mission;
 import net.sf.freecol.server.ai.mission.UnitSeekAndDestroyMission;
 import net.sf.freecol.server.ai.mission.UnitWanderHostileMission;
-import net.sf.freecol.server.model.ServerPlayer;
 
 
 /**
@@ -74,7 +80,7 @@ import net.sf.freecol.server.model.ServerPlayer;
  * The method {@link #startWorking} gets called by the
  * {@link AIInGameInputHandler} when it is this player's turn.
  */
-public class NativeAIPlayer extends AIPlayer {
+public final class NativeAIPlayer extends MissionAIPlayer {
 
     private static final Logger logger = Logger.getLogger(NativeAIPlayer.class.getName());
 
@@ -86,12 +92,6 @@ public class NativeAIPlayer extends AIPlayer {
 
     public static final int MAX_NUMBER_OF_DEMANDS = 1;
 
-    /**
-     * Stores temporary information for sessions (trading with another
-     * player etc).
-     */
-    private final HashMap<String, Integer> sessionRegister = new HashMap<>();
-
     /**
      * Debug helper to keep track of why/what the units are doing.
      * Do not serialize.
@@ -100,20 +100,20 @@ public class NativeAIPlayer extends AIPlayer {
 
 
     /**
-     * Creates a new <code>AIPlayer</code>.
+     * Creates a new {@code AIPlayer}.
      *
      * @param aiMain The main AI-class.
      * @param player The player that should be associated with this
-     *            <code>AIPlayer</code>.
+     *            {@code AIPlayer}.
      */
-    public NativeAIPlayer(AIMain aiMain, ServerPlayer player) {
+    public NativeAIPlayer(AIMain aiMain, Player player) {
         super(aiMain, player);
 
-        uninitialized = getPlayer() == null;
+        this.initialized = getPlayer() != null;
     }
 
     /**
-     * Creates a new <code>AIPlayer</code>.
+     * Creates a new {@code AIPlayer}.
      *
      * @param aiMain The main AI-object.
      * @param xr The input stream containing the XML.
@@ -123,7 +123,7 @@ public class NativeAIPlayer extends AIPlayer {
                           FreeColXMLReader xr) throws XMLStreamException {
         super(aiMain, xr);
 
-        uninitialized = getPlayer() == null;
+        this.initialized = getPlayer() != null;
     }
 
 
@@ -131,20 +131,16 @@ public class NativeAIPlayer extends AIPlayer {
      * Simple initialization of AI missions given that we know the starting
      * conditions.
      *
-     * @param lb A <code>LogBuilder</code>  to log to.
+     * @param lb A {@code LogBuilder}  to log to.
      */
     private void initializeMissions(LogBuilder lb) {
-        final AIMain aiMain = getAIMain();
         final Player player = getPlayer();
         lb.add("\n  Initialize");
 
         // Give defensive missions up to the minimum expected defence,
         // leave the rest with the default wander-hostile mission.
-        List<Unit> units = new ArrayList<>();
-        for (IndianSettlement is : player.getIndianSettlements()) {
-            units.clear();
-            units.addAll(is.getTile().getUnitList());
-            units.addAll(is.getUnitList());
+        for (IndianSettlement is : player.getIndianSettlementList()) {
+            List<Unit> units = is.getAllUnitsList();
             while (units.size() > is.getRequiredDefenders()) {
                 Unit u = units.remove(0);
                 AIUnit aiu = getAIUnit(u);
@@ -163,18 +159,17 @@ public class NativeAIPlayer extends AIPlayer {
      * Determines the stances towards each player.
      * That is: should we declare war?
      *
-     * @param lb A <code>LogBuilder</code> to log to.
+     * @param lb A {@code LogBuilder} to log to.
      */
     private void determineStances(LogBuilder lb) {
-        final ServerPlayer serverPlayer = (ServerPlayer)getPlayer();
+        final Player player = getPlayer();
         lb.mark();
 
-        for (Player p : getGame().getLivePlayers(serverPlayer)) {
+        for (Player p : getGame().getLivePlayerList(player)) {
             Stance newStance = determineStance(p);
-            if (newStance != serverPlayer.getStance(p)) {
+            if (newStance != player.getStance(p)) {
                 getAIMain().getFreeColServer().getInGameController()
-                    .changeStance(serverPlayer, newStance, 
-                                  (ServerPlayer)p, true);
+                    .changeStance(player, newStance, p, true);
                 lb.add(" ", p.getDebugName(), "->", newStance, ", ");
             }
         }
@@ -187,12 +182,12 @@ public class NativeAIPlayer extends AIPlayer {
      * units new missions.
      *
      * @param randoms An array of random settlement indexes.
-     * @param lb A <code>LogBuilder</code> to log to.
+     * @param lb A {@code LogBuilder} to log to.
      */
     private void secureSettlements(int[] randoms, LogBuilder lb) {
         int randomIdx = 0;
         List<IndianSettlement> settlements
-            = getPlayer().getIndianSettlements();
+            = getPlayer().getIndianSettlementList();
         for (IndianSettlement is : settlements) {
             // Spread arms and horses between camps
             // FIXME: maybe make this dependent on difficulty level?
@@ -212,24 +207,17 @@ public class NativeAIPlayer extends AIPlayer {
 
     /**
      * Greedily equips braves with horses and muskets.
+     *
      * Public for the test suite.
      *
-     * @param is The <code>IndianSettlement</code> where the equipping occurs.
-     * @param lb A <code>LogBuilder</code> to log to.
+     * @param is The {@code IndianSettlement} where the equipping occurs.
+     * @param lb A {@code LogBuilder} to log to.
      */
     public void equipBraves(IndianSettlement is, LogBuilder lb) {
-        final Specification spec = getSpecification();
-
-        // Find all the units
-        List<Unit> units = is.getUnitList();
-        units.addAll(is.getTile().getUnitList());
-
         // Prioritize promoting partially equipped units to full dragoon
-        Collections.sort(units,
-            getGame().getCombatModel().getMilitaryStrengthComparator());
-
-        boolean moreHorses = true, moreMuskets = true;
-        for (Unit u : units) {
+        final Comparator<Unit> comp = getGame().getCombatModel()
+            .getMilitaryStrengthComparator();
+        for (Unit u : sort(is.getAllUnitsList(), comp)) {
             Role r = is.canImproveUnitMilitaryRole(u);
             if (r != null) {
                 Role old = u.getRole();
@@ -242,10 +230,11 @@ public class NativeAIPlayer extends AIPlayer {
 
     /**
      * Takes the necessary actions to secure an indian settlement
+     *
      * Public for the test suite.
      *
-     * @param is The <code>IndianSettlement</code> to secure.
-     * @param lb A <code>LogBuilder</code> to log to.
+     * @param is The {@code IndianSettlement} to secure.
+     * @param lb A {@code LogBuilder} to log to.
      */
     public void secureIndianSettlement(final IndianSettlement is,
                                        LogBuilder lb) {
@@ -256,11 +245,9 @@ public class NativeAIPlayer extends AIPlayer {
         DefendSettlementMission dm;
 
         // Collect native units and defenders
-        List<Unit> units = new ArrayList<>();
+        List<Unit> units = is.getAllUnitsList();
         List<Unit> defenders = new ArrayList<>();
-        units.addAll(is.getUnitList());
-        units.addAll(is.getTile().getUnitList());
-        for (Unit u : is.getOwnedUnits()) {
+        for (Unit u : is.getOwnedUnitList()) {
             if (!units.contains(u)) units.add(u);
         }
 
@@ -313,9 +300,9 @@ public class NativeAIPlayer extends AIPlayer {
                     bonus = (float)tension.getLevel().ordinal()
                         - Tension.Level.CONTENT.ordinal();
                 }
-                value += t.getUnitList().stream()
-                    .filter(u -> cm.getOffencePower(u, is) > threshold)
-                    .mapToDouble(u -> cm.getOffencePower(u, is) + bonus).sum();
+                value += sumDouble(t.getUnits(),
+                                   u -> cm.getOffencePower(u, is) > threshold,
+                                   u -> cm.getOffencePower(u, is) + bonus);
                 if (value > 0.0) threats.put(t, value);
             }
         }
@@ -325,24 +312,16 @@ public class NativeAIPlayer extends AIPlayer {
         // Also favour units native to the settlement.
         final int homeBonus = 3;
         final Tile isTile = is.getTile();
-        final Comparator<Unit> isComparator
-            = new Comparator<Unit>() {
-                @Override
-                public int compare(Unit u1, Unit u2) {
-                    Tile t1 = u1.getTile();
-                    int s1 = t1.getDistanceTo(isTile);
-                    Tile t2 = u2.getTile();
-                    int s2 = t2.getDistanceTo(isTile);
-                    if (u1.getHomeIndianSettlement() == is) s1 -= homeBonus;
-                    if (u2.getHomeIndianSettlement() == is) s2 -= homeBonus;
-                    return s1 - s2;
-                }
-           };
+        final Comparator<Unit> isComparator = cachingIntComparator(u -> {
+                final Tile t = u.getTile();
+                return t.getDistanceTo(isTile)
+                    - ((u.getHomeIndianSettlement() == is) ? homeBonus : 0);
+            });
 
         // Do we need more or less defenders?
         int needed = minimumDefence + threats.size();
         if (defenders.size() < needed) { // More needed, call some in.
-            Collections.sort(units, isComparator);
+            units.sort(isComparator);
             while (!units.isEmpty()) {
                 Unit u = units.remove(0);
                 AIUnit aiu = aiMain.getAIUnit(u);
@@ -354,22 +333,16 @@ public class NativeAIPlayer extends AIPlayer {
                 }
             }
         } else if (defenders.size() > needed) { // Less needed, release them
-            Collections.sort(defenders, isComparator);
-            Collections.reverse(defenders);
+            defenders.sort(isComparator.reversed());
             while (defenders.size() > needed) {
                 units.add(defenders.remove(0));
             }
         }
 
         // Sort threat tiles by threat value.
-        List<Tile> threatTiles = new ArrayList<>(threats.keySet());
-        Collections.sort(threatTiles, new Comparator<Tile>() {
-                @Override
-                public int compare(Tile t1, Tile t2) {
-                    return Double.compare(threats.get(t2),
-                            threats.get(t1));
-                }
-            });
+        final Comparator<Tile> threatComp
+            = Comparator.<Tile>comparingDouble(t -> threats.get(t));
+        List<Tile> threatTiles = sort(threats.keySet(), threatComp);
 
         if (!defenders.isEmpty()) {
             lb.add(" defend with:");
@@ -381,18 +354,14 @@ public class NativeAIPlayer extends AIPlayer {
         // Assign units to attack the threats, greedily chosing closest unit.
         while (!threatTiles.isEmpty() && !units.isEmpty()) {
             Tile tile = threatTiles.remove(0);
-            int bestDistance = Integer.MAX_VALUE;
-            Unit unit = null;
-            for (Unit u : units) {
-                AIUnit aiu = aiMain.getAIUnit(u);
-                if (UnitSeekAndDestroyMission.invalidReason(aiu,
-                        tile.getDefendingUnit(u)) != null) continue;
-                int distance = u.getTile().getDistanceTo(tile);
-                if (bestDistance > distance) {
-                    bestDistance = distance;
-                    unit = u;
-                }
-            }
+            final ToIntFunction<Unit> score = cacheInt(u ->
+                u.getTile().getDistanceTo(tile));
+            final Predicate<Unit> validPred = u ->
+                UnitSeekAndDestroyMission.invalidMissionReason(aiMain.getAIUnit(u),
+                    tile.getDefendingUnit(u)) == null
+                && score.applyAsInt(u) >= 0;
+            final Comparator<Unit> scoreComp = Comparator.comparingInt(score);
+            Unit unit = minimize(units, validPred, scoreComp);
             if (unit == null) continue; // Declined to attack.
             units.remove(unit);
             AIUnit aiUnit = aiMain.getAIUnit(unit);
@@ -405,24 +374,20 @@ public class NativeAIPlayer extends AIPlayer {
     /**
      * Gives a mission to all units.
      *
-     * @param lb A <code>LogBuilder</code> to log to.
+     * @param lb A {@code LogBuilder} to log to.
      */
     private void giveNormalMissions(LogBuilder lb) {
-        final AIMain aiMain = getAIMain();
         final Player player = getPlayer();
-        final Specification spec = getSpecification();
-        final int turnNumber = getGame().getTurn().getNumber();
         List<AIUnit> aiUnits = getAIUnits();
 
         lb.mark();
-        List<AIUnit> done = new ArrayList<>();
+        List<AIUnit> done = new ArrayList<>(aiUnits.size());
         reasons.clear();
         for (AIUnit aiUnit : aiUnits) {
             final Unit unit = aiUnit.getUnit();
             Mission m = aiUnit.getMission();
-            String reason = null;
 
-            if (unit.isUninitialized() || unit.isDisposed()) {
+            if (!unit.isInitialized() || unit.isDisposed()) {
                 reasons.put(unit, "Invalid");
 
             } else if (m != null && m.isValid() && !m.isOneTime()) {
@@ -484,8 +449,7 @@ public class NativeAIPlayer extends AIPlayer {
             lb.add("\n  Free Land Units:");
             for (AIUnit aiu : aiUnits) lb.add(" ", aiu.getUnit());
         }
-        lb.add("\n  Missions(settlements=",
-            player.getNumberOfSettlements(), ")");
+        lb.add("\n  Missions(settlements=", player.getSettlementCount(), ")");
         logMissions(reasons, lb);
     }
 
@@ -493,19 +457,19 @@ public class NativeAIPlayer extends AIPlayer {
      * Brings gifts to nice players with nearby colonies.
      *
      * @param randoms An array of random percentages.
-     * @param lb A <code>LogBuilder</code> to log to.
+     * @param lb A {@code LogBuilder} to log to.
      */
     private void bringGifts(int[] randoms, LogBuilder lb) {
         final Player player = getPlayer();
         final CostDecider cd = CostDeciders.numberOfLegalTiles();
-        final int giftProbability = getSpecification()
-            .getInteger(GameOptions.GIFT_PROBABILITY);
+        final int giftPercent = getSpecification()
+            .getPercentage(GameOptions.GIFT_PROBABILITY);
         int randomIdx = 0;
         lb.mark();
 
-        for (IndianSettlement is : player.getIndianSettlements()) {
+        for (IndianSettlement is : player.getIndianSettlementList()) {
             // Do not bring gifts all the time.
-            if (randoms[randomIdx++] >= giftProbability) continue;
+            if (randoms[randomIdx++] >= giftPercent) continue;
 
             // Check if the settlement has anything to give.
             Goods gift = is.getRandomGift(getAIRandom());
@@ -515,7 +479,7 @@ public class NativeAIPlayer extends AIPlayer {
             // enough missions in operation.
             List<Unit> availableUnits = new ArrayList<>();
             int alreadyAssignedUnits = 0;
-            for (Unit ou : is.getOwnedUnits()) {
+            for (Unit ou : is.getOwnedUnitList()) {
                 AIUnit aiu = getAIUnit(ou);
                 if (aiu == null) {
                     continue;
@@ -542,8 +506,8 @@ public class NativeAIPlayer extends AIPlayer {
                         getAIRandom(), availableUnits.size()));
                 availableUnits.remove(u);
                 aiUnit = getAIUnit(u);
-                if (IndianBringGiftMission.invalidReason(aiUnit) == null
-                    && u.findPath(u.getTile(), home, null, cd) != null) {
+                if (IndianBringGiftMission.invalidMissionReason(aiUnit) == null
+                    && u.findPath(u.getTile(), home, null, cd, null) != null) {
                     unit = u;
                 }
             }
@@ -561,12 +525,12 @@ public class NativeAIPlayer extends AIPlayer {
                 PathNode path;
                 if (c == null
                     || !is.hasContacted(c.getOwner())
-                    || IndianBringGiftMission.invalidReason(aiUnit, c) != null
+                    || IndianBringGiftMission.invalidMissionReason(aiUnit, c) != null
                     || (path = unit.findPath(home, c.getTile(),
-                                             null, cd)) == null) continue;
+                                             null, cd, null)) == null) continue;
                 int alarm = Math.max(1, is.getAlarm(c.getOwner()).getValue());
                 nearbyColonies.add(new RandomChoice<>(c,
-                        1000000 / alarm / path.getTotalTurns()));
+                        1000000 / alarm / (1 + path.getTotalTurns())));
             }
 
             // If there are any suitable colonies, pick a random one
@@ -578,7 +542,7 @@ public class NativeAIPlayer extends AIPlayer {
             Colony target = RandomChoice.getWeightedRandom(logger,
                 "Choose gift colony", nearbyColonies, getAIRandom());
             if (target == null) {
-                throw new IllegalStateException("No gift target!?!");
+                throw new RuntimeException("No gift target: " + this);
             }
 
             // Send the unit.
@@ -593,25 +557,25 @@ public class NativeAIPlayer extends AIPlayer {
      * Demands tribute from nasty players with nearby colonies.
      *
      * @param randoms An array of random percentages.
-     * @param lb A <code>LogBuilder</code> to log to.
+     * @param lb A {@code LogBuilder} to log to.
      */
     private void demandTribute(int[] randoms, LogBuilder lb) {
         final Player player = getPlayer();
         final CostDecider cd = CostDeciders.numberOfLegalTiles();
-        final int demandProbability = getSpecification()
-            .getInteger(GameOptions.DEMAND_PROBABILITY);
+        final int demandPercent = getSpecification()
+            .getPercentage(GameOptions.DEMAND_PROBABILITY);
         int randomIdx = 0;
         lb.mark();
 
-        for (IndianSettlement is : player.getIndianSettlements()) {
+        for (IndianSettlement is : player.getIndianSettlementList()) {
             // Do not demand tribute all of the time.
-            if (randoms[randomIdx++] >= demandProbability) continue;
+            if (randoms[randomIdx++] >= demandPercent) continue;
 
             // Check if there are available units, and if there are already
             // enough missions in operation.
             List<Unit> availableUnits = new ArrayList<>();
             int alreadyAssignedUnits = 0;
-            for (Unit ou : is.getOwnedUnits()) {
+            for (Unit ou : is.getOwnedUnitList()) {
                 AIUnit aiu = getAIUnit(ou);
                 if (Mission.invalidNewMissionReason(aiu) == null) {
                     if (aiu.hasMission(IndianDemandMission.class)) {
@@ -638,8 +602,8 @@ public class NativeAIPlayer extends AIPlayer {
                         getAIRandom(), availableUnits.size()));
                 availableUnits.remove(u);
                 aiUnit = getAIUnit(u);
-                if (IndianDemandMission.invalidReason(aiUnit) == null
-                    && u.findPath(u.getTile(), home, null, cd) != null) {
+                if (IndianDemandMission.invalidMissionReason(aiUnit) == null
+                    && u.findPath(u.getTile(), home, null, cd, null) != null) {
                     unit = u;
                 }
             }
@@ -656,14 +620,14 @@ public class NativeAIPlayer extends AIPlayer {
                 PathNode path;
                 if (c == null
                     || !is.hasContacted(c.getOwner())
-                    || IndianDemandMission.invalidReason(aiUnit, c) != null
+                    || IndianDemandMission.invalidMissionReason(aiUnit, c) != null
                     || (path = unit.findPath(home, c.getTile(),
-                                             null, cd)) == null) continue;
+                                             null, cd, null)) == null) continue;
                 int alarm = is.getAlarm(c.getOwner()).getValue();
                 int defence = c.getUnitCount() + ((c.getStockade() == null) ? 1
                     : (c.getStockade().getLevel() * 10));
                 int weight = 1 + alarm * (1000000 / defence
-                                                  / path.getTotalTurns());
+                    / (1 + path.getTotalTurns()));
                 nearbyColonies.add(new RandomChoice<>(c, weight));
             }
             // If there are any suitable colonies, pick one to demand from.
@@ -694,16 +658,14 @@ public class NativeAIPlayer extends AIPlayer {
      * @param sense The sense to apply the modifiers.
      * @return The ship trade penalties.
      */
-    private Set<Modifier> getShipTradePenalties(boolean sense) {
+    private List<Modifier> getShipTradePenalties(boolean sense) {
         final Specification spec = getSpecification();
-        int penalty = spec.getInteger(GameOptions.SHIP_TRADE_PENALTY);
-        Set<Modifier> result = new HashSet<>();
-        for (Modifier m : spec.getModifiers(Modifier.SHIP_TRADE_PENALTY)) {
-            Modifier n = new Modifier(m);
-            n.setValue((sense) ? penalty : -penalty);
-            result.add(n);
-        }
-        return result;
+        final int penalty = ((sense) ? 1 : -1)
+            * spec.getInteger(GameOptions.SHIP_TRADE_PENALTY);
+        final Function<Modifier, Modifier> mapper
+            = m -> Modifier.makeModifier(m).setValue(penalty);
+        return transform(spec.getModifiers(Modifier.SHIP_TRADE_PENALTY),
+                         alwaysTrue(), mapper);
     }
 
     /**
@@ -722,12 +684,31 @@ public class NativeAIPlayer extends AIPlayer {
 
     // AIPlayer interface
     // Inherit:
-    //   indianDemand
     //   acceptDiplomaticTrade
     //   acceptTax
     //   acceptMercenaries
     //   selectFoundingFather
 
+    /**
+     * {@inheritDoc}
+     */
+    public IndianDemandAction indianDemand(Unit unit, Colony colony,
+                                           GoodsType type, int amount,
+                                           IndianDemandAction accept) {
+        final Player player = getPlayer();
+        AIUnit aiu;
+        IndianDemandMission mission;
+        if (unit.getOwner() == player) { // Its one of ours
+            if ((aiu = getAIUnit(unit)) != null // and its valid and demanding
+                && (mission = aiu.getMission(IndianDemandMission.class)) != null
+                && accept != IndianDemandAction.INDIAN_DEMAND_DONE) {
+                mission.setSucceeded(accept == IndianDemandAction.INDIAN_DEMAND_ACCEPT);
+            }
+        }
+        // Once we get here, the demand is settled
+        return IndianDemandAction.INDIAN_DEMAND_DONE;
+    }
+
     /**
      * {@inheritDoc}
      */
@@ -735,13 +716,12 @@ public class NativeAIPlayer extends AIPlayer {
     public void startWorking() {
         final Player player = getPlayer();
         final Turn turn = getGame().getTurn();
-        final int nSettlements = player.getNumberOfSettlements();
+        final int nSettlements = player.getSettlementCount();
         final Random air = getAIRandom();
 
         LogBuilder lb = new LogBuilder(1024);
         lb.add(player.getDebugName(), " in ", turn, "/", turn.getNumber());
 
-        sessionRegister.clear();
         clearAIUnits();
 
         determineStances(lb);
@@ -803,148 +783,203 @@ public class NativeAIPlayer extends AIPlayer {
     }
 
     /**
-     * {@inheritDoc}
+     * Price the goods to buy.
+     *
+     * @param nt The {@code NativeTrade} to update.
+     * @param anger A penalty to the native prices due to anger.
      */
-    @Override
-    public void registerSellGoods(Goods goods) {
-        String goldKey = "tradeGold#" + goods.getType().getId()
-            + "#" + goods.getAmount() + "#" + goods.getLocation().getId();
-        sessionRegister.put(goldKey, null);
+    private void updateTrade(NativeTrade nt, int anger) {
+        final Specification spec = getSpecification();
+        final Turn turn = getGame().getTurn();
+        final IndianSettlement is = nt.getIndianSettlement();
+        final Unit unit = nt.getUnit();
+        Set<Modifier> modifiers = new HashSet<>();
+
+        if (is.hasMissionary(unit.getOwner())
+            && spec.getBoolean(GameOptions.ENHANCED_MISSIONARIES)) {
+            Unit u = is.getMissionary();
+            modifiers.addAll(u.getMissionaryTradeModifiers(true));
+        }
+        if (unit.isNaval()) {
+            modifiers.addAll(getShipTradePenalties(true));
+        }
+        for (NativeTradeItem nti : nt.getUnitToSettlement()) {
+            int price = (int)FeatureContainer.applyModifiers(1.0f / anger
+                * is.getPriceToBuy(nti.getGoods()), turn, modifiers);
+            for (int h = nti.getHaggleCount(); h >= 0; h--) {
+                price = NativeTrade.haggleUp(price);
+            }
+            if (price <= NativeTradeItem.PRICE_UNSET) {
+                price = NativeTradeItem.PRICE_INVALID;
+            }
+            nti.setPrice(price);
+        }
+
+        modifiers.clear();
+        if (is.hasMissionary(unit.getOwner())
+            && spec.getBoolean(GameOptions.ENHANCED_MISSIONARIES)) {
+            Unit u = is.getMissionary();
+            modifiers.addAll(u.getMissionaryTradeModifiers(false));
+        }
+        if (unit.isNaval()) {
+            modifiers.addAll(getShipTradePenalties(false));
+        }
+        for (NativeTradeItem nti : nt.getSettlementToUnit()) {
+            int price = (int)FeatureContainer.applyModifiers((float)anger
+                * is.getPriceToSell(nti.getGoods()), turn, modifiers);
+            for (int h = nti.getHaggleCount(); h >= 0; h--) {
+                price = NativeTrade.haggleDown(price);
+            }
+            if (price <= NativeTradeItem.PRICE_UNSET) {
+                price = NativeTradeItem.PRICE_INVALID;
+            }
+            nti.setPrice(price);
+        }
     }
 
     /**
      * {@inheritDoc}
      */
     @Override
-    public int buyProposition(Unit unit, Settlement settlement,
-                              Goods goods, int gold) {
-        logger.finest("Entering method buyProposition");
-        Specification spec = getSpecification();
-        IndianSettlement is = (IndianSettlement) settlement;
-        Player buyer = unit.getOwner();
-        String goldKey = "tradeGold#" + goods.getType().getId()
-            + "#" + goods.getAmount() + "#" + settlement.getId();
-        String hagglingKey = "tradeHaggling#" + unit.getId();
-        int price;
-        Integer registered = sessionRegister.get(goldKey);
-        if (registered == null) {
-            price = is.getPriceToSell(goods);
-            switch (is.getAlarm(buyer).getLevel()) {
+    public NativeTradeAction handleTrade(NativeTradeAction action,
+                                         NativeTrade nt) {
+        final int HAGGLE_NUMBER = 3;
+        if (nt == null || !this.getPlayer().owns(nt.getIndianSettlement())) {
+            return NativeTradeAction.NAK_INVALID;
+        }
+        final IndianSettlement is = nt.getIndianSettlement();
+        final Unit unit = nt.getUnit();
+        final Player other = unit.getOwner();
+        NativeTradeItem ours;
+        int anger, haggle;
+        Tension tension = is.getAlarm(other);
+        if (tension == null) return NativeTradeAction.NAK_INVALID;
+
+        switch (action) {
+        case OPEN:
+            switch (tension.getLevel()) {
             case HAPPY: case CONTENT:
+                anger = 1;
                 break;
             case DISPLEASED:
-                price *= 2;
+                anger = 2;
+                break;
+            case ANGRY: default:
+                anger = -1;
                 break;
-            default:
-                return NetworkConstants.NO_TRADE_HOSTILE;
             }
-            Set<Modifier> modifiers = new HashSet<>();
-            if (is.hasMissionary(buyer)
-                && spec.getBoolean(GameOptions.ENHANCED_MISSIONARIES)) {
-                Unit u = is.getMissionary();
-                modifiers.addAll(u.getMissionaryTradeModifiers(false));
-            }
-            if (unit.isNaval()) {
-                modifiers.addAll(getShipTradePenalties(false));
-            }
-            price = (int)FeatureContainer.applyModifiers((float)price,
-                getGame().getTurn(), modifiers);
-            sessionRegister.put(goldKey, price);
-            return price;
-        }
-        price = registered;
-        if (price < 0 || price == gold) return price;
-        if (gold < (price * 9) / 10) {
-            logger.warning("Cheating attempt: sending offer too low");
-            sessionRegister.put(goldKey, -1);
-            return NetworkConstants.NO_TRADE;
-        }
+            if (anger < 0) return NativeTradeAction.NAK_HOSTILE;
+            updateTrade(nt, anger);
+            return (nt.getDone()) ? NativeTradeAction.NAK_NOSALE
+                : NativeTradeAction.ACK_OPEN;
 
-        int haggling = 1;
-        if (sessionRegister.containsKey(hagglingKey)) {
-            haggling = sessionRegister.get(hagglingKey);
-        }
-        if (randomInt(logger, "Haggle-buy", getAIRandom(), 3 + haggling) >= 3) {
-            sessionRegister.put(goldKey, -1);
-            return NetworkConstants.NO_TRADE_HAGGLE;
-        }
-        sessionRegister.put(goldKey, gold);
-        sessionRegister.put(hagglingKey, haggling + 1);
-        return gold;
-    }
-
-    /**
-     * {@inheritDoc}
-     */
-    @Override
-    public int sellProposition(Unit unit, Settlement settlement,
-                               Goods goods, int gold) {
-        logger.finest("Entering method sellProposition");
-        Specification spec = getSpecification();
-        IndianSettlement is = (IndianSettlement) settlement;
-        Player seller = unit.getOwner();
-        String goldKey = "tradeGold#" + goods.getType().getId()
-            + "#" + goods.getAmount() + "#" + unit.getId()
-            + "#" + settlement.getId();
-        String hagglingKey = "tradeHaggling#" + unit.getId();
-        int price;
-        if (sessionRegister.containsKey(goldKey)) {
-            price = sessionRegister.get(goldKey);
-        } else {
-            price = is.getPriceToBuy(goods);
-            switch (is.getAlarm(seller).getLevel()) {
+        case BUY:
+            switch (tension.getLevel()) {
             case HAPPY: case CONTENT:
+                anger = 1;
                 break;
             case DISPLEASED:
-                price /= 2;
+                anger = 2;
+                break;
+            case ANGRY: default:
+                anger = -1;
+                break;
+            }
+            if (anger < 0) return NativeTradeAction.NAK_HOSTILE;
+            updateTrade(nt, anger);
+            ours = find(nt.getSettlementToUnit(), nt.getItem().goodsMatcher());
+            if (ours == null) return NativeTradeAction.NAK_INVALID;
+            if (nt.getItem().priceIsSet()
+                && nt.getItem().getPrice() >= ours.getPrice()) {
+                return NativeTradeAction.ACK_BUY;
+            }
+            haggle = ours.getHaggleCount() + 1;
+            if (randomInt(logger, "Haggle-buy", getAIRandom(),
+                          HAGGLE_NUMBER + haggle) >= HAGGLE_NUMBER) {
+                return NativeTradeAction.NAK_HAGGLE;
+            }
+            ours.setHaggleCount(haggle);
+            updateTrade(nt, anger);
+            nt.setItem(ours);
+            return NativeTradeAction.ACK_BUY_HAGGLE;
+
+        case SELL:
+            switch (tension.getLevel()) {
+            case HAPPY: case CONTENT:
+                anger = 1;
+                break;
+            case DISPLEASED:
+                anger = 2;
                 break;
             case ANGRY:
-                if (!goods.getType().isMilitaryGoods())
-                    return NetworkConstants.NO_TRADE_HOSTILE;
-                price /= 2;
+                anger = (any(nt.getUnitToSettlement(),
+                             nti -> nti.getGoods().getType().getMilitary()))
+                    ? 3 : -1;
                 break;
             default:
-                return NetworkConstants.NO_TRADE_HOSTILE;
+                anger = -1;
+                break;
             }
-            Set<Modifier> modifiers = new HashSet<>();
-            if (is.hasMissionary(seller)
-                && spec.getBoolean(GameOptions.ENHANCED_MISSIONARIES)) {
-                Unit u = is.getMissionary();
-                modifiers.addAll(u.getMissionaryTradeModifiers(true));
+            if (anger < 0) return NativeTradeAction.NAK_HOSTILE;
+            updateTrade(nt, anger);
+            ours = find(nt.getUnitToSettlement(), nt.getItem().goodsMatcher());
+            if (ours == null) return NativeTradeAction.NAK_INVALID;
+            if (nt.getItem().priceIsSet()
+                && nt.getItem().getPrice() <= ours.getPrice()) {
+                return NativeTradeAction.ACK_SELL;
             }
-            if (unit.isNaval()) {
-                modifiers.addAll(getShipTradePenalties(true));
+            haggle = ours.getHaggleCount() + 1;
+            if (randomInt(logger, "Haggle-sell", getAIRandom(),
+                          HAGGLE_NUMBER + haggle) >= HAGGLE_NUMBER) {
+                return NativeTradeAction.NAK_HAGGLE;
             }
-            price = (int)FeatureContainer.applyModifiers((float)price,
-                getGame().getTurn(), modifiers);
-            if (price <= 0) return 0;
-            sessionRegister.put(goldKey, price);
+            ours.setHaggleCount(haggle);
+            updateTrade(nt, anger);
+            nt.setItem(ours);
+            return NativeTradeAction.ACK_SELL_HAGGLE;
+
+        case GIFT:
+            ours = find(nt.getUnitToSettlement(), nt.getItem().goodsMatcher());
+            if (ours == null) return NativeTradeAction.NAK_INVALID;
+            return (is.canAdd(nt.getItem().getGoods()))
+                ? NativeTradeAction.ACK_GIFT
+                : NativeTradeAction.NAK_GOODS;
+
+        default: // Invalid
+            return NativeTradeAction.NAK_INVALID;
         }
-        if (gold < 0 || price == gold) return price;
-        if (gold > (price * 11) / 10) {
-            logger.warning("Cheating attempt: haggling request too high");
-            sessionRegister.put(goldKey, -1);
-            return NetworkConstants.NO_TRADE;
-        }
-        int haggling = 1;
-        if (sessionRegister.containsKey(hagglingKey)) {
-            haggling = sessionRegister.get(hagglingKey);
-        }
-        if (randomInt(logger, "Haggle-sell", getAIRandom(), 3 + haggling) >= 3) {
-            sessionRegister.put(goldKey, -1);
-            return NetworkConstants.NO_TRADE_HAGGLE;
-        }
-        sessionRegister.put(goldKey, gold);
-        sessionRegister.put(hagglingKey, haggling + 1);
-        return gold;
     }
 
+    /**
+     * {@inheritDoc}
+     */
+    public int getNeededWagons(Tile tile) {
+        throw new RuntimeException("Can not happen: " + this);
+    }
+
+    /**
+     * {@inheritDoc}
+     */
+    public int pioneersNeeded() {
+        throw new RuntimeException("Can not happen: " + this);
+    }
+
+    /**
+     * {@inheritDoc}
+     */
+    public int scoutsNeeded() {
+        throw new RuntimeException("Can not happen: " + this);
+    }
+
+    /**
+     * {@inheritDoc}
+     */
+    public void completeWish(Wish w) {
+        throw new RuntimeException("Can not happen: " + this);
+    }
+    
 
     // Serialization
 
-    /**
-     * {@inheritDoc}
-     */
-    @Override
-    public String getXMLTagName() { return getXMLElementTagName(); }
+    // getXMLTagName not needed, uses parent
 }
