@@ -1,5 +1,5 @@
 /**
- *  Copyright (C) 2002-2015   The FreeCol Team
+ *  Copyright (C) 2002-2022   The FreeCol Team
  *
  *  This file is part of FreeCol.
  *
@@ -20,14 +20,19 @@
 package net.sf.freecol.server.ai;
 
 import java.util.ArrayList;
-import java.util.Collections;
 import java.util.Comparator;
 import java.util.HashMap;
+import java.util.HashSet;
 import java.util.Iterator;
 import java.util.List;
 import java.util.Map;
 import java.util.Map.Entry;
 import java.util.Random;
+import java.util.Set;
+import java.util.function.Function;
+import java.util.function.Predicate;
+import java.util.function.ToDoubleFunction;
+import java.util.function.ToIntFunction;
 import java.util.logging.Level;
 import java.util.logging.Logger;
 
@@ -35,7 +40,6 @@ import javax.xml.stream.XMLStreamException;
 
 import net.sf.freecol.common.io.FreeColXMLReader;
 import net.sf.freecol.common.model.Ability;
-import net.sf.freecol.common.model.AbstractGoods;
 import net.sf.freecol.common.model.Building;
 import net.sf.freecol.common.model.Colony;
 import net.sf.freecol.common.model.Location;
@@ -48,8 +52,9 @@ import net.sf.freecol.common.model.Unit;
 import net.sf.freecol.common.model.pathfinding.CostDeciders;
 import net.sf.freecol.common.model.pathfinding.GoalDecider;
 import net.sf.freecol.common.model.pathfinding.GoalDeciders;
-import net.sf.freecol.common.util.LogBuilder;
+import net.sf.freecol.common.util.CachingFunction;
 import static net.sf.freecol.common.util.CollectionUtils.*;
+import net.sf.freecol.common.util.LogBuilder;
 import static net.sf.freecol.common.util.RandomUtils.*;
 import net.sf.freecol.common.util.Utils;
 import net.sf.freecol.server.ai.mission.DefendSettlementMission;
@@ -57,7 +62,6 @@ import net.sf.freecol.server.ai.mission.Mission;
 import net.sf.freecol.server.ai.mission.TransportMission;
 import net.sf.freecol.server.ai.mission.PrivateerMission;
 import net.sf.freecol.server.ai.mission.UnitSeekAndDestroyMission;
-import net.sf.freecol.server.model.ServerPlayer;
 
 
 /**
@@ -65,13 +69,66 @@ import net.sf.freecol.server.model.ServerPlayer;
  *
  * For now, mostly just the EuropeanAIPlayer, with a few tweaks.
  */
-public class REFAIPlayer extends EuropeanAIPlayer {
+public final class REFAIPlayer extends EuropeanAIPlayer {
 
     private static final Logger logger = Logger.getLogger(REFAIPlayer.class.getName());
 
     /** Limit on the number of REF units chasing a single hostile unit. */
     private static final int UNIT_USAD_THRESHOLD = 5;
 
+    /** Goal decider class for the REF Navy */
+    private static class REFNavyGoalDecider implements GoalDecider {
+
+        /** The rebel player to attack. */
+        private Player rebel;
+
+        /** A container to fill with target units. */
+        private Set<Unit> rebelNavy;
+
+
+        /**
+         * Build a new goal decider for the REF to find rebel naval units.
+         *
+         * @param rebel The {@code Player} to attack.
+         * @param rebelNavy A container to fill with the naval units found.
+         */
+        public REFNavyGoalDecider(Player rebel, Set<Unit> rebelNavy) {
+            this.rebel = rebel;
+            this.rebelNavy = rebelNavy;
+        }
+        
+        /**
+         * {@inheritDoc}
+         */
+        public PathNode getGoal() {
+            return null;
+        }
+
+        /**
+         * {@inheritDoc}
+         */
+        public boolean hasSubGoals() {
+            return true;
+        }
+
+        /**
+         * {@inheritDoc}
+         */
+        public boolean check(Unit unit, PathNode pathNode) {
+            final Predicate<Unit> pred
+                = (u -> u.isOffensiveUnit() && u.isNaval()
+                    && !this.rebelNavy.contains(u));
+            Tile tile = pathNode.getTile();
+            if (tile != null && !tile.isEmpty()
+                && !tile.isLand()
+                && this.rebel.owns(tile.getFirstUnit())) {
+                this.rebelNavy.addAll(transform(tile.getUnits(), pred));
+                return true;
+            }
+            return false;
+        }
+    };
+
     /** Container class for REF target colony information. */
     private static class TargetTuple implements Comparable<TargetTuple> {
 
@@ -103,21 +160,23 @@ public class REFAIPlayer extends EuropeanAIPlayer {
 
         @Override
         public int compareTo(TargetTuple other) {
-            double cmp = other.score - score;
-            return (cmp < 0.0) ? -1 : (cmp > 0.0) ? 1 : 0;
+            return Double.compare(score, other.score);
         }
 
+
         // Override Object
 
         /**
          * {@inheritDoc}
          */
         @Override
-        public boolean equals(Object other) {
-            if (other instanceof TargetTuple) {
-                return this.compareTo((TargetTuple)other) == 0;
+        public boolean equals(Object o) {
+            if (o instanceof TargetTuple) {
+                TargetTuple other = (TargetTuple)o;
+                return this.compareTo(other) == 0
+                    && super.equals(other);
             }
-            return super.equals(other);
+            return false;
         }
 
         /**
@@ -139,58 +198,66 @@ public class REFAIPlayer extends EuropeanAIPlayer {
     /** Map of target to count. */
     private final Map<Location, Integer> targetMap = new HashMap<>();
 
+    /** Has the army been landed? */
+    private boolean landed = false;
+
 
     /**
-     * Creates a new <code>REFAIPlayer</code>.
+     * Creates a new {@code REFAIPlayer}.
      *
      * @param aiMain The main AI-class.
      * @param player The player that should be associated with this
-     *            <code>REFAIPlayer</code>.
+     *            {@code REFAIPlayer}.
      */
-    public REFAIPlayer(AIMain aiMain, ServerPlayer player) {
+    public REFAIPlayer(AIMain aiMain, Player player) {
         super(aiMain, player);
 
-        uninitialized = getPlayer() == null;
+        this.initialized = getPlayer() != null;
     }
 
     /**
-     * Creates a new <code>REFAIPlayer</code>.
+     * Creates a new {@code REFAIPlayer}.
      *
      * @param aiMain The main AI-object.
      * @param xr The input stream containing the XML.
      * @throws XMLStreamException if a problem was encountered during parsing.
      */
-    public REFAIPlayer(AIMain aiMain,
-                       FreeColXMLReader xr) throws XMLStreamException {
+    public REFAIPlayer(AIMain aiMain, FreeColXMLReader xr)
+        throws XMLStreamException {
         super(aiMain, xr);
 
-        uninitialized = getPlayer() == null;
+        this.initialized = getPlayer() != null;
     }
 
 
     /**
      * Find suitable colony targets.
      *
-     * @param aiu The <code>AIUnit</code> to search with.
+     * @param aiu The {@code AIUnit} to search with.
      * @param port If true, insist on the colonies being ports.
-     * @param aiCarrier The <code>AIUnit</code> to use as a carrier.
-     * @return A list of <code>TargetTuple</code> target choices.
+     * @param aiCarrier The {@code AIUnit} to use as a carrier.
+     * @return A list of {@code TargetTuple} target choices.
      */
     private List<TargetTuple> findColonyTargets(AIUnit aiu, boolean port,
                                                 AIUnit aiCarrier) {
         final Player player = getPlayer();
         final Unit unit = aiu.getUnit();
         final Unit carrier = aiCarrier.getUnit();
-        final List<TargetTuple> targets = new ArrayList<>();
-        for (Player p : player.getRebels()) {
-            for (Colony c : p.getColonies()) {
-                if (port && !c.isConnectedPort()) continue;
-                PathNode path = unit.findPath(carrier, c, carrier, null);
-                if (path == null) continue;
-                int score = UnitSeekAndDestroyMission.scorePath(aiu, path);
-                targets.add(new TargetTuple(c, path, score));
-            }
-        }
+        final CachingFunction<Colony, PathNode> pathMapper
+            = new CachingFunction<Colony, PathNode>(c ->
+                unit.findPath(carrier, c, carrier));
+        final Predicate<Colony> pathPred = c ->
+            pathMapper.apply(c) != null;
+        final Function<Colony, TargetTuple> newTupleMapper = c -> {
+            PathNode path = pathMapper.apply(c);
+            return new TargetTuple(c, path,
+                UnitSeekAndDestroyMission.scorePath(aiu, path));
+        };
+        final List<TargetTuple> targets
+            = transform(flatten(player.getRebels(),
+                    ((port) ? Player::getConnectedPorts :Player::getColonies)),
+                pathPred, newTupleMapper);
+        if (targets.isEmpty()) return targets; // Should not happen
 
         // Increase score for drydock/s, musket and tools suppliers,
         // but decrease for fortifications.
@@ -201,34 +268,37 @@ public class REFAIPlayer extends EuropeanAIPlayer {
                                    targets.size());
         int twidx = 0;
         for (TargetTuple t : targets) {
-            t.score *= 0.01 * (101 - Math.min(100, t.colony.getSoL()));
-            for (Building b : t.colony.getBuildings()) {
-                if (b.getLevel() > 1) {
-                    if (b.hasAbility(Ability.REPAIR_UNITS)) t.score *= 1.5;
-                    for (AbstractGoods ag : b.getOutputs()) {
-                        if (ag.getType().isMilitaryGoods()) {
-                            t.score *= 2.0;
-                        } else if (ag.getType().isBuildingMaterial()
-                            && ag.getType().isRefined()) {
-                            t.score *= 1.5;
-                        }
-                    }
-                }
-            }
-            int stockade = (!t.colony.hasStockade()) ? 0
-                : t.colony.getStockade().getLevel();
-            t.score *= (6 - stockade) / 6.0;
-            t.score *= 1.0 + 0.01 * (twiddle[twidx++] - percentTwiddle);
+            final ToDoubleFunction<Building> bdf = b ->
+                ((b.hasAbility(Ability.REPAIR_UNITS)) ? 1.5 : 1.0)
+                * product(b.getOutputs(), ag ->
+                    (ag.getType().getMilitary()) ? 2.0
+                    : (ag.getType().isBuildingMaterial()
+                        && ag.getType().isRefined()) ? 1.5
+                    : 1.0);
+            t.score *= 0.01 * (101 - Math.min(100, t.colony.getSoL()))
+                * product(t.colony.getBuildings(), b -> b.getLevel() > 1, bdf)
+                * ((6 - ((!t.colony.hasStockade()) ? 0
+                            : t.colony.getStockade().getLevel())) / 6.0)
+                * (1.0 + 0.01 * (twiddle[twidx++] - percentTwiddle));
         }
-        Collections.sort(targets);
-
-        LogBuilder lb = new LogBuilder(64);
-        lb.add("REF found colony targets:");
-        for (TargetTuple t : targets) lb.add(" ", t.colony, "(", t.score, ")");
-        lb.log(logger, Level.FINE);
+        targets.sort(Comparator.naturalOrder());
         return targets;
     }
 
+    /**
+     * Find suitable colony targets.
+     *
+     * @param aiu The {@code AIUnit} to search with.
+     * @param aiCarrier The {@code AIUnit} to use as a carrier.
+     * @return A list of {@code TargetTuple} target choices.
+     */
+    private List<TargetTuple> findColonyTargets(AIUnit aiu, AIUnit aiCarrier) {
+        List<TargetTuple> targets = findColonyTargets(aiu, true, aiCarrier);
+        if (targets.isEmpty())
+            targets = findColonyTargets(aiu, false, aiCarrier);
+        return targets;
+    }
+                
     /**
      * Initialize the REF.
      * - Find the initial target
@@ -276,11 +346,18 @@ public class REFAIPlayer extends EuropeanAIPlayer {
             return false;
         }
 
-        List<TargetTuple> targets = findColonyTargets(aiUnit, true, aiCarrier);
+        List<TargetTuple> targets = findColonyTargets(aiUnit, aiCarrier);
         if (targets.isEmpty()) {
             logger.warning("REF found no targets.");
             return false;
-        }
+        } else {
+            LogBuilder lb = new LogBuilder(64);
+            lb.add("REF found colony targets:");
+            for (TargetTuple t : targets) {
+                lb.add(" ", t.colony, "(", t.score, ")");
+            }
+            lb.log(logger, Level.FINE);
+        }            
 
         final Player rebel = targets.get(0).colony.getOwner();
         double ratio = getStrengthRatio(rebel);
@@ -293,8 +370,7 @@ public class REFAIPlayer extends EuropeanAIPlayer {
         // appear at this location, otherwise at the best entry
         // location for it.
         int fail = 0;
-        for (int i = 0; i < n; i++) {
-            final TargetTuple t = targets.get(i);
+        for (TargetTuple t : targets) {
             final GoalDecider gd = GoalDeciders
                 .getDisembarkGoalDecider(t.colony.getTile());
             PathNode path = unit.search(t.entry, gd, null, 10, carrier);
@@ -324,8 +400,7 @@ public class REFAIPlayer extends EuropeanAIPlayer {
                 }
                 lb.add(")");
             } else { // They were all bad, just use the existing simple path
-                for (int i = 0; i < n; i++) {
-                    final TargetTuple t = targets.get(i);
+                for (TargetTuple t : targets) {
                     t.disembarkTile = t.path.getTransportDropNode()
                         .previous.getTile();
                 }
@@ -355,7 +430,7 @@ public class REFAIPlayer extends EuropeanAIPlayer {
                     t.score *= 1.5; // Prefer invisible paths
                 }
             }
-            Collections.sort(targets); // Re-sort with new scores
+            targets.sort(Comparator.naturalOrder()); // Re-sort with new scores
         }
 
         // Give the land units seek-and-destroy missions for the
@@ -406,38 +481,21 @@ public class REFAIPlayer extends EuropeanAIPlayer {
 
         // Try to find some rebel naval units near the entry locations
         // for the targets.
-        final List<Unit> rebelNavy = new ArrayList<>();
-        final GoalDecider navyGD = new GoalDecider() {
-                @Override
-                public PathNode getGoal() { return null; }
-                @Override
-                public boolean hasSubGoals() { return true; }
-                @Override
-                public boolean check(Unit unit, PathNode pathNode) {
-                    Tile tile = pathNode.getTile();
-                    if (tile != null && !tile.isEmpty()
-                        && !tile.isLand()
-                        && rebel.owns(tile.getFirstUnit())) {
-                        for (Unit u : tile.getUnitList()) {
-                            if (u.isOffensiveUnit() && u.isNaval()
-                                && !rebelNavy.contains(u)) rebelNavy.add(u);
-                        }
-                    }
-                    return false;
-                }
-            };
+        final Set<Unit> rebelNavy = new HashSet<>();
+        final GoalDecider navyGD = new REFNavyGoalDecider(rebel, rebelNavy);
         for (int i = 0; i < n; i++) {
             carrier.search(targets.get(i).entry, navyGD, null,
                            carrier.getInitialMovesLeft() * 2, null);
         }
 
         // Attack naval targets.
-        final Comparator<Unit> militaryStrength
+        final Comparator<Unit> militaryStrengthComparator
             = getGame().getCombatModel().getMilitaryStrengthComparator();
-        Collections.sort(rebelNavy, militaryStrength);
-        Iterator<Unit> ui = rebelNavy.iterator();
+        List<Unit> rebelTargets = new ArrayList<>(rebelNavy);
+        rebelTargets.sort(militaryStrengthComparator);
+        Iterator<Unit> ui = rebelTargets.iterator();
         List<Tile> entries = new ArrayList<>();
-        entries.add(rebel.getEntryLocation().getTile());
+        entries.add(rebel.getEntryTile());
         while (!navy.isEmpty()) {
             final AIUnit aiu = navy.remove(0);
             final Unit u = aiu.getUnit();
@@ -469,12 +527,12 @@ public class REFAIPlayer extends EuropeanAIPlayer {
      * missions.
      *
      * @param nt The number of transport missions required.
-     * @param transports The list of <code>AIUnit</code>s with a transport
+     * @param transports The list of {@code AIUnit}s with a transport
      *     mission.
-     * @param privateers The list of <code>AIUnit</code>s with a privateer
+     * @param privateers The list of {@code AIUnit}s with a privateer
      *     mission.
-     * @param lb A <code>LogBuilder</code> to log to.
-     * @return A list of new <code>AIUnit</code>s with transport missions.
+     * @param lb A {@code LogBuilder} to log to.
+     * @return A list of new {@code AIUnit}s with transport missions.
      */
     private List<AIUnit> requireTransports(int nt, List<AIUnit> transports,
                                            List<AIUnit> privateers,
@@ -497,20 +555,15 @@ public class REFAIPlayer extends EuropeanAIPlayer {
         }
         if (transports.size() < nt) {
             // Sort by longest distance to target
-            Collections.sort(naval, new Comparator<AIUnit>() {
-                    @Override
-                    public int compare(AIUnit a1, AIUnit a2) {
-                        int d1 = a1.getMission(PrivateerMission.class)
-                            .getDistanceToTarget();
-                        int d2 = a2.getMission(PrivateerMission.class)
-                            .getDistanceToTarget();
-                        return d1 - d2;
-                    }
-                });
-            while (!naval.isEmpty()) {
-                AIUnit aiu = naval.remove(0);
-                int distance = aiu.getMission(PrivateerMission.class)
-                    .getDistanceToTarget();
+            final ToIntFunction<AIUnit> targetDistance
+                = cacheInt(aiu -> {
+                        Unit target = (Unit)aiu.getMission().getTarget();
+                        Tile tile = aiu.getUnit().getTile();
+                        return tile.getDistanceTo(target.getTile());
+                    });
+            for (AIUnit aiu : sort(naval,
+                                   Comparator.comparingInt(targetDistance))) {
+                int distance = targetDistance.applyAsInt(aiu);
                 if ((m = getTransportMission(aiu)) != null) {
                     lb.add(" REQUIRED ", distance, " ", m);
                     result.add(aiu);
@@ -622,11 +675,11 @@ public class REFAIPlayer extends EuropeanAIPlayer {
         // - defend the closest port
         // - go idle in a port
         for (AIUnit aiu : land) {
-            Location target = UnitSeekAndDestroyMission.findTarget(aiu, 
+            Location target = UnitSeekAndDestroyMission.findMissionTarget(aiu, 
                 seekAndDestroyRange, false);
             if (target != null) {
-                int count = (targetMap.containsKey(target))
-                    ? targetMap.get(target) : 0;
+                Integer count = targetMap.get(target);
+                if (count == null) count = 0;
                 if (target instanceof Unit
                     && count < UNIT_USAD_THRESHOLD
                     && (m = getSeekAndDestroyMission(aiu, target)) != null) {
@@ -655,8 +708,8 @@ public class REFAIPlayer extends EuropeanAIPlayer {
             }
 
             // Go defend the nearest colony needing defence
-            Colony best = u.getClosestColony(getBadlyDefended().stream()
-                .map(AIColony::getColony));
+            Colony best = u.getClosestColony(map(getBadlyDefended(),
+                                                 AIColony::getColony));
             if (best != null
                 && (m = getDefendSettlementMission(aiu, best)) != null) {
                 lb.add(" GO-DEFEND-", best.getName(), " " , m);
@@ -738,10 +791,10 @@ public class REFAIPlayer extends EuropeanAIPlayer {
                 AIUnit found = null;
                 Colony target = null;
                 for (AIUnit aiCarrier : aiCarriers) {
-                    for (Unit u : aiCarrier.getUnit().getUnitList()) {
-                        if (u.hasAbility(Ability.REF_UNIT)
-                            && (found = getAIUnit(u)) != null) break;
-                    }
+                    found = first(transform(aiCarrier.getUnit().getUnits(),
+                                            u -> u.hasAbility(Ability.REF_UNIT),
+                                            u -> getAIUnit(u),
+                                            toListNoNulls()));
                     if (found != null
                         && (m = found.getMission()) != null
                         && m.isValid()
@@ -753,14 +806,8 @@ public class REFAIPlayer extends EuropeanAIPlayer {
                 }
                 if (target == null) {
                     AIUnit aiCarrier = getAIUnit(found.getUnit().getCarrier());
-                    List<TargetTuple> ct = findColonyTargets(found, true,
-                                                             aiCarrier);
-                    if (ct.isEmpty()) {
-                        ct = findColonyTargets(found, false, aiCarrier);
-                    }
-                    if (!ct.isEmpty()) {
-                        target = ct.get(0).colony;
-                    }
+                    List<TargetTuple> ct = findColonyTargets(found, aiCarrier);
+                    if (!ct.isEmpty()) target = ct.get(0).colony;
                 }
                 if (target == null) continue; // No target for these idlers
 
@@ -783,8 +830,8 @@ public class REFAIPlayer extends EuropeanAIPlayer {
                 // Are there more idle units waiting here?
                 if (!e.getValue().isEmpty()) {
                     idlePorts.add(key);
-                    space += e.getValue().stream()
-                        .mapToInt(aiu -> aiu.getUnit().getSpaceTaken()).sum();
+                    space += sum(e.getValue(),
+                                 aiu -> aiu.getUnit().getSpaceTaken());
                 }
             }
 
@@ -803,12 +850,10 @@ public class REFAIPlayer extends EuropeanAIPlayer {
 
                 // Send transports to the idle ports, preferring the ones
                 // with the most units.
-                Collections.sort(idlePorts, new Comparator<Location>() {
-                        @Override
-                        public int compare(Location l1, Location l2) {
-                            return idlers.get(l1).size() - idlers.get(l2).size();
-                        }
-                    });
+                final Comparator<Location> portUnitComparator
+                    = Comparator.comparing(loc -> idlers.get(loc),
+                                           ascendingListLengthComparator);
+                idlePorts.sort(portUnitComparator);
                 boolean bad = false;
                 while (!bad && !todo.isEmpty()) {
                     for (Location l : idlePorts) {
@@ -865,7 +910,11 @@ public class REFAIPlayer extends EuropeanAIPlayer {
                     transport.add(aiu.getMission(TransportMission.class));
                 }
             } else {
-                if (u.isInEurope()) land.add(aiu);
+                if (u.isInEurope()) {
+                    land.add(aiu);
+                } else if (u.isOnTile()) {
+                    this.landed = true;
+                }
             }
         }
         if (!land.isEmpty() && !transport.isEmpty()) {
@@ -897,22 +946,23 @@ public class REFAIPlayer extends EuropeanAIPlayer {
                     if (((Settlement)target).isConnectedPort()) {
                         value += 500;
                     } else {
-                        if (getPlayer().getNumberOfSettlements() <= 0) {
+                        if (!getPlayer().hasSettlements()) {
                             return Integer.MIN_VALUE;
                         }
                     }
                 } else if (target instanceof Unit) {
                     // Do not chase units until at least one colony is captured.
-                    if (getPlayer().getNumberOfSettlements() <= 0) {
+                    if (!getPlayer().hasSettlements()) {
                         return Integer.MIN_VALUE;
                     }
                     // Do not chase the same unit!
-                    if (any(getAIUnits().stream()
-                            .filter(aiu -> aiu != aiUnit)
-                            .map(aiu -> aiu.getMission(UnitSeekAndDestroyMission.class)),
-                            m -> m != null
-                                && m.getTarget() instanceof Unit
-                                && (Unit)m.getTarget() == target))
+                    final Predicate<UnitSeekAndDestroyMission> unitPred = m ->
+                        m != null && m.getTarget() instanceof Unit
+                            && (Unit)m.getTarget() == target;
+                    final Function<AIUnit, UnitSeekAndDestroyMission> missionMapper = aiu ->
+                        aiu.getMission(UnitSeekAndDestroyMission.class);
+                    if (any(transform(getAIUnits(), aiu -> aiu != aiUnit,
+                                      missionMapper), unitPred))
                         return Integer.MIN_VALUE;
                     // The REF is more interested in colonies.
                     value /= 2;
@@ -921,4 +971,9 @@ public class REFAIPlayer extends EuropeanAIPlayer {
         }
         return value;
     }
+
+
+    // Serialization
+    
+    // getXMLTagName not needed, uses parent
 }
