@@ -1,5 +1,5 @@
 /**
- *  Copyright (C) 2002-2015   The FreeCol Team
+ *  Copyright (C) 2002-2022   The FreeCol Team
  *
  *  This file is part of FreeCol.
  *
@@ -19,6 +19,8 @@
 
 package net.sf.freecol.server.ai.mission;
 
+import java.util.Comparator;
+import java.util.function.Predicate;
 import java.util.logging.Level;
 import java.util.logging.Logger;
 
@@ -27,6 +29,7 @@ import javax.xml.stream.XMLStreamException;
 import net.sf.freecol.common.io.FreeColXMLReader;
 import net.sf.freecol.common.io.FreeColXMLWriter;
 import net.sf.freecol.common.model.Colony;
+import net.sf.freecol.common.model.Constants;
 import net.sf.freecol.common.model.Location;
 import net.sf.freecol.common.model.PathNode;
 import net.sf.freecol.common.model.Player;
@@ -36,7 +39,7 @@ import net.sf.freecol.common.model.pathfinding.CostDecider;
 import net.sf.freecol.common.model.pathfinding.CostDeciders;
 import net.sf.freecol.common.model.pathfinding.GoalDecider;
 import net.sf.freecol.common.model.pathfinding.GoalDeciders;
-import net.sf.freecol.common.networking.NetworkConstants;
+import static net.sf.freecol.common.util.CollectionUtils.*;
 import net.sf.freecol.common.util.LogBuilder;
 import static net.sf.freecol.common.util.RandomUtils.*;
 import net.sf.freecol.server.ai.AIColony;
@@ -47,14 +50,16 @@ import net.sf.freecol.server.ai.EuropeanAIPlayer;
 
 
 /**
- * Mission for building a <code>Colony</code>.
+ * Mission for building a {@code Colony}.
  *
  * @see net.sf.freecol.common.model.Colony Colony
  */
-public class BuildColonyMission extends Mission {
+public final class BuildColonyMission extends Mission {
 
     private static final Logger logger = Logger.getLogger(BuildColonyMission.class.getName());
 
+    public static final String TAG = "buildColonyMission";
+
     /** The tag for this mission. */
     private static final String tag = "AI colony builder";
 
@@ -65,27 +70,28 @@ public class BuildColonyMission extends Mission {
      */
     private Location target;
 
-    /** The value of a target <code>Tile</code>. */
+    /** The value of a target {@code Tile}. */
     private int colonyValue;
 
 
     /**
-     * Creates a mission for the given <code>AIUnit</code>.
+     * Creates a mission for the given {@code AIUnit}.
      *
      * @param aiMain The main AI-object.
-     * @param aiUnit The <code>AIUnit</code> this mission is created for.
-     * @param target The target <code>Location</code> for this mission.
+     * @param aiUnit The {@code AIUnit} this mission is created for.
+     * @param target The target {@code Location} for this mission.
      */
     public BuildColonyMission(AIMain aiMain, AIUnit aiUnit, Location target) {
-        super(aiMain, aiUnit, target);
+        super(aiMain, aiUnit);
+
+        setTarget(target);
     }
 
     /**
-     * Creates a new <code>BuildColonyMission</code> and reads the given
-     * element.
+     * Creates a new {@code BuildColonyMission} from a reader.
      *
      * @param aiMain The main AI-object.
-     * @param aiUnit The <code>AIUnit</code> this mission is created for.
+     * @param aiUnit The {@code AIUnit} this mission is created for.
      * @param xr The input stream containing the XML.
      * @throws XMLStreamException if a problem was encountered during parsing.
      * @see net.sf.freecol.server.ai.AIObject#readFromXML
@@ -101,7 +107,7 @@ public class BuildColonyMission extends Mission {
     /**
      * Get the colony value for a tile.
      *
-     * @param tile The <code>Tile</code> to test.
+     * @param tile The {@code Tile} to test.
      * @return The colony value for this player.
      */
     private int getColonyValue(Tile tile) {
@@ -112,8 +118,8 @@ public class BuildColonyMission extends Mission {
     /**
      * Extract a valid target for this mission from a path.
      *
-     * @param aiUnit A <code>AIUnit</code> to perform the mission.
-     * @param path A <code>PathNode</code> to extract a target from,
+     * @param aiUnit A {@code AIUnit} to perform the mission.
+     * @param path A {@code PathNode} to extract a target from,
      *     (uses the unit location if null).
      * @return A target for this mission, or null if none found.
      */
@@ -122,16 +128,16 @@ public class BuildColonyMission extends Mission {
         final Location loc = path.getLastNode().getLocation();
         Tile tile = loc.getTile();
         Colony colony = loc.getColony();
-        return (invalidReason(aiUnit, tile) == null) ? tile
-            : (invalidReason(aiUnit, colony) == null) ? colony
+        return (invalidMissionReason(aiUnit, tile) == null) ? tile
+            : (invalidMissionReason(aiUnit, colony) == null) ? colony
             : null;
     }
 
     /**
      * Gets the value of a path to a colony building site.
      *
-     * @param aiUnit The <code>AIUnit</code> to build the colony.
-     * @param path The <code>PathNode</code> to check.
+     * @param aiUnit The {@code AIUnit} to build the colony.
+     * @param path The {@code PathNode} to check.
      * @return A score for the target.
      */
     public static float scorePath(AIUnit aiUnit, PathNode path) {
@@ -147,12 +153,12 @@ public class BuildColonyMission extends Mission {
     }
 
     /**
-     * Gets a <code>GoalDecider</code> for finding the best colony
-     * <code>Tile</code>, optionally falling back to the nearest colony.
+     * Gets a {@code GoalDecider} for finding the best colony
+     * {@code Tile}, optionally falling back to the nearest colony.
      *
-     * @param aiUnit The <code>AIUnit</code> that is searching.
+     * @param aiUnit The {@code AIUnit} that is searching.
      * @param deferOK Enable colony fallback.
-     * @return A suitable <code>GoalDecider</code>.
+     * @return A suitable {@code GoalDecider}.
      */
     private static GoalDecider getGoalDecider(final AIUnit aiUnit,
                                               boolean deferOK) {
@@ -186,13 +192,13 @@ public class BuildColonyMission extends Mission {
     /**
      * Finds a site for a new colony.  Favour closer sites.
      *
-     * @param aiUnit The <code>AIUnit</code> to execute this mission.
+     * @param aiUnit The {@code AIUnit} to execute this mission.
      * @param range An upper bound on the number of moves.
      * @param deferOK Enables deferring to a fallback colony.
      * @return A path to the new target, or null if none found.
      */
-    public static PathNode findTargetPath(AIUnit aiUnit, int range,
-                                          boolean deferOK) {
+    private static PathNode findTargetPath(AIUnit aiUnit, int range,
+                                           boolean deferOK) {
         if (invalidAIUnitReason(aiUnit) != null) return null;
         final Unit unit = aiUnit.getUnit();
         final Location start = unit.getPathStartLocation();
@@ -208,13 +214,13 @@ public class BuildColonyMission extends Mission {
     /**
      * Finds a site for a new colony or a backup colony to go to.
      *
-     * @param aiUnit The <code>AIUnit</code> to find a colony site with.
+     * @param aiUnit The {@code AIUnit} to find a colony site with.
      * @param range An upper bound on the number of moves.
      * @param deferOK Enables deferring to a fallback colony.
      * @return A new target for this mission.
      */
-    public static Location findTarget(AIUnit aiUnit, int range,
-                                      boolean deferOK) {
+    public static Location findMissionTarget(AIUnit aiUnit, int range,
+                                             boolean deferOK) {
         PathNode path = findTargetPath(aiUnit, range, deferOK);
         return (path != null) ? extractTarget(aiUnit, path)
             : Location.upLoc(findCircleTarget(aiUnit,
@@ -224,11 +230,11 @@ public class BuildColonyMission extends Mission {
     /**
      * Why would this mission be invalid with the given unit?
      *
-     * @param aiUnit The <code>AIUnit</code> to test.
+     * @param aiUnit The {@code AIUnit} to test.
      * @return A reason why the mission would be invalid with the unit,
      *     or null if none found.
      */
-    private static String invalidMissionReason(AIUnit aiUnit) {
+    private static String invalidUnitReason(AIUnit aiUnit) {
         String reason = invalidAIUnitReason(aiUnit);
         return (reason != null)
             ? reason
@@ -242,8 +248,8 @@ public class BuildColonyMission extends Mission {
     /**
      * Why is this mission invalid with a given colony target?
      *
-     * @param aiUnit The <code>AIUnit</code> to check.
-     * @param colony The potential target <code>Colony</code>.
+     * @param aiUnit The {@code AIUnit} to check.
+     * @param colony The potential target {@code Colony}.
      * @return A reason for mission invalidity, or null if none found.
      */
     private static String invalidColonyReason(AIUnit aiUnit, Colony colony) {
@@ -253,8 +259,8 @@ public class BuildColonyMission extends Mission {
     /**
      * Why is this mission invalid with a given tile target?
      *
-     * @param aiUnit The <code>AIUnit</code> to check.
-     * @param tile The potential target <code>Tile</code>.
+     * @param aiUnit The {@code AIUnit} to check.
+     * @param tile The potential target {@code Tile}.
      * @return A reason for mission invalidity, or null if none found.
      */
     private static String invalidTileReason(AIUnit aiUnit, Tile tile) {
@@ -270,14 +276,24 @@ public class BuildColonyMission extends Mission {
         return "target-" + reason;
     }
 
+    /**
+     * Why would this mission be invalid with the given AI unit?
+     *
+     * @param aiUnit The {@code AIUnit} to check.
+     * @return A reason for mission invalidity, or null if none found.
+     */
+    public static String invalidMissionReason(AIUnit aiUnit) {
+        return invalidUnitReason(aiUnit);
+    }
+
     /**
      * Why would this mission be invalid with the given AI unit and location?
      *
-     * @param aiUnit The <code>AIUnit</code> to check.
-     * @param loc The <code>Location</code> to check.
+     * @param aiUnit The {@code AIUnit} to check.
+     * @param loc The {@code Location} to check.
      * @return A reason for invalidity, or null if none found.
      */
-    public static String invalidReason(AIUnit aiUnit, Location loc) {
+    public static String invalidMissionReason(AIUnit aiUnit, Location loc) {
         String reason = invalidMissionReason(aiUnit);
         return (reason != null) ? reason
             : (loc instanceof Colony)
@@ -287,16 +303,6 @@ public class BuildColonyMission extends Mission {
             : Mission.TARGETINVALID;
     }
 
-    /**
-     * Why would this mission be invalid with the given AI unit?
-     *
-     * @param aiUnit The <code>AIUnit</code> to check.
-     * @return A reason for mission invalidity, or null if none found.
-     */
-    public static String invalidReason(AIUnit aiUnit) {
-        return invalidMissionReason(aiUnit);
-    }
-
       
     // Implement Mission
     //   Inherit dispose, getTransportDestination, isOneTime
@@ -336,7 +342,7 @@ public class BuildColonyMission extends Mission {
      */
     @Override
     public Location findTarget() {
-        return findTarget(getAIUnit(), 5, true);
+        return findMissionTarget(getAIUnit(), 5, true);
     }
 
     /**
@@ -344,7 +350,7 @@ public class BuildColonyMission extends Mission {
      */
     @Override
     public String invalidReason() {
-        return invalidReason(getAIUnit(), target);
+        return invalidMissionReason(getAIUnit(), target);
     }
 
     /**
@@ -397,10 +403,12 @@ public class BuildColonyMission extends Mission {
             case MOVE: // Arrived
                 break;
 
-            case MOVE_HIGH_SEAS: case MOVE_NO_MOVES:
-            case MOVE_NO_REPAIR: case MOVE_ILLEGAL:
+            case MOVE_HIGH_SEAS: case MOVE_NO_MOVES: case MOVE_ILLEGAL:
                 return lbWait(lb);
 
+            case MOVE_NO_REPAIR:
+                return lbFail(lb, false, AIUNITDIED);
+
             case MOVE_NO_ACCESS_EMBARK: case MOVE_NO_TILE:
                 return this;
 
@@ -428,23 +436,18 @@ public class BuildColonyMission extends Mission {
 
                 // Find a real tile target?
                 Location newTarget;
-                if ((newTarget = findTarget(aiUnit, 5, false)) != null) {
+                if ((newTarget = findMissionTarget(aiUnit, 5, false)) != null) {
                     setTarget(newTarget);
                     return lbRetarget(lb);
                 }
 
                 // Go to the nearest smaller colony?
-                Colony best = null;
-                int bestValue = INFINITY;
-                for (Colony c : player.getColonies()) {
-                    if (c == colony) continue;
-                    if (c.getUnitCount() < colony.getUnitCount()) continue;
-                    PathNode path = unit.findPath(c);
-                    if (path != null && path.getTotalTurns() < bestValue) {
-                        bestValue = path.getTotalTurns();
-                        best = c;
-                    }
-                }
+                final Predicate<Colony> smallPred = c ->
+                    c != colony && c.getUnitCount() < colony.getUnitCount();
+                final Comparator<Colony> closeComp = cachingIntComparator(c ->
+                    unit.getTurnsToReach(c));
+                Colony best = minimize(player.getColonies(), smallPred,
+                                       closeComp);
                 if (best != null) {
                     lb.add(", going to smaller ", best.getUnitCount(), "<",
                         colony.getUnitCount(), " colony");
@@ -490,7 +493,7 @@ public class BuildColonyMission extends Mission {
                 if (price >= 0) {
                     fail = !AIMessage.askClaimLand(tile, aiUnit,
                         ((price == 0) ? 0 : (player.checkGold(price)) ? price
-                            : NetworkConstants.STEAL_LAND))
+                            : Constants.STEAL_LAND))
                         || !player.owns(tile);
                 }
                 if (fail) return retargetMission("tile-claim-at-" + tile, lb);
@@ -558,14 +561,5 @@ public class BuildColonyMission extends Mission {
      * {@inheritDoc}
      */
     @Override
-    public String getXMLTagName() { return getXMLElementTagName(); }
-
-    /**
-     * Gets the tag name of the root element representing this object.
-     *
-     * @return "buildColonyMission".
-     */
-    public static String getXMLElementTagName() {
-        return "buildColonyMission";
-    }
+    public String getXMLTagName() { return TAG; }
 }
