@@ -1,5 +1,5 @@
 /**
- *  Copyright (C) 2002-2015   The FreeCol Team
+ *  Copyright (C) 2002-2022   The FreeCol Team
  *
  *  This file is part of FreeCol.
  *
@@ -19,6 +19,7 @@
 
 package net.sf.freecol.server.ai.mission;
 
+import java.util.Comparator;
 import java.util.Random;
 import java.util.logging.Logger;
 
@@ -45,14 +46,16 @@ import net.sf.freecol.common.model.Unit;
 import net.sf.freecol.common.model.Unit.MoveType;
 import net.sf.freecol.common.model.pathfinding.CostDecider;
 import net.sf.freecol.common.model.pathfinding.GoalDecider;
+import static net.sf.freecol.common.util.CollectionUtils.*;
 import net.sf.freecol.common.util.LogBuilder;
 import static net.sf.freecol.common.util.StringUtils.*;
+
 import net.sf.freecol.server.ai.AIMain;
 import net.sf.freecol.server.ai.AIMessage;
 import net.sf.freecol.server.ai.AIObject;
-import net.sf.freecol.server.ai.AIPlayer;
 import net.sf.freecol.server.ai.AIUnit;
 import net.sf.freecol.server.ai.EuropeanAIPlayer;
+import net.sf.freecol.server.ai.MissionAIPlayer;
 import net.sf.freecol.server.ai.TransportableAIObject;
 
 
@@ -76,6 +79,7 @@ public abstract class Mission extends AIObject {
 
     // Common mission invalidity reasons.
     protected static final String AIUNITNULL = "aiUnit-null";
+    protected static final String AIUNITDIED = "aiUnit-died";
     protected static final String TARGETNULL = "target-null";
     protected static final String TARGETINVALID = "target-invalid";
     protected static final String TARGETOWNERSHIP = "target-ownership";
@@ -89,45 +93,36 @@ public abstract class Mission extends AIObject {
 
 
     /**
-     * Creates a mission for the given <code>AIUnit</code>.
+     * Creates a mission for the given {@code AIUnit}.
      *
      * Note that missions are attached to their units, and thus do
      * not need AI ids, hence the plain superclass constructor.
      *
      * @param aiMain The main AI-object.
-     * @param aiUnit The <code>AIUnit</code> this mission is created for.
+     * @param aiUnit The {@code AIUnit} this mission is created for.
      */
     protected Mission(AIMain aiMain, AIUnit aiUnit) {
         super(aiMain);
 
         this.aiUnit = aiUnit;
-        this.uninitialized = aiUnit == null;
-    }
+        setInitialized();
 
-    /**
-     * Creates a mission for the given <code>AIUnit</code> and target.
-     *
-     * Note that missions are attached to their units, and thus do
-     * not need AI ids, hence the plain superclass constructor.
-     *
-     * @param aiMain The main AI-object.
-     * @param aiUnit The <code>AIUnit</code> this mission is created for.
-     * @param target The initial target <code>Location</code>.
-     */
-    protected Mission(AIMain aiMain, AIUnit aiUnit, Location target) {
-        this(aiMain, aiUnit);
-
-        setTarget(target);
         if (aiUnit != null && aiUnit.getMission() != this) {
             aiUnit.changeMission(this);
         }
     }
 
+    /**
+     * {@inheritDoc}
+     */
+    public final void setInitialized() {
+        this.initialized = getAIUnit() != null;
+    }
 
     /**
      * Gets the AI-unit this mission has been created for.
      *
-     * @return The <code>AIUnit</code>.
+     * @return The {@code AIUnit}.
      */
     public final AIUnit getAIUnit() {
         return aiUnit;
@@ -136,7 +131,7 @@ public abstract class Mission extends AIObject {
     /**
      * Gets the unit this mission has been created for.
      *
-     * @return The <code>Unit</code>.
+     * @return The {@code Unit}.
      */
     public final Unit getUnit() {
         return (aiUnit == null) ? null : aiUnit.getUnit();
@@ -145,7 +140,7 @@ public abstract class Mission extends AIObject {
     /**
      * Convenience accessor for the owning player.
      *
-     * @return The <code>Player</code> that owns the mission unit.
+     * @return The {@code Player} that owns the mission unit.
      */
     protected final Player getPlayer() {
         return (getUnit() == null) ? null : getUnit().getOwner();
@@ -154,16 +149,16 @@ public abstract class Mission extends AIObject {
     /**
      * Convenience accessor for the owning AI player.
      *
-     * @return The <code>AIPlayer</code>.
+     * @return The {@code AIPlayer}.
      */
-    protected final AIPlayer getAIPlayer() {
-        return getAIMain().getAIPlayer(getUnit().getOwner());
+    protected final MissionAIPlayer getAIPlayer() {
+        return (MissionAIPlayer)getAIMain().getAIPlayer(getUnit().getOwner());
     }
 
     /**
      * Convenience accessor for the owning European AI player.
      *
-     * @return The <code>EuropeanAIPlayer</code>.
+     * @return The {@code EuropeanAIPlayer}.
      */
     protected final EuropeanAIPlayer getEuropeanAIPlayer() {
         Player player = getUnit().getOwner();
@@ -177,7 +172,7 @@ public abstract class Mission extends AIObject {
     /**
      * Convenience accessor for the unit/player PRNG.
      *
-     * @return A <code>Random</code> to use.
+     * @return A {@code Random} to use.
      */
     protected final Random getAIRandom() {
         return aiUnit.getAIRandom();
@@ -195,6 +190,7 @@ public abstract class Mission extends AIObject {
     /**
      * Is an invalidity reason due to a target failure?
      *
+     * @param reason The reason to check.
      * @return True if the reason starts with "target-".
      */
     public static boolean isTargetReason(String reason) {
@@ -204,12 +200,12 @@ public abstract class Mission extends AIObject {
     /**
      * Is a unit able to perform a mission of a particular type?
      *
-     * @param unit The <code>Unit</code> to check.
+     * @param unit The {@code Unit} to check.
      * @return A reason for mission invalidity, or null if none found.
      */
-    public static String invalidUnitReason(Unit unit) {
+    private static String invalidUnitReason(Unit unit) {
         return (unit == null) ? "unit-null"
-            : (unit.isUninitialized()) ? "unit-uninitialized"
+            : (!unit.isInitialized()) ? "unit-uninitialized"
             : (unit.isDisposed()) ? "unit-disposed"
             : (unit.isDamaged()) ? "unit-under-repair"
             : null;
@@ -218,7 +214,7 @@ public abstract class Mission extends AIObject {
     /**
      * Is an AI unit able to perform a mission of a particular type?
      *
-     * @param aiUnit The <code>AIUnit</code> to check.
+     * @param aiUnit The {@code AIUnit} to check.
      * @return A reason for mission invalidity, or null if none found.
      */
     public static String invalidAIUnitReason(AIUnit aiUnit) {
@@ -232,6 +228,7 @@ public abstract class Mission extends AIObject {
      * Is an AI unable to perform a new mission because it already has
      * a valid, non-onetime mission?
      *
+     * @param aiUnit The {@code AIUnit} to test.
      * @return "mission-exists" if a valid mission is found, or null
      *     if none found.
      */
@@ -246,7 +243,7 @@ public abstract class Mission extends AIObject {
     /**
      * Is a target a valid mission target?
      *
-     * @param target The target <code>Location</code> to check.
+     * @param target The target {@code Location} to check.
      * @return A reason for the target to be invalid, or null if none found.
      */
     public static String invalidTargetReason(Location target) {
@@ -258,8 +255,8 @@ public abstract class Mission extends AIObject {
     /**
      * Is a target a valid mission target?
      *
-     * @param target The target <code>Location</code> to check.
-     * @param owner A <code>Player</code> that should own
+     * @param target The target {@code Location} to check.
+     * @param owner A {@code Player} that should own
      *     the target.
      * @return A reason for the target to be invalid, or null if none found.
      */
@@ -277,7 +274,7 @@ public abstract class Mission extends AIObject {
     /**
      * Is there a reason to invalidate mission to move a transportable?
      *
-     * @param t The <code>TransportableAIObject</code> to check.
+     * @param t The {@code TransportableAIObject} to check.
      * @return A reason for the transport to be invalid, or null if
      *     none found.
      */
@@ -321,8 +318,8 @@ public abstract class Mission extends AIObject {
     /**
      * Is another player a valid attack target?
      *
-     * @param aiUnit The <code>AIUnit</code> that will attack.
-     * @param other The <code>Player</code> to attack.
+     * @param aiUnit The {@code AIUnit} that will attack.
+     * @param other The {@code Player} to attack.
      * @return A reason why the attack would be invalid, or null if none found.
      */
     public static String invalidAttackReason(AIUnit aiUnit, Player other) {
@@ -355,10 +352,10 @@ public abstract class Mission extends AIObject {
      * Implementations should usually start by calling this routine
      * (i.e. Mission.invalidReason(AIUnit)).
      *
-     * @param aiUnit The <code>AIUnit</code> to check.
+     * @param aiUnit The {@code AIUnit} to check.
      * @return A reason for mission invalidity, or null if none found.
      */
-    public static String invalidReason(AIUnit aiUnit) {
+    public static String invalidMissionReason(AIUnit aiUnit) {
         return invalidAIUnitReason(aiUnit);
     }
 
@@ -374,11 +371,11 @@ public abstract class Mission extends AIObject {
      * invalidAIUnitReason() or this routine if the target checking is
      * trivial.
      *
-     * @param aiUnit The <code>AIUnit</code> to check.
-     * @param loc The target <code>Location</code> to check.
+     * @param aiUnit The {@code AIUnit} to check.
+     * @param loc The target {@code Location} to check.
      * @return A reason for mission invalidity, or null if none found.
      */
-    public static String invalidReason(AIUnit aiUnit, Location loc) {
+    public static String invalidMissionReason(AIUnit aiUnit, Location loc) {
         String reason = invalidAIUnitReason(aiUnit);
         return (reason != null) ? reason : invalidTargetReason(loc);
     }
@@ -389,8 +386,8 @@ public abstract class Mission extends AIObject {
     /**
      * State where a unit is.
      *
-     * @param lb A <code>LogBuilder</code> to log to.
-     * @return This <code>Mission</code>.
+     * @param lb A {@code LogBuilder} to log to.
+     * @return This {@code Mission}.
      */
     protected Mission lbAt(LogBuilder lb) {
         final Unit unit = getUnit();
@@ -401,10 +398,10 @@ public abstract class Mission extends AIObject {
     /**
      * State that the unit has made an attack.
      *
-     * @param lb A <code>LogBuilder</code> to log to.
-     * @param what What is being attacked (a <code>Unit</code> or
-     *     <code>Settlement</code>).
-     * @return This <code>Mission</code>.
+     * @param lb A {@code LogBuilder} to log to.
+     * @param what What is being attacked (a {@code Unit} or
+     *     {@code Settlement}).
+     * @return This {@code Mission}.
      */
     protected Mission lbAttack(LogBuilder lb, Location what) {
         lb.add(", attacking ", what);
@@ -414,8 +411,8 @@ public abstract class Mission extends AIObject {
     /**
      * State that the unit is dodging.
      *
-     * @param lb A <code>LogBuilder</code> to log to.
-     * @return This <code>Mission</code>.
+     * @param lb A {@code LogBuilder} to log to.
+     * @return This {@code Mission}.
      */
     protected Mission lbDodge(LogBuilder lb) {
         final Unit unit = getUnit();
@@ -427,12 +424,12 @@ public abstract class Mission extends AIObject {
     /**
      * State that this mission has completed successfully.
      *
-     * @param lb A <code>LogBuilder</code> to log to.
+     * @param lb A {@code LogBuilder} to log to.
      * @param cont If true, the mission should continue, otherwise drop the
      *     mission.
      * @param reasons Reasons for the successful completion.
-     * @return The current <code>Mission</code> of the unit, which may
-     *     now be different from <code>this</code>.
+     * @return The current {@code Mission} of the unit, which may
+     *     now be different from {@code this}.
      */
     protected Mission lbDone(LogBuilder lb, boolean cont, Object... reasons) {
         lb.add(", COMPLETED: ", reasons);
@@ -442,9 +439,9 @@ public abstract class Mission extends AIObject {
     /**
      * Drop the current mission.
      *
-     * @param lb A <code>LogBuilder</code> to log to.
+     * @param lb A {@code LogBuilder} to log to.
      * @param reasons Optional reasons to drop the mission.
-     * @return Null, which is now the current <code>Mission</code> of
+     * @return Null, which is now the current {@code Mission} of
      *     this unit.
      */
     protected Mission lbDrop(LogBuilder lb, Object... reasons) {
@@ -455,12 +452,12 @@ public abstract class Mission extends AIObject {
     /**
      * The current mission has failed.
      *
-     * @param lb A <code>LogBuilder</code> to log to.
+     * @param lb A {@code LogBuilder} to log to.
      * @param cont If true, the mission should continue, otherwise drop the
      *     mission.
      * @param reasons Reasons for the successful completion.
-     * @return The current <code>Mission</code> of the unit, which may
-     *     not be different from <code>this</code>.
+     * @return The current {@code Mission} of the unit, which may
+     *     not be different from {@code this}.
      */
     protected Mission lbFail(LogBuilder lb, boolean cont, Object... reasons) {
         lb.add(", FAILED: ", reasons);
@@ -470,9 +467,9 @@ public abstract class Mission extends AIObject {
     /**
      * State that a bad move has occurred.
      *
-     * @param lb A <code>LogBuilder</code> to log to.
-     * @param mt The bad <code>MoveType</code>.
-     * @return This <code>Mission</code>.
+     * @param lb A {@code LogBuilder} to log to.
+     * @param mt The bad {@code MoveType}.
+     * @return This {@code Mission}.
      */     
     protected Mission lbMove(LogBuilder lb, Unit.MoveType mt) {
         lb.add(", bad move type at ", getUnit().getLocation(), ": ", mt);
@@ -482,8 +479,8 @@ public abstract class Mission extends AIObject {
     /**
      * State that the mission has been retargeted.
      *
-     * @param lb A <code>LogBuilder</code> to log to.
-     * @return This <code>Mission</code>.
+     * @param lb A {@code LogBuilder} to log to.
+     * @return This {@code Mission}.
      */
     protected Mission lbRetarget(LogBuilder lb) {
         lb.add(", retargeted ", getTarget());
@@ -493,9 +490,9 @@ public abstract class Mission extends AIObject {
     /**
      * State that the unit is waiting for something.
      *
-     * @param lb A <code>LogBuilder</code> to log to.
+     * @param lb A {@code LogBuilder} to log to.
      * @param reasons Reasons for the wait.
-     * @return This <code>Mission</code>.
+     * @return This {@code Mission}.
      */
     protected Mission lbWait(LogBuilder lb, Object... reasons) {
         lb.add(reasons);
@@ -510,11 +507,11 @@ public abstract class Mission extends AIObject {
      * can not find a target with the normal path finding routines,
      * and thus should consider targets that may require a carrier.
      *
-     * @param aiUnit The <code>AIUnit</code> that is searching.
-     * @param gd The <code>GoalDecider</code> that selects targets.
+     * @param aiUnit The {@code AIUnit} that is searching.
+     * @param gd The {@code GoalDecider} that selects targets.
      * @param radius A maximum radius from the unit location to search within.
      * @param deferOK If true, fall back to the nearest port to Europe.
-     * @return The best target <code>Tile</code> found, or null if none.
+     * @return The best target {@code Tile} found, or null if none.
      */
     protected static Location findCircleTarget(final AIUnit aiUnit,
                                                final GoalDecider gd,
@@ -534,8 +531,8 @@ public abstract class Mission extends AIObject {
      * We have been blocked on the way to a target.  Is it valid to
      * attack the blockage, or should it just be avoided?
      *
-     * @param aiUnit The <code>AIUnit</code> that was blocked.
-     * @param target The target <code>Location</code>.
+     * @param aiUnit The {@code AIUnit} that was blocked.
+     * @param target The target {@code Location}.
      * @return The blockage to attack, or null if not.
      */
     public static Location resolveBlockage(AIUnit aiUnit, Location target) {
@@ -547,7 +544,7 @@ public abstract class Mission extends AIObject {
             Settlement settlement = tile.getSettlement();
             Location blocker = (settlement != null) ? settlement
                 : tile.getDefendingUnit(unit);
-            if (UnitSeekAndDestroyMission.invalidReason(aiUnit, blocker)
+            if (UnitSeekAndDestroyMission.invalidMissionReason(aiUnit, blocker)
                 == null) return blocker;
         }
         return null;
@@ -557,7 +554,7 @@ public abstract class Mission extends AIObject {
      * Moves a unit one step randomly.
      *
      * @param logMe A string to log the random number generation with.
-     * @param direction An optional preferred <code>Direction</code>.
+     * @param direction An optional preferred {@code Direction}.
      * @return The direction of the move, or null if no move was made.
      */
     protected Direction moveRandomly(String logMe, Direction direction) {
@@ -588,7 +585,7 @@ public abstract class Mission extends AIObject {
      */
     protected void moveRandomlyTurn(String logMe) {
         Direction direction = null;
-        while ((direction = moveRandomly(logMe, direction)) != null);
+        while ((direction = moveRandomly(logMe, direction)) != null) {}
         getUnit().setMovesLeft(0);
     }
 
@@ -597,26 +594,21 @@ public abstract class Mission extends AIObject {
      * Useful for missions where the unit might be in Europe, but should
      * go to a safe spot in the New World and proceed from there.
      *
-     * @param player The <code>Player</code> that is searching.
+     * @param player The {@code Player} that is searching.
      * @return A good settlement to restart a Mission from.
      */
     protected static Settlement getBestSettlement(Player player) {
-        int bestValue = -1;
-        Settlement best = null;
-        for (Settlement settlement : player.getSettlements()) {
-            int value = settlement.getUnitCount()
-                + settlement.getTile().getUnitCount();
-            if (settlement instanceof Colony) {
-                Colony colony = (Colony)settlement;
-                value += ((colony.isConnectedPort()) ? 10 : 0) // Favour coastal
-                    + colony.getAvailableWorkLocations().size();
-            }
-            if (value > bestValue) {
-                bestValue = value;
-                best = settlement;
-            }
-        }
-        return (best == null) ? null : best;
+        final Comparator<Settlement> comp = cachingIntComparator(s -> {
+                int value = s.getUnitCount() + s.getTile().getUnitCount();
+                if (s instanceof Colony) {
+                    Colony colony = (Colony)s;
+                    // Favour coastal
+                    value += ((colony.isConnectedPort()) ? 10 : 0)
+                        + colony.getAvailableWorkLocationsList().size();
+                }
+                return value;
+            });
+        return maximize(player.getSettlements(), comp);
     }
 
     /**
@@ -648,10 +640,10 @@ public abstract class Mission extends AIObject {
      *   and are not performed here, the calling mission code must
      *   handle them.
      *
-     * @param target The destination <code>Location</code>.
-     * @param costDecider The <code>CostDecider</code> to use in any path
+     * @param target The destination {@code Location}.
+     * @param costDecider The {@code CostDecider} to use in any path
      *     finding.
-     * @param lb A <code>LogBuilder</code> to log to.
+     * @param lb A {@code LogBuilder} to log to.
      * @return The type of move the unit stopped at.
      */
     protected MoveType travelToTarget(Location target, CostDecider costDecider,
@@ -659,10 +651,10 @@ public abstract class Mission extends AIObject {
         if (target == null) return MoveType.MOVE_ILLEGAL;
         final Tile targetTile = target.getTile();
         if (!(target instanceof Europe) && targetTile == null) {
-            throw new IllegalStateException("Target neither Europe nor Tile");
+            throw new RuntimeException("Target neither Europe nor Tile: "
+                + target);
         }
         final Unit unit = getUnit();
-        final AIUnit aiUnit = getAIUnit();
         AIUnit aiCarrier = aiUnit.getTransport();
         final Map map = unit.getGame().getMap();
         PathNode path = null;
@@ -720,7 +712,7 @@ public abstract class Mission extends AIObject {
                     useTransport = true;
                 } else {
                     path = unit.findPath(unit.getLocation(), target,
-                                         null, costDecider);
+                                         null, costDecider, null);
                 }
             } else if (aiCarrier != null) {
                 // Transport already allocated.
@@ -734,7 +726,7 @@ public abstract class Mission extends AIObject {
             } else {
                 // Should not need transport within the same contiguity.
                 path = unit.findPath(unit.getLocation(), target,
-                                     null, costDecider);
+                                     null, costDecider, null);
             }
         }
 
@@ -770,13 +762,13 @@ public abstract class Mission extends AIObject {
                     waiting = true;
 
                 } else if ((path = unit.findPath(unit.getLocation(), pick,
-                                                 null, costDecider)) == null) {
+                            null, costDecider, null)) == null) {
                     // No path to the collection point.
                     lbAt(lb);
                     lb.add(", no path to meet ", aiCarrier.getUnit(),
                            " at ", pick);
                     path = unit.findPath(unit.getLocation(), target,
-                                         null, costDecider);
+                        null, costDecider, null);
                     if (path == null) {
                         // Unable to fall back to going direct.
                         // Return failure in the hope that it is a
@@ -790,7 +782,7 @@ public abstract class Mission extends AIObject {
                     useTransport = false;
 
                 } else if ((ownPath = unit.findPath(unit.getLocation(), 
-                            target, null, costDecider)) == null
+                            target, null, costDecider, null)) == null
                     || (ownTurns = ownPath.getTotalTurns())
                     > (pathTurns = path.getTotalTurns())) {
                     // Either there is no direct path to the target or
@@ -854,13 +846,12 @@ public abstract class Mission extends AIObject {
      * Follow a path that is on the map (except perhaps the last node)
      * and does not use a carrier.
      *
-     * @param path The <code>PathNode</code> to follow.
-     * @param lb A <code>LogBuilder</code> to log to.
+     * @param path The {@code PathNode} to follow.
+     * @param lb A {@code LogBuilder} to log to.
      * @return The type of move the unit stopped at.
      */
     private MoveType followMapPath(PathNode path, LogBuilder lb) {
         final Unit unit = getUnit();
-        final AIUnit aiUnit = getAIUnit();
         final Location target = path.getLastNode().getLocation();
 
         for (; path != null; path = path.next) {
@@ -906,8 +897,8 @@ public abstract class Mission extends AIObject {
      * Retarget a mission because of some problem.
      *
      * @param reason The reason for the retarget.
-     * @param lb A <code>LogBuilder</code> to log to.
-     * @return The current <code>Mission</code>, which has been set to
+     * @param lb A {@code LogBuilder} to log to.
+     * @return The current {@code Mission}, which has been set to
      *     null on failure to retarget.
      */
     public Mission retargetMission(String reason, LogBuilder lb) {
@@ -973,6 +964,9 @@ public abstract class Mission extends AIObject {
     /**
      * Sets the target of this mission, if any.
      *
+     * The actual target is handled in the missions that really have
+     * them, this is a helper to make sure the unit is updated.
+     *
      * @param target The new target of this mission, or null if none.
      */
     public abstract void setTarget(Location target);
@@ -1015,8 +1009,8 @@ public abstract class Mission extends AIObject {
     /**
      * Performs the mission.
      *
-     * @param lb A <code>LogBuilder</code> to log to.
-     * @return The <code>Mission</code> to continue with, or null
+     * @param lb A {@code LogBuilder} to log to.
+     * @return The {@code Mission} to continue with, or null
      *     if the current mission has completed.
      */
     public abstract Mission doMission(LogBuilder lb);
@@ -1055,6 +1049,9 @@ public abstract class Mission extends AIObject {
         // attribute.  Missions do not have ids.
     }
 
+
+    // Override Object
+
     /**
      * {@inheritDoc}
      */
