@@ -1,5 +1,5 @@
 /**
- *  Copyright (C) 2002-2015   The FreeCol Team
+ *  Copyright (C) 2002-2022   The FreeCol Team
  *
  *  This file is part of FreeCol.
  *
@@ -19,6 +19,8 @@
 
 package net.sf.freecol.server.ai.mission;
 
+import java.util.List;
+import java.util.Set;
 import java.util.logging.Logger;
 
 import javax.xml.stream.XMLStreamException;
@@ -40,7 +42,9 @@ import net.sf.freecol.common.model.pathfinding.CostDecider;
 import net.sf.freecol.common.model.pathfinding.CostDeciders;
 import net.sf.freecol.common.model.pathfinding.GoalDecider;
 import net.sf.freecol.common.model.pathfinding.GoalDeciders;
+import static net.sf.freecol.common.util.CollectionUtils.*;
 import net.sf.freecol.common.util.LogBuilder;
+import static net.sf.freecol.common.util.RandomUtils.*;
 import net.sf.freecol.server.ai.AIMain;
 import net.sf.freecol.server.ai.AIMessage;
 import net.sf.freecol.server.ai.AIUnit;
@@ -49,10 +53,12 @@ import net.sf.freecol.server.ai.AIUnit;
 /**
  * Mission for controlling a scout.
  */
-public class ScoutingMission extends Mission {
+public final class ScoutingMission extends Mission {
 
     private static final Logger logger = Logger.getLogger(ScoutingMission.class.getName());
 
+    public static final String TAG = "scoutingMission";
+
     /** The tag for this mission. */
     private static final String tag = "AI scout";
 
@@ -64,24 +70,28 @@ public class ScoutingMission extends Mission {
      * - An unexplored tile
      */
     private Location target;
+    /** A Tile to go to before going to the real target. */
+    private Location precedingTile;
 
 
     /**
-     * Creates a mission for the given <code>AIUnit</code>.
+     * Creates a mission for the given {@code AIUnit}.
      *
      * @param aiMain The main AI-object.
-     * @param aiUnit The <code>AIUnit</code> this mission is created for.
-     * @param target The target <code>Location</code>.
+     * @param aiUnit The {@code AIUnit} this mission is created for.
+     * @param target The target {@code Location}.
      */
     public ScoutingMission(AIMain aiMain, AIUnit aiUnit, Location target) {
-        super(aiMain, aiUnit, target);
+        super(aiMain, aiUnit);
+
+        setTarget(target);
     }
 
     /**
-     * Creates a new <code>ScoutingMission</code> and reads the given element.
+     * Creates a new {@code ScoutingMission} from a reader.
      *
      * @param aiMain The main AI-object.
-     * @param aiUnit The <code>AIUnit</code> this mission is created for.
+     * @param aiUnit The {@code AIUnit} this mission is created for.
      * @param xr The input stream containing the XML.
      * @throws XMLStreamException if a problem was encountered during parsing.
      * @see net.sf.freecol.server.ai.AIObject#readFromXML
@@ -97,7 +107,7 @@ public class ScoutingMission extends Mission {
     /**
      * Does a supplied unit have horses?
      *
-     * @param aiUnit The scout <code>AIUnit</code> to check.
+     * @param aiUnit The scout {@code AIUnit} to check.
      * @return True if the scout has horses.
      */
     private static boolean canScoutNatives(AIUnit aiUnit) {
@@ -108,8 +118,8 @@ public class ScoutingMission extends Mission {
     /**
      * Extract a valid target for this mission from a path.
      *
-     * @param aiUnit A <code>AIUnit</code> to perform the mission.
-     * @param path A <code>PathNode</code> to extract a target from,
+     * @param aiUnit A {@code AIUnit} to perform the mission.
+     * @param path A {@code PathNode} to extract a target from,
      *     (uses the unit location if null).
      * @return A target for this mission, or null if none found.
      */
@@ -128,8 +138,8 @@ public class ScoutingMission extends Mission {
      * Evaluate a potential scouting mission for a given unit and
      * path.
      *
-     * @param aiUnit The <code>AIUnit</code> to do the mission.
-     * @param path A <code>PathNode</code> to take to the target.
+     * @param aiUnit The {@code AIUnit} to do the mission.
+     * @param path A {@code PathNode} to take to the target.
      * @return A score for the proposed mission.
      */
     public static int scorePath(AIUnit aiUnit, PathNode path) {
@@ -138,20 +148,18 @@ public class ScoutingMission extends Mission {
             ? 12 / (path.getTotalTurns() + 1)
             : (loc instanceof IndianSettlement)
             ? 2000 / (path.getTotalTurns() + 1)
-            : (loc instanceof Tile)
-            ? ((((Tile)loc).hasLostCityRumour())
-                ? 1000 / (path.getTotalTurns() + 1)
-                : 50 / (path.getTotalTurns() + 1))
+            : (loc instanceof Tile && ((Tile)loc).hasLostCityRumour())
+            ? 1000 / (path.getTotalTurns() + 1)
             : Integer.MIN_VALUE;
     }
 
     /**
-     * Gets a <code>GoalDecider</code> for finding the best colony
-     * <code>Tile</code>, optionally falling back to the nearest colony.
+     * Gets a {@code GoalDecider} for finding the best colony
+     * {@code Tile}, optionally falling back to the nearest colony.
      *
-     * @param aiUnit The <code>AIUnit</code> that is searching.
+     * @param aiUnit The {@code AIUnit} that is searching.
      * @param deferOK Enable colony fallback.
-     * @return A suitable <code>GoalDecider</code>.
+     * @return A suitable {@code GoalDecider}.
      */
     private static GoalDecider getGoalDecider(final AIUnit aiUnit,
                                               boolean deferOK) {
@@ -186,13 +194,13 @@ public class ScoutingMission extends Mission {
     /**
      * Finds a suitable scouting target for the supplied unit.
      *
-     * @param aiUnit The <code>AIUnit</code> to execute this mission.
+     * @param aiUnit The {@code AIUnit} to execute this mission.
      * @param range An upper bound on the number of moves.
      * @param deferOK Enables deferring to a fallback colony.
      * @return A path to the new target, or null if none found.
      */
-    public static PathNode findTargetPath(AIUnit aiUnit, int range,
-                                          boolean deferOK) {
+    private static PathNode findTargetPath(AIUnit aiUnit, int range,
+                                           boolean deferOK) {
         if (invalidAIUnitReason(aiUnit) != null) return null;
         final Unit unit = aiUnit.getUnit();
         final Location start = unit.getPathStartLocation();
@@ -209,29 +217,30 @@ public class ScoutingMission extends Mission {
      * Finds a suitable scouting target for the supplied unit.
      * Falls back to the best settlement if a path is not found.
      *
-     * @param aiUnit The <code>AIUnit</code> to test.
+     * @param aiUnit The {@code AIUnit} to test.
      * @param range An upper bound on the number of moves.
      * @param deferOK Enables deferring to a fallback colony.
-     * @return A <code>PathNode</code> to the target, or null if none found.
+     * @return A {@code PathNode} to the target, or null if none found.
      */
-    public static Location findTarget(AIUnit aiUnit, int range,
-                                      boolean deferOK) {
+    public static Location findMissionTarget(AIUnit aiUnit, int range,
+                                             boolean deferOK) {
         PathNode path = findTargetPath(aiUnit, range, deferOK);
-        return (path != null) ? extractTarget(aiUnit, path)
-            : Location.upLoc(findCircleTarget(aiUnit,
-                    getGoalDecider(aiUnit, deferOK), range*3, deferOK));
+        return Location.upLoc((path != null)
+            ? extractTarget(aiUnit, path)
+            : findCircleTarget(aiUnit,
+                getGoalDecider(aiUnit, deferOK), range*3, deferOK));
     }
 
     /**
      * Prepare a unit for this mission.  Allow even experts to proceed
      * even if not mounted.
      *
-     * @param aiUnit The <code>AIUnit</code> to prepare.
+     * @param aiUnit The {@code AIUnit} to prepare.
      * @return A reason why the unit can not perform this mission, or null
      *     if none.
      */
     public static String prepare(AIUnit aiUnit) {
-        String reason = invalidReason(aiUnit);
+        String reason = invalidMissionReason(aiUnit);
         return (reason != null) ? reason
             : (canScoutNatives(aiUnit)
                 || aiUnit.equipForRole(aiUnit.getUnit().getSpecification()
@@ -243,11 +252,11 @@ public class ScoutingMission extends Mission {
     /**
      * Why would this mission be invalid with the given unit.
      *
-     * @param aiUnit The <code>AIUnit</code> to check.
+     * @param aiUnit The {@code AIUnit} to check.
      * @return A reason why the mission would be invalid with the unit,
      *     or null if none found.
      */
-    private static String invalidMissionReason(AIUnit aiUnit) {
+    private static String invalidUnitReason(AIUnit aiUnit) {
         String reason = invalidAIUnitReason(aiUnit);
         return (reason != null) ? reason
             : (!canScoutNatives(aiUnit)) ? "unit-not-a-SCOUT"
@@ -257,8 +266,8 @@ public class ScoutingMission extends Mission {
     /**
      * Why is this mission invalid with a given colony target?
      *
-     * @param aiUnit The <code>AIUnit</code> to check.
-     * @param colony The potential target <code>Colony</code>.
+     * @param aiUnit The {@code AIUnit} to check.
+     * @param colony The potential target {@code Colony}.
      * @return A reason for mission invalidity, or null if none found.
      */
     private static String invalidColonyReason(AIUnit aiUnit, Colony colony) {
@@ -268,8 +277,8 @@ public class ScoutingMission extends Mission {
     /**
      * Why is this mission invalid with a given native settlement target?
      *
-     * @param aiUnit The <code>AIUnit</code> to check.
-     * @param is The potential target <code>IndianSettlement</code>.
+     * @param aiUnit The {@code AIUnit} to check.
+     * @param is The potential target {@code IndianSettlement}.
      * @return A reason for mission invalidity, or null if none found.
      */
     private static String invalidIndianSettlementReason(AIUnit aiUnit,
@@ -291,8 +300,8 @@ public class ScoutingMission extends Mission {
      * Is this a valid scouting target because it is a suitable native
      * settlement or an intermediate colony.
      *
-     * @param aiUnit The <code>AIUnit</code> to test.
-     * @param settlement The <code>Settlement</code> to test.
+     * @param aiUnit The {@code AIUnit} to test.
+     * @param settlement The {@code Settlement} to test.
      * @return A reason why the mission would be invalid, or null if none found.
      */
     private static String invalidSettlementReason(AIUnit aiUnit,
@@ -307,8 +316,8 @@ public class ScoutingMission extends Mission {
     /**
      * Is this a valid scouting target because it is a suitable tile.
      *
-     * @param aiUnit The <code>AIUnit</code> to test.
-     * @param tile The <code>Tile</code> to test.
+     * @param aiUnit The {@code AIUnit} to test.
+     * @param tile The {@code Tile} to test.
      * @return A reason why the mission would be invalid, or null if none found.
      */
     private static String invalidTileReason(AIUnit aiUnit, Tile tile) {
@@ -321,21 +330,21 @@ public class ScoutingMission extends Mission {
     /**
      * Why would this mission be invalid with the given AI unit?
      *
-     * @param aiUnit The <code>AIUnit</code> to check.
+     * @param aiUnit The {@code AIUnit} to check.
      * @return A reason for mission invalidity, or null if none found.
      */
-    public static String invalidReason(AIUnit aiUnit) {
-        return invalidMissionReason(aiUnit);
+    public static String invalidMissionReason(AIUnit aiUnit) {
+        return invalidUnitReason(aiUnit);
     }
 
     /**
      * Why would this mission be invalid with the given AI unit and location?
      *
-     * @param aiUnit The <code>AIUnit</code> to check.
-     * @param loc The <code>Location</code> to check.
+     * @param aiUnit The {@code AIUnit} to check.
+     * @param loc The {@code Location} to check.
      * @return A reason for invalidity, or null if none found.
      */
-    public static String invalidReason(AIUnit aiUnit, Location loc) {
+    public static String invalidMissionReason(AIUnit aiUnit, Location loc) {
         String reason = invalidMissionReason(aiUnit);
         return (reason != null)
             ? reason
@@ -363,7 +372,7 @@ public class ScoutingMission extends Mission {
      */
     @Override
     public Location getTarget() {
-        return this.target;
+        return (this.precedingTile != null) ? this.precedingTile : this.target;
     }
 
     /**
@@ -371,8 +380,26 @@ public class ScoutingMission extends Mission {
      */
     @Override
     public void setTarget(Location target) {
-        if (target == null
-            || target instanceof Settlement || target instanceof Tile) {
+        if (target == null || target instanceof Tile
+            || target instanceof Colony) {
+            this.target = target;
+            this.precedingTile = null;
+        } else if (target instanceof IndianSettlement) {
+            AIUnit aiUnit = getAIUnit();
+            Unit unit = aiUnit.getUnit();
+            Tile t = target.getTile();
+            if (!unit.hasTile()
+                || unit.getTile().getContiguity() != t.getContiguity()) {
+                // Not going to be able to walk there, so we will need to take
+                // a carrier, but that means going first to a tile adjacent
+                // to the settlement, which we were not doing in BR#3228.
+                List<Tile> tiles = toList(t.getContiguityAdjacent(t.getContiguity()));
+                List<Tile> coast = transform(tiles, Tile::isCoastland);
+                this.precedingTile = getRandomMember(logger, "scout-tile",
+                    (coast.isEmpty()) ? tiles : coast, aiUnit.getAIRandom());
+            } else {
+                this.precedingTile = null;
+            }
             this.target = target;
         }
     }
@@ -382,7 +409,7 @@ public class ScoutingMission extends Mission {
      */
     @Override
     public Location findTarget() {
-        return findTarget(getAIUnit(), 20, true);
+        return findMissionTarget(getAIUnit(), 20, true);
     }
 
     /**
@@ -390,7 +417,7 @@ public class ScoutingMission extends Mission {
      */
     @Override
     public String invalidReason() {
-        return invalidReason(getAIUnit(), getTarget());
+        return invalidMissionReason(getAIUnit(), getTarget());
     }
 
     /**
@@ -413,13 +440,21 @@ public class ScoutingMission extends Mission {
         Unit.MoveType mt = travelToTarget(getTarget(),
             CostDeciders.avoidSettlementsAndBlockingUnits(), lb);
         switch (mt) {
-        case MOVE: // Arrived at a colony
+        case MOVE: // Arrived
+            if (this.precedingTile != null) {
+                // At a tile next to the settlement, stop pretending that
+                // is where we are going.
+                this.precedingTile = null;
+                return doMission(lb);
+            }
             break;
 
-        case MOVE_HIGH_SEAS: case MOVE_NO_MOVES:
-        case MOVE_NO_REPAIR: case MOVE_ILLEGAL:
+        case MOVE_HIGH_SEAS: case MOVE_NO_MOVES: case MOVE_ILLEGAL:
             return lbWait(lb);
 
+        case MOVE_NO_REPAIR:
+            return lbFail(lb, false, AIUNITDIED);
+
         case MOVE_NO_ACCESS_EMBARK:
             return this;
 
@@ -437,8 +472,9 @@ public class ScoutingMission extends Mission {
 
         case ENTER_INDIAN_SETTLEMENT_WITH_SCOUT:
             d = unit.getTile().getDirection(getTarget().getTile());
-            assert d != null;
-            if (AIMessage.askScoutSpeakToChief(aiUnit, d)) {
+            assert d != null && getTarget() instanceof IndianSettlement;
+            if (AIMessage.askScoutSpeakToChief(aiUnit,
+                                               (IndianSettlement)getTarget())) {
                 lbDone(lb, true, "speak-with-chief at ", getTarget());
             } else {
                 lbFail(lb, true, "unexpected failure to speak at ", getTarget());
@@ -465,7 +501,7 @@ public class ScoutingMission extends Mission {
         // the mission.
         lbAt(lb);
         Location completed = getTarget();
-        Location newTarget = findTarget(aiUnit, 20, false);
+        Location newTarget = findMissionTarget(aiUnit, 20, false);
         if (newTarget == null
             || (completed instanceof Colony && newTarget == completed)) {
             if (completed instanceof Colony && canScoutNatives(aiUnit)) {
@@ -502,21 +538,12 @@ public class ScoutingMission extends Mission {
     protected void readAttributes(FreeColXMLReader xr) throws XMLStreamException {
         super.readAttributes(xr);
 
-        target = xr.getLocationAttribute(getGame(), TARGET_TAG, false);
+        setTarget(xr.getLocationAttribute(getGame(), TARGET_TAG, false));
     }
 
     /**
      * {@inheritDoc}
      */
     @Override
-    public String getXMLTagName() { return getXMLElementTagName(); }
-
-    /**
-     * Gets the tag name of the root element representing this object.
-     *
-     * @return "scoutingMission".
-     */
-    public static String getXMLElementTagName() {
-        return "scoutingMission";
-    }
+    public String getXMLTagName() { return TAG; }
 }
