@@ -1,5 +1,5 @@
 /**
- *  Copyright (C) 2002-2015   The FreeCol Team
+ *  Copyright (C) 2002-2022   The FreeCol Team
  *
  *  This file is part of FreeCol.
  *
@@ -21,7 +21,6 @@ package net.sf.freecol.server.ai.mission;
 
 import java.util.ArrayList;
 import java.util.List;
-import java.util.Iterator;
 import java.util.logging.Logger;
 
 import javax.xml.stream.XMLStreamException;
@@ -29,8 +28,10 @@ import javax.xml.stream.XMLStreamException;
 import net.sf.freecol.common.FreeColException;
 import net.sf.freecol.common.io.FreeColXMLReader;
 import net.sf.freecol.common.io.FreeColXMLWriter;
+import net.sf.freecol.common.model.AbstractGoods;
 import net.sf.freecol.common.model.Colony;
 import net.sf.freecol.common.model.CombatModel;
+import static net.sf.freecol.common.model.Constants.*;
 import net.sf.freecol.common.model.Europe;
 import net.sf.freecol.common.model.Goods;
 import net.sf.freecol.common.model.GoodsType;
@@ -54,18 +55,22 @@ import net.sf.freecol.server.ai.Cargo;
 import net.sf.freecol.server.ai.EuropeanAIPlayer;
 import net.sf.freecol.server.ai.TransportableAIObject;
 
+import edu.umd.cs.findbugs.annotations.SuppressFBWarnings;
+
 
 /**
  * Mission for transporting units and goods on a carrier.
  *
  * @see net.sf.freecol.common.model.Unit Unit
  */
-public class TransportMission extends Mission {
+public final class TransportMission extends Mission {
 
     private static final Logger logger = Logger.getLogger(TransportMission.class.getName());
 
-    private static final String tag = "AI transport";
+    public static final String TAG = "transportMission";
 
+    private static final String tag = "AI transport";
+    
     private static enum CargoResult {
         TCONTINUE,  // Cargo should continue
         TDONE,      // Cargo completed successfully
@@ -82,7 +87,7 @@ public class TransportMission extends Mission {
 
     private static final int MINIMUM_GOLD_TO_STAY_IN_EUROPE = 600;
 
-    /** A list of <code>Cargo</code>s to work on. */
+    /** A list of {@code Cargo}s to work on. */
     private final List<Cargo> cargoes = new ArrayList<>();
 
     /** The current target location to travel to. */
@@ -90,21 +95,22 @@ public class TransportMission extends Mission {
 
 
     /**
-     * Creates a mission for the given <code>AIUnit</code>.
+     * Creates a mission for the given {@code AIUnit}.
      *
      * @param aiMain The main AI-object.
-     * @param aiUnit The <code>AIUnit</code> this mission is created for.
+     * @param aiUnit The {@code AIUnit} this mission is created for.
      */
     public TransportMission(AIMain aiMain, AIUnit aiUnit) {
-        super(aiMain, aiUnit, aiUnit.getTrivialTarget());
+        super(aiMain, aiUnit);
+
+        setTarget(aiUnit.getTrivialTarget());
     }
 
     /**
-     * Creates a new <code>TransportMission</code> and reads the given
-     * element.
+     * Creates a new {@code TransportMission} from a reader.
      *
      * @param aiMain The main AI-object.
-     * @param aiUnit The <code>AIUnit</code> this mission is created for.
+     * @param aiUnit The {@code AIUnit} this mission is created for.
      * @param xr The input stream containing the XML.
      * @throws XMLStreamException if a problem was encountered during parsing.
      * @see net.sf.freecol.server.ai.AIObject#readFromXML
@@ -118,13 +124,11 @@ public class TransportMission extends Mission {
 
 
     /**
-     * Disposes of this <code>Mission</code>.
+     * Disposes of this {@code Mission}.
      */
     @Override
     public void dispose() {
-        logger.finest(tag + " disposing (" + clearCargoes() + "): " + this
-            //+ "\n" + net.sf.freecol.common.debug.FreeColDebugger.stackTraceToString()
-            );
+        logger.finest(tag + " disposing (" + clearCargoes() + "): " + this);
         super.dispose();
     }
 
@@ -133,9 +137,9 @@ public class TransportMission extends Mission {
 
     /**
      * Checks if the carrier using this mission is carrying the given
-     * <code>TransportableAIObject</code>.
+     * {@code TransportableAIObject}.
      *
-     * @param t The <code>TransportableAIObject</code> to check.
+     * @param t The {@code TransportableAIObject} to check.
      * @return True if the carrier is carrying the transportable.
      */
     private boolean isCarrying(TransportableAIObject t) {
@@ -145,7 +149,7 @@ public class TransportMission extends Mission {
     /**
      * Is a transportable waiting for delivery on the cargoes list?
      *
-     * @param t The <code>TransportableAIObject</code> to check.
+     * @param t The {@code TransportableAIObject} to check.
      * @return True if the transportable is queued in this mission.
      */
     public boolean isTransporting(TransportableAIObject t) {
@@ -158,7 +162,7 @@ public class TransportMission extends Mission {
      *
      * FIXME: magic numbers to the spec.
      *
-     * @param other The other <code>Unit</code> to attack.
+     * @param other The other {@code Unit} to attack.
      * @return True if the attack should proceed.
      */
     private boolean shouldAttack(Unit other) {
@@ -213,9 +217,7 @@ public class TransportMission extends Mission {
         List<Cargo> old = tCopy();
         synchronized (cargoes) {
             cargoes.clear();
-            for (Cargo c : nxt) {
-                if (c.isValid()) cargoes.add(c);
-            }
+            cargoes.addAll(transform(nxt, Cargo::isValid));
             if (setSpace) tSpace();
         }
         tRetarget();
@@ -236,11 +238,11 @@ public class TransportMission extends Mission {
     }
 
     /**
-     * Find a <code>Cargo</code> with the given
-     * <code>TransportableAIObject</code>.
+     * Find a {@code Cargo} with the given
+     * {@code TransportableAIObject}.
      *
-     * @param t The <code>TransportableAIObject</code> to look for.
-     * @return The <code>Cargo</code> found, or null if none found.
+     * @param t The {@code TransportableAIObject} to look for.
+     * @return The {@code Cargo} found, or null if none found.
      */
     private Cargo tFind(TransportableAIObject t) {
         synchronized (cargoes) {
@@ -262,7 +264,7 @@ public class TransportMission extends Mission {
     /**
      * Adds a cargo to the cargoes list.
      *
-     * @param cargo The <code>Cargo</code> to add.
+     * @param cargo The {@code Cargo} to add.
      * @param index The position to add it.
      * @return True if the addition succeeded or the cargo was already present.
      */
@@ -286,7 +288,7 @@ public class TransportMission extends Mission {
     /**
      * Remove a cargo from the cargoes list.
      *
-     * @param cargo The <code>Cargo</code> to remove.
+     * @param cargo The {@code Cargo} to remove.
      * @return True if the remove succeeded.
      */
     private boolean tRemove(Cargo cargo) {
@@ -346,9 +348,10 @@ public class TransportMission extends Mission {
         Location now = null;
         int ret = 0;
         for (Cargo cargo : tCopy()) {
-            if (!Map.isSameLocation(now, cargo.getCarrierTarget())) {
+            Location t = cargo.getCarrierTarget();
+            if (!Map.isSameLocation(now, t)) {
                 ret++;
-                now = cargo.getCarrierTarget();
+                now = t;
             }
         }
         return ret;
@@ -367,7 +370,7 @@ public class TransportMission extends Mission {
     /**
      * If this carrier is the current carrier of a transportable, drop it.
      *
-     * @param t The <code>TransportableAIObject</code> to check.
+     * @param t The {@code TransportableAIObject} to check.
      */
     private void dropTransportable(TransportableAIObject t) {
         AIUnit carrier = getAIUnit();
@@ -378,7 +381,7 @@ public class TransportMission extends Mission {
      * If this carrier is the not the current carrier of a
      * transportable, make it so.
      *
-     * @param t The <code>TransportableAIObject</code> to check.
+     * @param t The {@code TransportableAIObject} to check.
      */
     private void takeTransportable(TransportableAIObject t) {
         AIUnit carrier = getAIUnit();
@@ -391,8 +394,8 @@ public class TransportMission extends Mission {
      * Public so that mobile transportables (units) can move to the
      * collection point.
      *
-     * @param t The <code>TransportableAIObject</code> to collect.
-     * @return The collection <code>Location<code>, or null if not found.
+     * @param t The {@code TransportableAIObject} to collect.
+     * @return The collection {@code Location}, or null if not found.
      */
     public Location getTransportTarget(TransportableAIObject t) {
         if (isCarrying(t)) return null;
@@ -407,7 +410,7 @@ public class TransportMission extends Mission {
      * transport if they find themselves better able to get there
      * themselves.
      *
-     * @param t The <code>TransportableAIObject</code> to collect.
+     * @param t The {@code TransportableAIObject} to collect.
      * @return The expected transport turns.
      */
     public int getTransportTurns(TransportableAIObject t) {
@@ -436,7 +439,7 @@ public class TransportMission extends Mission {
     /**
      * Unwrap a wrapped list of cargoes.
      *
-     * @param ts The list of <code>Cargo</code>s to unwrap.
+     * @param ts The list of {@code Cargo}s to unwrap.
      * @return The unwrapped list of cargoes.
      */
     private List<Cargo> unwrapCargoes(List<Cargo> ts) {
@@ -457,13 +460,14 @@ public class TransportMission extends Mission {
      * @return A message about the cargoes being cleared.
      */
     private String clearCargoes() {
-        String log = "cargoes cleared: ";
+        StringBuilder sb = new StringBuilder(64);
+        sb.append("cargoes cleared: ");
         for (Cargo cargo : tClear()) {
             dropTransportable(cargo.getTransportable());
-            log += " " + cargo;
+            sb.append(' ').append(cargo);
         }
         tRetarget();
-        return log;
+        return sb.toString();
     }
 
     /**
@@ -478,15 +482,15 @@ public class TransportMission extends Mission {
     /**
      * For a given transportable, work out where the carrier has to go to
      * advance the cargo (target), and what to do there (mode), allowing
-     * a new <code>Cargo</code> to be defined.
+     * a new {@code Cargo} to be defined.
      *
      * AIUnit cargo is harder than AIGoods, because AIUnits might have their
      * own inland paths, and thus we need to consider drop nodes.
      *
-     * @param t The <code>TransportableAIObject</code> to consider.
-     * @param lb A <code>LogBuilder</code> to log to.
-     * @return A new <code>Cargo</code> defining the action to take
-     *     with the <code>TransportableAIObject</code>, or null if impossible.
+     * @param t The {@code TransportableAIObject} to consider.
+     * @param lb A {@code LogBuilder} to log to.
+     * @return A new {@code Cargo} defining the action to take
+     *     with the {@code TransportableAIObject}, or null if impossible.
      */
     public Cargo makeCargo(TransportableAIObject t, LogBuilder lb) {
         final Unit carrier = getUnit();
@@ -535,9 +539,9 @@ public class TransportMission extends Mission {
     /**
      * Add the given Cargo to the cargoes list.
      *
-     * @param cargo The <code>Cargo</code> to add.
+     * @param cargo The {@code Cargo} to add.
      * @param index The index of where to add the cargo.
-     * @param lb A <code>LogBuilder</code> to log to.
+     * @param lb A {@code LogBuilder} to log to.
      * @return True if the cargo was added.
      */
     private boolean addCargo(Cargo cargo, int index, LogBuilder lb) {
@@ -556,7 +560,7 @@ public class TransportMission extends Mission {
     /**
      * Removes the given Cargo from the cargoes list.
      *
-     * @param cargo The <code>Cargo</code> to remove.
+     * @param cargo The {@code Cargo} to remove.
      */
     private void removeCargo(Cargo cargo) {
         if (!tRemove(cargo)) {
@@ -568,7 +572,7 @@ public class TransportMission extends Mission {
     /**
      * Is there space available for a new cargo?
      *
-     * @param cargo The <code>Cargo</code> to check.
+     * @param cargo The {@code Cargo} to check.
      * @return True if there is space available for this cargo.
      */
     public boolean spaceAvailable(Cargo cargo) {
@@ -578,7 +582,7 @@ public class TransportMission extends Mission {
     /**
      * Is there space available for a new cargo?
      *
-     * @param t The <code>TransportableAIObject</code> to check.
+     * @param t The {@code TransportableAIObject} to check.
      * @return True if there is space available for this transportable.
      */
     public boolean spaceAvailable(TransportableAIObject t) {
@@ -600,17 +604,15 @@ public class TransportMission extends Mission {
      * the same target, but do not break the space restrictions.  If
      * this does not work, it has to go at the end.
      *
-     * @param cargo The new <code>Cargo</code> to add.
+     * @param cargo The new {@code Cargo} to add.
      * @param requireMatch Fail if an existing destination is not matched.
-     * @param lb A <code>LogBuilder</code> to log to.
+     * @param lb A {@code LogBuilder} to log to.
      * @return True if the cargo was queued.
      */
     private boolean queueCargo(Cargo cargo, boolean requireMatch,
                                LogBuilder lb) {
         final Unit carrier = getUnit();
-        final int maxHolds = carrier.getCargoCapacity();
         final List<Cargo> ts = tCopy();
-        final int newSpace = cargo.getNewSpace();
         int candidate = -1;
 
         if (ts.isEmpty() // Trivial case
@@ -621,11 +623,11 @@ public class TransportMission extends Mission {
         }
 
         if (candidate < 0) { // Match an existing target?
-            outer: for (int i = 0; i < ts.size(); i++) {
+            for (int i = 0; i < ts.size(); i++) {
                 Cargo tr = ts.get(i);
                 if (Map.isSameLocation(tr.getCarrierTarget(),
                                        cargo.getCarrierTarget())) {
-                    if (!cargo.canQueueAt(carrier, i, ts)) continue outer;
+                    if (!cargo.canQueueAt(carrier, i, ts)) continue;
                     candidate = i;
                     break;
                 }
@@ -642,12 +644,12 @@ public class TransportMission extends Mission {
     /**
      * Dump a currently carried cargo.
      *
-     * @param cargo The <code>Cargo</code> to dump.
-     * @param lb A <code>LogBuilder</code> to log to.
+     * @param cargo The {@code Cargo} to dump.
+     * @param lb A {@code LogBuilder} to log to.
      * @return True if the cargo is no longer on board and not on the
      *     transport list, or is on board but is scheduled to be dumped.
      */
-    public boolean dumpCargo(Cargo cargo, LogBuilder lb) {
+    private boolean dumpCargo(Cargo cargo, LogBuilder lb) {
         TransportableAIObject t = cargo.getTransportable();
         if (isCarrying(t)) t.leaveTransport();
         if (!isCarrying(t) && tFind(t) != null) removeCargo(cargo);
@@ -666,11 +668,11 @@ public class TransportMission extends Mission {
     /**
      * Requeue an existing cargo.  Typically done when the target changes.
      *
-     * @param cargo The <code>Cargo</code> to requeue.
-     * @param lb A <code>LogBuilder</code> to log to.
+     * @param cargo The {@code Cargo} to requeue.
+     * @param lb A {@code LogBuilder} to log to.
      * @return True if the queuing succeeded.
      */
-    public boolean requeueCargo(Cargo cargo, LogBuilder lb) {
+    private boolean requeueCargo(Cargo cargo, LogBuilder lb) {
         final TransportableAIObject t = cargo.getTransportable();
         boolean ret = false;
         assert tFind(t) == cargo;
@@ -699,7 +701,7 @@ public class TransportMission extends Mission {
      * cargo on board should be on the cargoes list but the list is
      * not necessarily going to be in a sensible order.
      *
-     * @param lb A <code>LogBuilder</code> to log to.
+     * @param lb A {@code LogBuilder} to log to.
      */
     private void checkCargoes(LogBuilder lb) {
         final Unit carrier = getUnit();
@@ -707,7 +709,7 @@ public class TransportMission extends Mission {
         final AIUnit aiCarrier = getAIUnit();
 
         List<Unit> unitsPresent = carrier.getUnitList();
-        List<Goods> goodsPresent = carrier.getCompactGoods();
+        List<Goods> goodsPresent = carrier.getCompactGoodsList();
         List<TransportableAIObject> todo = new ArrayList<>();
         List<TransportableAIObject> drop = new ArrayList<>();
 
@@ -718,7 +720,7 @@ public class TransportMission extends Mission {
         for (Cargo cargo : tCopy()) {
             dump = false;
             TransportableAIObject t = cargo.getTransportable();
-            reason = invalidReason(aiCarrier, cargo.getCarrierTarget());
+            reason = invalidMissionReason(aiCarrier, cargo.getCarrierTarget());
             if (reason != null || (reason = cargo.check(aiCarrier)) != null) {
                 // Just remove, it is invalid
                 removeCargo(cargo);
@@ -759,14 +761,8 @@ public class TransportMission extends Mission {
             } else if (t instanceof AIGoods) {
                 Goods goods = ((AIGoods)t).getGoods();
                 if (goods != null) {
-                    Iterator<Goods> gi = goodsPresent.iterator();
-                    while (gi.hasNext()) {
-                        Goods g = gi.next();
-                        if (g.getType() == goods.getType()) {
-                            gi.remove();
-                            break;
-                        }
-                    }
+                    removeInPlace(goodsPresent,
+                                  AbstractGoods.matches(goods.getType()));
                 }
             }
         }
@@ -825,17 +821,18 @@ public class TransportMission extends Mission {
     }
 
     /**
-     * Check a <code>Cargo</code> for continued validity and
+     * Check a {@code Cargo} for continued validity and
      * whether action is needed at the current location.
      *
-     * @param cargo The <code>Cargo</code> to check.
-     * @param lb A <code>LogBuilder</code> to log to.
-     * @return TCONTINUE if the <code>Cargo</code> should continue,
+     * @param cargo The {@code Cargo} to check.
+     * @param lb A {@code LogBuilder} to log to.
+     * @return TCONTINUE if the {@code Cargo} should continue,
      *     TDONE if it has completed,
      *     TFAIL if it has failed,
      *     TNEXT if it has progressed to the next stage,
      *     TRETRY if a blockage has occurred and it should be retried,
      */
+    @SuppressFBWarnings(value="SF_SWITCH_FALLTHROUGH")
     private CargoResult tryCargo(Cargo cargo, LogBuilder lb) {
         final Unit carrier = getUnit();
         final Location here = carrier.getLocation();
@@ -861,7 +858,7 @@ public class TransportMission extends Mission {
             if ((d = cargo.getJoinDirection()) == null) {
                 logger.warning("Null pickup direction"
                     + " for " + cargo.toShortString()
-                    + " at " + t.getLocation().toString()
+                    + " at " + t.getLocation()
                     + " to " + carrier);
                 return CargoResult.TFAIL;
             }                    
@@ -927,7 +924,6 @@ public class TransportMission extends Mission {
             // Fall through
         case UNLOAD:
             if (isCarrying(t) && !t.leaveTransport(d)) {
-                //lb.add(", ", t, " NO-LEAVE");
                 PathNode pn = t.getDeliveryPath(carrier, t.getTransportDestination());
                 lb.add(", ", t, " NO-LEAVE(", here, "~", cargo.getLeaveDirection(), "~", t.getTransportDestination(), " ", ((pn == null) ? "no-path" : pn.fullPathToString()));
                 return CargoResult.TRETRY;
@@ -960,7 +956,7 @@ public class TransportMission extends Mission {
      * Perform the transport load/unload operations on arrival at the
      * target for the top cargo.
      *
-     * @param lb A <code>LogBuilder</code> to log to.
+     * @param lb A {@code LogBuilder} to log to.
      */
     private void doTransport(LogBuilder lb) {
         final Unit unit = getUnit();
@@ -998,7 +994,7 @@ public class TransportMission extends Mission {
                     cargo.clear();
                     break;
                 case TNEXT: default:
-                    throw new IllegalStateException("Can not happen");
+                    throw new RuntimeException("Can not happen: " + result);
                 }
             }
             curr.clear();
@@ -1015,10 +1011,7 @@ public class TransportMission extends Mission {
                     ? tryCargo(cargo, lb)
                     : CargoResult.TCONTINUE;
                 switch (result) {
-                case TCONTINUE:
-                    cont.add(cargo);
-                    break;
-                case TNEXT:
+                case TCONTINUE: case TNEXT:
                     cont.add(cargo);
                     break;
                 case TRETRY:
@@ -1032,7 +1025,7 @@ public class TransportMission extends Mission {
                     cargo.clear();
                     break;
                 default:
-                    throw new IllegalStateException("Can not happen");
+                    throw new RuntimeException("Can not happen: " + result);
                 }
             }
 
@@ -1071,7 +1064,7 @@ public class TransportMission extends Mission {
 
     /**
      * Calculates a score for a proposed list of
-     * <code>Cargo</code>s using the current unit.  Disallows
+     * {@code Cargo}s using the current unit.  Disallows
      * routes that would overfill the carrier.
      *
      * Useful for comparing proposed cargo delivery routes.  The
@@ -1079,8 +1072,8 @@ public class TransportMission extends Mission {
      * to break ties we also consider the hold*turn product to reduce
      * the risk of losses due to enemy action.
      *
-     * @param initialLocation The initial <code>Location</code>.
-     * @param order An ordering of <code>Cargo</code>s.
+     * @param initialLocation The initial {@code Location}.
+     * @param order An ordering of {@code Cargo}s.
      * @return A score for the cargo ordering.
      */
     private float scoreCargoOrder(Location initialLocation, List<Cargo> order) {
@@ -1111,7 +1104,7 @@ public class TransportMission extends Mission {
      * Leaves the cargoes in the order they are expected to
      * execute, with valid spaceLeft values.
      *
-     * @param lb A <code>LogBuilder</code> to log to.
+     * @param lb A {@code LogBuilder} to log to.
      */
     private void optimizeCargoes(LogBuilder lb) {
         lb.add(", optimize");
@@ -1152,8 +1145,8 @@ public class TransportMission extends Mission {
     /**
      * What is the best transportable for a carrier to collect?
      *
-     * @param carrier The carrier <code>Unit</code> to consider.
-     * @return The best available new <code>Cargo</code>, or null if
+     * @param carrier The carrier {@code Unit} to consider.
+     * @return The best available new {@code Cargo}, or null if
      *     none found.
      */
     private Cargo getBestCargo(Unit carrier) {
@@ -1162,7 +1155,6 @@ public class TransportMission extends Mission {
         float bestDirectValue = 0.0f, bestFallbackValue = 0.0f;
         for (TransportableAIObject t : euaip.getUrgentTransportables()) {
             if (t.isDisposed() || !t.carriableBy(carrier)) continue;
-            Location loc = t.getTransportSource();
             Cargo cargo;
             try {
                 cargo = Cargo.newCargo(t, carrier);
@@ -1191,11 +1183,11 @@ public class TransportMission extends Mission {
     /**
      * Why would an TransportMission be invalid with the given unit?
      *
-     * @param aiUnit The <code>AIUnit</code> to test.
+     * @param aiUnit The {@code AIUnit} to test.
      * @return A reason why the mission would be invalid with the unit,
      *     or null if none found.
      */
-    private static String invalidMissionReason(AIUnit aiUnit) {
+    private static String invalidUnitReason(AIUnit aiUnit) {
         String reason = invalidAIUnitReason(aiUnit);
         return (reason != null)
             ? reason
@@ -1212,7 +1204,7 @@ public class TransportMission extends Mission {
      *   - step is disposed
      *   - step is a captured settlement
      *
-     * @param cargo The <code>Cargo</code> to test.
+     * @param cargo The {@code Cargo} to test.
      * @return A reason why the mission would be invalid,
      *     or null if none found.
      */
@@ -1224,14 +1216,24 @@ public class TransportMission extends Mission {
             : null;
     }
             
+    /**
+     * Why would this mission be invalid with the given AI unit?
+     *
+     * @param aiUnit The {@code AIUnit} to check.
+     * @return A reason for mission invalidity, or null if none found.
+     */
+    public static String invalidMissionReason(AIUnit aiUnit) {
+        return invalidUnitReason(aiUnit);
+    }
+
     /**
      * Why would this mission be invalid with the given AI unit and location?
      *
-     * @param aiUnit The <code>AIUnit</code> to check.
-     * @param loc The <code>Location</code> to check.
+     * @param aiUnit The {@code AIUnit} to check.
+     * @param loc The {@code Location} to check.
      * @return A reason for invalidity, or null if none found.
      */
-    public static String invalidReason(AIUnit aiUnit, Location loc) {
+    public static String invalidMissionReason(AIUnit aiUnit, Location loc) {
         String reason;
         return ((reason = invalidMissionReason(aiUnit)) != null)
             ? reason
@@ -1242,25 +1244,16 @@ public class TransportMission extends Mission {
             : Mission.TARGETINVALID;
     }
 
-    /**
-     * Why would this mission be invalid with the given AI unit?
-     *
-     * @param aiUnit The <code>AIUnit</code> to check.
-     * @return A reason for mission invalidity, or null if none found.
-     */
-    public static String invalidReason(AIUnit aiUnit) {
-        return invalidMissionReason(aiUnit);
-    }
 
     // End of cargoes handling.
 
     // Publically accessible routines to manipulate a TransportableAIObject.
 
     /**
-     * Removes the given <code>TransportableAIObject</code> from the
+     * Removes the given {@code TransportableAIObject} from the
      * cargo list.
      *
-     * @param t The <code>TransportableAIObject</code> to remove.
+     * @param t The {@code TransportableAIObject} to remove.
      * @return True if the removal succeeded.
      */
     public boolean removeTransportable(TransportableAIObject t) {
@@ -1271,8 +1264,8 @@ public class TransportMission extends Mission {
     /**
      * Retargets a transportable that should already be on board the carrier.
      *
-     * @param t The <code>TransportableAIObject</code> to retarget.
-     * @param lb A <code>LogBuilder</code> to log to.
+     * @param t The {@code TransportableAIObject} to retarget.
+     * @param lb A {@code LogBuilder} to log to.
      * @return True if the retargeting succeeded.
      */
     public boolean requeueTransportable(TransportableAIObject t,
@@ -1287,9 +1280,9 @@ public class TransportMission extends Mission {
      * Public for the benefit of EuropeanAIPlayer.allocateTransportables
      * and CashInTreasureTrain.doMission.
      *
-     * @param t The <code>TransportableAIObject</code> to add.
+     * @param t The {@code TransportableAIObject} to add.
      * @param requireMatch Fail if an existing destination is not matched.
-     * @param lb A <code>LogBuilder</code> to log to.
+     * @param lb A {@code LogBuilder} to log to.
      * @return True if the transportable was queued.
      */
     public boolean queueTransportable(TransportableAIObject t,
@@ -1301,8 +1294,8 @@ public class TransportMission extends Mission {
     /**
      * Dump a transportable.
      *
-     * @param t The <code>TransportableAIObject</code> to dump.
-     * @param lb A <code>LogBuilder</code> to log to.
+     * @param t The {@code TransportableAIObject} to dump.
+     * @param lb A {@code LogBuilder} to log to.
      * @return True if the transportable is no longer on board, queued, or
      *     was reset to be dumped at the next stop.
      */
@@ -1322,8 +1315,8 @@ public class TransportMission extends Mission {
      * collect this unit.  Useful for prioritizing treasure
      * collection.
      *
-     * @param aiu The <code>AIUnit</code> to collect.
-     * @param lb A <code>LogBuilder</code> to log to.
+     * @param aiu The {@code AIUnit} to collect.
+     * @param lb A {@code LogBuilder} to log to.
      * @return True if the unit was queued.
      */
     public boolean forceCollection(AIUnit aiu, LogBuilder lb) {
@@ -1337,8 +1330,8 @@ public class TransportMission extends Mission {
      * Suppress European trade in a type of goods which is about to be
      * boycotted.
      *
-     * @param type The <code>GoodsType</code> to suppress.
-     * @param lb A <code>LogBuilder</code> to log to.
+     * @param type The {@code GoodsType} to suppress.
+     * @param lb A {@code LogBuilder} to log to.
      */
     public void suppressEuropeanTrade(GoodsType type, LogBuilder lb) {
         for (Cargo c : tCopy()) {
@@ -1391,7 +1384,7 @@ public class TransportMission extends Mission {
     @Override
     public String invalidReason() {
         final AIUnit aiUnit = getAIUnit();
-        String reason = invalidReason(aiUnit, getTarget());
+        String reason = invalidMissionReason(aiUnit, getTarget());
         Cargo cargo;
         return (reason != null) ? reason
             : ((cargo = tFirst()) == null) ? null
@@ -1433,10 +1426,12 @@ public class TransportMission extends Mission {
                 }
                 break;
 
-            case MOVE_HIGH_SEAS: case MOVE_NO_MOVES:
-            case MOVE_NO_REPAIR: case MOVE_ILLEGAL:
+            case MOVE_HIGH_SEAS: case MOVE_NO_MOVES: case MOVE_ILLEGAL:
                 return lbWait(lb);
 
+            case MOVE_NO_REPAIR:
+                return lbFail(lb, false, AIUNITDIED);
+
             case MOVE_NO_TILE: // Another unit is blocking a river?
                 moveRandomly(tag, null);
                 return lbDodge(lb);
@@ -1467,13 +1462,22 @@ public class TransportMission extends Mission {
         }
     }
 
+    /**
+     * More verbose version of toString().
+     *
+     * @return A summary of this mission including its transportables.
+     */
+    public String toFullString() {
+        LogBuilder lb = new LogBuilder(64);
+        lb.add(this);
+        for (Cargo cargo : tCopy()) lb.add("\n  ->", cargo);
+        return lb.toString();
+    }
+
 
     // Serialization
 
     private static final String TARGET_TAG = "target";
-    // @compat 0.10.5
-    private static final String OLD_TRANSPORTABLE_TAG = "transportable";
-    // end @compat
 
 
     /**
@@ -1534,44 +1538,17 @@ public class TransportMission extends Mission {
     protected void readChild(FreeColXMLReader xr) throws XMLStreamException {
         final String tag = xr.getLocalName();
 
-        if (Cargo.getXMLElementTagName().equals(tag)) {
+        if (Cargo.TAG.equals(tag)) {
             tAdd(new Cargo(getAIMain(), xr), -1);
 
-        // @compat 0.10.5
-        } else if (OLD_TRANSPORTABLE_TAG.equals(tag)) {
-            // Ignore the old format, let checkCargoes sort it out
-            xr.closeTag(OLD_TRANSPORTABLE_TAG);
-        // end @compat
-
         } else {
             super.readChild(xr);
         }
     }
 
-    /**
-     * More verbose version of toString().
-     *
-     * @return A summary of this mission including its transportables.
-     */
-    public String toFullString() {
-        LogBuilder lb = new LogBuilder(64);
-        lb.add(this);
-        for (Cargo cargo : tCopy()) lb.add("\n  ->", cargo);
-        return lb.toString();
-    }
-
     /**
      * {@inheritDoc}
      */
     @Override
-    public String getXMLTagName() { return getXMLElementTagName(); }
-
-    /**
-     * Gets the tag name of the root element representing this object.
-     *
-     * @return "transportMission".
-     */
-    public static String getXMLElementTagName() {
-        return "transportMission";
-    }
+    public String getXMLTagName() { return TAG; }
 }
