@@ -1,5 +1,5 @@
 /**
- *  Copyright (C) 2002-2015   The FreeCol Team
+ *  Copyright (C) 2002-2022   The FreeCol Team
  *
  *  This file is part of FreeCol.
  *
@@ -28,58 +28,61 @@ import net.sf.freecol.common.io.FreeColXMLWriter;
 import net.sf.freecol.common.model.Ability;
 import net.sf.freecol.common.model.Colony;
 import net.sf.freecol.common.model.CombatModel;
-import net.sf.freecol.common.model.IndianSettlement;
 import net.sf.freecol.common.model.Location;
 import net.sf.freecol.common.model.Direction;
 import net.sf.freecol.common.model.PathNode;
 import net.sf.freecol.common.model.Settlement;
-import net.sf.freecol.common.model.Tension;
 import net.sf.freecol.common.model.Tile;
 import net.sf.freecol.common.model.Unit;
 import net.sf.freecol.common.model.pathfinding.CostDeciders;
 import net.sf.freecol.common.model.pathfinding.GoalDecider;
+import static net.sf.freecol.common.util.CollectionUtils.*;
 import net.sf.freecol.common.util.LogBuilder;
 import net.sf.freecol.server.ai.AIMain;
 import net.sf.freecol.server.ai.AIMessage;
 import net.sf.freecol.server.ai.AIUnit;
+import net.sf.freecol.server.ai.MissionAIPlayer;
 
 
 /**
  * Mission for attacking a specific target, be it a Unit or a Settlement.
  */
-public class UnitSeekAndDestroyMission extends Mission {
+public final class UnitSeekAndDestroyMission extends Mission {
 
     private static final Logger logger = Logger.getLogger(UnitSeekAndDestroyMission.class.getName());
 
+    public static final String TAG = "unitSeekAndDestroyMission";
+
     /** The tag for this mission. */
     private static final String tag = "AI seek+destroyer";
 
     /**
      * The object we are trying to destroy. This can be a
-     * either <code>Settlement</code> or a <code>Unit</code>.
+     * either {@code Settlement} or a {@code Unit}.
      */
     private Location target, transportTarget;
 
 
     /**
-     * Creates a mission for the given <code>AIUnit</code>.
+     * Creates a mission for the given {@code AIUnit}.
      *
      * @param aiMain The main AI-object.
-     * @param aiUnit The <code>AIUnit</code> this mission is created for.
+     * @param aiUnit The {@code AIUnit} this mission is created for.
      * @param target The object we are trying to destroy.  This can be
-     *     either a <code>Settlement</code> or a <code>Unit</code>.
+     *     either a {@code Settlement} or a {@code Unit}.
      */
     public UnitSeekAndDestroyMission(AIMain aiMain, AIUnit aiUnit,
                                      Location target) {
-        super(aiMain, aiUnit, target);
+        super(aiMain, aiUnit);
+
+        setTarget(target);
     }
 
     /**
-     * Creates a new <code>UnitSeekAndDestroyMission</code> and reads
-     * the given element.
+     * Creates a new {@code UnitSeekAndDestroyMission} from a reader.
      *
      * @param aiMain The main AI-object.
-     * @param aiUnit The <code>AIUnit</code> this mission is created for.
+     * @param aiUnit The {@code AIUnit} this mission is created for.
      * @param xr The input stream containing the XML.
      * @throws XMLStreamException if a problem was encountered
      *      during parsing.
@@ -96,8 +99,8 @@ public class UnitSeekAndDestroyMission extends Mission {
     /**
      * Extract a valid target for this mission from a path.
      *
-     * @param aiUnit The <code>AIUnit</code> to perform the mission.
-     * @param path A <code>PathNode</code> to extract a target from.
+     * @param aiUnit The {@code AIUnit} to perform the mission.
+     * @param path A {@code PathNode} to extract a target from.
      * @return A target for this mission, or null if none found.
      */
     public static Location extractTarget(AIUnit aiUnit, PathNode path) {
@@ -107,10 +110,10 @@ public class UnitSeekAndDestroyMission extends Mission {
         Unit u;
         return (loc == null || aiUnit == null || aiUnit.getUnit() == null) 
             ? null
-            : (invalidReason(aiUnit, loc.getSettlement()) == null)
+            : (invalidSettlementReason(aiUnit, loc.getSettlement()) == null)
             ? loc.getSettlement()
             : ((t = loc.getTile()) != null
-                && invalidReason(aiUnit,
+                && invalidTargetReason(aiUnit,
                     u = t.getDefendingUnit(aiUnit.getUnit())) == null)
             ? u
             : null;
@@ -125,9 +128,9 @@ public class UnitSeekAndDestroyMission extends Mission {
      * FIXME: if we are the REF and there is a significant Tory
      * population inside, assume traitors have briefed us.
      *
-     * @param aiUnit The <code>AIUnit</code> to do the mission.
-     * @param path The <code>PathNode</code> to take to the settlement.
-     * @param settlement The <code>Settlement</code> to attack.
+     * @param aiUnit The {@code AIUnit} to do the mission.
+     * @param path The {@code PathNode} to take to the settlement.
+     * @param settlement The {@code Settlement} to attack.
      * @return A score of the desirability of the mission.
      */
     private static int scoreSettlementPath(AIUnit aiUnit, PathNode path,
@@ -144,34 +147,23 @@ public class UnitSeekAndDestroyMission extends Mission {
         final double off = combatModel.getOffencePower(unit, settlement);
         value += (int)Math.round(off * 50);
 
-        if (settlement instanceof Colony) {
-            // Favour high population (more loot:-).
-            Colony colony = (Colony) settlement;
-            value += 50 * colony.getUnitCount();
-            if (colony.hasStockade()) { // Avoid fortifications.
-                value -= 200 * colony.getStockade().getLevel();
-            }
-        } else if (settlement instanceof IndianSettlement) {
-            // Favour the most hostile settlements
-            IndianSettlement is = (IndianSettlement) settlement;
-            Tension tension = is.getAlarm(unit.getOwner());
-            if (tension != null) value += tension.getValue() / 2;
-        }
-        return aiUnit.getAIOwner().adjustMission(aiUnit, path, 
-            UnitSeekAndDestroyMission.class, value);
+        value = settlement.calculateSettlementValue(value, unit);
+
+        return ((MissionAIPlayer)aiUnit.getAIOwner()).adjustMission(
+            aiUnit, path, UnitSeekAndDestroyMission.class, value);
     }
 
     /**
      * Scores a potential attack on a unit.
      *
-     * @param aiUnit The <code>AIUnit</code> to do the mission.
-     * @param path The <code>PathNode</code> to take to the settlement.
-     * @param defender The <code>Unit</code> to attack.
+     * @param aiUnit The {@code AIUnit} to do the mission.
+     * @param path The {@code PathNode} to take to the settlement.
+     * @param defender The {@code Unit} to attack.
      * @return A score of the desirability of the mission.
      */
     private static int scoreUnitPath(AIUnit aiUnit, PathNode path,
                                      Unit defender) {
-        if (invalidUnitReason(aiUnit, defender) != null) {
+        if (invalidMissionReason(aiUnit, defender) != null) {
             return Integer.MIN_VALUE;
         }
         final Unit unit = aiUnit.getUnit();
@@ -183,13 +175,12 @@ public class UnitSeekAndDestroyMission extends Mission {
         if (tile == null || off <= 0) return Integer.MIN_VALUE;
 
         int value = 1020 - turns * 100;
-        value += 100 * (off - def);
+        value += (int)(100 * (off - def));
 
         // Add a big bonus for treasure trains on the tile.
         // Do not cheat and look at the value.
-        value += 1000 * tile.getUnitList().stream()
-            .filter(u -> u.canCarryTreasure() && u.getTreasureAmount() > 0)
-            .count();
+        value += 1000 * count(tile.getUnits(),
+            u -> u.canCarryTreasure() && u.getTreasureAmount() > 0);
 
         if (defender.isNaval()) {
             if (tile.isLand()) value += 500; // Easy win
@@ -197,16 +188,16 @@ public class UnitSeekAndDestroyMission extends Mission {
             if (defender.hasAbility(Ability.EXPERT_SOLDIER)
                 && !defender.isArmed()) value += 100;
         }
-        return aiUnit.getAIOwner().adjustMission(aiUnit, path,
-            UnitSeekAndDestroyMission.class, value);
+        return ((MissionAIPlayer)aiUnit.getAIOwner()).adjustMission(
+            aiUnit, path, UnitSeekAndDestroyMission.class, value);
     }
 
     /**
      * Evaluate a potential seek and destroy mission for a given unit
      * to a given tile.
      *
-     * @param aiUnit The <code>AIUnit</code> to do the mission.
-     * @param path A <code>PathNode</code> to take to the target.
+     * @param aiUnit The {@code AIUnit} to do the mission.
+     * @param path A {@code PathNode} to take to the target.
      * @return A score for the proposed mission.
      */
     public static int scorePath(AIUnit aiUnit, PathNode path) {
@@ -219,15 +210,15 @@ public class UnitSeekAndDestroyMission extends Mission {
     }
 
     /**
-     * Gets a <code>GoalDecider</code> for finding the best colony
-     * <code>Tile</code>, optionally falling back to the nearest colony.
+     * Gets a {@code GoalDecider} for finding the best colony
+     * {@code Tile}, optionally falling back to the nearest colony.
      *
-     * @param aiUnit The <code>AIUnit</code> that is searching.
+     * @param aiUnit The {@code AIUnit} that is searching.
      * @param deferOK Not used in this mission.
-     * @return A suitable <code>GoalDecider</code>.
+     * @return A suitable {@code GoalDecider}.
      */
     private static GoalDecider getGoalDecider(final AIUnit aiUnit,
-                                              boolean deferOK) {
+        @SuppressWarnings("unused") boolean deferOK) {
         return new GoalDecider() {
                 private PathNode bestPath = null;
                 private int bestValue = Integer.MIN_VALUE;
@@ -252,13 +243,13 @@ public class UnitSeekAndDestroyMission extends Mission {
     /**
      * Finds a suitable seek-and-destroy target path for an AI unit.
      *
-     * @param aiUnit The <code>AIUnit</code> to find a target for.
+     * @param aiUnit The {@code AIUnit} to find a target for.
      * @param range An upper bound on the number of moves.
      * @param deferOK Not implemented in this mission.
      * @return A path to the target, or null if none found.
      */
-    public static PathNode findTargetPath(AIUnit aiUnit, int range,
-                                          boolean deferOK) {
+    private static PathNode findTargetPath(AIUnit aiUnit, int range,
+        @SuppressWarnings("unused") boolean deferOK) {
         if (invalidAIUnitReason(aiUnit) != null) return null;
         final Unit unit = aiUnit.getUnit();
         final Location start = unit.getPathStartLocation();
@@ -272,13 +263,13 @@ public class UnitSeekAndDestroyMission extends Mission {
     /**
      * Finds a suitable seek-and-destroy target for an AI unit.
      *
-     * @param aiUnit The <code>AIUnit</code> to find a target for.
+     * @param aiUnit The {@code AIUnit} to find a target for.
      * @param range An upper bound on the number of moves.
      * @param deferOK Not implemented in this mission.
      * @return A suitable target, or null if none found.
      */
-    public static Location findTarget(AIUnit aiUnit, int range,
-                                      boolean deferOK) {
+    public static Location findMissionTarget(AIUnit aiUnit, int range,
+                                             boolean deferOK) {
         PathNode path = findTargetPath(aiUnit, range, deferOK);
         return (path != null) ? extractTarget(aiUnit, path)
             : null;
@@ -287,11 +278,11 @@ public class UnitSeekAndDestroyMission extends Mission {
     /**
      * Why would a UnitSeekAndDestroyMission be invalid with the given unit.
      *
-     * @param aiUnit The <code>AIUnit</code> to test.
+     * @param aiUnit The {@code AIUnit} to test.
      * @return A reason why the mission would be invalid, or null if
      *     none found.
      */
-    private static String invalidMissionReason(AIUnit aiUnit) {
+    private static String invalidUnitReason(AIUnit aiUnit) {
         String reason = invalidAIUnitReason(aiUnit);
         return (reason != null)
             ? reason
@@ -306,8 +297,8 @@ public class UnitSeekAndDestroyMission extends Mission {
      * Why would a UnitSeekAndDestroyMission be invalid with the given unit
      * and settlement.
      *
-     * @param aiUnit The <code>AIUnit</code> to seek-and-destroy with.
-     * @param settlement The <code>Settlement</code> to test.
+     * @param aiUnit The {@code AIUnit} to seek-and-destroy with.
+     * @param settlement The {@code Settlement} to test.
      * @return A reason why the mission would be invalid, or null if
      *     none found.
      */
@@ -327,12 +318,12 @@ public class UnitSeekAndDestroyMission extends Mission {
      * Why would a UnitSeekAndDestroyMission be invalid with the given unit
      * and target unit.
      *
-     * @param aiUnit The <code>AIUnit</code> to seek-and-destroy with.
-     * @param unit The target <code>Unit</code> to test.
+     * @param aiUnit The {@code AIUnit} to seek-and-destroy with.
+     * @param unit The target {@code Unit} to test.
      * @return A reason why the mission would be invalid, or null if
      *     none found.
      */
-    private static String invalidUnitReason(AIUnit aiUnit, Unit unit) {
+    private static String invalidTargetReason(AIUnit aiUnit, Unit unit) {
         String reason = invalidTargetReason(unit);
         if (reason != null) return reason;
         final Tile tile = unit.getTile();
@@ -350,28 +341,28 @@ public class UnitSeekAndDestroyMission extends Mission {
     /**
      * Why would this mission be invalid with the given AI unit?
      *
-     * @param aiUnit The <code>AIUnit</code> to check.
+     * @param aiUnit The {@code AIUnit} to check.
      * @return A reason for mission invalidity, or null if none found.
      */
-    public static String invalidReason(AIUnit aiUnit) {
-        return invalidMissionReason(aiUnit);
+    public static String invalidMissionReason(AIUnit aiUnit) {
+        return invalidUnitReason(aiUnit);
     }
 
     /**
      * Why would this mission be invalid with the given AI unit and location?
      *
-     * @param aiUnit The <code>AIUnit</code> to check.
-     * @param loc The <code>Location</code> to check.
+     * @param aiUnit The {@code AIUnit} to check.
+     * @param loc The {@code Location} to check.
      * @return A reason for invalidity, or null if none found.
      */
-    public static String invalidReason(AIUnit aiUnit, Location loc) {
+    public static String invalidMissionReason(AIUnit aiUnit, Location loc) {
         String reason = invalidMissionReason(aiUnit);
         return (reason != null)
             ? reason                
             : (loc instanceof Settlement)
             ? invalidSettlementReason(aiUnit, (Settlement)loc)
             : (loc instanceof Unit)
-            ? invalidUnitReason(aiUnit, (Unit)loc)
+            ? invalidTargetReason(aiUnit, (Unit)loc)
             : Mission.TARGETINVALID;
     }
 
@@ -437,7 +428,7 @@ public class UnitSeekAndDestroyMission extends Mission {
      */
     @Override
     public Location findTarget() {
-        return findTarget(getAIUnit(), 4, false);
+        return findMissionTarget(getAIUnit(), 4, false);
     }
 
     /**
@@ -445,7 +436,7 @@ public class UnitSeekAndDestroyMission extends Mission {
      */
     @Override
     public String invalidReason() {
-        return invalidReason(getAIUnit(), getTarget());
+        return invalidMissionReason(getAIUnit(), getTarget());
     }
 
     /**
@@ -475,7 +466,7 @@ public class UnitSeekAndDestroyMission extends Mission {
         // Is there a target-of-opportunity?
         final Unit unit = getUnit();
         Location nearbyTarget = (unit.isOnCarrier()) ? null
-            : findTarget(aiUnit, 1, false);
+            : findMissionTarget(aiUnit, 1, false);
         if (nearbyTarget != null) {
             if (getTarget() == null) {
                 setTarget(nearbyTarget);
@@ -504,10 +495,12 @@ public class UnitSeekAndDestroyMission extends Mission {
         Unit.MoveType mt = travelToTarget(currentTarget,
             CostDeciders.avoidSettlementsAndBlockingUnits(), lb);
         switch (mt) {
-        case MOVE_HIGH_SEAS: case MOVE_NO_MOVES:
-        case MOVE_NO_REPAIR: case MOVE_ILLEGAL:
+        case MOVE_HIGH_SEAS: case MOVE_NO_MOVES: case MOVE_ILLEGAL:
             return lbWait(lb);
 
+        case MOVE_NO_REPAIR:
+            return lbFail(lb, false, AIUNITDIED);
+
         case MOVE_NO_ACCESS_EMBARK: case MOVE_NO_TILE:
             return this;
 
@@ -523,7 +516,12 @@ public class UnitSeekAndDestroyMission extends Mission {
                               settlement);
             }
             Direction d = unitTile.getDirection(currentTarget.getTile());
-            assert d != null;
+            if (d == null) {
+                logger.warning("SDDM bogus " + mt + " with " + unit
+                    + " from " + unitTile + " to " + currentTarget
+                    + " at " + currentTarget.getTile());
+                return lbWait(lb);
+            }
             AIMessage.askAttack(aiUnit, d);
             return lbAttack(lb, currentTarget);
 
@@ -564,14 +562,5 @@ public class UnitSeekAndDestroyMission extends Mission {
      * {@inheritDoc}
      */
     @Override
-    public String getXMLTagName() { return getXMLElementTagName(); }
-
-    /**
-     * Gets the tag name of the root element representing this object.
-     *
-     * @return "unitSeekAndDestroyMission".
-     */
-    public static String getXMLElementTagName() {
-        return "unitSeekAndDestroyMission";
-    }
+    public String getXMLTagName() { return TAG; }
 }
