@@ -1,5 +1,5 @@
 /**
- *  Copyright (C) 2002-2015  The FreeCol Team
+ *  Copyright (C) 2002-2021  The FreeCol Team
  *
  *  This file is part of FreeCol.
  *
@@ -19,17 +19,27 @@
 
 package net.sf.freecol.common.model;
 
+import static net.sf.freecol.common.util.CollectionUtils.any;
+import static net.sf.freecol.common.util.CollectionUtils.count;
+import static net.sf.freecol.common.util.CollectionUtils.find;
+import static net.sf.freecol.common.util.CollectionUtils.first;
+import static net.sf.freecol.common.util.CollectionUtils.forEach;
+import static net.sf.freecol.common.util.CollectionUtils.iterable;
+import static net.sf.freecol.common.util.CollectionUtils.matchKey;
+import static net.sf.freecol.common.util.CollectionUtils.none;
+import static net.sf.freecol.common.util.CollectionUtils.toList;
+import static net.sf.freecol.common.util.CollectionUtils.transform;
+
+import java.io.IOException;
 import java.io.StringWriter;
 import java.util.Iterator;
 import java.util.List;
-import java.util.Set;
+
+import javax.xml.stream.XMLStreamException;
 
 import net.sf.freecol.common.io.FreeColXMLWriter;
-import net.sf.freecol.common.model.Modifier;
 import net.sf.freecol.common.model.Modifier.ModifierType;
 import net.sf.freecol.common.model.UnitLocation.NoAddReason;
-import net.sf.freecol.common.model.Direction;
-import net.sf.freecol.common.model.Tile;
 import net.sf.freecol.server.model.ServerBuilding;
 import net.sf.freecol.server.model.ServerUnit;
 import net.sf.freecol.util.test.FreeColTestCase;
@@ -39,28 +49,62 @@ public class BuildingTest extends FreeColTestCase {
 
     private static final BuildingType armoryType
         = spec().getBuildingType("model.building.armory");
-    private static final BuildingType blacksmithType
+    private static final BuildingType arsenalType
+        = spec().getBuildingType("model.building.arsenal");
+    private static final BuildingType blacksmithHouseType
         = spec().getBuildingType("model.building.blacksmithHouse");
+    private static final BuildingType blacksmithShopType
+        = spec().getBuildingType("model.building.blacksmithShop");
     private static final BuildingType carpenterHouseType
         = spec().getBuildingType("model.building.carpenterHouse");
+    private static final BuildingType cathedralType
+        = spec().getBuildingType("model.building.cathedral");
     private static final BuildingType chapelType
         = spec().getBuildingType("model.building.chapel");
+    private static final BuildingType churchType
+        = spec().getBuildingType("model.building.church");
+    private static final BuildingType cigarFactoryType
+        = spec().getBuildingType("model.building.cigarFactory");
     private static final BuildingType countryType
         = spec().getBuildingType("model.building.country");
     private static final BuildingType depotType
         = spec().getBuildingType("model.building.depot");
+    private static final BuildingType distillerHouseType
+        = spec().getBuildingType("model.building.distillerHouse");
     private static final BuildingType fortType
         = spec().getBuildingType("model.building.fort");
     private static final BuildingType fortressType
         = spec().getBuildingType("model.building.fortress");
+    private static final BuildingType furTraderHouseType
+        = spec().getBuildingType("model.building.furTraderHouse");
+    private static final BuildingType furTradingPostType
+        = spec().getBuildingType("model.building.furTradingPost");
+    private static final BuildingType furFactoryType
+        = spec().getBuildingType("model.building.furFactory");
+    private static final BuildingType ironWorksType
+        = spec().getBuildingType("model.building.ironWorks");
+    private static final BuildingType lumberMillType
+        = spec().getBuildingType("model.building.lumberMill");
+    private static final BuildingType magazineType
+        = spec().getBuildingType("model.building.magazine");
     private static final BuildingType newspaperType
         = spec().getBuildingType("model.building.newspaper");
     private static final BuildingType printingPressType
         = spec().getBuildingType("model.building.printingPress");
+    private static final BuildingType rumDistilleryType
+        = spec().getBuildingType("model.building.rumDistillery");
+    private static final BuildingType rumFactoryType
+        = spec().getBuildingType("model.building.rumFactory");
     private static final BuildingType schoolType
         = spec().getBuildingType("model.building.schoolhouse");
     private static final BuildingType stockadeType
         = spec().getBuildingType("model.building.stockade");
+    private static final BuildingType textileMillType
+        = spec().getBuildingType("model.building.textileMill");
+    private static final BuildingType tobacconistHouseType
+        = spec().getBuildingType("model.building.tobacconistHouse");
+    private static final BuildingType tobacconistShopType
+        = spec().getBuildingType("model.building.tobacconistShop");
     private static final BuildingType townHallType
         = spec().getBuildingType("model.building.townHall");
     private static final BuildingType universityType
@@ -69,6 +113,8 @@ public class BuildingTest extends FreeColTestCase {
         = spec().getBuildingType("model.building.warehouse");
     private static final BuildingType weaverHouseType
         = spec().getBuildingType("model.building.weaverHouse");
+    private static final BuildingType weaverShopType
+        = spec().getBuildingType("model.building.weaverShop");
 
     private static final GoodsType bellsType
         = spec().getGoodsType("model.goods.bells");
@@ -113,6 +159,8 @@ public class BuildingTest extends FreeColTestCase {
         = spec().getUnitType("model.unit.indenturedServant");
     private static final UnitType indianConvertType
         = spec().getUnitType("model.unit.indianConvert");
+    private static final UnitType masterBlacksmithType
+        = spec().getUnitType("model.unit.masterBlacksmith");
     private static final UnitType masterCarpenterType
         = spec().getUnitType("model.unit.masterCarpenter");
     private static final UnitType masterDistillerType
@@ -120,15 +168,101 @@ public class BuildingTest extends FreeColTestCase {
     private static final UnitType pettyCriminalType
         = spec().getUnitType("model.unit.pettyCriminal");
 
+    private static final Integer[] bonusLevels = new Integer[] { -2,-1,0,1,2 };
 
+
+    private void clearBuilding(Building b) {
+        final Colony colony = b.getColony();
+        for (Unit u : b.getUnitList()) {
+            u.setLocation(colony.getTile());
+            u.setWorkType(grainType);
+            colony.joinColony(u);
+        }
+        assertEquals(0, b.getUnitCount());
+    }
+        
+    private void productionTest(BuildingType[] buildings, int[][][]values) {
+        Game game = getGame();
+        game.changeMap(getTestMap(true));
+
+        // Enable factory
+        Colony colony = getStandardColony(8); // Big enough for factory
+        colony.getOwner().addFather(spec()
+            .getFoundingFather("model.foundingFather.adamSmith"));
+
+        Building building = colony.getBuilding(buildings[0]);
+        if (building == null) {
+            building = new ServerBuilding(game, colony, buildings[0]);
+            colony.addBuilding(building);
+        }
+        while (building.getType() != buildings[0]) building.upgrade();
+        clearBuilding(building);
+        // Add goods so production is viable
+        ProductionType pt = building.getBestProductionType(false, null);
+        assertNotNull(pt);
+        forEach(pt.getInputs(), ag -> colony.addGoods(ag.getType(), 50));
+        GoodsType outputType = first(pt.getOutputs()).getType();
+        UnitType[] unitTypes = new UnitType[] {
+            indianConvertType, pettyCriminalType, indenturedServantType,
+            freeColonistType, building.getExpertUnitType() };
+        Tile tile = colony.getTile();
+        for (UnitType ut : unitTypes) {
+            if (none(colony.getUnits(), matchKey(ut, Unit::getType))) {
+                Unit u = find(colony.getUnits(),
+                              matchKey(freeColonistType, Unit::getType));
+                u.changeType(ut);
+            }
+        }
+
+        StringBuilder sb = new StringBuilder(64);
+        for (int i = 0; i < values.length; i++) {
+            final BuildingType bt = buildings[i];
+
+            for (int j = 0; j < values[0].length; j++) {
+                final int level = bonusLevels[j];
+
+                for (int k = 0; k < values[0][0].length; k++) {
+                    final UnitType ut = unitTypes[k];
+
+                    // Set up conditions for bt,level,ut
+                    while (building.getType() != bt) building.upgrade();
+                    assertEquals(building.getType(), bt);
+                    clearBuilding(building);
+                    colony.setProductionBonus(level);
+                    Unit worker = find(colony.getUnits(),
+                                       matchKey(ut, Unit::getType));
+                    assertNotNull(worker);
+                    worker.setWorkType(outputType);
+                    worker.setLocation(building);
+                    assertEquals(1, building.getUnitCount());
+                    assertEquals(worker, building.getUnitList().get(0));
+                    assertEquals(outputType, worker.getWorkType());
+                    colony.invalidateCache();
+                    assertEquals(colony.getProductionBonus(), level);
+
+                    //System.err.println("PM " + outputType.getSuffix()+"/"+bt.getSuffix() + "/" + ut.getSuffix() + " PI=" + building.getProductionInfo());
+                    //forEach(building.getProductionModifiers(outputType, ut), m -> System.err.println("M=" + m));
+
+                    // At last! The test.
+                    sb.setLength(0);
+                    sb.append(outputType.getSuffix())
+                        .append(" produced in ").append(bt.getSuffix())
+                        .append(" at bonus ").append(String.valueOf(level))
+                        .append(" with ").append(ut.getSuffix());
+                    assertEquals(sb.toString(), values[i][j][k],
+                                 building.getTotalProductionOf(outputType));
+                }
+            }
+        }
+    }                            
+        
     public void testCanBuildNext() {
         Game game = getGame();
-        game.setMap(getTestMap(true));
-
-        Colony colony = getStandardColony();
+        game.changeMap(getTestMap(true));
 
         // First check with a building that can be fully built with a
         // normal colony
+        Colony colony = getStandardColony();
         Building warehouse = new ServerBuilding(getGame(), colony, depotType);
         colony.addBuilding(warehouse);
         assertTrue(warehouse.canBuildNext());
@@ -153,7 +287,8 @@ public class BuildingTest extends FreeColTestCase {
 
     public void testStockadeRequiresMinimumPopulation() {
         Game game = getGame();
-        game.setMap(getTestMap(true));
+        game.changeMap(getTestMap(true));
+
         Colony colony = getStandardColony(2);
         assertEquals(Colony.NoBuildReason.POPULATION_TOO_SMALL,
                      colony.getNoBuildReason(stockadeType, null));
@@ -167,7 +302,8 @@ public class BuildingTest extends FreeColTestCase {
 
     public void testFortRequiresMinimumPopulation() {
         Game game = getGame();
-        game.setMap(getTestMap(true));
+        game.changeMap(getTestMap(true));
+
         Colony colony = getStandardColony(2);
         assertEquals(Colony.NoBuildReason.POPULATION_TOO_SMALL,
                      colony.getNoBuildReason(fortType, null));
@@ -182,7 +318,7 @@ public class BuildingTest extends FreeColTestCase {
 
     public void testFortressRequiresMinimumPopulation() {
         Game game = getGame();
-        game.setMap(getTestMap(true));
+        game.changeMap(getTestMap(true));
 
         Colony colony = getStandardColony(7);
         colony.addBuilding(new ServerBuilding(game, colony, stockadeType));
@@ -200,10 +336,9 @@ public class BuildingTest extends FreeColTestCase {
 
     public void testInitialColony() {
         Game game = getGame();
-        game.setMap(getTestMap(true));
+        game.changeMap(getTestMap(true));
 
         Colony colony = getStandardColony();
-
         Building warehouse = colony.getBuilding(warehouseType);
 
         // Is build as depot...
@@ -219,7 +354,7 @@ public class BuildingTest extends FreeColTestCase {
 
     public void testChurch() {
         Game game = getGame();
-        game.setMap(getTestMap(true));
+        game.changeMap(getTestMap(true));
 
         // Make a colony big enough to build a cathedral, and add enough
         // liberty to zero the production bonus.
@@ -288,7 +423,7 @@ public class BuildingTest extends FreeColTestCase {
         assertEquals("Church total cross production, 3 x free colonist", 11,
             church.getTotalProductionOf(crossesType));
 
-        Building smithy = colony.getBuilding(blacksmithType);
+        Building smithy = colony.getBuilding(blacksmithHouseType);
         assertTrue(unit0.setLocation(smithy));
         assertTrue(unit1.setLocation(smithy));
         assertTrue(unit2.setLocation(smithy));
@@ -366,7 +501,7 @@ public class BuildingTest extends FreeColTestCase {
 
     public void testCanAddToBuilding() {
         Game game = getGame();
-        game.setMap(getTestMap(true));
+        game.changeMap(getTestMap(true));
 
         Colony colony = getStandardColony(6);
         Tile tile = colony.getTile();
@@ -410,11 +545,10 @@ public class BuildingTest extends FreeColTestCase {
      */
     public void testCanAddToSchool() {
         Game game = getGame();
-        game.setMap(getTestMap(true));
+        game.changeMap(getTestMap(true));
 
         Colony colony = getStandardColony(10);
-
-        Iterator<Unit> units = colony.getUnitIterator();
+        Iterator<Unit> units = colony.getUnitList().iterator();
 
         Unit farmer = units.next();
         farmer.setType(expertFarmerType);
@@ -537,66 +671,58 @@ public class BuildingTest extends FreeColTestCase {
 
     public void testSerialize() {
         Game game = getGame();
-        game.setMap(getTestMap(true));
+        game.changeMap(getTestMap(true));
 
         Colony colony = getStandardColony(6);
         for (Building building : colony.getBuildings()) {
-            try {
-                StringWriter sw = new StringWriter();
+            try (StringWriter sw = new StringWriter();
                 FreeColXMLWriter xw = new FreeColXMLWriter(sw,
-                    FreeColXMLWriter.WriteScope.toSave());
-
+                    FreeColXMLWriter.WriteScope.toSave(), false)) {
                 building.toXML(xw);
-
-                xw.close();
-
-            } catch (Exception e) {
-                fail(e.toString());
+            } catch (IOException|XMLStreamException ex) {
+                fail(ex.toString());
             }
         }
     }
 
     public void testStockade() {
         final Game game = getGame();
-        final Turn turn = game.getTurn();
-        game.setMap(getTestMap(true));
-        Set<Modifier> modifierSet;
+        game.changeMap(getTestMap(true));
 
+        final Turn turn = game.getTurn();
+        List<Modifier> modifiers;
         Colony colony = getStandardColony(2);
-        modifierSet = colony.getModifiers(Modifier.DEFENCE);
-        assertEquals(1, modifierSet.size());
-        Modifier modifier = modifierSet.iterator().next();
+        modifiers = toList(colony.getModifiers(Modifier.DEFENCE));
+        assertEquals(1, modifiers.size());
+        Modifier modifier = first(modifiers);
         assertEquals(50f, modifier.getValue());
         assertEquals(ModifierType.PERCENTAGE, modifier.getType());
 
-        modifierSet = stockadeType.getModifiers(Modifier.DEFENCE);
-        assertEquals(1, modifierSet.size());
-        modifier = modifierSet.iterator().next();
+        modifiers = toList(stockadeType.getModifiers(Modifier.DEFENCE));
+        assertEquals(1, modifiers.size());
+        modifier = first(modifiers);
         assertEquals(100f, modifier.getValue());
         assertEquals(ModifierType.PERCENTAGE, modifier.getType());
-        assertEquals(0f, stockadeType.applyModifiers(0f, turn,
-                Modifier.MINIMUM_COLONY_SIZE));
+        assertEquals(0f, stockadeType.apply(0f, turn, Modifier.MINIMUM_COLONY_SIZE));
 
-        modifierSet = fortType.getModifiers(Modifier.DEFENCE);
-        assertEquals(1, modifierSet.size());
-        modifier = modifierSet.iterator().next();
+        modifiers = toList(fortType.getModifiers(Modifier.DEFENCE));
+        assertEquals(1, modifiers.size());
+        modifier = first(modifiers);
         assertEquals(150f, modifier.getValue());
         assertEquals(ModifierType.PERCENTAGE, modifier.getType());
-        assertEquals(0f, stockadeType.applyModifiers(0f, turn,
-                Modifier.MINIMUM_COLONY_SIZE));
+        assertEquals(0f, stockadeType.apply(0f, turn, Modifier.MINIMUM_COLONY_SIZE));
 
-        modifierSet = fortressType.getModifiers(Modifier.DEFENCE);
-        assertEquals(1, modifierSet.size());
-        modifier = modifierSet.iterator().next();
+        modifiers = toList(fortressType.getModifiers(Modifier.DEFENCE));
+        assertEquals(1, modifiers.size());
+        modifier = first(modifiers);
         assertEquals(200f, modifier.getValue());
         assertEquals(ModifierType.PERCENTAGE, modifier.getType());
-        assertEquals(0f, stockadeType.applyModifiers(0f, turn,
-                Modifier.MINIMUM_COLONY_SIZE));
+        assertEquals(0f, stockadeType.apply(0f, turn, Modifier.MINIMUM_COLONY_SIZE));
     }
 
     public void testCottonClothProduction() {
         Game game = getGame();
-        game.setMap(getTestMap(true));
+        game.changeMap(getTestMap(true));
 
         Colony colony = getStandardColony(2);
         List<Unit> units = colony.getUnitList();
@@ -606,9 +732,9 @@ public class BuildingTest extends FreeColTestCase {
         final Building weaver = colony.getBuilding(weaverHouseType);
 
         assertTrue(colonist.getLocation() instanceof ColonyTile);
-        assertEquals(plainsType, ((ColonyTile)colonist.getLocation()).getWorkTile().getType());
+        assertEquals(plainsType, colonist.getWorkTile().getType());
         assertTrue(worker.getLocation() instanceof ColonyTile);
-        assertEquals(plainsType, ((ColonyTile)worker.getLocation()).getWorkTile().getType());
+        assertEquals(plainsType, worker.getWorkTile().getType());
 
         assertTrue(weaver.add(worker));
         assertEquals(worker, weaver.getUnitList().get(0));
@@ -634,7 +760,7 @@ public class BuildingTest extends FreeColTestCase {
 
     public void testPasture() {
         Game game = getGame();
-        game.setMap(getTestMap(true));
+        game.changeMap(getTestMap(true));
 
         Colony colony = getStandardColony(1);
         Building country = colony.getBuilding(countryType);
@@ -703,9 +829,9 @@ public class BuildingTest extends FreeColTestCase {
 
     public void testTownhallProduction() {
         final Game game = getGame();
-        game.setMap(getTestMap(true));
-        final Turn turn = game.getTurn();
+        game.changeMap(getTestMap(true));
 
+        final Turn turn = game.getTurn();
         Colony colony = getStandardColony(6);
         Player owner = colony.getOwner();
         Unit colonist = colony.getUnitList().get(0);
@@ -716,8 +842,8 @@ public class BuildingTest extends FreeColTestCase {
         Building building = colony.getBuilding(townHallType);
         for (Unit u : building.getUnitList()) u.setLocation(tile);
 
-        Set<Modifier> modifiers = colony.getModifiers("model.goods.bells");
-        assertTrue("No initial modifiers", modifiers.isEmpty());
+        assertTrue("No initial modifiers",
+                   none(colony.getModifiers("model.goods.bells")));
         assertEquals("Initial bell production", 1,
                      building.getTotalProductionOf(bellsType));
 
@@ -732,29 +858,29 @@ public class BuildingTest extends FreeColTestCase {
         // Add Jefferson.
         FoundingFather jefferson = spec()
             .getFoundingFather("model.foundingFather.thomasJefferson");
-        modifiers = jefferson.getModifiers("model.goods.bells");
+        List<Modifier> modifiers = toList(jefferson.getModifiers("model.goods.bells"));
         assertEquals("Jefferson modifier size", 1, modifiers.size());
-        Modifier bellsModifier = modifiers.iterator().next();
+        Modifier bellsModifier = first(modifiers);
         owner.addFather(jefferson);
 
         // Jefferson is a property of the player...
         assertTrue("Jefferson should be present in player",
-            owner.getModifiers("model.goods.bells")
-                .contains(bellsModifier));
+            any(owner.getModifiers("model.goods.bells"),
+                m -> m == bellsModifier));
         assertTrue("Jefferson should be present in player in building scope",
-            owner.getModifiers("model.goods.bells", townHallType, turn)
-                .contains(bellsModifier));
+            any(owner.getModifiers("model.goods.bells", townHallType, turn),
+                m -> m ==  bellsModifier));
         assertFalse("Jefferson should not be present in player in unit scope",
-            owner.getModifiers("model.goods.bells", freeColonistType, turn)
-                .contains(bellsModifier));
+            any(owner.getModifiers("model.goods.bells", freeColonistType, turn),
+                m -> m == bellsModifier));
         // ...not the colony,
         assertFalse("Jefferson modifier should not be present in colony",
-            colony.getModifiers("model.goods.bells")
-                  .contains(bellsModifier));
+            any(colony.getModifiers("model.goods.bells"),
+                m -> m == bellsModifier));
         // ...and the building modifiers do not have it.
         assertFalse("Jefferson modifier should not be present in building modifiers",
-            building.getModifiers("model.goods.bells")
-                    .contains(bellsModifier));
+            any(building.getModifiers("model.goods.bells"),
+                m -> m == bellsModifier));
 
         // 3(colonist)
         assertEquals("Production(Colonist/Jefferson)", 3,
@@ -799,7 +925,7 @@ public class BuildingTest extends FreeColTestCase {
 
     public void testPrintingPressBonus() {
         Game game = getGame();
-        game.setMap(getTestMap(true));
+        game.changeMap(getTestMap(true));
 
         Colony colony = getStandardColony(6);
         Unit unit = colony.getUnitList().get(0);
@@ -826,7 +952,7 @@ public class BuildingTest extends FreeColTestCase {
 
     public void testNewspaperBonus() {
         Game game = getGame();
-        game.setMap(getTestMap(true));
+        game.changeMap(getTestMap(true));
 
         Colony colony = getStandardColony(6);
         Unit unit = colony.getUnitList().get(0);
@@ -857,46 +983,130 @@ public class BuildingTest extends FreeColTestCase {
 
     public void testUnitProduction() {
         Game game = getGame();
-        game.setMap(getTestMap(true));
+        game.changeMap(getTestMap(true));
 
         Colony colony = getStandardColony(4);
         Unit unit = colony.getUnitList().get(0);
         for (Building building : colony.getBuildings()) {
             clearWorkLocation(building);
             unit.setLocation(building);
-            for (AbstractGoods output : building.getOutputs()) {
+            for (AbstractGoods output : iterable(building.getOutputs())) {
                 GoodsType outputType = output.getType();
-                for (UnitType type : spec().getUnitTypeList()) {
-                    if (!building.getType().canAdd(type)
-                        || !type.isAvailableTo(colony.getOwner())) continue;
+                for (UnitType type : transform(spec().getUnitTypeList(),
+                        ut -> (building.getType().canAdd(ut)
+                            && ut.isAvailableTo(colony.getOwner())))) {
                     unit.changeType(type);
                     if (output != null) {
                         int productivity = building.getUnitProduction(unit, outputType);
-                        int expected = output.getAmount();
-                        if (type == building.getExpertUnitType()) {
-                            expected = 6;
-                        } else if (type == indenturedServantType) {
-                            expected = 2;
-                        } else if (type == indianConvertType) {
-                            expected = 1;
-                        } else if (type == pettyCriminalType) {
-                            expected = 1;
-                        }
-                        if (expected != output.getAmount()) {
-                            assertFalse("ModifierSet should not be empty!",
-                                        type.getModifiers(outputType.getId()).isEmpty());
+                        int expect = (type == building.getExpertUnitType()) ? 6
+                            : (type == indenturedServantType) ? 2
+                            : (type == indianConvertType) ? 1
+                            : (type == pettyCriminalType) ? 1
+                            : output.getAmount();
+                        if (expect != output.getAmount()) {
+                            assertTrue("Modifiers expected",
+                                count(type.getModifiers(outputType.getId())) > 0);
                         }
                         assertEquals("Wrong productivity for " + type
-                            + " in " + building, expected, productivity);
+                            + " in " + building, expect, productivity);
                     }
                 }
             }
         }
     }
 
+    // Lumber and cross production data contributed by Lone_Wolf in BR#2981.
+    private static int lumberProd[][][] = {
+        { // carpenterHouse
+            { 0, 0, 0, 1, 4 }, // -2
+            { 0, 0, 1, 2, 5 }, // -1
+            { 1, 1, 2, 3, 6 }, // 0
+            { 2, 2, 3, 4, 7 }, // +1
+            { 3, 3, 4, 5, 8 }, // +2
+        }, { // lumber mill
+            { 0, 0, 0, 2, 8 }, // -2
+            { 0, 0, 2, 4, 10 }, // -1
+            { 2, 2, 4, 6, 12 }, // 0
+            { 4, 4, 6, 8, 14 }, // +1
+            { 6, 6, 8, 10, 16 }, // +2
+        }
+    };
+    private static final BuildingType[] lumberBuildings = new BuildingType[] {
+        carpenterHouseType, lumberMillType };
+    public void testLumberProduction() {
+        productionTest(lumberBuildings, lumberProd);
+    }
+
+    // Cross production is very like lumber production, with the addition of
+    // the unattended production.
+    private static int crossProd[][][] = {
+        { // church
+            { 2, 2, 2, 3, 6 }, // -2
+            { 2, 2, 3, 4, 7 }, // -1
+            { 3, 3, 4, 5, 8 }, // 0
+            { 4, 4, 5, 6, 9 }, // +1
+            { 5, 5, 6, 7, 10 }, // +2
+        }, { // cathedral
+            { 3, 3, 3, 5, 11 }, // -2
+            { 3, 3, 5, 7, 13 }, // -1
+            { 5, 5, 7, 9, 15 }, // 0
+            { 7, 7, 9, 11, 17 }, // +1
+            { 9, 9, 11, 13, 19 }, // +2
+        }
+    };
+    private static final BuildingType[] crossBuildings = new BuildingType[] {
+        churchType, cathedralType };
+    public void testCrossProduction() {
+        productionTest(crossBuildings, crossProd);
+    }
+    
+    // Factory production data contributed by Lone_Wolf in BR#2979.
+    private static int factoryProd[][][] = {
+        { // house
+            { 0, 0, 0, 1, 2 }, // -2
+            { 0, 0, 1, 2, 4 }, // -1
+            { 1, 1, 2, 3, 6 }, // 0
+            { 2, 2, 3, 4, 8 }, // 1
+            { 3, 3, 4, 5, 10 }, // 2
+        }, { // shop
+            { 0, 0, 2, 4, 8 }, // -2
+            { 1, 1, 3, 5, 10 }, // -1
+            { 2, 2, 4, 6, 12 }, // 0
+            { 3, 3, 5, 7, 14 }, // 1
+            { 4, 4, 6, 8, 16 }, // 2
+        }, { // factory
+            { 0, 0, 3, 6, 12 }, // -2
+            { 1, 1, 4, 7, 14 }, // -1
+            { 3, 3, 6, 9, 18 }, // 0
+            { 4, 4, 7, 10, 20 }, // 1
+            { 6, 6, 9, 12, 24 }, // 2
+        }
+    };
+    private static final BuildingType[] clothBuildings = new BuildingType[] {
+        weaverHouseType, weaverShopType, textileMillType };
+    private static final BuildingType[] furBuildings = new BuildingType[] {
+        furTraderHouseType, furTradingPostType, furFactoryType };
+    private static final BuildingType[] sugarBuildings = new BuildingType[] {
+        distillerHouseType, rumDistilleryType, rumFactoryType };
+    private static final BuildingType[] tobaccoBuildings = new BuildingType[] {
+        tobacconistHouseType, tobacconistShopType, cigarFactoryType };
+    private static final BuildingType[] toolBuildings = new BuildingType[] {
+        blacksmithHouseType, blacksmithShopType, ironWorksType };
+    private static final BuildingType[] musketBuildings = new BuildingType[] {
+        armoryType, magazineType, arsenalType };
+    public void testFactoryProduction() {
+        productionTest(clothBuildings, factoryProd);
+        productionTest(clothBuildings, factoryProd);
+        productionTest(sugarBuildings, factoryProd);
+        productionTest(tobaccoBuildings, factoryProd);
+        productionTest(toolBuildings, factoryProd);
+        productionTest(musketBuildings, factoryProd);
+    }
+        
     public void testToolsMusketProduction() {
+        // Test the interaction between tools and muskets
         Game game = getGame();
-        game.setMap(getTestMap(true));
+        game.changeMap(getTestMap(true));
 
         Colony colony = getStandardColony(8);
         Tile tile = colony.getTile();
@@ -920,7 +1130,7 @@ public class BuildingTest extends FreeColTestCase {
         assertTrue(colony.getBuilding(townHallType)
             .setWorkFor(units.get(7)));
 
-        Building smithy = colony.getBuilding(blacksmithType);
+        Building smithy = colony.getBuilding(blacksmithHouseType);
         assertFalse(smithy.setWorkFor(units.get(0)));
         colony.addGoods(oreType, 50); // Add ore so the smithy becomes viable
         assertTrue(smithy.setWorkFor(units.get(0)));
