@@ -1,5 +1,5 @@
 /**
- *  Copyright (C) 2002-2015  The FreeCol Team
+ *  Copyright (C) 2002-2022  The FreeCol Team
  *
  *  This file is part of FreeCol.
  *
@@ -21,6 +21,8 @@ package net.sf.freecol.common.model;
 
 import java.util.List;
 
+import net.sf.freecol.common.option.GameOptions;
+import static net.sf.freecol.common.util.CollectionUtils.*;
 import net.sf.freecol.server.model.ServerBuilding;
 import net.sf.freecol.server.model.ServerColony;
 import net.sf.freecol.server.model.ServerUnit;
@@ -68,9 +70,9 @@ public class ColonyProductionTest extends FreeColTestCase {
     public void testProductionSoldier() {
         Game game = getStandardGame();
         Map map = getTestMap();
-        game.setMap(map);
-        Player dutch = game.getPlayerByNationId("model.nation.dutch");
+        game.changeMap(map);
 
+        Player dutch = game.getPlayerByNationId("model.nation.dutch");
         Tile tile = map.getTile(5, 8);
         Resource grain = new Resource(game, tile, grainResource);
         tile.addResource(grain);
@@ -116,9 +118,9 @@ public class ColonyProductionTest extends FreeColTestCase {
     public void testProductionPioneer() {
         Game game = getStandardGame();
         Map map = getTestMap();
-        game.setMap(map);
-        Player dutch = game.getPlayerByNationId("model.nation.dutch");
+        game.changeMap(map);
 
+        Player dutch = game.getPlayerByNationId("model.nation.dutch");
         Tile tile = map.getTile(5, 8);
         Resource grain = new Resource(game, tile, grainResource);
         tile.addResource(grain);
@@ -161,7 +163,7 @@ public class ColonyProductionTest extends FreeColTestCase {
 
     public void testBellNetProduction() {
         Game game = getStandardGame();
-        game.setMap(getTestMap());
+        game.changeMap(getTestMap());
 
         // Get a minimal colony so that the units-that-use-no-bells
         // parameter will not be relevant.
@@ -236,20 +238,21 @@ public class ColonyProductionTest extends FreeColTestCase {
     /**
      * Tests that there is no over production of horses, to avoid them
      * being thrown out.  A test of the proper production of horses is
-     * in <code>BuildingTest</code>
+     * in {@code BuildingTest}
      */
     public void testNoHorsesOverProduction() {
         Game game = getGame();
-        game.setMap(getTestMap());
+        game.changeMap(getTestMap());
 
         Colony colony = getStandardColony(1);
         Building pasture = colony.getBuilding(countryType);
         Unit unit = colony.getFirstUnit();
         unit.setLocation(colony.getWorkLocationFor(unit, bellsType));
 
-        List<AbstractGoods> outputs = pasture.getOutputs();
+        List<AbstractGoods> outputs = toList(pasture.getOutputs());
         assertEquals(1, outputs.size());
         assertEquals(horsesType, outputs.get(0).getType());
+
         assertEquals("Wrong warehouse capacity in colony",
             GoodsContainer.CARGO_SIZE, colony.getWarehouseCapacity());
 
@@ -282,10 +285,8 @@ public class ColonyProductionTest extends FreeColTestCase {
 
 
     public void testConsumers() {
-
         Game game = getGame();
-        game.setMap(getTestMap());
-
+        game.changeMap(getTestMap());
 
         Colony colony = getStandardColony(3);
         int units = colony.getUnitCount();
@@ -371,7 +372,7 @@ public class ColonyProductionTest extends FreeColTestCase {
 
     public void testProduction() {
         Game game = getGame();
-        game.setMap(getTestMap());
+        game.changeMap(getTestMap());
 
         Colony colony = getStandardColony(3);
         ColonyTile tile = colony.getColonyTile(colony.getTile());
@@ -491,7 +492,7 @@ public class ColonyProductionTest extends FreeColTestCase {
 
     public void testGetPotentialProduction() {
         Game game = getGame();
-        game.setMap(getTestMap());
+        game.changeMap(getTestMap());
 
         Colony colony = getStandardColony(1);
         ColonyTile colonyTile = colony.getColonyTile(colony.getTile());
