@@ -1,5 +1,5 @@
 /**
- *  Copyright (C) 2002-2015  The FreeCol Team
+ *  Copyright (C) 2002-2022  The FreeCol Team
  *
  *  This file is part of FreeCol.
  *
@@ -19,10 +19,8 @@
 
 package net.sf.freecol.common.model;
 
-import net.sf.freecol.common.model.Direction;
-import net.sf.freecol.common.model.Player;
 import net.sf.freecol.common.model.Player.NoClaimReason;
-import net.sf.freecol.common.model.Role;
+import static net.sf.freecol.common.util.CollectionUtils.*;
 import net.sf.freecol.server.model.ServerUnit;
 import net.sf.freecol.server.model.ServerPlayer;
 import net.sf.freecol.util.test.FreeColTestCase;
@@ -86,7 +84,7 @@ public class ColonyTest extends FreeColTestCase {
 
     public void testCurrentlyBuilding() {
         Game game = getGame();
-        game.setMap(getTestMap(true));
+        game.changeMap(getTestMap(true));
 
         Colony colony = getStandardColony();
         colony.setCurrentlyBuilding(warehouseType);
@@ -100,7 +98,7 @@ public class ColonyTest extends FreeColTestCase {
 
     public void testBuildQueueDoesNotAcceptBuildingDoubles() {
         Game game = getGame();
-        game.setMap(getTestMap(true));
+        game.changeMap(getTestMap(true));
 
         Colony colony = getStandardColony();
 
@@ -123,7 +121,7 @@ public class ColonyTest extends FreeColTestCase {
 
     public void testBuildQueueAcceptsUnitDoubles() {
         Game game = getGame();
-        game.setMap(getTestMap(true));
+        game.changeMap(getTestMap(true));
 
         Colony colony = getStandardColony();
         colony.setCurrentlyBuilding(wagonTrainType);
@@ -137,10 +135,10 @@ public class ColonyTest extends FreeColTestCase {
     }
 
     public void testOccupationWithFood() {
-        int population = 1;
-
         Game game = getGame();
-        game.setMap(getTestMap(true));
+        game.changeMap(getTestMap(true));
+
+        int population = 1;
         Colony colony = getStandardColony(population);
 
         assertTrue("colony should produce enough food",
@@ -183,25 +181,21 @@ public class ColonyTest extends FreeColTestCase {
     }
 
     private int countParties(Colony colony) {
-        int modifierCount = 0;
-        for (Modifier existingModifier
-                 : colony.getModifiers("model.goods.bells")) {
-            if (Specification.COLONY_GOODS_PARTY_SOURCE
-                .equals(existingModifier.getSource())) modifierCount++;
-        }
-        return modifierCount;
+        return count(colony.getModifiers("model.goods.bells"),
+            m -> Specification.COLONY_GOODS_PARTY_SOURCE.equals(m.getSource()));
     }
 
     public void testTeaParty() {
         Game game = getGame();
-        game.setMap(getTestMap(true));
+        game.changeMap(getTestMap(true));
+
         Colony colony = getStandardColony(5);
-        ServerPlayer serverPlayer = (ServerPlayer)colony.getOwner();
+        Player player = colony.getOwner();
 
         assertEquals(0, countParties(colony));
-        colony.addModifier(serverPlayer.makeTeaPartyModifier());
+        colony.addModifier(((ServerPlayer)player).makeTeaPartyModifier());
         assertEquals(1, countParties(colony));
-        Modifier m = serverPlayer.makeTeaPartyModifier();
+        Modifier m = ((ServerPlayer)player).makeTeaPartyModifier();
         m.setFirstTurn(new Turn(game.getTurn().getNumber() + 1));
         colony.addModifier(m);
         assertEquals(2, countParties(colony));
@@ -209,7 +203,7 @@ public class ColonyTest extends FreeColTestCase {
 
     public void testAddUnitToColony() {
         Game game = getGame();
-        game.setMap(getTestMap(arcticTileType, true));
+        game.changeMap(getTestMap(arcticTileType, true));
 
         int population = 1;
         Colony colony = getStandardColony(population);
@@ -296,14 +290,14 @@ public class ColonyTest extends FreeColTestCase {
     public void testFoundColony() {
         Game game = getGame();
         Map map = getCoastTestMap(plainsTileType);
-        game.setMap(map);
+        game.changeMap(map);
 
         Colony colony = getStandardColony(3, 1, 8);
         Player dutch = colony.getOwner();
         Tile colonyTile = colony.getTile();
         assertEquals(colonyTile.getType(), plainsTileType);
 
-        Unit colonist = colony.getUnitList().get(0);
+        Unit colonist = first(colony.getUnits());
         Tile workedTile = null;
         for (ColonyTile ct : colony.getColonyTiles()) {
             if (ct.isColonyCenterTile()) continue;
@@ -429,7 +423,8 @@ public class ColonyTest extends FreeColTestCase {
 
     public void testUnderSiege() {
         Game game = getGame();
-        game.setMap(getTestMap(true));
+        game.changeMap(getTestMap(true));
+
         Colony colony = getStandardColony(5);
         Tile tile = colony.getTile().getNeighbourOrNull(Direction.N);
         Player iroquois = game.getPlayerByNationId("model.nation.iroquois");
@@ -468,7 +463,8 @@ public class ColonyTest extends FreeColTestCase {
 
     public void testUpkeep() {
         Game game = getGame();
-        game.setMap(getTestMap(true));
+        game.changeMap(getTestMap(true));
+
         Colony colony = getStandardColony(5);
 
         assertEquals("New colonies should not require upkeep.",
@@ -485,7 +481,8 @@ public class ColonyTest extends FreeColTestCase {
 
     public void testCopyColony() {
         Game game = getGame();
-        game.setMap(getTestMap(true));
+        game.changeMap(getTestMap(true));
+
         Colony colony = getStandardColony(2);
         Player player = colony.getOwner();
 
@@ -498,7 +495,7 @@ public class ColonyTest extends FreeColTestCase {
         assertEquals(colony.getName(), copied.getName());
         // Note: we can not check that player.hasSettlement(copied) is false
         // because it too will use FCGO.equals().
-        for (Settlement s : player.getSettlements()) {
+        for (Settlement s : player.getSettlementList()) {
             assertFalse(s == copied);
         }
 
@@ -512,14 +509,14 @@ public class ColonyTest extends FreeColTestCase {
         assertEquals(oct.getColony(), copied);
         assertEquals(oct.getOwningSettlement(), copied);
 
-        for (WorkLocation wl : colony.getAllWorkLocations()) {
+        for (WorkLocation wl : colony.getAllWorkLocationsList()) {
             WorkLocation owl = copied.getCorresponding(wl);
             assertNotNull(owl);
             assertFalse(wl == owl);
             assertEquals(wl.getId(), owl.getId());
-            if (wl instanceof ColonyTile) {
-                Tile wt = ((ColonyTile)wl).getWorkTile();
-                Tile owt = ((ColonyTile)owl).getWorkTile();
+            Tile wt = wl.getWorkTile();
+            if (wt != null) {
+                Tile owt = owl.getWorkTile();
                 assertFalse(wt == owt);
                 assertEquals(wt.getId(), owt.getId());
                 assertEquals(wt.getType(), owt.getType());
