@@ -1,5 +1,5 @@
 /**
- *  Copyright (C) 2002-2015  The FreeCol Team
+ *  Copyright (C) 2002-2022  The FreeCol Team
  *
  *  This file is part of FreeCol.
  *
@@ -23,10 +23,7 @@ import java.util.HashMap;
 import java.util.List;
 import java.util.Map;
 
-import net.sf.freecol.common.model.BuildingType;
-import net.sf.freecol.common.model.GoodsType;
-import net.sf.freecol.common.model.TileType;
-import net.sf.freecol.common.model.UnitType;
+import static net.sf.freecol.common.util.CollectionUtils.*;
 import net.sf.freecol.util.test.FreeColTestCase;
 
 
@@ -215,14 +212,14 @@ public class ProductionTypeTest extends FreeColTestCase {
                                 Map<GoodsType, Integer> outputs,
                                 List<ProductionType> productionTypes) {
         for (ProductionType productionType : productionTypes) {
-            for (AbstractGoods ag : productionType.getInputs()) {
+            for (AbstractGoods ag : toList(productionType.getInputs())) {
                 Integer i = inputs.get(ag.getType());
                 assertNotNull("Input expected for " + ag.getType(), i);
                 assertEquals("Input amount mismatch for " + ag.getType(),
                     i.intValue(), ag.getAmount());
                 inputs.remove(ag.getType());
             }
-            for (AbstractGoods ag : productionType.getOutputs()) {
+            for (AbstractGoods ag : toList(productionType.getOutputs())) {
                 Integer i = outputs.get(ag.getType());
                 assertNotNull("Output expected for " + ag.getType(), i);
                 assertEquals("Output amount mismatch for " + ag.getType(),
