@@ -1,5 +1,5 @@
 /**
- *  Copyright (C) 2002-2015  The FreeCol Team
+ *  Copyright (C) 2002-2022  The FreeCol Team
  *
  *  This file is part of FreeCol.
  *
@@ -20,7 +20,9 @@
 package net.sf.freecol.common.model;
 
 import java.util.List;
+import java.util.stream.Stream;
 
+import static net.sf.freecol.common.util.CollectionUtils.*;
 import net.sf.freecol.util.test.FreeColTestCase;
 import net.sf.freecol.util.test.FreeColTestUtils;
 
@@ -241,7 +243,8 @@ public class TileTest extends FreeColTestCase {
 
     public void testPrimarySecondaryGoods() {
         Game game = getStandardGame();
-        game.setMap(getTestMap(true));
+        game.changeMap(getTestMap(true));
+
         Colony colony = getStandardColony();
         Tile tile = colony.getTile();
         ColonyTile center = colony.getColonyTile(tile);
@@ -255,7 +258,7 @@ public class TileTest extends FreeColTestCase {
         assertEquals(cotton, secondaryProduction.getType());
         assertEquals(2, secondaryProduction.getAmount());
 
-        TileImprovement ti = new TileImprovement(game, tile, plow);
+        TileImprovement ti = new TileImprovement(game, tile, plow, null);
         ti.setTurnsToComplete(0);
         tile.add(ti);
         colony.invalidateCache();
@@ -380,7 +383,8 @@ public class TileTest extends FreeColTestCase {
     public void testImprovements() throws Exception {
         Game game = getStandardGame();
         Map map = getTestMap(plains);
-        game.setMap(map);
+        game.changeMap(map);
+
         Tile tile1 = map.getTile(5, 8);
         Tile tile2 = map.getTile(4, 8);
 
@@ -411,10 +415,9 @@ public class TileTest extends FreeColTestCase {
 
     public void testProductionModifiers() throws Exception {
         Game game = getGame();
-        game.setMap(getTestMap(true));
+        game.changeMap(getTestMap(true));
 
         Colony colony = getStandardColony();
-
         List<ColonyTile> colonyTiles = colony.getColonyTiles();
         ColonyTile colonyTile1 = null;
         ColonyTile colonyTile2 = null;
@@ -501,6 +504,8 @@ public class TileTest extends FreeColTestCase {
         assertFalse(tile2.canProduce(silver, null));
         oldBase = tile2.getBaseProduction(null, silver, null);
         Resource addedSilver = new Resource(game, tile2, mineralsResource);
+        assertTrue(addedSilver.canProduce(silver, null));
+        assertTrue(addedSilver.canProduce(silver, colonistType));
         tile2.addResource(addedSilver);
         newBase = tile2.getBaseProduction(null, silver, null);
         assertTrue(tile2.canProduce(silver, null));
@@ -521,28 +526,21 @@ public class TileTest extends FreeColTestCase {
                                 mineralsResource));
     }
 
-    private boolean hasBonusFrom(List<Modifier> modifierSet,
-                                 FreeColGameObjectType source) {
-        for (Modifier modifier : modifierSet) {
-            if (source.equals(modifier.getSource())) {
-                return true;
-            }
-        }
-        return false;
+    private boolean hasBonusFrom(Stream<Modifier> modifierSet,
+                                 FreeColSpecObjectType source) {
+        return any(modifierSet, matchKeyEquals(source, Modifier::getSource));
     }
 
 
     public void testColonyImprovements() throws Exception {
         Game game = getStandardGame();
         Map map = getTestMap(plains);
-        game.setMap(map);
+        game.changeMap(map);
 
         Colony colony = FreeColTestUtils.getColonyBuilder().build();
-
         assertTrue(colony.getTile().hasRoad());
 
         colony.exciseSettlement();
-
         assertFalse(colony.getTile().hasRoad());
     }
 
@@ -550,7 +548,7 @@ public class TileTest extends FreeColTestCase {
     public void testSortedPotential() {
         Game game = getStandardGame();
         Map map = getTestMap(plains);
-        game.setMap(map);
+        game.changeMap(map);
         Player dutch = game.getPlayer("model.nation.dutch");
         Market market = dutch.getMarket();
         UnitType sugarPlanter = spec().getUnitType("model.unit.masterSugarPlanter");
@@ -603,7 +601,7 @@ public class TileTest extends FreeColTestCase {
     public void testConiferForest() {
         Map map = getTestMap(coniferForest);
         Game game = getGame();
-        game.setMap(map);
+        game.changeMap(map);
 
         Colony colony = getStandardColony(1);
         final Tile tile = colony.getTile();
@@ -678,7 +676,7 @@ public class TileTest extends FreeColTestCase {
     public void testMinerals() {
         Game game = getGame();
         Map map = getTestMap(tundra);
-        game.setMap(map);
+        game.changeMap(map);
 
         Colony colony = getStandardColony();
         Tile tile = colony.getTile().getNeighbourOrNull(Direction.N);
@@ -695,11 +693,41 @@ public class TileTest extends FreeColTestCase {
         assertTrue(silver.isFarmed());
         assertEquals(0, tundra.getPotentialProduction(silver, colonistType));
         assertEquals(1, tile.getPotentialProduction(silver, colonistType));
-        assertEquals(1, tile.getProductionModifiers(silver, colonistType).size());
+        assertEquals(1, count(tile.getProductionModifiers(silver, colonistType)));
         assertEquals(1, colonyTile.getPotentialProduction(silver, unit.getType()));
         assertTrue(colonyTile.canBeWorked());
         assertTrue(colonyTile.canAdd(unit));
         assertEquals(colonyTile, colony.getWorkLocationFor(unit, silver));
+
+        // However minerals are *necessary* to produce silver
+        Tile marshTile = colony.getTile().getNeighbourOrNull(Direction.S);
+        marshTile.setType(marsh);
+        ColonyTile marshColonyTile = colony.getColonyTile(marshTile);
+        for (Unit u : marshColonyTile.getUnitList()) {
+            u.setLocation(colony.getBuilding(townHallType));
+        }
+        // No production yet
+        assertTrue(marshColonyTile.isEmpty());
+        assertEquals(marshColonyTile.getWorkTile().getOwningSettlement(),
+                     colony);
+        assertEquals(0, marsh.getPotentialProduction(silver, colonistType));
+        assertEquals(0, marshTile.getPotentialProduction(silver, colonistType));
+        assertEquals(0, count(marshTile.getProductionModifiers(silver, colonistType)));
+        assertEquals(0, marshColonyTile.getPotentialProduction(silver, unit.getType()));        
+        // Adding the road should not trigger any production
+        TileImprovement road = marshTile.addRoad();
+        road.setTurnsToComplete(0);
+        assertTrue(road.isComplete());
+        assertEquals(0, marsh.getPotentialProduction(silver, colonistType));
+        assertEquals(0, marshTile.getPotentialProduction(silver, colonistType));
+        assertEquals(0, count(marshTile.getProductionModifiers(silver, colonistType)));
+        assertEquals(0, marshColonyTile.getPotentialProduction(silver, unit.getType()));        
+        // Add the minerals and production should be available
+        marshTile.addResource(new Resource(game, marshTile, mineralsResource));
+        assertEquals(0, marsh.getPotentialProduction(silver, colonistType));
+        assertEquals(2, marshTile.getPotentialProduction(silver, colonistType));
+        assertEquals(2, count(marshTile.getProductionModifiers(silver, colonistType)));
+        assertEquals(2, marshColonyTile.getPotentialProduction(silver, unit.getType()));        
     }
 
     public void testDefenceModifiers() {
@@ -710,7 +738,7 @@ public class TileTest extends FreeColTestCase {
                 || "model.tile.mountains".equals(tileType.getId())
                 || "model.tile.swamp".equals(tileType.getId());
             assertEquals("Defence for " + tileType.getId(), present,
-                !tileType.getDefenceModifiers().isEmpty());
+                count(tileType.getDefenceModifiers()) != 0);
         }
     }
 
@@ -726,11 +754,12 @@ public class TileTest extends FreeColTestCase {
 
     public void testCopy() {
         Game game = getStandardGame();
-        game.setMap(getTestMap(plains));
+        game.changeMap(getTestMap(plains));
+
         Colony colony = getStandardColony();
         Tile tile = colony.getTile();
 
-        Tile otherTile = tile.copy(game, tile.getClass());
+        Tile otherTile = tile.copy(game);
         assertNotNull(otherTile);
         assertFalse(otherTile == tile);
         assertEquals(tile.getId(), otherTile.getId());
@@ -750,7 +779,7 @@ public class TileTest extends FreeColTestCase {
     public void testGetBestDisembarkTile() {
         Game game = getStandardGame();
         Map map = getCoastTestMap(plains, true);
-        game.setMap(map);
+        game.changeMap(map);
 
         Player dutch = game.getPlayerByNationId("model.nation.dutch");
         Tile settlementTile = map.getTile(9, 2);
