@@ -1,5 +1,5 @@
 /**
- *  Copyright (C) 2002-2015  The FreeCol Team
+ *  Copyright (C) 2002-2021  The FreeCol Team
  *
  *  This file is part of FreeCol.
  *
@@ -19,9 +19,11 @@
 
 package net.sf.freecol.server.model;
 
+import static net.sf.freecol.common.util.CollectionUtils.matchKey;
+import static net.sf.freecol.common.util.CollectionUtils.transform;
+
 import java.util.Iterator;
 import java.util.List;
-import java.util.stream.Collectors;
 
 import net.sf.freecol.common.model.Ability;
 import net.sf.freecol.common.model.Building;
@@ -121,8 +123,7 @@ public class ServerBuildingTest extends FreeColTestCase {
      * @return A list of all the units of the given type in this colony.
      */
     private List<Unit> getUnitList(Colony colony, UnitType type) {
-        return colony.getUnitList().stream()
-            .filter(u -> u.getType() == type).collect(Collectors.toList());
+        return transform(colony.getUnits(), matchKey(type, Unit::getType));
     }
 
     /**
@@ -138,7 +139,7 @@ public class ServerBuildingTest extends FreeColTestCase {
         Colony colony = getSchoolColony(4, SchoolLevel.SCHOOLHOUSE);
         Building school = colony.getBuilding(schoolType);
         assertTrue(school.canTeach());
-        Iterator<Unit> units = colony.getUnitIterator();
+        Iterator<Unit> units = colony.getUnitList().iterator();
 
         Unit colonist = units.next();
         colonist.setType(freeColonistType);
@@ -171,7 +172,7 @@ public class ServerBuildingTest extends FreeColTestCase {
 
         Colony colony = getSchoolColony(4, SchoolLevel.COLLEGE);
         Building college = colony.getBuilding(collegeType);
-        Iterator<Unit> units = colony.getUnitIterator();
+        Iterator<Unit> units = colony.getUnitList().iterator();
 
         Unit colonist = units.next();
         colonist.setType(freeColonistType);
@@ -199,7 +200,7 @@ public class ServerBuildingTest extends FreeColTestCase {
         assertEquals(4, colony.getUnitCount());
         Building university = colony.getBuilding(universityType);
         assertNotNull(university);
-        Iterator<Unit> units = colony.getUnitIterator();
+        Iterator<Unit> units = colony.getUnitList().iterator();
 
         Unit colonist = units.next();
         colonist.setType(freeColonistType);
@@ -257,7 +258,7 @@ public class ServerBuildingTest extends FreeColTestCase {
         Building townHall = colony.getBuilding(townHallType);
         Building lumberMill = colony.getBuilding(lumberMillType);
         Building college = colony.getBuilding(collegeType);
-        Iterator<Unit> units = colony.getUnitIterator();
+        Iterator<Unit> units = colony.getUnitList().iterator();
 
         clearWorkLocation(townHall);
         Unit colonist1 = units.next();
@@ -356,7 +357,7 @@ public class ServerBuildingTest extends FreeColTestCase {
         Building university = colony.getBuilding(universityType);
         Building townHall = colony.getBuilding(townHallType);
         clearWorkLocation(townHall);
-        Iterator<Unit> units = colony.getUnitIterator();
+        Iterator<Unit> units = colony.getUnitList().iterator();
 
         Unit colonist1 = units.next();
         colonist1.setType(freeColonistType);
@@ -410,7 +411,7 @@ public class ServerBuildingTest extends FreeColTestCase {
 
         Colony colony = getSchoolColony(4, SchoolLevel.UNIVERSITY);
         Building university = colony.getBuilding(universityType);
-        Iterator<Unit> units = colony.getUnitIterator();
+        Iterator<Unit> units = colony.getUnitList().iterator();
 
         Unit colonist = units.next();
         colonist.setType(freeColonistType);
@@ -455,7 +456,7 @@ public class ServerBuildingTest extends FreeColTestCase {
 
         Colony colony = getSchoolColony(4, SchoolLevel.UNIVERSITY);
         Building university = colony.getBuilding(universityType);
-        Iterator<Unit> units = colony.getUnitIterator();
+        Iterator<Unit> units = colony.getUnitList().iterator();
 
         Unit colonist = units.next();
         colonist.setType(freeColonistType);
@@ -495,7 +496,7 @@ public class ServerBuildingTest extends FreeColTestCase {
 
         Colony colony = getSchoolColony(4, SchoolLevel.UNIVERSITY);
         Building university = colony.getBuilding(universityType);
-        Iterator<Unit> units = colony.getUnitIterator();
+        Iterator<Unit> units = colony.getUnitList().iterator();
 
         Unit colonist = units.next();
         colonist.setType(freeColonistType);
@@ -553,7 +554,7 @@ public class ServerBuildingTest extends FreeColTestCase {
 
         Colony colony = getSchoolColony(4, SchoolLevel.UNIVERSITY);
         Building university = colony.getBuilding(universityType);
-        Iterator<Unit> units = colony.getUnitIterator();
+        Iterator<Unit> units = colony.getUnitList().iterator();
 
         Unit criminal = units.next();
         criminal.setType(pettyCriminalType);
@@ -585,7 +586,7 @@ public class ServerBuildingTest extends FreeColTestCase {
 
         Colony colony = getSchoolColony(4, SchoolLevel.UNIVERSITY);
         Building university = colony.getBuilding(universityType);
-        Iterator<Unit> units = colony.getUnitIterator();
+        Iterator<Unit> units = colony.getUnitList().iterator();
 
         Unit criminal = units.next();
         criminal.setType(pettyCriminalType);
@@ -614,7 +615,7 @@ public class ServerBuildingTest extends FreeColTestCase {
 
         Colony colony = getSchoolColony(4, SchoolLevel.UNIVERSITY);
         Building university = colony.getBuilding(universityType);
-        Iterator<Unit> units = colony.getUnitIterator();
+        Iterator<Unit> units = colony.getUnitList().iterator();
 
         Unit indenturedServant = units.next();
         indenturedServant.setType(indenturedServantType);
@@ -646,7 +647,7 @@ public class ServerBuildingTest extends FreeColTestCase {
 
         Colony colony = getSchoolColony(4, SchoolLevel.UNIVERSITY);
         Building university = colony.getBuilding(universityType);
-        Iterator<Unit> units = colony.getUnitIterator();
+        Iterator<Unit> units = colony.getUnitList().iterator();
         Unit outsider = new ServerUnit(game, colony.getTile(),
                                        colony.getOwner(), freeColonistType);
 
@@ -690,7 +691,7 @@ public class ServerBuildingTest extends FreeColTestCase {
         colony.addGoods(foodType, 100);
 
         Building university = colony.getBuilding(universityType);
-        Iterator<Unit> units = colony.getUnitIterator();
+        Iterator<Unit> units = colony.getUnitList().iterator();
 
         Unit colonist = units.next();
         colonist.setType(freeColonistType);
@@ -741,19 +742,25 @@ public class ServerBuildingTest extends FreeColTestCase {
 
         // Move teachers back to school, miner first to pick up the criminal
         teacher1.setLocation(university);
+        assertNotNull(teacher1.getStudent());
+        assertEquals(teacher1, teacher1.getStudent().getTeacher());
+        assertEquals(pettyCriminalType, teacher1.getStudent().getType());
         teacher2.setLocation(university);
+        assertNotNull(teacher2.getStudent());
+        assertEquals(teacher2, teacher2.getStudent().getTeacher());
+        assertEquals(freeColonistType, teacher2.getStudent().getType());
 
         ServerTestHelper.newTurn();
-        assertEquals(1, teacher1.getTurnsOfTraining());
-        assertEquals(1, teacher2.getTurnsOfTraining());
+        ServerTestHelper.newTurn();
+        ServerTestHelper.newTurn();
+        assertEquals(3, teacher1.getTurnsOfTraining());
+        assertEquals(3, teacher2.getTurnsOfTraining());
+        assertEquals(4, teacher1.getNeededTurnsOfTraining());
+        assertEquals(4, teacher2.getNeededTurnsOfTraining());
         assertEquals(1, getUnitList(colony, freeColonistType).size());
         assertEquals(1, getUnitList(colony, pettyCriminalType).size());
 
         ServerTestHelper.newTurn();
-        ServerTestHelper.newTurn();
-        ServerTestHelper.newTurn();
-        assertEquals(0, teacher1.getTurnsOfTraining());
-        assertEquals(0, teacher2.getTurnsOfTraining());
 
         // Teacher1's student (criminal) should be a servant now
         // Teacher2's student (colonist) should be a carpenter now
@@ -763,6 +770,9 @@ public class ServerBuildingTest extends FreeColTestCase {
         assertEquals(1, getUnitList(colony, expertOreMinerType).size());
         assertEquals(2, getUnitList(colony, masterCarpenterType).size());
 
+        assertEquals(0, teacher1.getTurnsOfTraining());
+        assertEquals(0, teacher2.getTurnsOfTraining());
+
         FreeColTestUtils.setStudentSelection(selection);
     }
 
@@ -772,7 +782,7 @@ public class ServerBuildingTest extends FreeColTestCase {
 
         Colony colony = getSchoolColony(4, SchoolLevel.UNIVERSITY);
         Building university = colony.getBuilding(universityType);
-        Iterator<Unit> units = colony.getUnitIterator();
+        Iterator<Unit> units = colony.getUnitList().iterator();
 
         Unit colonist = units.next();
         colonist.setType(freeColonistType);
@@ -826,7 +836,7 @@ public class ServerBuildingTest extends FreeColTestCase {
         colony.addGoods(foodType, 100);
 
         Building university = colony.getBuilding(universityType);
-        Iterator<Unit> units = colony.getUnitIterator();
+        Iterator<Unit> units = colony.getUnitList().iterator();
 
         Unit colonist = units.next();
         colonist.setType(freeColonistType);
@@ -914,7 +924,7 @@ public class ServerBuildingTest extends FreeColTestCase {
 
         Colony colony = getSchoolColony(4, SchoolLevel.UNIVERSITY);
         Building university = colony.getBuilding(universityType);
-        Iterator<Unit> units = colony.getUnitIterator();
+        Iterator<Unit> units = colony.getUnitList().iterator();
         Unit colonist = units.next();
         colonist.setType(freeColonistType);
         colonist.setLocation(colony.getBuilding(townHallType));
@@ -946,7 +956,7 @@ public class ServerBuildingTest extends FreeColTestCase {
 
         Colony colony = getSchoolColony(4, SchoolLevel.UNIVERSITY);
         Building university = colony.getBuilding(universityType);
-        Iterator<Unit> units = colony.getUnitIterator();
+        Iterator<Unit> units = colony.getUnitList().iterator();
 
         Unit colonist = units.next();
         colonist.setType(freeColonistType);
@@ -995,7 +1005,7 @@ public class ServerBuildingTest extends FreeColTestCase {
         colony.addGoods(foodType, 100);
 
         Building university = colony.getBuilding(universityType);
-        Iterator<Unit> units = colony.getUnitIterator();
+        Iterator<Unit> units = colony.getUnitList().iterator();
 
         Unit colonist = units.next();
         colonist.setType(freeColonistType);
@@ -1066,7 +1076,7 @@ public class ServerBuildingTest extends FreeColTestCase {
         colony.addGoods(foodType, 100);
 
         Building university = colony.getBuilding(universityType);
-        Iterator<Unit> units = colony.getUnitIterator();
+        Iterator<Unit> units = colony.getUnitList().iterator();
 
         Unit indenturedServant = units.next();
         indenturedServant.setType(indenturedServantType);
@@ -1123,7 +1133,7 @@ public class ServerBuildingTest extends FreeColTestCase {
         Colony colony = getSchoolColony(4, SchoolLevel.UNIVERSITY);
         Building university = colony.getBuilding(universityType);
         colony.getOwner().addAbility(new Ability(Ability.INDEPENDENCE_DECLARED));
-        Iterator<Unit> units = colony.getUnitIterator();
+        Iterator<Unit> units = colony.getUnitList().iterator();
 
         Unit regular = units.next();
         regular.setType(colonialRegularType);
@@ -1134,7 +1144,8 @@ public class ServerBuildingTest extends FreeColTestCase {
 
         regular.setLocation(university);
         assertEquals(colonist, regular.getStudent());
-        trainForTurns(colony, freeColonistType.getEducationTurns(veteranSoldierType));
+        trainForTurns(colony,
+            spec().getNeededTurnsOfTraining(veteranSoldierType, freeColonistType));
 
         assertEquals(veteranSoldierType, colonist.getType());
 
@@ -1147,7 +1158,7 @@ public class ServerBuildingTest extends FreeColTestCase {
 
         Colony colony = getSchoolColony(2, SchoolLevel.SCHOOLHOUSE);
         Building school = colony.getBuilding(schoolType);
-        Iterator<Unit> units = colony.getUnitIterator();
+        Iterator<Unit> units = colony.getUnitList().iterator();
 
         Unit lumber = units.next();
         lumber.setType(expertLumberJackType);
@@ -1186,7 +1197,7 @@ public class ServerBuildingTest extends FreeColTestCase {
         Colony colony = getSchoolColony(6, SchoolLevel.UNIVERSITY);
         Building university = colony.getBuilding(universityType);
 
-        Iterator<Unit> units = colony.getUnitIterator();
+        Iterator<Unit> units = colony.getUnitList().iterator();
 
         Unit carpenter = units.next();
         carpenter.setType(masterCarpenterType);
