@@ -1,5 +1,5 @@
 /**
- *  Copyright (C) 2002-2015  The FreeCol Team
+ *  Copyright (C) 2002-2021  The FreeCol Team
  *
  *  This file is part of FreeCol.
  *
@@ -19,7 +19,8 @@
 
 package net.sf.freecol.server.model;
 
-import java.util.List;
+import static net.sf.freecol.common.util.CollectionUtils.first;
+import static net.sf.freecol.common.util.CollectionUtils.transform;
 
 import net.sf.freecol.common.model.AbstractGoods;
 import net.sf.freecol.common.model.Building;
@@ -89,7 +90,7 @@ public class ServerColonyTest extends FreeColTestCase {
 
     public void testFoodConsumption() {
         Game game = ServerTestHelper.startServerGame(getTestMap(plains));
-        ServerPlayer dutch = (ServerPlayer)game.getPlayerByNationId("model.nation.dutch");
+        ServerPlayer dutch = getServerPlayer(game, "model.nation.dutch");
         // Setting test colony and colonist
         Colony colony = FreeColTestUtils.getColonyBuilder()
             .colonyTile(game.getMap().getTile(5, 8)).build();
@@ -119,15 +120,11 @@ public class ServerColonyTest extends FreeColTestCase {
 
         // Set the food production of the center tile of the colony to 2
         // This will be the only food production of the colony
-        List<AbstractGoods> colonyTileProduction
-            = colonyTile.getType().getPossibleProduction(true);
-        for (int i = 0; i < colonyTileProduction.size(); i++) {
-            AbstractGoods production = colonyTileProduction.get(i);
-            if (production.getType() == foodGoodsType) {
-                production.setAmount(2);
-                break;
-            }
-        }
+        AbstractGoods foodGoods
+            = first(transform(colonyTile.getType().getPossibleProduction(true),
+                              AbstractGoods.matches(foodGoodsType)));
+        if (foodGoods != null) foodGoods.setAmount(2);
+
         Unit unit = colony.getUnitList().get(0);
         unit.setLocation(colony.getWorkLocationFor(unit, bellsType));
 
@@ -263,7 +260,7 @@ public class ServerColonyTest extends FreeColTestCase {
         Unit unit = new ServerUnit(game, colony.getTile(), colony.getOwner(),
                                    colonistType);
         unit.setLocation(colony);
-        for (AbstractGoods ag : lumberMillType.getRequiredGoods()) {
+        for (AbstractGoods ag : lumberMillType.getRequiredGoodsList()) {
             GoodsType type = ag.getType();
             int amount = ag.getAmount() + 1;
             colony.addGoods(type, amount);
@@ -282,7 +279,7 @@ public class ServerColonyTest extends FreeColTestCase {
 
     public void testNoBuildingMaterialsProductionWhenBuildingNothing() {
         Game game = getGame();
-        game.setMap(getTestMap(true));
+        game.changeMap(getTestMap(true));
 
         Colony colony = getStandardColony(4);
         Building carpenterHouse = colony.getBuilding(carpenterHouseType);
@@ -322,7 +319,7 @@ public class ServerColonyTest extends FreeColTestCase {
         Unit statesman = colony.getUnitList().get(0);
         townHall.setWorkFor(statesman);
         assertEquals(bellsType, statesman.getWorkType());
-
+        
         ServerBuilding church
             = (ServerBuilding)colony.getBuilding(chapelType);
         church.upgrade();
@@ -330,23 +327,31 @@ public class ServerColonyTest extends FreeColTestCase {
         church.setWorkFor(preacher);
         assertEquals(crossesType, preacher.getWorkType());
 
-        assertEquals(0, colony.getGoodsCount(bellsType));
-        ServerTestHelper.newTurn();
-
-        int bells = 3;
         assertEquals(population, colony.getUnitCount());
-        assertEquals(bells, colony.getNetProductionOf(bellsType));
-        assertEquals(bells, colony.getGoodsCount(bellsType));
+        assertEquals(1, townHall.getUnitCount());
+        assertEquals(1, church.getUnitCount());
+
+        int bells0 = 3;
+        assertEquals(0, colony.getGoodsCount(bellsType));
+        assertEquals(bells0, colony.getNetProductionOf(bellsType));
+        assertEquals(0, colony.getLiberty());
+
+        ServerTestHelper.newTurn();
+        assertEquals(bells0, colony.getGoodsCount(bellsType));
+        assertEquals(bells0, colony.getNetProductionOf(bellsType));
+        assertEquals(bells0, colony.getLiberty());
 
         colony.addGoods(bellsType, 7);
-        bells += 7;
-        assertEquals(bells, colony.getGoodsCount(bellsType));
-        assertEquals(bells, colony.getLiberty());
+        int bells1 = bells0 + 7;
+        assertEquals(bells1, colony.getGoodsCount(bellsType));
+        assertEquals(bells0, colony.getNetProductionOf(bellsType));
+        assertEquals(bells1, colony.getLiberty());
 
         colony.removeGoods(bellsType, 5);
-        bells -= 5;
-        assertEquals(bells, colony.getGoodsCount(bellsType));
-        assertEquals(bells, colony.getLiberty());
+        int bells2 = bells1 - 5;
+        assertEquals(bells2, colony.getGoodsCount(bellsType));
+        assertEquals(bells0, colony.getNetProductionOf(bellsType));
+        assertEquals(bells2, colony.getLiberty());
 
         int crosses = colony.getTotalProductionOf(crossesType)
             - colony.getConsumptionOf(crossesType);
