@@ -1,5 +1,5 @@
 /**
- *  Copyright (C) 2002-2015  The FreeCol Team
+ *  Copyright (C) 2002-2021  The FreeCol Team
  *
  *  This file is part of FreeCol.
  *
@@ -27,7 +27,6 @@ import net.sf.freecol.common.model.Colony;
 import net.sf.freecol.common.model.ColonyTile;
 import net.sf.freecol.common.model.Direction;
 import net.sf.freecol.common.model.Game;
-import net.sf.freecol.common.model.GameOptions;
 import net.sf.freecol.common.model.GoodsType;
 import net.sf.freecol.common.model.Map;
 import net.sf.freecol.common.model.Modifier;
@@ -38,9 +37,11 @@ import net.sf.freecol.common.model.TileImprovement;
 import net.sf.freecol.common.model.TileImprovementType;
 import net.sf.freecol.common.model.TileType;
 import net.sf.freecol.common.model.Unit;
+import net.sf.freecol.common.model.UnitChangeType;
 import net.sf.freecol.common.model.UnitType;
-import net.sf.freecol.common.model.UnitTypeChange.ChangeType;
 import net.sf.freecol.common.model.WorkLocation;
+import net.sf.freecol.common.option.GameOptions;
+import static net.sf.freecol.common.util.CollectionUtils.*;
 import net.sf.freecol.server.ServerTestHelper;
 import net.sf.freecol.server.control.InGameController;
 import net.sf.freecol.util.test.FreeColTestCase;
@@ -53,6 +54,8 @@ public class ServerUnitTest extends FreeColTestCase {
         = spec().getBuildingType("model.building.carpenterHouse");
     private static final BuildingType townHallType
         = spec().getBuildingType("model.building.townHall");
+    private static final BuildingType warehouseType
+        = spec().getBuildingType("model.building.depot");
 
     private static final GoodsType foodType
         = spec().getPrimaryFoodType();
@@ -93,9 +96,9 @@ public class ServerUnitTest extends FreeColTestCase {
     /**
      * Does a tile have a completed improvement of the given type?
      *
-     * @param tile The <code>Tile</code> to check.
-     * @param type The <code>TileImprovementType</code> to check.
-     * @return True if this <code>Tile</code> has the improvement type and
+     * @param tile The {@code Tile} to check.
+     * @param type The {@code TileImprovementType} to check.
+     * @return True if this {@code Tile} has the improvement type and
      *      it is complete.
      */
     private static boolean hasImprovement(Tile tile, TileImprovementType type) {
@@ -140,7 +143,7 @@ public class ServerUnitTest extends FreeColTestCase {
         Game game = ServerTestHelper.startServerGame(getTestMap(plains));
         InGameController igc = ServerTestHelper.getInGameController();
 
-        ServerPlayer dutch = (ServerPlayer)game.getPlayerByNationId("model.nation.dutch");
+        ServerPlayer dutch = getServerPlayer(game, "model.nation.dutch");
         Tile plain = game.getMap().getTile(5, 8);
         plain.setExplored(dutch, true);
         plain.setOwner(dutch);
@@ -184,7 +187,7 @@ public class ServerUnitTest extends FreeColTestCase {
         Game game = ServerTestHelper.startServerGame(getTestMap(plains));
         InGameController igc = ServerTestHelper.getInGameController();
 
-        ServerPlayer dutch = (ServerPlayer)game.getPlayerByNationId("model.nation.dutch");
+        ServerPlayer dutch = getServerPlayer(game, "model.nation.dutch");
         Map map = game.getMap();
         map.getTile(5, 8).setExplored(dutch, true);
         map.getTile(6, 8).setExplored(dutch, true);
@@ -268,7 +271,7 @@ public class ServerUnitTest extends FreeColTestCase {
         Game game = ServerTestHelper.startServerGame(getTestMap(savannahForest));
         InGameController igc = ServerTestHelper.getInGameController();
 
-        ServerPlayer dutch = (ServerPlayer)game.getPlayerByNationId("model.nation.dutch");
+        ServerPlayer dutch = getServerPlayer(game, "model.nation.dutch");
         Map map = game.getMap();
         Tile tile = map.getTile(5, 8);
         map.getTile(5, 8).setExplored(dutch, true);
@@ -395,9 +398,8 @@ public class ServerUnitTest extends FreeColTestCase {
         // Can colonist be upgraded
         UnitType learn = spec().getExpertForProducing(colonist.getWorkType());
         assertNotNull(learn);
-        assertTrue(learn != colonist.getType());
-        assertTrue(colonist.getType().canBeUpgraded(learn,
-                                                    ChangeType.EXPERIENCE));
+        assertFalse(learn == colonist.getType());
+        assertNotNull(colonist.getUnitChange(UnitChangeType.EXPERIENCE));
 
         // set some experience
         int expectXP = 10;
@@ -415,9 +417,8 @@ public class ServerUnitTest extends FreeColTestCase {
         boolean isExpert = colonist.getType() == expertFarmerType;
         assertFalse("Unit should not be an expert", isExpert);
 
-        // Make upgrade
+        // Unit should upgrade
         ServerTestHelper.newTurn();
-        assertTrue(colonist.getExperience() > expectXP);
 
         // verify upgrade
         isExpert = colonist.getType() == expertFarmerType;
@@ -428,7 +429,7 @@ public class ServerUnitTest extends FreeColTestCase {
         Game game = ServerTestHelper.startServerGame(getTestMap(savannahForest));
         InGameController igc = ServerTestHelper.getInGameController();
 
-        ServerPlayer dutch = (ServerPlayer)game.getPlayerByNationId("model.nation.dutch");
+        ServerPlayer dutch = getServerPlayer(game, "model.nation.dutch");
         Map map = game.getMap();
         Tile tile = map.getTile(5, 8);
         tile.setOwner(dutch);
@@ -491,17 +492,17 @@ public class ServerUnitTest extends FreeColTestCase {
         ServerTestHelper.newTurn();
 
         // Lumber should be delivered on this turn
-        int lumber = colony.getGoodsCount(lumberType);
+        colony.removeGoods(lumberType);
         ServerTestHelper.newTurn();
-        assertEquals("Lumber delivery with hardy pioneer", lumber + 20 * 2,
+        assertEquals("Lumber delivery with hardy pioneer", 20 * 2,
                      colony.getGoodsCount(lumberType));
 
-        // Upgrade to lumber mill
-        assertEquals(0,
-            colony.getModifiers(Modifier.TILE_TYPE_CHANGE_PRODUCTION).size());
+        // Upgrade to lumber mill and full warehouse
+        assertTrue(none(colony.getModifiers(Modifier.TILE_TYPE_CHANGE_PRODUCTION)));
         colony.getBuilding(carpenterHouseType).upgrade();
+        colony.getBuilding(warehouseType).upgrade();
         assertEquals(1,
-            colony.getModifiers(Modifier.TILE_TYPE_CHANGE_PRODUCTION).size());
+            count(colony.getModifiers(Modifier.TILE_TYPE_CHANGE_PRODUCTION)));
 
         // Almost clear another tile
         Tile tile2 = tile.getNeighbourOrNull(Direction.N);
@@ -514,9 +515,9 @@ public class ServerUnitTest extends FreeColTestCase {
         ServerTestHelper.newTurn();
 
         // Lumber should be delivered on this turn
-        lumber = colony.getGoodsCount(lumberType);
+        colony.removeGoods(lumberType);
         ServerTestHelper.newTurn();
         assertEquals("Lumber delivered with hardy pioneer and mill",
-                     lumber + 20 * 2 * 3, colony.getGoodsCount(lumberType));
+                     20 * 2 * 3, colony.getGoodsCount(lumberType));
     }
 }
